/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _ALSAAUDIO_H_
#define _ALSAAUDIO_H_

#include <alsa/asoundlib.h>

#include "Plugin.h"
#include "Thread.h"
#include "AudioClock.h"
#include "Pollable.h"
#include "Pipe.h"

class AlsaClock;

extern Type * AlsaStream_type;
extern Type * AlsaIn_type;
extern Type * AlsaOut_type;
extern Type * AlsaClock_type;

class AlsaStream
:	public Plugin,
	public Pollable
{
	friend class AlsaClock;

	private:
		/* ALSA access */
		snd_pcm_t	* handle;
		
		snd_pcm_hw_params_t * hw_params;
		snd_pcm_sw_params_t * sw_params;
		
		snd_pcm_status_t * status;	

		const snd_pcm_channel_area_t * areas;

		/* framecount declared accessible by mmap_begin */		
		int frames;

	public:
		/* if not relaxed: driver stop on xrun */
		bool relaxed;
		
		/* frames ready for IO in interface ringbuffer */		
		int available;
		
		snd_pcm_sframes_t delay;
		snd_pcm_uframes_t offset;

		bool interleaved, configured;
		uint bytes_per_sample, frames_per_cycle, sample_rate;
		int channels;
					
		d_sample * gain; /* gain per channel */

	public:
		/* type is overridden with AlsaIn_type or AlsaOut_type. */
		AlsaStream (Type * type = AlsaStream_type);
		~AlsaStream();

		void open (char * _name);

		/* Plugin::* */
		virtual void pre_remove() 
			{ 
				if (state >= Plugin::Prepared)
					throw RuntimeError() << "stop the graph first.";
			}
		
	public: /* Pollable:: */
		virtual void get_pollfd (InstanceList<struct pollfd> & list) 
			{
				int n = snd_pcm_poll_descriptors_count (handle);
				int i = list.insert (n);
				snd_pcm_poll_descriptors (handle, &(list[i]), n);
			}
		
		virtual operator Object * ()
			{
				return this;
			}

	private:
		void link (AlsaStream *);
		void configure (uint srate, uint fpc, uint cycles);

		void prepare();
		void start();
		void stop();
		
		/* updates delay and available; returns false on xrun. */
		bool update_status();
		void dump_status();

	protected:
		/* silence all channels. */
		void complete_silence (uint frames);

		/* returns number of frames accessible */
		int begin_cycle (int _frames);
		
		/* commits */
		void end_cycle (int _frames);
		
		void silence (int channel);

		void stream_in (int channel, d_sample *);
		void stream_out (int channel, d_sample *);

		/* called by configure to initialize the ports. */
		void setup_ports();
};

/* dedicated to the two directions a PCM stream can represent --
 * the actual behaviour differences are coded into AlsaStream. 
 */
class AlsaIn
: public AlsaStream
{
	public:
		AlsaIn()
			: AlsaStream (AlsaIn_type)
			{ }
		
		virtual void process_audio (int frames);
};

class AlsaOut
: public AlsaStream
{
	public:
		AlsaOut()
			: AlsaStream (AlsaOut_type)
			{ }
		
		~AlsaOut()
			{
				if (gain) free (gain);
			}

		virtual void process_audio (int frames);
};

class AlsaClockStats
{
	public:
		struct 
		{
			int min, max;
			int64 sum;
		} frames;
		
		struct
		{
			d_time min, max, sum, then;
		} time;
	
		void reset()
		{
			time.then = gettimeofday();

			frames.min = 100000;
			frames.max = 
			frames.sum = 0;

			time.min = 1000000.;
			time.max = 
			time.sum = 0.;
		}

		int64 cycles;
		int subcycles;

		AlsaClockStats() 
			{
				reset();
			}

		void subcycle (int _frames);
		
		d_time begin_cycle();
		void end_cycle (int sample_rate, int show_stats);
};

class RTSentinel;

class AlsaClock
:	public AudioClock,
	public Thread
{
	private:
		Poll poll;
		Pipe state_pipe;

		TypedObjectList<AlsaStream> streams;

	public:
		int sample_rate, frames_per_cycle, buffer_cycles;

		/* age in cycles */
		unsigned long cycles; 

		/* show cycle() statistics */
		int show_stats; 
		
		/* last xrun happened at this cycle */
		unsigned long last_xrun; 
		
		/* consecutive xruns faster than this stop the clock. */
		unsigned long max_xrun_cycles; 
		
		/* don't ever process less than this # of frames in one cycle. */
		d_frame min_available;

		AlsaClockStats stats;
		bool quitting;

		RTSentinel * sentinel;

		AlsaClock();
		~AlsaClock();

		/* Plugin:: */
		virtual void set_state (Graph *, int _state);
		virtual void wake_up();

		virtual void pre_insert (Graph * g);
		
		virtual void pre_remove_other (Plugin * p)
			{
				if (state < Prepared)
					return;
				if (AlsaStream_type->is_derived (p))
					throw RuntimeError() << this << " needs " << p;
			}

	private:
		virtual void * thread();
		
		void prepare();
		void start_rolling();
		void stop_rolling();

		void cycle();
		void xrun (int delay);
};

#endif /* _ALSAAUDIO_H_ */
