/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "Graph.h"
#include "AlsaAudio.h"
#include "RTSentinel.h"

/* /////////////////////////////////////////////////////////////////////// */

AlsaClock::AlsaClock()
:	AudioClock (AlsaClock_type, 0),
	Thread ("AlsaClock"),
	poll ("alsa clock poll"),
	state_pipe ("state"),
	streams ("alsa streams")
{
	sample_rate = SAMPLE_RATE;
	frames_per_cycle = FRAMES_PER_CYCLE;
	buffer_cycles = 2;

	show_stats = 0;
	last_xrun = 0;
	max_xrun_cycles = 4;
	min_available = 16;

	quitting = 0;

	poll.add (state_pipe);

	set_object_name ("ALSA audio clock");

	sentinel = 0;
	if (start_thread_rt (15))
		sentinel = new RTSentinel ("alsa clock sentinel", 16);
}

AlsaClock::~AlsaClock()
{
	quitting = 1;
	wake_up();

	join_thread();

	if (sentinel)
		delete sentinel;

	streams.clear();
}

void AlsaClock::pre_insert (Graph * g)
{
	if (state != Off)
		throw AssertionError() << "stop me first.";
	if (g->state != Off)
		throw AssertionError() << "cannot join a graph in motion.";
	
	g->sample_rate = sample_rate;
	g->frames_per_cycle = frames_per_cycle;
}

/* interrupt the poll on alsa file handles. */
void
AlsaClock::wake_up()
{
	state_pipe.write (this, 1);
}

void *
AlsaClock::thread()
{
	cycles = 0;

	/* because the graph relies on the audio thread being active to process 
	 * requests, this loop should run as robustly as possible. */
	while (1)
	{
		if (quitting)
		{
			if (state >= Activated)
				stop_rolling();
			break;
		}

		if (sentinel)
			sentinel->give_livesign();

		int nfds = (state >= Activated) ? poll.length : 1;

		int r = poll.poll (10, nfds);

		if (r < 0)
			continue;
		
		/* can't have the graph destructor set graph to 0 */
		ScopedLock lock (mutex);

		/* check state fd poll */
		if (poll[0].revents & POLLHUP)
			throw StopIteration() << "state socket hung up.";
		if (poll[0].revents & POLLERR) 
			throw IOError() << "poll error on state socket.";

		/* state fd wake up */
		if (poll[0].revents & POLLIN) 
		{
			char dummy;
			while (state_pipe.read (&dummy, 1) == 1)
				;

			/* woken, check for graph modifications. */
			if (graph)
				graph->check_for_requests();

			if (quitting)
				break;
		}
		else if (r == 0 && graph) 
		{
			/* timed out, check for graph requests. */				
			graph->check_for_requests();
		}

		if (state < Plugin::Activated)
			continue;

		int do_cycle = 0;
		for (int i = 1; i < nfds; i++)
		{
			struct pollfd & pfd = poll[i];
			
			AlsaStream * stream = streams[i - 1];
			
			if (pfd.revents & POLLHUP)
				throw IOError() << stream << " hung up.";

			else if (pfd.revents & POLLERR)
			{
				msg(), "poll error on ", get_name (stream);

				if (stream->update_status())
				{
					stream->dump_status();
					msg(), get_name (stream), " status seems OK. puzzled & stopping.";
					graph->set_state (Off);
				}
				else
				{
					xrun (stream->delay);
				}

				do_cycle = 0;
				break;
			}

			if (pfd.revents & (POLLIN | POLLOUT))
				do_cycle = 1;
		}

		if (do_cycle)
		{
			try 
			{
				cycle();
			} 
			catch (blewit & b) 
			{
				msg(), "cycle() blew it: " , b.what();

				stop_rolling();
	
				state = Off;
			}

			graph->check_for_requests();
		}
	}
	
	if (sentinel)
		sentinel->active = 0;

	return 0;
}

/* AlsaClock::cycle() in a nutshell: 
 *
 *  check if we are late, 
 *  determine the # of frames to process,
 *  call graph->process_audio().
 */
void
AlsaClock::cycle()
{
	d_time dt = stats.begin_cycle();
	d_time cycle_time = graph->audio_cycle_time;

	if (cycles && dt > 1.3 * cycle_time)
		msg(), 1000. * (dt - cycle_time), "ms late at frame ", graph->frame;
					
	/* xrun detection */
	for (int i = 0; i < streams; i++)
	{
		if (streams[i]->update_status())
			continue;
		xrun (streams[i]->delay);
		return;
	}

	/* frames to process calculation */
	int available = INT_MAX;
	for (int i = 0; i < streams; i++)
		available = min (available, streams[i]->available);

	/* this loop runs only once under normal circumstances. */
	while (available >= min_available)
	{
		int frames = min (available, frames_per_cycle);

		if (show_stats == 1)
		{
			_msg m;
			
			for (int i = 0; i < streams; i++)
				m, get_type_name (streams[i]), ": ", streams[i]->available;
		}

		graph->process_audio (frames);

		available -= frames;
		stats.subcycle (frames);
	}

	stats.end_cycle (sample_rate, show_stats);

	++cycles;
}

void
AlsaClock::set_state (Graph * g, int _state)
{
	switch (_state) 
	{
		case Off:
			if (state == Prepared)
				throw ValueError() << this << "can not stop from prepared state.";
			if (state >= Activated)
				stop_rolling();
			break;

		case Prepared:
			if (state != Off)
				throw ValueError() << this << "can only prepare in stopped state.";
			prepare();
			break;

		case Activated:
			if (state >= Activated)
				break;
			if (state != Prepared)
				throw ValueError() << this << "can only activate from prepared state.";
			last_xrun = 0;
			start_rolling();
			break;

		case Rolling:
			if (state != Activated)
				throw ValueError() << this << "can only roll from activated state.";
			break;
	}

	state = _state;
}

void
AlsaClock::prepare()
{
	if (graph->frames_per_cycle != frames_per_cycle)
		throw AssertionError() << "Graph->frames_per_cycle = " 
				<< graph->frames_per_cycle << "!= my " << frames_per_cycle;

	if (graph->sample_rate != sample_rate)
		throw AssertionError() << "Graph->sample_rate = " 
				<< graph->sample_rate << "!= my " << sample_rate;

	/* enforce full cycles of processing only. */
	min_available = frames_per_cycle;

	/* rebuild poll fd list, start with state fd */
	poll.clear();
	poll.add (state_pipe);

	/* collect all AlsaStreams from the graph */
	streams.clear();
	for (int i = 0; i < graph->plugins; i++)
	{
		AlsaStream * s = (AlsaStream *) graph->plugins[i];
		if (!AlsaStream_type->is_derived (s))
			continue;
			
		poll.add (*s);

		if (get_type (s) == AlsaOut_type)
			streams.add_object (s, 0);
		else
			streams.add_object (s);
	}

	if (streams.length == 0)
		throw AssertionError() << "no streams to run.";

	/* this should rather be in AlsaStream::set_state, but it's, alas, a
	 * little more complicated because we have to link them, and feed
	 * preroll data to the output streams. 
	 */
	for (int i = 0; i < streams; i++)
		streams[i]->configure (sample_rate, frames_per_cycle, buffer_cycles);
	
	for (int i = 1; i < streams; i++)
		streams[i]->link (streams[0]);
	
	streams[0]->prepare();

	/* the playback streams need some preroll data, we feed a full buffer worth. */
	for (int i = 0; i < buffer_cycles; i++)
	{
		int frames = frames_per_cycle;

		for (int i = 0; i < streams; i++)
			if (get_type (streams[i]) == AlsaOut_type)
				frames = min (frames, streams[i]->begin_cycle (frames));

		msg(), "prerolling ", frames, "frames";

		graph->process_audio (frames);
	}
}

void
AlsaClock::start_rolling()
{
	streams[0]->start();

	cycles	= 0;
}

void
AlsaClock::stop_rolling()
{
	streams[0]->stop();

	for (int i = 1; i < streams; i++)
		streams[i]->link (0);
}

void
AlsaClock::xrun (int delay)
{
	stop_rolling();

	if (cycles < last_xrun + max_xrun_cycles)
	{
		msg(), this, "is xrunning too fast. FULL STOP.";

		/* should we take the graph lock here? works fine so far ... */
		graph->set_state (Off);
		return;
	}

	prepare();

	/* do 'virtual' processing for a little more than the timespan lost
	 * in order to smoothly keep up with midi time. 
	 */ 
	while (delay >= 0)
	{
		graph->process_audio (frames_per_cycle);
		delay -= frames_per_cycle;
	}
	
	start_rolling();
}

/* /////////////////////////////////////////////////////////////////////// */

d_time
AlsaClockStats::begin_cycle()
{
	d_time t = gettimeofday();
	d_time dt = t - time.then;
	time.then = t;
	return dt;
}

void
AlsaClockStats::subcycle (int f)
{
	frames.min = min (frames.min, f);
	frames.max = max (frames.max, f);
	frames.sum += (int64) f;
}

void
AlsaClockStats::end_cycle (int sample_rate, int show_stats)
{
	d_time dt = gettimeofday() - time.then;
	time.min = min (time.min, dt);
	time.max = max (time.max, dt);
	time.sum += dt;

	cycles++;

	if (show_stats <= 1) return;
	if (cycles % show_stats) return;
	
	double cpu = time.sum / ((double) frames.sum / (double) sample_rate);
	msg(), "cpu=", cpu, "f=", frames.min, frames.max, "t=", time.min, time.max;
	reset();
}


