/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "AlsaMidi.h"
#include "MidiStream.h"
#include "AlsaError.h"
#include "Graph.h"
#include "PyType.h"

Type * AlsaMidi_type;
Type * AlsaMidiIn_type;
Type * AlsaMidiOut_type;

AlsaMidi::AlsaMidi (Type * t, int _flags)
: Plugin (t, _flags)
{
	set_alsa_error_handler();

	handle = 0;
	q = -1;
	port = -1;
	client = -1;
	direction = -1;
}

AlsaMidi::~AlsaMidi()
{
	if (handle == 0)
		return;
	
	if (port != -1)
		snd_seq_delete_simple_port (handle, port);
	
	if (q != -1)
		snd_seq_free_queue (handle, q);
	
	snd_seq_close (handle);
}

void
AlsaMidi::open (char * name, int _direction)
{
	set_object_name (name);
	direction = _direction;

	int error = snd_seq_open (&handle, "hw", direction, SND_SEQ_NONBLOCK);
	if (error)
		throw IOError (abs (error)) << "snd_seq_open fails.";
	
	q = snd_seq_alloc_queue (handle);

	client = snd_seq_client_id (handle);

	int port_caps = 0;

	if (direction == SND_SEQ_OPEN_OUTPUT)
		port_caps |= SND_SEQ_PORT_CAP_READ | SND_SEQ_PORT_CAP_SUBS_READ;
	else
		port_caps |= SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_SUBS_WRITE;

	port = snd_seq_create_simple_port (handle, name, port_caps,
			SND_SEQ_PORT_TYPE_MIDI_GENERIC);
	
	if (port < 0)
		throw IOError (-port);
}

void
AlsaMidi::subscribe (int major, int minor, bool connect)
{
	snd_seq_port_subscribe_t * subscription;
	snd_seq_port_subscribe_alloca (&subscription);
	
	snd_seq_port_subscribe_set_queue (subscription, q);

	snd_seq_addr_t 
		me = {client, port},
		he = {major, minor};
		
	if (direction == SND_SEQ_OPEN_OUTPUT)
	{
		snd_seq_port_subscribe_set_sender (subscription, &me);
		snd_seq_port_subscribe_set_dest (subscription, &he);
	}
	else
	{
		snd_seq_port_subscribe_set_sender (subscription, &he);
		snd_seq_port_subscribe_set_dest (subscription, &me);
	}

	int error;

	if (connect)
		error = snd_seq_subscribe_port ((snd_seq_t *) handle, subscription);
	else
		error = snd_seq_unsubscribe_port ((snd_seq_t *) handle, subscription);

	if (error)
		throw IOError (abs (error));
}

/* /////////////////////////////////////////////////////////////////////// */

AlsaMidiIn::AlsaMidiIn()
: AlsaMidi (AlsaMidiIn_type, 0)
{
	add_out (new Out (this, MidiEvent_type, "midi"));
	add_out (new Out (this, SysexEvent_type, "sysex"));
}

AlsaMidiIn::~AlsaMidiIn()
{
	if (handle)
		MidiInThread::remove (this);
}

void
AlsaMidiIn::init (char * name)
{
	open (name, SND_SEQ_OPEN_INPUT);

	MidiInThread::add (this);
}

void
AlsaMidiIn::get_pollfd (InstanceList<struct pollfd> & list)
{
	if (handle == 0)
		throw IOError() << this << "isn't open yet.";

	int n = snd_seq_poll_descriptors_count (handle, POLLIN);

	if (n < 1 || n > 16)
		throw IOError() << this << "has " << n << " poll fds.";
		
	int i = list.insert (n);
	snd_seq_poll_descriptors (handle, &list[i], n, POLLIN);
}

void
AlsaMidiIn::do_io()
{
	/* 1 = fetch events from kernel. */
	int n = snd_seq_event_input_pending (handle, 1);
	if (n <= 0)
		return;

	while (n--)
	{
		snd_seq_event_t * packet;
		int error = snd_seq_event_input ((snd_seq_t *) handle, &packet);
		
		if (error < 0)
		{
			msg(), this, "could not fetch an event: ", strerror (-error);
			return;
		}
		
		if (packet->type == SND_SEQ_EVENT_SYSEX)
		{
			SysexEvent * e = (SysexEvent *) outs[1]->cache->pop();
			e->tick = graph ? graph->tick : 0;
			e->set_data (
					(unsigned char *) packet + sizeof (snd_seq_event_t),
					packet->data.ext.len
				);
			outs[1]->push (e);
		}
		else
		{
			MidiEvent * e = (MidiEvent *) outs[0]->cache->pop();
			e->length = 0;

			switch (packet->type)
			{
				case SND_SEQ_EVENT_NOTEOFF:
					e->length = 3;
					e->data[0] = 0x80 | packet->data.note.channel;
					e->data[1] = packet->data.note.note;
					e->data[2] = packet->data.note.velocity;
					break;

				case SND_SEQ_EVENT_NOTEON:
					e->length = 3;
					e->data[0] = 0x90 | packet->data.note.channel;
					e->data[1] = packet->data.note.note;
					e->data[2] = packet->data.note.velocity;
					break;

				case SND_SEQ_EVENT_KEYPRESS:
					e->length = 3;
					e->data[0] = 0xA0 | packet->data.note.channel;
					e->data[1] = packet->data.note.note;
					e->data[2] = packet->data.note.velocity;
					break;

				case SND_SEQ_EVENT_CONTROLLER:
					e->length = 3;
					e->data[0] = 0xB0 | packet->data.control.channel;
					e->data[1] = packet->data.control.param;
					e->data[2] = packet->data.control.value;
					break;

				case SND_SEQ_EVENT_PGMCHANGE:
					e->length = 2;
					e->data[0] = 0xC0 | packet->data.control.channel;
					e->data[1] = packet->data.control.value;
					break;

				case SND_SEQ_EVENT_CHANPRESS:
					e->length = 2;
					e->data[0] = 0xD0 | packet->data.control.channel;
					e->data[1] = packet->data.control.value;
					break;

				case SND_SEQ_EVENT_PITCHBEND:
					e->length = 3;
					e->data[0] = 0xE0 | packet->data.control.channel;
					e->data[1] = packet->data.control.value >> 7;
					e->data[2] = packet->data.control.value & 0x7F;
					break;

				default:
					break;
			}

			if (e->length == 0)
				::decref (e);
			else
			{
				e->tick = graph ? graph->tick : 0;
				outs[0]->push (e);
			}
		}

		/* clean up. */
		snd_seq_free_event (packet);
	}
}

/* /////////////////////////////////////////////////////////////////////// */

AlsaMidiOut::AlsaMidiOut()
: AlsaMidi (AlsaMidiOut_type, Time)
{
	add_in (new In (this, MidiEvent_type, IN_SLOTS, "midi"));
}

void
AlsaMidiOut::init (char * name)
{
	open (name, SND_SEQ_OPEN_OUTPUT);

	snd_seq_ev_clear (&packet);
	snd_seq_ev_set_source (&packet, port);

	packet.queue = q;

	packet.type = 0;
	packet.flags = SND_SEQ_TIME_STAMP_TICK | SND_SEQ_TIME_MODE_REL;
	packet.time.tick = 0;

	snd_seq_ev_set_subs (&packet);
}

void
AlsaMidiOut::send (MidiEvent * e)
{
	snd_seq_ev_set_fixed (&packet);

	if (SysexEvent_type->is_derived (e))
	{
		snd_seq_ev_set_sysex (&packet, e->length, e->data);
	}
	else switch (e->status())
	{
		case 0x80:
			packet.type = SND_SEQ_EVENT_NOTEOFF;
			packet.data.note.channel = e->channel();
			packet.data.note.note = e->data[1];
			packet.data.note.velocity = e->data[2];
			packet.data.note.off_velocity = e->data[2];
			break;

		case 0x90:
			packet.type = SND_SEQ_EVENT_NOTEON;
			packet.data.note.channel = e->channel();
			packet.data.note.note = e->data[1];
			packet.data.note.velocity = e->data[2];
			break;
	
		case 0xA0:
			packet.type = SND_SEQ_EVENT_KEYPRESS;
			packet.data.note.channel = e->channel();
			packet.data.note.note = e->data[1];
			packet.data.note.velocity = e->data[2];
			break;
		
		case 0xB0:
			packet.type = SND_SEQ_EVENT_CONTROLLER;
			packet.data.control.channel = e->channel();
			packet.data.control.param = e->data[1];
			packet.data.control.value = e->data[2];
			break;

		case 0xC0:
			packet.type = SND_SEQ_EVENT_PGMCHANGE;
			packet.data.control.channel = e->channel();
			packet.data.control.value = e->data[1];
			break;

		case 0xD0:
			packet.type = SND_SEQ_EVENT_CHANPRESS;
			packet.data.control.channel = e->channel();
			packet.data.control.value = e->data[1];
			break;
		
		case 0xE0:
			packet.type = SND_SEQ_EVENT_PITCHBEND;
			packet.data.control.channel = e->channel();
			packet.data.control.value = 0x7FFF + (int) (1 + e->get_pitch());
			break;

		default:
			msg(), this, "dropping event ", e->data[0];
			break;
	}

	if (packet.type)
	{
		int r = snd_seq_event_output_direct (handle, &packet);
		if (r < 0)
			msg(), this, "sending event fails: ", strerror (-r);
	}
}

void
AlsaMidiOut::process_time (d_time dt, int clocks)
{
	MidiEvent * e;

	d_tick tick = graph->time2tick (graph->time + dt);

	while ((e = (MidiEvent *) ins[0]->pop_before (tick)))
	{
		send (e);
		::decref (e);
	}

	if (graph->is_loop_time (graph->time, graph->time + dt))
	{
		++loop_cycle;

		/* send all-notes-off. */
		packet.type = SND_SEQ_EVENT_CONTROLLER;
		packet.data.control.param = 123;
		packet.data.control.value = 0;

		for (int i = 0; i < 16; ++i)
		{
			packet.data.control.channel = i;
			int r = snd_seq_event_output_direct (handle, &packet);
			if (r < 0)
				msg(), this, "sending all-notes-off channel ", i, "fails: ", 
						strerror (-r);
		}
	}
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject * 
AlsaMidi_get_client (AlsaMidi * m)
{
	return PyInt (m->client);
}

static PyGetSetDef 
AlsaMidi_getsets [] = 
{
	{"client", (getter) AlsaMidi_get_client, (setter) 0,
	 "the ALSA sequencer client ID, major part (minor is 0)."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
AlsaMidi_subscribe (AlsaMidi * p, PyObject * args)
Py_TRY
{
	int major, minor;

	if (!PyArg_ParseTuple (args, "ii", &major, &minor))
		return 0;
	
	p->subscribe (major, minor, true);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AlsaMidi_unsubscribe (AlsaMidi * p, PyObject * args)
Py_TRY
{
	int major, minor;

	if (!PyArg_ParseTuple (args, "ii", &major, &minor))
		return 0;
	
	p->subscribe (major, minor, false);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
AlsaMidi_methods [] = 
{
	{"subscribe", (PyCFunction) AlsaMidi_subscribe, METH_VARARGS,
	 "subscribe (major, minor) -- connect to ALSA sequencer "
	 "client 'major:minor'."},
	{"unsubscribe", (PyCFunction) AlsaMidi_unsubscribe, METH_VARARGS,
	 "unsubscribe (major, minor) -- disconnect from ALSA sequencer "
	 "client 'major:minor'."},
	{ }
};

void
Py<AlsaMidi>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = AlsaMidi_getsets;
	t->tp_methods = AlsaMidi_methods;
}

char * Py<AlsaMidi>::__doc__ = 
"AlsaMidi() -- abstract base type for ALSA sequencer MIDI ports.";

void
Py<AlsaMidi>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* /////////////////////////////////////////////////////////////////////// */

void
Py<AlsaMidiIn>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = AlsaMidi_type;
}

char * Py<AlsaMidiIn>::__doc__ = 
"AlsaMidiIn ('name') -- an ALSA sequencer client that streams in MIDI. "
"It cannot receive system-exclusive data (Sysex).";

void
Py<AlsaMidiIn>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"name", 0};
	char * name = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s", keys, &name))
		return;
		
	init (name);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
AlsaMidiOut_send (AlsaMidiOut * p, MidiEvent * e)
Py_TRY
{
	if (!MidiEvent_type->is_derived (e))
		throw TypeError() << "need a MidiEvent, not " << get_type_name (e);

	p->send (e);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
AlsaMidiOut_methods [] = 
{
	{"send", (PyCFunction) AlsaMidiOut_send, METH_O,
	 "send (event) -- send a MIDI event right now."},
	{ }
};

void
Py<AlsaMidiOut>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = AlsaMidi_type;
	t->tp_methods = AlsaMidiOut_methods;
}

char * Py<AlsaMidiOut>::__doc__ = 
"AlsaMidiOut ('name') -- an ALSA sequencer client that streams out MIDI. "
"It cannot send system-exclusive (Sysex) events. Neither does it send MIDI "
"clock.";

void
Py<AlsaMidiOut>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"name", 0};
	char * name = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s", keys, &name))
		return;
		
	init (name);
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_AlsaMidi_types (PyObject * dict)
{
	AlsaMidi_type = new PyType<AlsaMidi> ("AlsaMidi", dict);
	AlsaMidiIn_type = new PyType<AlsaMidiIn> ("AlsaMidiIn", dict);
	AlsaMidiOut_type = new PyType<AlsaMidiOut> ("AlsaMidiOut", dict);
}
