/* (c) 2001-4 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "AlsaAudio.h"
#include "Graph.h"

static void
alsa_error_msg (
	const char * file, int line, const char * function, 
	int err, const char * fmt, ...)
{
	char * s = 0;
	va_list args;
	
	va_start (args, fmt);
	vasprintf (&s, fmt, args);
	va_end (args);
	
	msg(), file, ":", line, "(", function, "): ", s, ", err = ", strerror (err);

	free (s);
}

void set_alsa_error_handler()
{
	static int set = 0;

	if (++set == 1)
		snd_lib_error_set_handler (alsa_error_msg);
}

AlsaStream::AlsaStream (Type * t)
:	Plugin (t, Plugin::Audio)
{
	set_alsa_error_handler();

	frames = 0;
	handle = 0;

	gain = 0;

	relaxed = 1;

	int error = 
		snd_pcm_hw_params_malloc (&hw_params) |
		snd_pcm_sw_params_malloc (&sw_params) |
		snd_pcm_status_malloc (&status);
	
	if (error)
		throw MemoryError() << "cannot create an AlsaStream.";
}

AlsaStream::~AlsaStream()
{
	if (handle)
		snd_pcm_close (handle);
	
	snd_pcm_hw_params_free (hw_params);
	snd_pcm_sw_params_free (sw_params);

	snd_pcm_status_free (status);

	if (gain) free (gain);
}

void
AlsaStream::open (char * _name)
{
	int error;
	int mode = SND_PCM_NONBLOCK;

	if (get_type (this) == AlsaIn_type)
		error = snd_pcm_open (&handle, _name, SND_PCM_STREAM_CAPTURE, mode); 
	else
		error = snd_pcm_open (&handle, _name, SND_PCM_STREAM_PLAYBACK, mode); 
	
	if (error)
		throw IOError (-error) << "opening " << _name;

	set_object_name ("%s-%s", _name, 
			get_type (this) == AlsaIn_type ? "in" : "out");
}

void
AlsaStream::link (AlsaStream * other)
{
	if (other)
	{
		int error = snd_pcm_link (handle, other->handle);

		if (error && error != -EALREADY)
			throw IOError (-error) << "in AlsaStream::link() to " << other;
	}
	else
	{
		int error = snd_pcm_unlink (handle);

		if (error && error != -EALREADY)
			msg(), strerror (-error), "in AlsaStream::link (0).";
		/*
			throw IOError (-error) << "in AlsaStream::link (0).";
		 */
	}
}

void
AlsaStream::prepare()
{
	int error = snd_pcm_prepare (handle);
	if (error)
		throw IOError (-error) << "preparing " << get_object_name();
}

void
AlsaStream::start()
{
	int error = snd_pcm_start (handle);
	if (error)
		throw IOError (-error) << "starting " << get_object_name();

	msg(), get_name (this), " started.";
}

void
AlsaStream::stop()
{
	int error = snd_pcm_drop (handle);
	if (error)
		throw IOError (-error) << "stopping " << get_object_name();
}

bool
AlsaStream::update_status()
{
	int error = snd_pcm_status (handle, status);
	if (error)
		throw IOError (-error) << "getting status on " << get_object_name();

	int state = snd_pcm_status_get_state (status);

	available = (int) snd_pcm_status_get_avail (status);
	delay = snd_pcm_status_get_delay (status);

	if (state != SND_PCM_STATE_XRUN && delay < 2 * (int) frames_per_cycle)
		return true;

	msg(), 
			abs (delay), "samples ", 
			(char *) (get_type (this) == AlsaIn_type ? "over" : "under"),
			"run on ", get_object_name(), " at graph frame ", graph->frame;

	return false;
}

void
AlsaStream::dump_status()
{
	msg(), "Status of ", get_name (this), " at frame ", graph->frame, ":";
	msg(), "state: ", snd_pcm_state_name (snd_pcm_status_get_state (status));
	msg(), "delay: ", delay;
	msg(), "avail: ", available;
	msg(), "plugin state: ", state;
}

int
AlsaStream::begin_cycle (int _frames)
{
	frames = _frames;

	areas = 0;

	int error = snd_pcm_mmap_begin (
			handle, &areas, 
			(snd_pcm_uframes_t *) &offset, 
			(snd_pcm_uframes_t *) &frames
		);

	if (error == 0)
		return frames;
		
	msg(), "mmap_begin failed for ", 
			get_object_name(), " because ", snd_strerror (error);

	return 0;
}

void
AlsaStream::end_cycle (int _frames)
{
	int error = snd_pcm_mmap_commit (
			handle, offset, (snd_pcm_uframes_t) _frames);

	/* remember that the cycle has ended. */
	frames = 0;
	areas = 0;

	/* the ALSA doc says it returns 0 on success, but >= 0 is more like it. */
	if (error < 0)
		throw IOError (-error) << "mmap_commit'ing on " << get_name (this); 
}

/* a note on the memory-mapped ringbuffer access in AlsaStreams:
 *
 * if the device ring buffer is not open, it is probably because we are
 * currently in xrun state, and doing 'virtual' processing for the time
 * lost in the xrun.
 * if the ring buffer is open, we should be the ones dictating _frames, 
 * the # of frames to process this cycle; so if these don't match there's 
 * something gone awry.
 */
void
AlsaIn::process_audio (int _frames)
{
	/* the timestamp to put on the output is two cycles earlier. */
	d_frame stamp = graph->frame - 2 * graph->frames_per_cycle;

	/* the timestamp is incorrect during two cycles after a loop jump: */
	if (graph->loop.cycles && stamp < graph->loop.frame[0])
		stamp += graph->loop.frame[1] - graph->loop.frame[0];

	/* prepare an array of audio events */
	CyclicAudio * e [channels];
	
	for (int i = 0; i < channels; i++)
	{
		/* not connected, skip this channel. */
		if (!outs[i]->connected)
			e[i] = 0;
		else
		{
			e[i] = (CyclicAudio *) outs[i]->cache->pop();
			e[i]->frame = stamp;
		}
	}

	int written = 0;
	
	while (_frames)
	{
		int n = begin_cycle (_frames);

		if (n <= 0)
			break;

		for (int i = 0; i < channels; ++i)
			if (e[i])
				stream_out (i, e[i]->data + written);

		end_cycle (n);
	
		_frames -= n;
		written += n;
	}

	for (int i = 0; i < channels; ++i)
		if (e[i])
			outs[i]->push (e[i]);
}

void
AlsaOut::process_audio (int _frames)
{
	/* prepare an array of audio events */
	CyclicAudio * e [channels];
	
	for (int i = 0; i < channels; i++)
		e[i] = ins[i]->pop_mix();

	int read = 0;
	
	while (_frames)
	{
		int n = begin_cycle (_frames);
		
		if (n <= 0)
			break;

		for (int i = 0; i < channels; ++i)
		{
			if (e[i])
				stream_in (i, e[i]->data + read);
			else
				silence (i);
		}

		end_cycle (n);
	
		_frames -= n;
		read += n;
	}

	for (int i = 0; i < channels; ++i)
		if (e[i])
			e[i]->return_to_cache();
}

/* silence on all channels. */
void
AlsaStream::complete_silence (uint frames)
{
	for (int i = 0; i < channels; i++)
		silence (i);
}

/* silence on one channel. */
void
AlsaStream::silence (int channel)
{
	if (!frames) 
		throw blewit() << "ringbuffer not open.";

	const snd_pcm_channel_area_t & area = areas[channel];
	
	int8 * ptr = (int8 *) area.addr + ((area.first + area.step * offset) >> 3);

	int step = area.step >> 3;

	if (bytes_per_sample == 4)
	{
		for (int i = 0; i < frames; ++i, ptr += step)
			*((int32 *) ptr) = 0;
	}
	else if (bytes_per_sample == 3)
	{
		/* untested. */
		for (int i = 0; i < frames; ++i, ptr += step)
		{
			ptr[0] = 0;
			ptr[1] = 0;
			ptr[2] = 0;
		}
	}
	else if (bytes_per_sample == 2)
	{
		for (int i = 0; i < frames; ++i, ptr += step)
			*((int16 *) ptr) = 0;
	}
	else if (bytes_per_sample == 1)
	{
		for (int i = 0; i < frames; ++i, ptr += step)
			*ptr = 0;
	}
}

#define ftoi32(f,i) \
	__asm__ __volatile__ \
	("fistpl %0" : "=m" (i) : "t" (f) : "st");

void
AlsaStream::stream_in (int channel, d_sample * data)
{
	if (!frames) 
		throw blewit() << "ringbuffer not open.";

	d_sample g = gain[channel];

	const snd_pcm_channel_area_t & area = areas[channel];
	
	int8 * ptr = (int8 *) area.addr + ((area.first + area.step * offset) >> 3);

	int step = area.step >> 3;

	if (bytes_per_sample == 4)
	{
		for (int i = 0; i < frames; ++i, ptr += step)
			ftoi32 ((d_sample) 0x7FFFFFFF * g * data[i], *ptr);
	}
	else if (bytes_per_sample == 3)
	{
		/* untested. */
		for (int i = 0; i < frames; ++i, ptr += step)
		{
			int32 t = (int32) lrintf ((d_sample) 0x7FFFFF * g * data[i]);

			#if __BYTE_ORDER == __LITTLE_ENDIAN
				ptr[2] = (int8) (t & 0xFF);
				t >>= 8;
				ptr[1] = (int8) (t & 0xFF);
				t >>= 8;
				ptr[0] = (int8) (t);
			#else
				ptr[0] = (int8) (t & 0xFF);
				t >>= 8;
				ptr[1] = (int8) (t & 0xFF);
				t >>= 8;
				ptr[2] = (int8) (t);
			#endif
		}
	}
	else if (bytes_per_sample == 2)
	{
		for (int i = 0; i < frames; ++i, ptr += step)
			*((int16 *) ptr) = (int16) lrintf ((d_sample) 0x7FFF * g * data[i]);
	}
	else if (bytes_per_sample == 1)
	{
		for (int i = 0; i < frames; ++i, ptr += step)
			*ptr = (int8) lrintf ((d_sample) 0x7F * g * data[i]);
	}
}

void
AlsaStream::stream_out (int channel, d_sample * data)
{
	if (!frames) 
		throw blewit() << "ringbuffer not open.";

	const snd_pcm_channel_area_t & area = areas[channel];
	
	int8 * ptr = (int8 *) area.addr + ((area.first + area.step * offset) >> 3);

	uint step = area.step >> 3;

	if (bytes_per_sample == 4)
	{
		for (int i = 0; i < frames; ++i, ptr += step)
			data[i] = (d_sample) *((int32 *) ptr) / (d_sample) 0x7FFFFFFF;
	}
	else if (bytes_per_sample == 3)
	{
		/* untested. */
		for (int i = 0; i < frames; ++i, ptr += step)
		{
			/* unfortunately we can't simply read a 32 bit value and mask it
			 * because that can cause SIGSEGV near the end of the buffer. */
			#if __BYTE_ORDER == __LITTLE_ENDIAN
				int32 t = ptr[2];
				t <<= 8;
				t |= ptr[1];
				t <<= 8;
				t |= ptr[0];
			#else
				int32 t = ptr[0];
				t <<= 8;
				t |= ptr[1];
				t <<= 8;
				t |= ptr[2];
			#endif
			
			data[i] = (d_sample) t / (d_sample) 0x7FFFFF;
		}
	}
	else if (bytes_per_sample == 2)
	{
		for (int i = 0; i < frames; ++i, ptr += step)
			data[i] = (d_sample) *((int16 *) ptr) / (d_sample) 0x7FFF;
	}
	else if (bytes_per_sample == 1)
	{
		for (int i = 0; i < frames; ++i, ptr += step)
			data[i] = (d_sample) *ptr / (d_sample) 0x7F;
	}
}

/* after configuring the stream, this creates the ports that the stream
 * presents as a Plugin. */
void
AlsaStream::setup_ports()
{
	/* we can return if we have set them up already; we trust the number of
	 * channels will stay the same across configure() calls -- this actually
	 * may not be true if I remember right about the RME H-DSP which halves
	 * the number of channels when doubling the sample rate. FIXME? */
	if (ins.length)
		return;
		
	if (gain)
		return;
	
	gain = (d_sample *) malloc (sizeof (d_sample) * channels);

	for (int i = 0; i < channels; i++)
	{
		/* unity */
		gain[i] = 1;

		NamedObject * port;

		if (get_type (this) == AlsaIn_type)
			port = add_out (new Out (this, CyclicAudio_type));
		else
			port = add_in (new In (this, CyclicAudio_type, IN_SLOTS));
		
		port->set_object_name ("%d", i);
	}
}

/* the biiiiiiiiiig configure. 'we brake for nobody'. */

#define CONFIGURE(what,ifnot) \
	{ \
		int error = what; \
		if (error) \
			throw IOError() << ifnot << " ( "__FILE__":" << __LINE__ <<")"; \
	}

void
AlsaStream::configure (uint _sample_rate, uint _frames_per_cycle, uint periods)
{
	if (configured)
		if (sample_rate == _sample_rate)
			if (frames_per_cycle == _frames_per_cycle)
				return;

	configured = false;

	sample_rate = _sample_rate;
	frames_per_cycle = _frames_per_cycle;
	
	CONFIGURE (snd_pcm_hw_params_any (handle, hw_params), 
		"no 'any' configuration to start with.");

	CONFIGURE (snd_pcm_hw_params_set_periods_integer (handle, hw_params),
		"no integer period size.");
	
	int error = snd_pcm_hw_params_set_access 
			(handle, hw_params, SND_PCM_ACCESS_MMAP_NONINTERLEAVED);
			
	if (error == 0)
		interleaved = false;
	else
	{
		CONFIGURE (snd_pcm_hw_params_set_access 
				(handle, hw_params, SND_PCM_ACCESS_MMAP_INTERLEAVED),
				"memory-mapped access method refused.");
		interleaved = true;
	}

	/* determine sample format, native endianness. */
	snd_pcm_format_t formats [4] = {
		SND_PCM_FORMAT_S32,
		SND_PCM_FORMAT_S24,
		SND_PCM_FORMAT_S16,
		SND_PCM_FORMAT_S8
	};

	bytes_per_sample = ~0;
	for (int i = 0; i < 4; i++)
	{
		error = snd_pcm_hw_params_set_format (handle, hw_params, formats[i]);
		if (error == 0)
		{
			bytes_per_sample = 4 - i;
			break;
		}
	}
			
	if (bytes_per_sample > 4)
		throw blewit() << "none of (32, 24, 16, 8) bit sample sizes available.";

	CONFIGURE (snd_pcm_hw_params_set_rate (handle, hw_params, sample_rate, 0),
		"sample rate " << (int) sample_rate << "is unsupported.");

	/* f..k the ants. */
#if 1
	error = snd_pcm_hw_params_get_channels_max (hw_params, (unsigned int *) &channels);
	if (error)
		throw OSError (abs (error)) << "retrieving channel count from " << this;
#else
	channels = (int) snd_pcm_hw_params_get_channels_max (hw_params);
#endif
	
	/* some software devices can have millions of channels. nonsense to
	 * continue regarding how we handle things. */
	if (channels > 256)
	{
		msg(), this, "offers ", channels, "channels. using just 2.";
		channels = 2;
	}
	
	error = snd_pcm_hw_params_set_channels (handle, hw_params, channels);
	if (error)
		throw blewit() << channels << " channels are not available.";

	CONFIGURE (snd_pcm_hw_params_set_period_size 
		(handle, hw_params, (uint) frames_per_cycle, 0),
		"hardware refuses " << (int) frames_per_cycle << " frames per cycle.");

	CONFIGURE (snd_pcm_hw_params_set_periods 
		(handle, hw_params, periods, 0),
		"hardware refuses " << (int) periods << " cycles per buffer.");
/*
	CONFIGURE (snd_pcm_hw_params_set_buffer_size 
		(handle, hw_params, periods * frames_per_cycle),
		"hardware refuses a buffer size of " 
		<< (int) (periods * frames_per_cycle) << ".");
*/
	CONFIGURE (snd_pcm_hw_params (handle, hw_params),
		"pindown of hardware setup failed.");

	/* sw params */
	snd_pcm_sw_params_current (handle, sw_params);
	
#if 0
	CONFIGURE (snd_pcm_sw_params_set_start_threshold (handle, sw_params, ~0u),
		"cannot turn off start threshold.");
#else
	CONFIGURE (snd_pcm_sw_params_set_start_threshold (handle, sw_params, 
				0u),
			"setting start threshold fails.");
#endif

	/* jaroslav says:
		 As you can see, the stop_threshold is set to UMAX_INT. In case when
		 stop_threshold is above ring buffer size, then it means for drivers: I
		 don't care about xruns, don't report them. The stop threshold must be
		 equal to ring buffer size to get xruns when no more frames are in ring
		 buffer.
	 */
	/*
		we have a 'relaxed' setting now.
	*/
	CONFIGURE (snd_pcm_sw_params_set_stop_threshold 
			(handle, sw_params, relaxed ? ~0 : periods * frames_per_cycle),
			"cannot set stop threshold.");

	CONFIGURE (snd_pcm_sw_params_set_silence_threshold 
			(handle, sw_params, 0),
			"cannot set silence threshold.");

	CONFIGURE (snd_pcm_sw_params_set_silence_size (handle, sw_params, 0),
		"cannot set 0 silence size.");

	uint avail_min = frames_per_cycle;
	CONFIGURE (snd_pcm_sw_params_set_avail_min (handle, sw_params, avail_min),
		"software setup for minimum available frames failed.");

	CONFIGURE (snd_pcm_sw_params (handle, sw_params),
		"software setup failed.");

	setup_ports();

	Plugin::set_frames_per_cycle (frames_per_cycle);

	configured = true;

	msg(), get_object_name(), ":", sample_rate, frames_per_cycle,
			(8 * bytes_per_sample), "bit ",
			(char *) (interleaved ? "interleaved" : "non-interleaved"),
			(char *) (relaxed ? ", relaxed" : "");
}
