/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "Graph.h"
#include "Cycler.h"
#include "PyType.h"

static Type * Decycler_type;
static Type * Recycler_type;

/* /////////////////////////////////////////////////////////////////////// */

void
DecycledAudio::reset()
{
	played = used = 0;
}

int
DecycledAudio::record (d_sample * src, int frames)
{
	frames = min (used + frames, size) - used;

	if (frames <= 0)
		return 0;
	
	memcpy (data + used, src, frames * sizeof (d_sample));
	used += frames;

	return frames;
}

int
DecycledAudio::play (d_sample * dest, int frames)
{
	frames = min (played + frames, used) - played;

	if (frames <= 0)
		return 0;
	
	memcpy (dest, data + played, frames * sizeof (d_sample));
	played += frames;

	return frames;
}

/* /////////////////////////////////////////////////////////////////////// */

Decycler::Decycler()
: Plugin (Decycler_type, Audio)
{
	current = 0;
	collect = 0;
}

Decycler::~Decycler()
{
	::decref (current, true);
}

void
Decycler::init (int _collect)
{
	/* arbitrary limit. */
	if (_collect <= 2)
		throw ValueError() << "'collect' must be > 2";

	collect = _collect;
	
	add_in (new In (this, CyclicAudio_type, IN_SLOTS, "in"));
	add_out (new Out (this, CyclicAudio_type, "out"));
	add_out (new Out (this, DecycledAudio_type, "decycled"));

	/* size all DecycledAudio in the output cache. */
	EventCache * cache = outs[1]->cache;

	int j = cache->have();
	while (j--)
	{
		DecycledAudio * s = (DecycledAudio *) cache->pop_object();
		s->set_size (collect);
		cache->push_object (s);
	}
}

void 
Decycler::set_state (Graph * g, int s)
{
	/* will shortly blow up in process_audio, so a warning is sufficient. */
	if (g && collect < g->frames_per_cycle)
		msg(), this, " collects less than ", g->frames_per_cycle;

	/* prepare. */
	if (s <= Plugin::Prepared && current)
	{
		::decref (current);
		current = 0;
	}

	Plugin::set_state (g, s);
}

void 
Decycler::process_audio (int frames)
{
	if (state < Activated)
		return;

	if (!outs[0]->connected && !outs[1]->connected)
		return;

	CyclicAudio * e = ins[0]->pop_mix();

	/* it's better to silence to the outgoing stream. */
	if (e == 0)
		e = graph->silence;

	int written = 0;

	while (written < frames)
	{
		if (current == 0)
		{
			/* this must not throw because the reference to 'e' would never be
			 * released. should actually issue a warning if the cache is empty.
			 */
			if (outs[1]->cache->have() < 1)
				break;

			current = (DecycledAudio *) outs[1]->cache->pop();
			
			current->reset();
			current->frame = graph->frame + written;
		}

		int n = current->record (e->data + written, frames - written);

		written += n;
		
		if (n)
			continue;

		/* is full, push and advance. */
		outs[1]->push (current);
		current = 0;
	}

	if (e != graph->silence)
		outs[0]->push (e);
}

/* /////////////////////////////////////////////////////////////////////// */

Recycler::Recycler()
: Plugin (Recycler_type, Audio)
{
	current = 0;
	latency = 0;

	add_in (new In (this, DecycledAudio_type, 1, "in"));
	add_out (new Out (this, CyclicAudio_type, "out"));
}

Recycler::~Recycler()
{
	::decref (current, true);
}

void 
Recycler::set_state (Graph * g, int s)
{
	/* prepare. */
	if (s <= Plugin::Prepared && current)
	{
		::decref (current);
		current = 0;
	}

	Plugin::set_state (g, s);
}

void 
Recycler::process_audio (int frames)
{
	/* prepare our output event. */
	CyclicAudio * e = (CyclicAudio *) outs[0]->cache->pop();

	/* the timestamp could be calculated from the inbound stream's 'frame' 
	 * and the dynamically updated latency.
	 * 
	 * seems to be more of a philosophical problem; so until somebody really
	 * cares we simply say 'the signal is created now'.
	 */
	e->frame = graph->frame;

	int written = 0;

	while (written < frames)
	{
		if (current == 0)
		{
			current = (DecycledAudio *) ins[0]->pop();

			if (current == 0)
			{
				/* silence the rest of this event. */
				memset (e->data + written, 0, (frames - written) * sizeof (d_sample));
				break;
			}

			if (current->frame < graph->frame + written)
				latency = graph->frame + written - current->frame;
		}

		int n = current->play (e->data + written, frames - written);

		written += n;
		
		if (n)
			continue;

		/* 'current' is through, advance to next. */

		/* it would be nice to return 'current' to its cache right away but
		 * that is far too risky for various reasons (origin could be a 
		 * ScriptPlugin out cache so it's not safe to write to from the audio
		 * thread; python code could hold a reference to current etc).
		 */
		::decref (current);
		current = 0;
	}

	outs[0]->push (e);
}

/* /////////////////////////////////////////////////////////////////////// */

/* Py<*> */

/* DecycledAudio sequence methods */

static int 
DecycledAudio_sq_length (DecycledAudio * e)
{
	return e->used;
}

static PyObject *
DecycledAudio_sq_item (DecycledAudio * e, int i) 
Py_TRY
{
	if (i < 0 || i >= e->size)
		throw IndexError() << "no sample " << i << " in " << e;
	return PyFloat (e->data [i]); 
} 
Py_CATCH (0);

static int
DecycledAudio_sq_ass_item (DecycledAudio * e, int i, PyObject * value) 
Py_TRY
{
	if (i < 0 || i >= e->size)
		throw IndexError() << "no sample " << i << " in " << e;
	
	if (!PyArg_Parse (value, "f", e->data + i))
		return -1;

	return 0;
} 
Py_CATCH (-1);

static PySequenceMethods 
DecycledAudio_as_sequence = 
{
	(inquiry) DecycledAudio_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) DecycledAudio_sq_item, /* sq_item */
	(intintargfunc) 0, /* sq_slice */
	(intobjargproc) DecycledAudio_sq_ass_item, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

static int
DecycledAudio_getreadbuffer (DecycledAudio * e, int i, void ** data)
Py_TRY
{
	if (i != 0)
		throw IndexError() << " no such buffer segment.";
	
	if (data)
		*data = e->data;

	return e->used * sizeof (d_sample);
}
Py_CATCH (-1);

static int
DecycledAudio_getsegcount (DecycledAudio * e, int * length)
{
	if (length)
		*length = e->used * sizeof (d_sample);

	return 1;
}

static PyBufferProcs 
DecycledAudio_as_buffer =
{
	(getreadbufferproc) DecycledAudio_getreadbuffer,
	(getwritebufferproc) DecycledAudio_getreadbuffer, /* sic. */
	(getsegcountproc) DecycledAudio_getsegcount,
	(getcharbufferproc) DecycledAudio_getreadbuffer /* sic. */
};

char * Py<DecycledAudio>::__doc__ = 
"DecycledAudio() -- sequence of sampled audio signal as produced by a "
"Decycler plugin.";

Type * DecycledAudio_type = 0;

void
Py<DecycledAudio>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<DecycledAudio>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = AudioSignal_type;
	t->tp_as_sequence = &DecycledAudio_as_sequence;
	t->tp_as_buffer = &DecycledAudio_as_buffer;
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
Decycler_get_collect (Decycler * d)
{
	return PyInt (d->collect);
}

static PyGetSetDef 
Decycler_getsets [] =
{
	{"collect", (getter) Decycler_get_collect, (setter) 0, 
	 "number of audio samples to collect into one DecycledAudio."
	},
	{ }
};

void
Py<Decycler>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = Decycler_getsets;
}

char * Py<Decycler>::__doc__ = 
"Decycler (collect) -- collects realtime cyclic audio signal into "
"DecycledAudio. "
"It can be used to 'decouple' an audio signal from cyclic low-latency "
"audio processing; the resulting stream of DecycledAudio can then be processed "
"by a ScriptPlugin.\n"
"A Recycler plugin can be used to feed the DecycledAudio stream "
"back into the audio processing cycle.\n"
"The Decycler produces a stream of silent "
"DecycledAudio if its input is not "
"connected.";

void
Py<Decycler>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"collect", 0};

	int n;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "i", keys, &n))
		return;
	
	init (n);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
Recycler_get_latency (Recycler * r)
{
	return PyInt (r->latency);
}

static int
Recycler_set_latency (Recycler * r, PyObject * value)
{
	if (!PyArg_Parse (value, "i", &r->latency))
		return -1;
	return 0;
}

static PyGetSetDef 
Recycler_getsets [] =
{
	{"latency", (getter) Recycler_get_latency, (setter) Recycler_set_latency, 
	 "the difference, in audio frames, between the timing of the incoming "
	 "DecycledAudio stream and the main audio clock."},
	{ }
};

void
Py<Recycler>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = Recycler_getsets;
	t->tp_base = Plugin_type;
}

char * Py<Recycler>::__doc__ = 
"Recycler() -- feeds a stream of incoming Samples into cyclic audio signal. "
"See the Decycler documentation.";

void
Py<Recycler>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		return;
}

/* //////////////////////////////////////////////////////////////////// */

void
init_Cycler_types (PyObject * dict)
{
	DecycledAudio_type = new PyType<DecycledAudio> ("DecycledAudio", dict);

	Decycler_type = new PyType<Decycler> ("Decycler", dict);
	Recycler_type = new PyType<Recycler> ("Recycler", dict);
}
