#include "basics.h"
#include "Graph.h"
#include "PyType.h"

#define SIZE 65536

static Type * Delay_type;

class Delay
: public Plugin
{
	public:
		d_sample * line;
		int index;

	public:
		double t, wet, fb;

		Delay();
		~Delay();

		virtual void set_state (Graph * g, int state);
		virtual void process_audio (int frames);
};

Delay::Delay()
: Plugin (Delay_type, Audio)
{
	line = 0;
	t = .8;
	wet = .5;
	fb = .6;

	add_in (new In (this, CyclicAudio_type, IN_SLOTS, "in"));
	add_out (new Out (this, CyclicAudio_type, "out"));
}

Delay::~Delay()
{
	if (line) free (line);
}

void
Delay::set_state (Graph * g, int state)
{
	if (state == Plugin::Prepared)
	{
		if (line) line = (d_sample *) realloc (line, SIZE * sizeof (d_sample));
		else line = (d_sample *) malloc (SIZE * sizeof (d_sample));
		memset (line, 0, SIZE * sizeof (d_sample));
		index = 0;
	}
	
	Plugin::set_state (g, state);
}

void
Delay::process_audio (int frames)
{
	CyclicAudio * event = ins[0]->pop_mix();
	
	if (event == 0)
		return;

	d_sample * s = event->data;
	
	int ti = max (1, (int) (SIZE * t));

	double dry = 1. - wet;

	for (int j = 0; j < frames; ++j)
	{
		register d_sample & t = line [(index + j) % ti];
		register d_sample v = t;
		t = v * fb + s[j];
		s[j] = wet * v + dry * s[j];
	}

	index = (index + frames) % ti;

	outs[0]->push (event);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
Delay_get_t (Delay * d)
{
	return PyFloat (d->t);
}

static int
Delay_set_t (Delay * d, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &d->t))
		return -1;
	return 0;
}

static PyObject *
Delay_get_wet (Delay * d)
{
	return PyFloat (d->wet);
}

static int
Delay_set_wet (Delay * d, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &d->wet))
		return -1;
	return 0;
}

static PyObject *
Delay_get_fb (Delay * d)
{
	return PyFloat (d->fb);
}

static int
Delay_set_fb (Delay * d, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &d->fb))
		return -1;
	return 0;
}

static PyGetSetDef 
Delay_getsets [] =
{
	{"t", (getter) Delay_get_t, (setter) Delay_set_t, 
	 "delay time, from [0 .. 1]."},
	{"wet", (getter) Delay_get_wet, (setter) Delay_set_wet, 
	 "dry/wet ratio, from [0 .. 1]."},
	{"fb", (getter) Delay_get_fb, (setter) Delay_set_fb, 
	 "feedback amount, from [0 .. 1]."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<Delay>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = Delay_getsets;
}

char * Py<Delay>::__doc__ = 
"Delay ([t = .5, wet = .5]) -- Delay.";

void
Py<Delay>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"t", "wet", "fb", 0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|ddd", keys,
				&t, &wet, &fb))
		return;
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_Delay_types (PyObject * dict)
{
	Delay_type = new PyType<Delay> ("Delay", dict);
}
