/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "DiskStream.h"
#include "SoundFile.h"
#include "profiler.h"

#include "PyType.h"

/* larger segment sizes mean less but longer disk accesses. segment size
 * is measured in frames. 
 */
#define SEGMENT_SIZE	(1 * 65536) 

/* more segments mean less chance of one disk stream 'suffocating' another. 
 * must be less than MAX_SEGMENTS as per DiskStream.h 
 */
#define SEGMENTS			3	

/* toggle verbose messages about segment states. */
#define DBG_SEGMENTS	0

/* A SoundFileSegment holds 1+ channels of audio signal data. They are
 * used as tiles dividing SoundFiles into larger chunks suitable for disk IO. 
 */
class SoundFileSegment
{
	public:
		int channels;
		
		int accessed;				/* 1 = has been accessed, read mode only */
		atomic ready;				/* 1 = full, 0 = empty or partial */
		d_frame frame;			/* start frame (in SoundFile time) */
		
		int size;						/* in frames, not samples */
		int used;						/* last write_window stopped here */

		d_sample ** data;

		SoundFileSegment (int _size, int _channels)
			{
				channels = _channels;
				size = _size;
				ready = 0;
				frame = 0;
				used = 0;

				data = (d_sample **) calloc (channels, sizeof (d_sample *));
				if (data == 0)
					throw MemoryError() << "no SoundFileSegment::data.";

				for (int i = 0; i < channels; ++i)
				{
					data[i] = (d_sample *) malloc (size * sizeof (d_sample));
					if (data[i] == 0)
						throw MemoryError() << "no SoundFileSegment::data [" << i << "].";
					touch_memory (data[i], size * sizeof (d_sample));
				}
			}

		~SoundFileSegment()
			{
				if (data == 0)
					return;

				for (int i = 0; i < channels; i++)
					if (data[i])
						free (data[i]);
				free (data);
			}
			
		void 
		read (SoundFile * file, d_frame f)
			{
				frame = f;

				used = min (size, (int) (file->frames - frame));

				if (used <= 0)
					used = 0;
				else
				{
					ScopedLock lock (file->mutex);

					file->seek (frame);
					file->read (data, used);
				}

				/* silent past the end of the file. */
				int silent = (int) (size - used);

				if (silent)
				{
					for (int i = 0; i < channels; ++i)
					{
						d_sample * s = data[i] + used;
						memset (s, 0, sizeof (d_sample) * silent);
					}
				}

				accessed = 0;
			}

		void 
		write (SoundFile * file)
			{
				ScopedLock lock (file->mutex);

				file->seek (frame);
				file->write (data, used);
				used = 0;
			}

		/* intersect [frame, frame + size] with [f, f + n] to determine the
		 * data window for reading/writing.
		 */
		d_frame 
		calculate_window (d_frame f, int * n)
			{
				d_frame f0 = min (max (frame, f), frame + size);
				d_frame f1 = min (frame + size, f + *n);

				*n = (f1 <= f0) ? 0 : (int) (f1 - f0);

				return f0;
			}

		/* copy n frames into events, starting at f */
		int 
		read_window (CyclicAudio ** events, d_frame f, int n)
			{
				if (ready == 0) /* no data. */
					return 0;

				d_frame f0 = calculate_window (f, &n);

				if (n <= 0) /* outside this segment */
				{
					if (accessed)
						ready = 0;
					return 0;
				}

				accessed = 1;

				int offset = (int) (f0 - f); /* into event->data */
				
				for (int i = 0; i < channels; i++)
				{
					d_sample * src = data[i] + (int) (f0 - frame);
					d_sample * dest = events[i]->data + offset;

					for (int s = 0; s < n; s++)
						*dest++ = *src++;
				}

				/* hit the end of the road. */
				if (f0 + n >= frame + size)
					ready = 0;

				return n;
			}

		/* copy n frames from events, starting at f */
		int 
		write_window (CyclicAudio ** events, d_frame f, int n)
			{
				if (ready)
					return 0;

				d_frame f0 = calculate_window (f, &n);
				if (n <= 0)
					return 0;

				int offset = (int) (f0 - f); /* into event->data */
				
				for (int i = 0; i < channels; i++)
				{
					d_sample * dest = data[i] + (int) (f0 - frame);
					d_sample * src = events[i]->data + offset;

					for (int s = 0; s < n; s++)
						*dest++ = *src++;
				}

				/* we always write consecutive data, so just set the fill level. */
				used = (f0 - frame) + n;
				
				/* hit the end of the road. */
				if (used == size)
					ready = 1;

				return n;
			}	
};

/* //////////////////////////////////////////////////////////////////////// */

DiskStream::DiskStream()
: Plugin (DiskStream_type, Plugin::Disk | Plugin::Audio)
{
	monitor = 1;
	path = 0;
	file = 0;
	channels = 1;
	sample_rate = SAMPLE_RATE;

	to_play_on_loop = 1;
	to_play_on_seek = 1;

	requested = -1;
	short_warning = -1;

	offset = 0;
	next_frame = 0;

	for (int i = 0; i < SEGMENTS; i++)
		segments[i] = 0;
}

DiskStream::~DiskStream()
{
	/* if process_disk is executing right now, wait for it to finish. */
	mutex.lock();

	/* stow away leftovers */
	if (file) 
	{
		flush();
		::decref (file, true);
	}

	for (int i = 0; i < SEGMENTS; i++)
		if (segments[i])
			delete segments[i];

	mutex.unlock();
}

/* set SoundFile. */
void
DiskStream::set_file (SoundFile * _file)
{
	if (file) 
		throw AssertionError() << this << " already has a file.";

	::incref (file = _file);

	set_object_name ("%s", file->get_path());

	path = file->get_path();
	channels = file->channels;
	sample_rate = file->sample_rate;

	if (file->mode == O_RDONLY)
		mode = Read;
	else
		mode = Write;

	for (int i = 0; i < channels; i++)
	{
		if (mode == Write)
		{
			In * in = add_in (new In (this, CyclicAudio_type, IN_SLOTS));
			in->set_object_name ("rec %d", i);
		}

		Out * out = add_out (new Out (this, CyclicAudio_type));
		out->set_object_name ("play %d", i);
	}

	for (int i = 0; i < SEGMENTS; i++)
		segments[i] = new SoundFileSegment (SEGMENT_SIZE, channels);
}

/* write partially filled segments. */
void
DiskStream::flush()
{
	if (mode == Read || file == 0)
		return;
	
	for (int i = 0; i < SEGMENTS; i++)
	{
		SoundFileSegment * s = segments[i];

		if (s && (s->ready || s->used))
			s->write (file);
	}
}

/* set segments to unused and recalculate next_frame based on 'f'. */
void
DiskStream::reset_segments (d_frame f)
{
	if (mode == Read)
	{
		/* calculate the next segment frame. */
		next_frame = next_segment (f);

		for (int i = 0; i < SEGMENTS; i++)
			segments[i]->ready = 0;
	}
	else
	{
		/* in Write mode, segments need to have frame offsets assigned before
		 * processing so we know where to store the signal. 
		 */
		/* align on SEGMENT_SIZE */
		next_frame = f - (f % SEGMENT_SIZE);
		for (int i = 0; i < SEGMENTS; i++)
		{
			#if DBG_SEGMENTS
			msg(), "segment: ", i, "@ ", next_frame;
			#endif

			segments[i]->ready = 0;
		
			segments[i]->frame = next_frame;
			next_frame += SEGMENT_SIZE;
		}
	}
}

/* calculate the frame at which the next segment starts. */
d_frame
DiskStream::next_segment (d_frame frame)
{
	d_frame f = frame + SEGMENT_SIZE;
	f = f - (f % SEGMENT_SIZE) + offset;

	/* loop case */
	if (graph && graph->loop.use && f >= graph->loop.frame[1])
	{
		frame = graph->loop.frame[0] - offset;
		frame -= (frame % SEGMENT_SIZE);
	}
	else
	{
		frame += SEGMENT_SIZE;
	}

	return frame;
}

/* seek. */
void
DiskStream::seek()
{
	if (mode == Write && to_play_on_seek)
		mode = ReadWrite;
	else if (mode == Read)
		msg(), "reset to ", graph->frame;
		
	/* will probably need to do i/o now. */
	request_io();
}

/* notify a disk scheduler that we need disk access. */
void
DiskStream::request_io()
{
	if (graph == 0 || graph->disk_scheduler == 0)
		return;

	/* one notification should do. */
	if (requested.inc_is_zero())
		graph->disk_scheduler->process_disk (this);	
}

/* Plugin:: */
void
DiskStream::pre_insert (Graph * g)
{
	Plugin::pre_insert (g);

	reset_segments (g->frame);

	/* ready some data */
	if (mode == Read)
		process_disk_read (g->frame);
}

void
DiskStream::set_state (Graph * g, int s)
{
	/* switching up into activated state, preroll. */
	if (mode != ReadWrite && s == Activated && state < s)
	{
		reset_segments (g->frame);
		
		if (mode == Read)
			request_io();
	}

	Plugin::set_state (g, s);
}

/* two helpers for read mode that frame audio data with silence;
 * we need this for events that span around signal start and end.
 */
inline void
prepend_silence (CyclicAudio ** events, int channels, int frames)
{
	for (int i = 0; i < channels; i++)
	{
		d_sample * s = events[i]->data;
		
		for (int j = frames; j; j--)
			*s++ = (d_sample) 0;
	}
}

inline void
append_silence (CyclicAudio ** events, int channels, int frame, int frames)
{
	for (int i = 0; i < channels; i++)
	{
		d_sample * s = events[i]->data + frame;
		
		for (int j = frames; j; j--)
			*s++ = (d_sample) 0;
	}
}

/* DiskStream::process_audio() in a nutshell:
 *
 *  mode == Read:  move audio data from segments into events,
 *  mode == Write: move audio data from events into segments,
 *  check for segments marked ready and request disk processing if necessary.
 */
void
DiskStream::process_audio (int frames)
{
	if (mode == Read)
		process_read (frames);
	else if (mode == Write)
		process_write (frames);
	else if (mode == ReadWrite)
		;
}

void
DiskStream::process_read (int frames)
{
	if (state < Rolling)
		return;

	/* a write stream that has turned into a reader needs to drop the
	 * input signal. */
	for (int i = 0; i < ins; ++i)
	{
		In * in = ins[i];
		
		/* we may add other ports later, so safeguard against non-audio ins. */
		if (!CyclicAudio_type->is_derived (in->type))
			continue;
			
		CyclicAudio * e = ins[i]->pop_mix();
		
		if (e) e->return_to_cache();
	}

	/* translate current frame into file time and check for intersection. */
	d_frame f = graph->frame - offset;

	if (f + frames < 0 || f > file->frames)
		return;

	#if DBG_SEGMENTS
	for (int i = 0; i < SEGMENTS; ++i)
	{
		SoundFileSegment * s = segments[i];
		msg(), "segment ", i, s->ready, s->frame, "@ ", f;
	}
	#endif

	/* request io if necessary */
	if (!have_segment (f) || 
			(f + SEGMENT_SIZE < file->frames && !have_segment (f + SEGMENT_SIZE)))
		request_io();

	/* prepare events for all outputs. */
	CyclicAudio * events [channels];

	for (int i = 0; i < channels; i++)
	{
		events[i] = (CyclicAudio *) outs[i]->cache->pop();
		events[i]->frame = graph->frame;
	}

	/* playback can start or end in the middle of this cycle, compute the
	 * number of silent frames before/after the signal. */
	int prepend = (f < 0) ? (int) -f : 0;
	int append = (f + frames > file->frames) ? 
		(int) (f + frames - file->frames) : 0;

	/* copy signal from segments. */
	int written = 0;

	for (int i = 0; i < SEGMENTS; i++)
		written += segments[i]->read_window (events, f, (int) frames);

	/* check whether we could provide enough data. */
	if (written < frames - (prepend + append))
	{
		if (requested != -1 && (
				short_warning.inc_is_zero() || !(short_warning & 0x1F)))
			msg(), get_name (this), " is short on audio data.";

		for (int i = 0; i < channels; i++)
			events[i]->silence();
	}
	else 
	{
		/* otherwise, put silence around the signal as has been computed above. */
		if (prepend) 
			prepend_silence (events, channels, (int) -f);
		if (append)
			append_silence (events, channels, frames - append, append);
	}

	/* move signal to outputs. */
	for (int i = 0; i < channels; i++)
		outs[i]->push (events[i]);
}

void
DiskStream::process_write (int frames)
{
	/* default frame if no audio input is provided. */
	d_frame f = graph->frame - 2 * graph->frames_per_cycle;
	
	/* fetch signal input. */
	CyclicAudio * events [channels];

	for (int i = 0; i < channels; i++)
	{
		CyclicAudio * e = ins[i]->pop_mix();
		
		if (e == 0)
		{
			events[i] = graph->silence;
			continue;
		}

		events[i] = e;
		f = e->frame;
	}

	/* at the end of a loop, switch to read mode. */
	bool is_loop = graph->is_loop_frame (f, f + frames);
	
	f -= offset;

	if (f >= 0 && state >= Rolling)
	{
		/* copy signal into segments. */
		for (int i = 0; i < SEGMENTS; i++)
		{
			SoundFileSegment * s = segments[i];
			
			s->write_window (events, f, (int) frames);
			
			if (s->ready == 1)
				request_io();
		}

		if (is_loop && to_play_on_loop)
		{
			mode = ReadWrite;
			request_io();
		}
	}

	/* move signal to outputs. */
	if (monitor)
	{
		for (int i = 0; i < channels; i++)
			if (events[i] != graph->silence)
				outs[i]->push (events[i]);
	}
	else
	{
		for (int i = 0; i < channels; i++)
			if (events[i] != graph->silence)
				events[i]->return_to_cache();
	}
}

/* do some disk IO. */
void
DiskStream::process_disk (Plugin * p)
{
	if (file == 0)
		throw AssertionError() << this << "::process_disk without a file.";

	#if DBG_SEGMENTS
	msg(), this, "process_disk @ ", graph->frame, mode;
	#endif

	/* switching to Read mode. */
	if (mode == ReadWrite)
	{
		flush();
		
		mode = Read;

		if (graph == 0)
			return;

		reset_segments (graph->frame);
	}

	if (mode == Read)
		process_disk_read (graph->frame + graph->frames_per_cycle);
	else 
		process_disk_write();

	/* reset requested state. */
	requested = -1;
	short_warning = -1;
}

/* Read mode; returns whether we have a segment at frame f. */
bool
DiskStream::have_segment (d_frame f)
{
	f = f - (f % SEGMENT_SIZE);

	for (int i = 0; i < SEGMENTS; ++i)
	{
		SoundFileSegment * s = segments[i];
		
		if (s->ready && s->frame == f)
			return true;
	}

	return 0;
}

void
DiskStream::process_disk_read (d_frame f)
{
	f -= offset;
	next_frame = f - (f % SEGMENT_SIZE);

	#if DBG_SEGMENTS
	msg(), this, "process_disk_read @ ", f;
	#endif

	/* reevaluate segment validity */
	for (int i = 0; i < SEGMENTS; ++i)
	{
		SoundFileSegment * s = segments[i];
		
		#if DBG_SEGMENTS
		msg(), s, " ", s->ready, "@ ", s->frame;
		#endif
		
		if (s->ready)
		{
			/* valid and next frame, ok. */
			if (s->frame == next_frame)
			{
				next_frame = next_segment (next_frame);
				continue;
			}

			/* contains current frame, ok. */
			else if (f >= s->frame && f < s->frame + SEGMENT_SIZE)
				continue;
		}

		/* mark as invalid and advance to next segment */
		for (int j = 0; j < SEGMENTS && have_segment (next_frame); ++j)
			next_frame = next_segment (next_frame);

		#if DBG_SEGMENTS
		msg(), s, " marked for ", next_frame;
		#endif

		s->ready = 0;
		s->frame = next_frame;
		next_frame = next_segment (next_frame);
	}
	
	for (int i = 0; i < SEGMENTS; ++i)
	{
		SoundFileSegment * s = segments[i];
	
		if (s->ready || s->frame > file->frames)
			continue;

		#if DBG_SEGMENTS
		msg(), "read ", i, "@ ", s->frame;
		#endif

		s->read (file, s->frame);
		s->ready = 1;
	}
}

void
DiskStream::process_disk_write()
{
	for (int i = 0; i < SEGMENTS; i++)
	{
		SoundFileSegment * s = segments[i];

		if (!s->ready)
			continue;

		s->write (file);
		s->frame += SEGMENTS * SEGMENT_SIZE;
		s->ready = 0;	/* order is important. */
	}
}

/* //////////////////////////////////////////////////////////////////////// */

Type * DiskStream_type = 0;

char * Py<DiskStream>::__doc__ = 
"DiskStream (file) -- recording and playback plugin for sound files. Pass a "
"SoundFile derivative (currently WaveFile or MpegFile) to the constructor. "
"If the file has been opened for write access, the input signal will be "
"recorded into that file until the graph is stopped, a seek is performed "
"or the end of a loop is hit. Afterwards, the recording will be played back "
"and the input signal dumped.\n"
"This plugin depends on the presence of a DiskScheduler in the graph it "
"resides in to work as expected.";

static PyObject * 
DiskStream_get_file (DiskStream * p)
{
	return ::incref (p->file);
}

static PyObject * 
DiskStream_get_monitor (DiskStream * p)
{
	return PyInt (p->monitor);
}

static PyObject * 
DiskStream_get_mode (DiskStream * p)
{
	char * m = "r";

	if (p->mode == DiskStream::ReadWrite)
		m = "rw";
	else if (p->mode == DiskStream::Write)
		m = "w";

	return PyString (m);
}

static PyObject * 
DiskStream_get_to_play_on_loop (DiskStream * p)
{
	return PyInt (p->to_play_on_loop);
}

static PyObject * 
DiskStream_get_to_play_on_seek (DiskStream * p)
{
	return PyInt (p->to_play_on_seek);
}

static int
DiskStream_set_monitor (DiskStream * p, PyObject * m)
{
	if (!PyArg_Parse (m, "i", &p->monitor))
		return -1;
	return 0;
}

static int
DiskStream_set_to_play_on_loop (DiskStream * p, PyObject * m)
{
	if (!PyArg_Parse (m, "i", &p->to_play_on_loop))
		return -1;
	return 0;
}

static int
DiskStream_set_to_play_on_seek (DiskStream * p, PyObject * m)
{
	if (!PyArg_Parse (m, "i", &p->to_play_on_seek))
		return -1;
	return 0;
}

static PyGetSetDef 
DiskStream_getsets [] = 
{
	{"file", (getter) DiskStream_get_file, 0, "the sound file."},
	{"mode", (getter) DiskStream_get_mode, 0, "the mode (read/write)."},
	{"monitor", (getter) DiskStream_get_monitor, (setter) DiskStream_set_monitor, 
	 "whether to copy the recorded signal to the plugin's outputs."},
	{"to_play_on_loop", (getter) DiskStream_get_to_play_on_loop, 
	 (setter) DiskStream_set_to_play_on_loop, 
	 "whether to switch from record to play in looped time."},
	{"to_play_on_seek", (getter) DiskStream_get_to_play_on_seek, 
	 (setter) DiskStream_set_to_play_on_seek, 
	 "whether to switch from record to play when jumping in time."},
	{ },
};

void
Py<DiskStream>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = DiskStream_getsets;
}

void
Py<DiskStream>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"file", 0};
	SoundFile * file = 0;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!", keys, 
			SoundFile_type, &file))
		throw blewit();

	this->set_file (file);
}

/* //////////////////////////////////////////////////////////////////////// */

void
init_DiskStream_types (PyObject * dict)
{
	DiskStream_type = new PyType<DiskStream> ("DiskStream", dict);
}
