/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "Event.h"
#include "AudioSignal.h"

#include "PyType.h"

#include "sysex-ids.h"

/* getsets */
static PyObject * 
Event_get_tick (Event * e)
{
	return PyFloat (e->tick);
}

static int 
Event_set_tick (Event * e, PyObject * arg)
{
	if (!PyArg_Parse (arg, "d", &e->tick))
		return -1;
	return 0;
}

static PyObject * 
Event_get_loop (Event * e)
{
	return PyInt (e->loop);
}

static int 
Event_set_loop (Event * e, PyObject * arg)
{
	if (!PyArg_Parse (arg, "i", &e->loop))
		return -1;
	return 0;
}

static PyGetSetDef 
Event_getsets [] = 
{
	{"tick", (getter) Event_get_tick, (setter) Event_set_tick, 
	 "'tick' is a timestamp in musical time."},
	{"loop", (getter) Event_get_loop, (setter) Event_set_loop, 
	 "'loop' stores the loop cycle number, needed for accurate event timing "
	 "around the loop points."},

	{ }
};

char * Py<Event>::__doc__ = 
"Event (tick[, loop]) -- an Object with a musical timestamp.";

Type * Event_type = 0;

void
Py<Event>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"tick", "loop", 0};
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|di", keys, &tick, &loop))
		throw blewit();
}

void
Py<Event>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Object_type;
	t->tp_getset = Event_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

char * Py<FloatEvent>::__doc__ = 
"FloatEvent (value[, tick]) -- an Event that carries a floating point value.";

Type * FloatEvent_type = 0;

/* getsets */
static PyObject * 
FloatEvent_get_value (FloatEvent * self)
{
	return PyFloat (self->value);
}

static int 
FloatEvent_set_value (FloatEvent * self, PyObject * arg)
{
	if (!PyArg_Parse (arg, "d", &self->value))
		return -1;
	return 0;
}

static PyGetSetDef 
FloatEvent_getsets [] = 
{
	{"value", (getter) FloatEvent_get_value, (setter) FloatEvent_set_value, 
	 "."},
	{}
};

void
Py<FloatEvent>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"value", "tick", 0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|dd", keys, 
			&value, &tick))
		throw blewit();
}

void
Py<FloatEvent>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Event_type;
	t->tp_getset = FloatEvent_getsets;
} 

/* // TextEvent ///////////////////////////////////////////////////////// */

/* getsets */
static PyObject * 
TextEvent_get_text (TextEvent * self)
{
	return PyString (self->text);
}

static int 
TextEvent_set_text (TextEvent * self, PyObject * arg) 
Py_TRY
{
	char * s;
	
	if (!PyArg_Parse (arg, "s", &s))
		return -1;
	if (strlen (s) >= MARK_EVENT_TEXT_MAX)
		throw OverflowError() << "only" << MARK_EVENT_TEXT_MAX << "chars supported.";

	self->set_text (s);

	return 0;
} 
Py_CATCH (-1);

static PyGetSetDef 
TextEvent_getsets [] = 
{
	{"text", (getter) TextEvent_get_text, (setter) TextEvent_set_text, 
	 "the label."},
	{}
};

Type * TextEvent_type = 0;

char * Py<TextEvent>::__doc__ = 
"TextEvent ([text = '']) -- an Event that carries a text.";

void
Py<TextEvent>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"text", "tick", 0};
	PyObject * text = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|Od", keys, 
			&text, &tick))
		throw blewit();
	
	if (text)
		TextEvent_set_text (this, text);
}

void
Py<TextEvent>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Event_type;
	t->tp_getset = TextEvent_getsets;
} 

/* // TempoEvent ///////////////////////////////////////////////////////// */

/* getsets */
static PyObject * 
TempoEvent_get_bpm (TempoEvent * self)
{
	return PyFloat (self->bpm);
}

static PyObject * 
TempoEvent_get_sec_per_tick (TempoEvent * self)
{
	return PyFloat (self->sec_per_tick);
}

static int 
TempoEvent_set_bpm (TempoEvent * self, PyObject * arg) 
Py_TRY
{
	double bpm;
	
	if (!PyArg_Parse (arg, "d", &bpm))
		return -1;
	if (bpm < 1.0 || bpm > 360.)
		throw ValueError() << "bpm must be >= 1 and <= 360.";

	self->bpm = bpm;
	return 0;
} 
Py_CATCH (-1);

static PyGetSetDef 
TempoEvent_getsets [] = 
{
	{"bpm", (getter) TempoEvent_get_bpm, (setter) TempoEvent_set_bpm, 
	 "the beats (quarter notes) per minute."},
	{"sec_per_tick", (getter) TempoEvent_get_sec_per_tick, (setter) 0, 
	 "seconds per tick."},
	{}
};

Type * TempoEvent_type = 0;

char * Py<TempoEvent>::__doc__ = 
"TempoEvent ([bpm = 96.0, tick = 0]) -- "
"an Event that describes musical tempo in quarter beats per minute.";

void
Py<TempoEvent>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"bpm", "tick", 0};
	PyObject * bpm = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|Od", keys, 
			&bpm, &tick))
		throw blewit();
	
	if (bpm)
		TempoEvent_set_bpm (this, bpm);
}

void
Py<TempoEvent>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Event_type;
	t->tp_getset = TempoEvent_getsets;
} 

/* /////////////////////////////////////////////////////////////////////// */

/* gets */
static PyObject * 
TimeEvent_get_n (TimeEvent * self)
{
	return PyInt (self->n);
}

static PyObject * 
TimeEvent_get_d (TimeEvent * self)
{
	return PyInt (self->d);
}

static PyObject * 
TimeEvent_get_ticks (TimeEvent * self)
{
	return PyInt (self->ticks);
}

static PyObject * 
TimeEvent_get_bbt (TimeEvent * e)
{
	return Py_BuildValue ("iif", e->bbt.bar, e->bbt.beat, (double) e->bbt.tick);
}

/* sets */
static int 
TimeEvent_set_n (TimeEvent * self, PyObject * arg) 
Py_TRY
{
	int n;
	
	if (!PyArg_Parse (arg, "i", &n))
		return -1;

	if (n < 1 || n > 1024) 
		throw ValueError() << "n must be in range (1, 1024).";
	
	self->n = n;
	return 0;
} 
Py_CATCH (-1);

static int 
TimeEvent_set_d (TimeEvent * self, PyObject * arg) 
Py_TRY
{
	int d;
	
	if (!PyArg_Parse (arg, "i", &d))
		return -1;

	if (d < 1 || d > 1024) 
		throw ValueError() << "d must be in range (1, 1024).";
	
	self->d = d;
	return 0;
} 
Py_CATCH (-1);

static int 
TimeEvent_set_ticks (TimeEvent * self, PyObject * arg) 
Py_TRY
{
	int ticks;
	
	if (!PyArg_Parse (arg, "i", &ticks))
		return -1;

	if (ticks < 1 || ticks > 1024) 
		throw ValueError() << "ticks must be in range (1, 1024).";
	
	self->ticks = ticks;
	return 0;
} 
Py_CATCH (-1);

static PyGetSetDef 
TimeEvent_getsets [] = 
{
	{"n", (getter) TimeEvent_get_n, (setter) TimeEvent_set_n, 
	 "the numerator of time measure."},
	{"d", (getter) TimeEvent_get_d, (setter) TimeEvent_set_d, 
	 "the denominator of time measure."},
	{"ticks", (getter) TimeEvent_get_ticks, (setter) TimeEvent_set_ticks, 
	 "the number of 'tick' time measuring units per denominator of "
	 "the measure."},
	{"bbt", (getter) TimeEvent_get_bbt, (setter) 0, 
	 "bar/beat/tick."},
	 
	{ }
};

char * Py<TimeEvent>::__doc__ = 
"TimeEvent ([n = 4, d = 4, ticks = 24, tick = 0]) -- "
"an Event that describes how to measure musical time.";

Type * TimeEvent_type = 0;

void
Py<TimeEvent>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"n", "d", "ticks", "tick", 0};
	PyObject * n = 0, * d = 0, * ticks = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|OOOd", keys, 
			&n, &d, &ticks, &tick))
		throw blewit();
	
	if (n && TimeEvent_set_n (this, n)) throw blewit();
	if (d && TimeEvent_set_d (this, d)) throw blewit();
	if (ticks && TimeEvent_set_ticks (this, ticks)) throw blewit();
}

void
Py<TimeEvent>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Event_type;
	t->tp_getset = TimeEvent_getsets;
}

/* // Py<MidiEvent> ////////////////////////////////////////////////////// */

static PyObject * 
MidiEvent_get_data (MidiEvent * e)
{
	return PyString (e->data, e->length);
}

static int
MidiEvent_set_data (MidiEvent * e, PyObject * value) 
Py_TRY
{
	byte * bytes;
	int length;

	if (!PyArg_Parse (value, "s#", &bytes, &length))
		return -1;

	e->set_data (bytes, length);
	return 0;
} 
Py_CATCH (-1);

/* status */

static PyObject * 
MidiEvent_get_status (MidiEvent * e)
{
	if (e->data[0] < 0x80 || e->data[0] >= 0xF0)
		return PyInt (e->data[0]);
	
	return PyInt (e->data[0] & 0xF0);
}

static int
MidiEvent_set_status (MidiEvent * e, PyObject * value) 
Py_TRY
{
	/* protect derived classes from being compromised. */
	if (get_type (e) != MidiEvent_type)
		throw TypeError() << "status is immutable on this event.";

	long s = PyInt_AsLong (value);
	
	if (s < 0) return -1;
	if (s > 255) 
		throw ValueError() << "status must be <= 0xFF.";
	
	e->data[0] = (byte) s;
	
	/* adjust the # of bytes */
	s &= 0xF0;
	
	if (s < 0x80) /* unknown */
		;
	else if (s < 0xC0 || s == 0xE0) /* all and pitch */
		e->length = 3;
	else if (s < 0xE0) /* program and channel press */
		e->length = 2;

	return 0;
} 
Py_CATCH (-1);

/* channel */

static PyObject * 
MidiEvent_get_channel (MidiEvent * e)
{
	if (e->has_channel() == 0)
		return PyNone();
	
	return PyInt ((e->data[0] & 0xF));
}

static int
MidiEvent_set_channel (MidiEvent * e, PyObject * value) 
Py_TRY
{
	long c = PyInt_AsLong (value);
	
	if (c < 0) return -1;
	if (c > 15) 
		throw ValueError() << "channel must be in (0 .. 15).";
	
	if (e->has_channel() == 0)
		throw TypeError() << "this event has no channel.";

	e->set_channel (c);

	return 0;
} 
Py_CATCH (-1);

/* /////////////////////////////////////////////////////////////////////// */

/* note */

d_note
MidiEvent::string2note (char * buffer)
{
	static int letter_notes[] = {9, 11, 0, 2, 4, 5, 7}; /* abcdefg */
	
	if (buffer == 0)
		throw AssertionError() << "string2note: string is null.";

	char * b = buffer;

	char c = tolower (*b);
	if (c < 'a' || c > 'g')
		throw ValueError() << "note identifier must be in 'abcdefg'";

	int note = letter_notes [c - 'a'];
	int octave = islower (*b) ? 4 : 3;
	int sign = 0;

	/* parse sign modifiers: */
	++b;
	while (*b)
	{
		if (*b == '#') 
			++sign;
		else if (*b == 'b') 
			--sign;
		else
			break;
		++b;
	}
	
	/* octave number following: */
	if (isdigit (*b))
	{
		char * end = b;
		octave = strtol (b, &end, 10);

		if (*end) 
			throw ValueError() << "trailing garbage in note string.";
	}
	else
	{
		/* parse octave modifiers: */
		while (*b)
		{
			if (*b == '`' || *b == '\'') 
				++octave;
			else if (*b == ',') 
				--octave;
			else
				throw ValueError() << "octave modifier must be one of \",'`\".";
			++b;
		}
	}

	d_note n = octave * 12 + note + sign;

	if (n < 0 || n > 127)
		throw ValueError() << "notes must be >= 0 and <= 127.";
	
	return n;
}

static PyObject *
MidiEvent_get_note (MidiEvent * e)
{
	if (e->has_note() == 0)
		return PyNone();
	
	return PyInt (e->data[1]);
}

char * 
MidiEvent::note2string (d_note n, char * buffer, bool flat)
{
	static char letters_flat[] = "CDDEEFGGAABB";
	static char letters_sharp[] = "CCDDEFFGGAAB";

	/* into integer space */
	int i = (int) (n + 0.5);
	i %= 127;

	int octave = i / 12 - 4; /* lower case above 4th octave */
	i %= 12;

	/* basic characters */
	char note, sign; 
	
	if (flat)
	{
		note = letters_flat[i % 12];
		sign = 'b';
	} else {
		note = letters_sharp[i % 12];
		sign = '#';
	}
	
	if (octave >= 0)
		note = tolower (note);

	/* go fill the buffer. */
	char * s = buffer;

	*s++ = note;

	/* append sign if needed */
	switch (i)
	{
		case 1:
		case 3:
		case 6:
		case 8:
		case 10:
			*s++ = sign; 
	}

	while (octave < -1)
		*s++ = ',',
		octave++;
	
	while (octave > 0)
		*s++ = '`',
		octave--;
		
	*s = 0;
	return buffer;
}

/* /////////////////////////////////////////////////////////////////////// */

static int
MidiEvent_set_note (MidiEvent * e, PyObject * value) 
Py_TRY
{
	if (e->has_note() == 0)
		throw TypeError() << "this event has no note.";

	d_note n = PyNote2n (value);

	/* silently truncate. */
	e->set_note (lrintf (n));

	return 0;
} 
Py_CATCH (-1);

static PyObject *
MidiEvent_get_controller (MidiEvent * e)
{
	if (e->status() != 0xB0)
		return PyNone();
	
	return PyInt (e->data[1]);
}

static int
MidiEvent_set_controller (MidiEvent * e, PyObject * value) 
Py_TRY
{
	if (e->status() != 0xB0)
		throw TypeError() << "this event has no controller.";

	long c;
	
	if (!PyArg_Parse (value, "i", &c))
		return -1;

	e->set_note (c);

	return 0;
} 
Py_CATCH (-1);

static PyObject *
MidiEvent_get_program (MidiEvent * e)
{
	if (e->status() != 0xC0)
		return PyNone();
	
	return PyInt (e->data[1]);
}

static int
MidiEvent_set_program (MidiEvent * e, PyObject * value) 
Py_TRY
{
	if (e->status() != 0xC0)
		throw TypeError() << "this event has no program.";

	long c;
	
	if (!PyArg_Parse (value, "i", &c))
		return -1;

	e->set_note (c);

	return 0;
} 
Py_CATCH (-1);

static PyObject *
MidiEvent_get_v (MidiEvent * e)
{
	if (e->has_channel() == 0)
		return PyNone();
	
	switch (e->status())
	{
		case 0x80: /* note off */
		case 0x90: /* note on */
		case 0xA0: /* key press */
		case 0xB0: /* controller */
			return PyInt (e->data[2]);

		case 0xC0: /* program change */
		case 0xD0: /* channel press */
			return PyInt (e->data[1]);
		
		case 0xE0:
			/* return PyFloat (e->get_pitch()); */
			break;
	}

	return PyNone();
}

static int
MidiEvent_set_v (MidiEvent * e, PyObject * value) 
Py_TRY
{
	if (e->has_channel() == 0)
		throw TypeError() << "this event has no value. :)";
	
	if (e->status() == 0xE0) /* pitch */
		throw TypeError() << "this event has no value. :)";

	long v;

	if (!PyArg_Parse (value, "i", &v))
	{
		PyErr_Clear();
		throw TypeError() << "v must be in range (0, 127)";
	}
		
	/* quietly saturate the value */
	if (v > 0x7f) v = 0x7f;
	else if (v < 0) v = 0;

	switch (e->status())
	{
		case 0xC0: /* program change */
		case 0xD0: /* channel press */
			e->data[1] = v;
			break;

		default:
			e->data[2] = v;
			break;
	}

	return 0;
} 
Py_CATCH (-1);

static PyObject *
MidiEvent_get_pitch (MidiEvent * e)
{
	if (e->status() != 0xE0)
		return PyNone();
	return PyFloat (e->get_pitch());
}

static int
MidiEvent_set_pitch (MidiEvent * e, PyObject * value)
Py_TRY
{
	if (e->status() != 0xE0)
		throw TypeError() << "not a pitch event";
	
	d_float pitch;
	if (!PyArg_Parse (value, "d", &pitch))
		return -1;

	int p = (int) ((pitch * 0x2000) + 0x2000);

	/* quietly saturate the pitch value */
	if (p < 0) p = 0;
	else if (p > 0x3fff) p = 0x3fff;

	e->data[1] = p & 0x7f;
	e->data[2] = (p >> 7);

	return 0;
}
Py_CATCH (-1);

static PyObject *
MidiEvent_get_sysex (MidiEvent * e)
{
	if (!SysexEvent_type->is_derived (e))
		return PyNone();

	byte * d = e->data;

	if (d[0] != 0xF0)
		return PyString ("malformed");
	++d;

	SysexID * id = sysex_ids;
	while (1)
	{
		if (id->d0) /* single-byte mfg. */
		{
			if (d[0] == id->d0)
				break;
		}
		else if (d[0] == id->d0 && d[1] == id->d1 && d[2] == id->d2)
			break;
		
		if (id->d0 == 0xFF) /* last */
			break;

		++id;
	}

	return PyString (id->manufacturer);
}

static PyObject *
MidiEvent_get_is_note_on (MidiEvent * e)
{
	return PyInt (e->is_note_on());
}

static PyObject *
MidiEvent_get_is_note_off (MidiEvent * e)
{
	return PyInt (e->is_note_off());
}

static PyGetSetDef 
MidiEvent_getsets [] = 
{
	{"data", (getter) MidiEvent_get_data, (setter) MidiEvent_set_data, 
	 "the raw event data bytes."},
	 
	{"status", (getter) MidiEvent_get_status, (setter) MidiEvent_set_status, 
	 "the status (unshifted hi nibble of the first data byte if < 0xF0)." },
	{"channel", (getter) MidiEvent_get_channel, (setter) MidiEvent_set_channel, 
	 "the midi channel. None if status >= 0xF0." },
	 
	{"note", (getter) MidiEvent_get_note, (setter) MidiEvent_set_note, 
	 "the note (60 is middle c). None if status not in (0x80, 0x90, 0xA0).\n"
	 "When setting the note, a string can be used instead of a number. You can "
	 "either follow common practice, ie. use upper/lower case characters with "
	 "tick marks (\",'`\" are recognized) or specify the octave by number.\n"
	 "In either case, any number of signs (\"b#\") following the note "
	 "and preceding the octave modifier are accepted."},

	{"controller", (getter) MidiEvent_get_controller, (setter) MidiEvent_set_controller, 
	 "the controller. None if status != 0xB0." },
	{"v", (getter) MidiEvent_get_v, (setter) MidiEvent_set_v, 
	 "the velocity, or value. None if status < 0x80 or >= 0xE0." },

	{"program", (getter) MidiEvent_get_program, (setter) MidiEvent_set_program, 
	 "the program. None if status != 0xC0." },

	{"pitch", (getter) MidiEvent_get_pitch, (setter) MidiEvent_set_pitch, 
	 "the velocity, or value. None if status != 0xE0." },

	{"sysex", (getter) MidiEvent_get_sysex, (setter) 0,
	 "the manufacturer if the event is Sysex, None otherwise."},

	{"is_note_on", (getter) MidiEvent_get_is_note_on, 0, 
	 "whether the event is a note-on. Can return negative even if status is "
	 "0x90, see 'is_note_off' documentation." },
	{"is_note_off", (getter) MidiEvent_get_is_note_off, 0, 
	 "whether the event is a note-off. Can return positive even if status is "
	 "0x90 if the velocity 'v' is 0; this is a peculiarity of the MIDI "
	 "protocol that reduces bandwidth need." },
	{ }
};

/* sequence methods */
static int 
MidiEvent_sq_length (MidiEvent * e)
{
	return e->length;
}

static PyObject *
MidiEvent_sq_item (MidiEvent * e, int i) 
Py_TRY
{
	return PyInt ((*e)[i]); 
} 
Py_CATCH (0);

static int
MidiEvent_sq_ass_item (MidiEvent * e, int i, PyObject * value) 
Py_TRY
{
	long l = -1;
	
	if (!PyArg_Parse (value, "i", &l))
		return -1;
	
	if (i > 0 && l > 127)
		l = 127;
	else if (i == 0 && l > 255)
		l = 255;
	else if (l < 0)
		l = 0;
		
	(*e)[i] = (byte) l; /* may throw */
	return 0;
} 
Py_CATCH (-1);

static PySequenceMethods 
MidiEvent_as_sequence = 
{
	(inquiry) MidiEvent_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) MidiEvent_sq_item, /* sq_item */
	(intintargfunc) 0, /* sq_slice */
	(intobjargproc) MidiEvent_sq_ass_item, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

Type * MidiEvent_type = 0;

char * Py<MidiEvent>::__doc__ = 
"MidiEvent (bytes or attribute = value pairs) -- "
"an Event that holds a short MIDI message in up to four "
"bytes. The message bytes are accessible through the "
"sequence protocol [ ].\n"
"Depending on the message contents, there's also access "
"through the more meaningful attributes.\n"
"Although some quiet saturation mechanisms are applied to the message bytes, "
"the message is not strictly enforced to adher to the MIDI standard. "
"For sending "
"arbitrary data to a MIDI device, the SysexEvent is a better candidate.";

void
Py<MidiEvent>::__init__ (PyObject * args, PyObject * kwds)
{
	if (args)
	{
		int n = PyTuple_Size (args);
		
		if (n < 0)
			throw blewit();
		if (n > 4)
			throw ValueError() << "too many bytes.";
		
		/* consume data bytes from args, followed by tick. */
		int i = 0;
		while (i < n)
		{
			/* enough bytes, this is the tick. */
			if (i && i == length)
			{
				if (!PyArg_Parse (PyTuple_GetItem (args, i), "d", &tick))
					throw blewit();
				
				++i;
				break;
			}

			long b;
			if (!PyArg_Parse (PyTuple_GetItem (args, i), "i", &b))
				throw blewit();
				
			/* quietly saturate data bytes. */
			if (b < 0) 
				b = 0;
			else if (i == 0)
			{
				if (b > 255) b = 255;
			}
			else if (b > 127) b = 127;
				
			data [i] = b;

			if (++i == 1)
			{
				/* set length after first byte. */
				length = get_length();

				/* note-on default velocity */
				if (data[0] == 0x90)
					data[2] = VELOCITY;
			}
		}

		/* anything in excess is garbage. */
		if (i != n) throw OverflowError() << n - i << " too many value(s).";
	}

	if (kwds)
	{
		PyObject * name, * value;

		int n = 0;
		while (PyDict_Next (kwds, &n, &name, &value))
			if (PyObject_SetAttr (this, name, value))
				break;
		
		if (PyErr_Occurred())
			throw blewit();
	}
}

static struct { int status; char * name; } _status [] =
{
	{0x80, "NoteOff"},
	{0x90, "NoteOn"},
	{0xa0, "NotePressure"},
	{0xb0, "Controller"},
	{0xc0, "Program"},
	{0xd0, "ChannelPressure"},
	{0xe0, "Pitch"},
	{0xf0, "Sysex"},

	{ }
};

void
Py<MidiEvent>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Event_type;
	t->tp_getset = MidiEvent_getsets;
	t->tp_as_sequence = &MidiEvent_as_sequence;

	PyType_Ready (t);

	int i = 0;
	while (_status[i].status)
	{
		PyDict_SetItemString (
				t->tp_dict, _status[i].name, PyInt (_status[i].status));
		++i;
	}
}

/* /////////////////////////////////////////////////////////////////////// */

Type * SysexEvent_type = 0;

char * Py<SysexEvent>::__doc__ = 
"SysexEvent ([data, tick]) -- "
"a MidiEvent that can store any number of message bytes.";

void
Py<SysexEvent>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"data", "tick", 0};
	
	byte * _data = {0,};
	int _length = 1;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|s#d", keys, 
			&_data, &_length, &tick))
		throw blewit();

	set_data (_data, _length);
}

void
Py<SysexEvent>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = MidiEvent_type;
}

/* /////////////////////////////////////////////////////////////////////// */

String & 
Note::operator >> (String & s)
{
	s << "Note (note = " << (int) data[1] 
		<< ", v = " << (int) data[2]
		<< ", l = " << l 
		<< ", tick = " << tick << ")";

	return s;
}

char * Py<Note>::__doc__ = 
"Note ([note = 'Eb4', v = 100, l = 24, tick = 0]) -- "
"a MidiEvent with status 0x90 and a length.\n"
"NB: all plugins except Part will mistake this event for a single "
"note-on. Parts emit a note-off when the Note ends.";

Type * Note_type = 0;

static PyObject * 
Note_get_l (Note * self)
{
	return PyFloat (self->l);
}

static int Note_set_l (Note * self, PyObject * arg) 
Py_TRY
{
	double l;
	
	if (!PyArg_Parse (arg, "d", &l))
		return -1;

	if (l < 0)
		throw ValueError() << "l must be positive.";
	
	/* silently force minimum note length. */
	if (l == 0)
		l = 1;

	self->l = (int) l;
	return 0;
} 
Py_CATCH (-1);

static PyGetSetDef Note_getsets [] = {
	{"l", (getter) Note_get_l, (setter) Note_set_l, 
	 "the length of the note, in ticks."},
	{ }
};

void
Py<Note>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"note", "v", "l", "tick", 0};
	PyObject * note = 0, * v = 0, * l = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|OOOd", keys, 
			&note, &v, &l, &tick))
		throw blewit();
	
	if (note && MidiEvent_set_note (this, note)) return;
	if (v && MidiEvent_set_v (this, v)) return;
	if (l && Note_set_l (this, l)) return;
}

void
Py<Note>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = MidiEvent_type;
	t->tp_getset = Note_getsets;
}


/* //////////////////////////////////////////////////////////////////// */

void
init_Event_types (PyObject * dict)
{
	Event_type = new PyType<Event> ("Event", dict);

	MidiEvent_type = new PyType<MidiEvent> ("MidiEvent", dict);
	SysexEvent_type = new PyType<SysexEvent> ("SysexEvent", dict);
	Note_type = new PyType<Note> ("Note", dict);
	
	FloatEvent_type = new PyType<FloatEvent> ("FloatEvent", dict);
	TextEvent_type = new PyType<TextEvent> ("TextEvent", dict);
	
	TempoEvent_type = new PyType<TempoEvent> ("TempoEvent", dict);
	TimeEvent_type = new PyType<TimeEvent> ("TimeEvent", dict);
}
