/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _EVENT_H_
#define _EVENT_H_

#include "basics.h"

class Event
:	public Object
{
	public:
		/* when the event was received or is to be scheduled. */
		d_tick tick;

		/* the sequential number of the loop cycle this event belongs to. */
		int loop;

		Event (d_tick _t = 0, Type * type = Event_type)
			: Object (type)
			{
				tick = _t;
				loop = 0;
			}

		Event (Event & e)
			: Object (get_type (&e)) /* same type construction */
			{
				*this = e;
			}

		virtual ~Event() 
			{ }
		
		virtual void clone (Event * e) 
			{ 
				*this = *e;
			}
			
		Event & operator = (Event & e)
			{
				tick = e.tick;
				loop = e.loop;

				return * this;
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "Event (tick = " << tick
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class FloatEvent
:	public Event
{
	public:
		d_float value;

		FloatEvent (double v = 0., d_tick t = 0)
			: Event (t, FloatEvent_type)
			{
				value = v;
			}
					
		FloatEvent (FloatEvent & f)
			: Event (f)
			{
				*this = f;
			}
			
		virtual void clone (Event * e) 
			{ 
				if (!FloatEvent_type->is_derived (e))
					throw TypeError() << "not the same type.";
				
				*this = *((FloatEvent *) e);
			}
	
		FloatEvent & operator = (FloatEvent & f)
			{
				tick = f.tick;
				loop = f.loop;
				value = f.value;
		
				return *this;
			}
			
		operator float () 
			{
				return (float) value;
			}

		operator double & () 
			{
				return (double &) value;
			}
		
		virtual String & operator >> (String & s)
			{
				return s 
					<< "FloatEvent (value = " << value 
					<< ", tick = " << tick 
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

/* describing a tempo change in beats per minute */

class TempoEvent
:	public Event
{
	public:
		d_float bpm;

		/* dynamically recalculated by TempoMaps */
		d_time time, sec_per_tick;

		TempoEvent (double _bpm = BPM, d_tick t = 0)
			: Event (t, TempoEvent_type)
			{
				bpm = _bpm;
				time = 0;
				sec_per_tick = 0;
			}
					
		TempoEvent (TempoEvent & t)
			: Event (t)
			{
				*this = t;
			}
			
		virtual void clone (Event * e) 
			{ 
				if (!TempoEvent_type->is_derived (e))
					throw TypeError() << "not the same type.";
				
				*this = *((TempoEvent *) e);
			}

		TempoEvent & operator = (TempoEvent & t)
			{
				tick = t.tick;
				loop = t.loop;
				bpm = t.bpm;
				time = t.time;
				sec_per_tick = t.sec_per_tick;

				return *this;
			}
			
		virtual String & operator >> (String & s)
			{
				return s 
					<< "TempoEvent (bpm = " << bpm 
					<< ", tick = " << tick 
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class BBT
{
	public:
		int bar, beat;
		d_tick tick;

		BBT (int br = 0, int bt = 0, d_tick t = 0)
			{
				bar = br;
				beat = bt;
				tick = t;
			}

		int compare (BBT & other)
			{
				/* how to do this with less branches, anyone? */
				if (bar > other.bar) return 1;
				if (bar < other.bar) return -1;
				if (beat > other.beat) return 1;
				if (beat < other.beat) return -1;
				if (tick > other.tick) return 1;
				if (tick < other.tick) return -1;
				return 0;
			}
};

class TimeEvent
: public Event
{
	public:
		/* numerator, denominator */
		int n, d;
		
		/* ticks per denominator unit */
		int ticks; 

		struct {
			int bar, beat;
		} length;

		/* cached info */
		BBT bbt;

		TimeEvent (int _n = 4, int _d = 4, int _ticks = QUARTER, d_tick t = 0)
			: Event (t, TimeEvent_type)
			{
				n = _n;
				d = _d;
				ticks = _ticks;

				update_length();
			}
		
		TimeEvent (TimeEvent & t)
			: Event (t)
			{
				*this = t;
			}

		virtual void clone (Event * e) 
			{ 
				if (!TimeEvent_type->is_derived (e))
					throw TypeError() << "not the same type.";
				
				*this = *((TimeEvent *) e);
			}

		void update_length(); /* call after n/d/ticks have changed */

		/* fill 'bbt' from 't' */
		void map_tick (d_tick t, BBT & bbt);
		/* return tick from bbt */
		d_tick map_bbt (BBT & bbt);

		TimeEvent & operator = (TimeEvent & t)
			{
				tick = t.tick;
				loop = t.loop;
				
				n = t.n;
				d = t.d;
				ticks = t.ticks;

				bbt = t.bbt;

				update_length();

				return *this;
			}
			
		virtual String & operator >> (String & s)
			{
				return s
					<< "TimeEvent (n = " << n
					<< ", d = " << d
					<< ", ticks = " << ticks
					<< ", tick = " << tick
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

#define MARK_EVENT_TEXT_MAX	96

class TextEvent
:	public Event
{
	public:
		char text [MARK_EVENT_TEXT_MAX];
											
		TextEvent (char * s = 0, d_tick t = 0)
			: Event (t, TextEvent_type)
			{
				if (s)
					set_text (s);
				else
					text[0] = 0;
			}
		
		TextEvent (TextEvent & c)
			: Event (c)
			{
				*this = c;
			}

		virtual void clone (Event * e) 
			{ 
				if (!TextEvent_type->is_derived (e))
					throw TypeError() << "not the same type.";
				
				*this = *((TextEvent *) e);
			}

		TextEvent & operator = (TextEvent & c)
			{
				tick = c.tick;
				loop = c.loop;
				set_text (c.text);

				return *this;
			}
			
		void set_text (char * s)
			{
				strncpy (text, s, MARK_EVENT_TEXT_MAX);
				text [MARK_EVENT_TEXT_MAX - 1] = 0;
			}
		
		virtual String & operator >> (String & s)
			{
				return s << "TextEvent (" << tick << ", \"" << text << "\")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class ClickEvent
:	public Event
{
	public:
		int bar, beat;
		d_float fraction; /* 0.0 is on the beat, 1.0 is on the next. */
											
		ClickEvent (int b, int bt, int f, d_tick t)
			: Event (t, ClickEvent_type)
			{
				bar = b;
				beat = bt;
				fraction = f;
			}
		
		ClickEvent (ClickEvent & c)
			: Event (c)
			{
				*this = c;
			}

		virtual void clone (Event * e) 
			{ 
				if (!ClickEvent_type->is_derived (e))
					throw TypeError() << "not the same type.";
				
				*this = *((ClickEvent *) e);
			}

		ClickEvent & operator = (ClickEvent & c)
			{
				tick = c.tick;
				loop = c.loop;
				bar = c.bar;
				beat = c.beat;
				fraction = c.fraction;

				return *this;
			}
			
		virtual String & operator >> (String & s) 
			{
				return s 
					<< "ClickEvent (bar = " << bar
					<< ", beat = " << beat
					<< ", fraction = " << fraction
					<< ", tick = " << tick << ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

/* an Event holding raw MIDI data. */
class MidiEvent
:	public Event
{
	protected:
		byte bytes[4];

	public:
		int length;
		byte * data;

		MidiEvent (d_tick t = 0, Type * type = MidiEvent_type)
			: Event (t, type) 
			{
				length = 0;
				data = bytes;
			}

		MidiEvent (MidiEvent & e)
			: Event (e)
			{
				length = 0;
				data = bytes;
				*this = e;
			}

		/* return -1 if this before e, 0 if equal, 1 if this after e */
		int compare (MidiEvent * e);
		
		virtual void clone (Event * e) 
			{ 
				if (!MidiEvent_type->is_derived (e))
					throw TypeError() << "not the same type.";

				MidiEvent & m = *((MidiEvent *) e);
				
				if (m.length > 4)
					throw TypeError() << "too many MIDI bytes.";

				*this = m;
			}
			
		MidiEvent & operator = (MidiEvent & m)
			{
				tick = m.tick;
				loop = m.loop;
				set_data (m.data, m.length);

				return *this;
			}

		virtual void set_data (byte * b, int l) 
			{
				if (l < 0 || l > 4)
					throw blewit() << "length must be from 1 to 4.";
				
				length = l;
				
				while (l-- > 0)
					bytes[l] = b[l];
			}

		int channel()
			{
				if (length < 1)
					return -1;
				if (data[0] < 0x80 || data[0] >= 0xF0)
					return -1;
				
				return data[0] & 0xF;
			}

		int status() { return data[0] & 0xF0; }
		
		int has_channel() { return data[0] >= 0x80 && data[0] < 0xF0; }
		int set_channel (int c) { return data[0] = (data[0] & 0xF0) | c; }

		int has_note() { return data[0] >= 0x80 && data[0] <= 0xAf; }
		int set_note (int n) { return data[1] = (byte) n; }

		static d_note string2note (char *);
		/* returns the string passed in for convenience- */
		static char * note2string (d_note, char *, bool flat = 1);

		int is_note_on() 
			{ 
				return status() == 0x90 && data[2]; 
			}
		int is_note_off() 
			{ 
				return status() == 0x80 || 
							(status() == 0x90 && data[2] == 0);
			}
		bool is_note_off_for (MidiEvent * e)
			{
				return is_note_off() && e->is_note_on() &&
						(e->data[0] & 0xF) == (data[0] & 0xF) && /* same channel */
						e->data[1] == data[1]; /* same pitch */
			}
		bool is_double_note (MidiEvent * e)
			{
				return is_note_on() && e->is_note_on() &&
						(e->data[0] & 0xF) == (data[0] & 0xFF) && /* same channel */
						e->data[1] == data[1]; /* same pitch */
			}
		
		int is_key_press() { return status() == 0xa0; }
		int is_controller() { return status() == 0xb0; }
		int is_program() { return status() == 0xc0; }
		int is_chn_press() { return status() == 0xd0; }
		int is_pitch() { return status() == 0xe0; }
		int is_sysex() { return SysexEvent_type->is_derived (this); }

		int get_length() 
			{
				int s = status();
				
				if (s < 0x80) return 1;
				if (s < 0xc0) return 3;
				if (s < 0xe0) return 2;
				if (s < 0xf0) return 3;
				
				return length;
			}
						
		d_float get_pitch() 
			{
				return (d_float) (((int) data[1] | (int) data[2] << 7) - 0x2000) 
																				 /
														(d_float) (0x2000);
			}
		
		byte & operator [] (int i) 
			{
				if (i < 0)
					i += length;
				if (i < 0 || i >= length)
					throw IndexError() << "no such byte.";
				return data[i];
			}

		virtual String & operator >> (String & s) 
			{
				s << "MidiEvent (data = '"; 

				for (int i = 0; i < length; i++)
					s << data[i];

				return s
					<< "', tick = " << tick << ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class SysexEvent
: public MidiEvent
{
	public:
		SysexEvent (d_tick t = 0)
			: MidiEvent (t, SysexEvent_type) 
			{ }

		SysexEvent (SysexEvent & e)
			: MidiEvent (e.tick, SysexEvent_type) 
			{
				length = 0;
				set_data (e.data, e.length);
			}

		~SysexEvent()
			{
				free_data();
			}

		virtual void clone (Event * e) 
			{ 
				if (!SysexEvent_type->is_derived (e))
					throw TypeError() << "not the same type.";

				*this = *((SysexEvent *) e);
			}
			
		SysexEvent & operator = (SysexEvent & s)
			{
				tick = s.tick;
				loop = s.loop;
				set_data (s.data, s.length);

				return *this;
			}
		
		void free_data()
			{
				if (data && data != bytes)
					free (data);
				data = bytes;
			}

		virtual void set_data (byte * b, int l) 
			{
				if (b == 0 || l == 0)
				{
					if (data && data != bytes) 
						free (data);

					length = 0;
					data = bytes;
					return;
				}
				
				length = l;

				if (data && data != bytes)
					data = (byte *) realloc (data, l);
				else
					data = (byte *) malloc (l);

				if (!data) 
					throw MemoryError();

				memcpy (data, b, l);
			}

		virtual String & operator >> (String & s) 
			{
				s << "SysexEvent (data = '"; 

				for (int i = 0; i < length && i < 10; i++)
					s << data[i];
				
				if (length > 10)
					s << "...";

				return s
					<< "', tick = " << tick << ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

/* Note */

class Note
: public MidiEvent
{
	public:
		d_tick l; /* note length */
		
		Note (d_tick t = 0)
			: MidiEvent (t, Note_type) 
			{
				length = 3;

				data[0] = 0x90; /* note on, channel 0 */
				data[1] = NOTE; 
				data[2] = VELOCITY;

				l = QUARTER / 8;
			}

		Note (Note & n)
			: MidiEvent (n)
			{
				*this = n;
			}

		int note()
			{
				return data[1];
			}

		virtual void clone (Event * e) 
			{ 
				if (!Note_type->is_derived (e))
					throw TypeError() << "not the same type.";

				*this = *((Note *) e);
			}
			
		Note & operator = (Note & n)
			{
				tick = n.tick;
				loop = n.loop;
				l = n.l;
				set_data (n.data, n.length);

				return *this;
			}

		virtual String & operator >> (String & s);
};

#endif /* _EVENT_H_ */
