/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "EventList.h"
#include "NoteOffs.h"

#include "PyType.h"

EventList::EventList (Type * t, int granularity)
: NamedObject (EventList_type),
	TypedObjectList<Event> ("event list", granularity),
	mutex (Mutex::Recursive)
{
	set_event_type (t);
	color = 0x7b8242;
	allow_doubles = false;
	quiet = false;
}

EventList::EventList (EventList & l, int lo, int hi)
: NamedObject (EventList_type),
	TypedObjectList<Event> ("event list", 510)
{
	set_event_type (l.type);

	if (hi < 0) hi = length;

	while (lo < hi)
		add_event (l[lo++]);

	color = l.color;
}

void
EventList::set_event_type (Type * t)
{
	if (length) /* already hold events */
		throw AssertionError() << "too late to set the event type.";
		
	if (!Event_type->is_derived (t))
		throw TypeError() << "can only hold events.";
	
	type = t;
}

/* //////////////////////////////////////////////////////////////////////// */

/* midi event precedence code */

static int 
_status_priority (MidiEvent * e)
{
	static int table[] = {
		5, /* 0x80, note off */
		4, /* 0x90, note */
		6, /* 0xA0, note press */
		2, /* 0xB0, cc */
		3, /* 0xC0, program change */
		7, /* 0xD0, channel press */
		1, /* 0xE0, pitch */
		0  /* 0xF0, sysex/rt */
	};

	return table [e->data[0] >> 5];
}

int
MidiEvent::compare (MidiEvent * e)
{
	/* check tick difference */
	d_tick dt = tick - e->tick;
	
	if (dt < 0)
		return -1;
	if (dt > 0)
		return +1;

	/* check status priorities */
	int dp = _status_priority (e) - _status_priority (this);

	if (dp < 0)
		return -1;
	if (dp > 0)
		return +1;

	/* note event, high notes first (velocity: dontcare) */
	if (data[0] < 0xA0)
		return (int) e->data[1] - (int) data[1];

	/* other events, low values first, ensures MSB -> LSB order for CC */
	return (int) data[1] - (int) e->data[1];
}

/* //////////////////////////////////////////////////////////////////////// */

int
EventList::add_event (Event * e)
{
	if (!type->is_derived (e))
		throw TypeError() << "can only handle " 
											<< type->tp_name << " objects.";

	if (MidiEvent_type->is_derived (type))
		return add_midi_event ((MidiEvent *) e);

	int i = index_of_tick (e->tick);
	TypedObjectList<Event>::add_object (e, i);
	return i;
}

int
EventList::add_midi_event (MidiEvent * e)
{
	if (e->is_note_off())
		return add_note_off (e);

	int i = index_of_tick (e->tick);

	MidiEvent ** list = (MidiEvent **) items;

	/* adjust insertion index based on midi event precedence */
	while (i >= 0 && i < length)
	{
		if (e->compare (list[i]) <= 0)
			break;

		++i;
	}

	/* check for doubles */
	if (e->is_note_on() && !allow_doubles)
	{
		while (i < length - 1)
		{
			int diff = e->compare (list[i + 1]);
			if (diff)
				break;

			++i;

			/* same note */
			if (e->is_double_note (list[i]))
			{
				if (!quiet)
					msg(), this, "ignoring double note ", e;
				return -1;
			}
		}
	}

	TypedObjectList<Event>::add_object (e, i);
	return i;
}

/* must be called with the PL held. */
int 
EventList::add_note_off (MidiEvent * e)
{
	MidiEvent ** list = (MidiEvent **) items;

	int i = index_of_tick (e->tick);

	while (--i >= 0)
	{
		if (!e->is_note_off_for (list[i]))
			continue;

		MidiEvent * on = list[i];

		/* construct Note event */
		Note * note = NEW (Note);
		
		note->tick = on->tick;
		note->l = e->tick - on->tick;

		note->data[0] = on->data[0];
		note->data[1] = on->data[1];
		note->data[2] = on->data[2];

		/* swap in */
		list[i] = note;
		::decref (on, true);

		return i;
	}

	/* no corresponding note-on found. to add anyway: */
#if 0
	TypedObjectList<Event>::add_object (off, i);
	return i;
#else
	if (!quiet)
		msg(), "no note-on in ", this, "for ", e;
#endif
	
	return -1;
}

/* binary search for a given tick index; return the next index if we couldn't
 * resolve.
 */
int
EventList::index_of_tick (d_tick t)
{
	int hi = length, lo = -1;

	while (hi - lo > 1)
	{
		int probe = (hi + lo) / 2;

		if ((*this)[probe]->tick < t)
			lo = probe;
		else
			hi = probe;
	}

	return hi;
}

/* Standard MIDI File encoding //////////////////////////////////////////// */

class SMFString
: public String
{
	public:
		NoteOffs playing;
		d_tick tick;

		SMFString (MidiEvent **, int);
		
		void bytes (const void * s, int n);

		template <class T>
			SMFString & operator += (const T & t)
				{
					bytes (&t, sizeof (T));
					return *this;
				}

		void varlen (int n);
		void event (MidiEvent *);
		void note_offs (d_tick);

		void push (Note *);
};

void
SMFString::bytes (const void * s, int n)
{
	int l = length;
	inflate (n);
	memcpy (string + l, s, n);
	string [length] = 0;
}

void SMFString::varlen (int n)
{
	uint32 buffer = n & 0x7F;

	int i = 1;
	while ((n >>= 7) > 0) 
	{
		buffer <<= 8;
		buffer |= 0x80 | (n & 0x7F);
		++i;
	}

	while (i) 
	{
		*this += (byte) (buffer & 0xFF);
		buffer >>= 8;
		--i;
	}
}

void
SMFString::event (MidiEvent * e)
{
	/* SMF loses timing precision */
	varlen ((int) (e->tick - tick));
	tick = e->tick;

	if (e->is_sysex())
		varlen (e->length);
	
	bytes (e->data, e->length);
}

void
SMFString::note_offs (d_tick t)
{
	MidiEvent * e;
	
	while ((e = playing.peek()) && e->tick <= t)
	{
		event (e);
		playing.pop();
	}
}

void
SMFString::push (Note * n)
{
	Note * e = NEW (Note);
	e->clone (n);
	
	e->tick += e->l; /* ending tick */
	(*e)[2] = 0; /* velocity zero = note off */
	
	playing.push (e);

	::decref (e, true);
}

SMFString::SMFString (MidiEvent ** e, int n)
{
	tick = 0;
	
	for (int i = 0; i < n; ++i)
	{
		note_offs (e[i]->tick);
		event (e[i]);

		if (get_type (e[i]) == Note_type)
			push ((Note *) e[i]);
	}

	note_offs (D_TICK_MAX);

	MidiEvent eot (tick);
	byte EOT[] = {0xFF, 0x2F, 0};
	eot.set_data (EOT, 3);
	
	event (&eot);
}

/* //////////////////////////////////////////////////////////////////////// */

static PyObject * 
EventList_get_type (EventList * l)
{
	return ::incref (l->type);
}

static PyObject * 
EventList_get_color (EventList * l)
{
	return PyInt (l->color);
}

static int
EventList_set_color (EventList * l, PyObject * arg)
{
	if (!PyArg_Parse (arg, "i", &l->color))
		return -1;
	return 0;
}

static PyObject * 
EventList_get_allow_doubles (EventList * l)
{
	return PyInt (l->allow_doubles);
}

static int
EventList_set_allow_doubles (EventList * l, PyObject * arg)
{
	if (!PyArg_Parse (arg, "i", &l->allow_doubles))
		return -1;
	return 0;
}

static PyObject * 
EventList_get_quiet (EventList * l)
{
	return PyInt (l->quiet);
}

static int
EventList_set_quiet (EventList * l, PyObject * arg)
{
	if (!PyArg_Parse (arg, "i", &l->quiet))
		return -1;
	return 0;
}

static PyGetSetDef 
EventList_getsets [] = 
{
	{"type", (getter) EventList_get_type, 0, "the type of events accepted."},
	{"color", (getter) EventList_get_color, (setter) EventList_set_color, 
	 "the color to use when drawing the events from this list."},
	{"allow_doubles", (getter) EventList_get_allow_doubles, (setter) EventList_set_allow_doubles, 
	 "if 0, don't add double notes (same tick, same note)."},
	{"quiet", (getter) EventList_get_quiet, (setter) EventList_set_quiet, 
	 "if 0, print warnings about ignored notes and note-offs to stdmsg."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

/* sequence methods */

static int 
EventList_sq_length (EventList * p)
{
	return p->length;
}

static PyObject *
EventList_sq_item (EventList * p, int i) 
Py_TRY
{
	if (i >= 0 && i < *p)
		return ::incref ((*p)[i]); 
	throw IndexError() << "no item " << i << " in " << p;
} 
Py_CATCH (0);

static int
EventList_sq_ass_item (EventList * p, int i, PyObject * arg) 
Py_TRY
{
	if (i < 0 || i > *p)
		throw IndexError() << "no such item";

	if (arg && arg != Py_None)
		throw NotImplementedError() << "sorry, can't replace items.";
	
	p->remove_items (i, 1);
	
	return 0;
} 
Py_CATCH (-1);

static PyObject *
EventList_sq_slice (EventList * p, int m, int n) 
Py_TRY
{
	return new EventList (*p, m, n);
} 
Py_CATCH (0);

static PySequenceMethods 
EventList_as_sequence = 
{
	(inquiry) EventList_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) EventList_sq_item, /* sq_item */
	(intintargfunc) EventList_sq_slice, /* sq_slice */
	(intobjargproc) EventList_sq_ass_item, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

static PyObject *
EventList_add (EventList * self, PyObject * args)
Py_TRY
{
	Event * event = 0;
	d_tick t = D_TICK_MAX;
	
	if (!PyArg_ParseTuple (args, "O!|d", Event_type, &event, &t))
		return 0;

	/* clone it. */
	Event * e = (Event *) get_type (event)->__new__();
	e->clone (event);
	if (t != D_TICK_MAX)
		e->tick = t;

	try 
	{
		self->add_event (e);
	} 
	catch (blewit & b) 
	{
		::decref (e);
		throw;
	}

	return e;
} 
Py_CATCH (0);

static PyObject *
EventList_insert (EventList * self, PyObject * args)
Py_TRY
{
	Event * e;
	d_tick t = D_TICK_MAX;
	
	if (!PyArg_ParseTuple (args, "O!|d", Event_type, &e, &t))
		return 0;

	if (t != D_TICK_MAX)
		e->tick = t;

	int i;
	
	try 
	{
		i = self->add_event (e);
	} 
	catch (blewit & b) 
	{
		throw;
	}

	return PyInt (i);
} 
Py_CATCH (0);

static PyObject *
EventList_remove (EventList * self, PyObject * args)
Py_TRY
{
	Event * e = 0;
	
	if (!PyArg_ParseTuple (args, "O!", Event_type, &e))
		return 0;

	self->remove (e);
	return PyNone();
} 
Py_CATCH (0);

static PyObject * 
EventList_lock (EventList * p)
{
	p->mutex.lock();
	return PyNone();
}

static PyObject * 
EventList_unlock (EventList * p)
{
	p->mutex.unlock();
	return PyNone();
}

PyObject *
EventList_index_of_tick (EventList * el, PyObject * tick)
{
	d_tick t;

	if (!PyArg_Parse (tick, "d", &t))
		return 0;

	ScopedLock (el->mutex);

	return PyInt (el->index_of_tick (t));
}


/* make a MIDI file string */
PyObject *
EventList_smf_track (EventList * el)
{
	ScopedLock (el->mutex);

	SMFString s ((MidiEvent **) el->items, *el);

	return PyString (s, s.length);
}

static PyMethodDef 
EventList_methods [] = 
{
  {"add", (PyCFunction) EventList_add, METH_VARARGS,
	 "add (event[, t]) -- add a copy of 'event' at optional tick 't'."},
  {"insert", (PyCFunction) EventList_insert, METH_VARARGS,
	 "insert (event[, t]) -- insert 'event' at optional tick 't'. Note that "
	 "this doesn't insert a copy, use EventList.add() for that."},
	{"remove", (PyCFunction) EventList_remove, METH_VARARGS,
	 "remove (event) -- remove 'event'."},
	{"lock", (PyCFunction) EventList_lock, METH_NOARGS,
	 "lock() -- by acquiring this lock, the part is protected from other threads' "
	 "attempts at reading or modifying its contents."},
	{"unlock", (PyCFunction) EventList_unlock, METH_NOARGS,
	 "unlock() -- releases the protecting lock."},
	{"index_of_tick", (PyCFunction) EventList_index_of_tick, METH_O,
	 "index_of_tick (tick) -- returns the list insertion point for 'tick'."},
	{"smf_track", (PyCFunction) EventList_smf_track, METH_NOARGS,
	 "smf_track() -- returns events encoded as a Standard MIDI File track."},
	{}
};

/* //////////////////////////////////////////////////////////////////////// */

Type * EventList_type = 0;

char * Py<EventList>::__doc__ = 
"EventList (type) -- a list of a certain type of Events, "
"sorted by tick.";

void
Py<EventList>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"type", 0};
	Type * t;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!", keys,
			&PyType_Type, &t))
		throw blewit();
	
	set_event_type (t);
}

void
Py<EventList>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;

	t->tp_base = NamedObject_type;
	t->tp_getset = EventList_getsets;
	t->tp_methods = EventList_methods;
	t->tp_as_sequence = &EventList_as_sequence;
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_EventList_types (PyObject * dict)
{
	EventList_type = new PyType<EventList> ("EventList", dict);
}

