/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "Graph.h"
#include "PyType.h"

#include "dsp/FIR.h"

static Type * FIR_type;

class FIR
: public Plugin
{
	public:
		DSP::FIR * fir;

		FIR();
		~FIR();
		
		/* Plugin:: */
		virtual void set_state (Graph * g, int s);
		virtual void process_audio (int frames);
};

FIR::FIR()
: Plugin (FIR_type, Audio)
{
	fir = 0;

	add_in (new In (this, CyclicAudio_type, IN_SLOTS, "in"));
	add_out (new Out (this, CyclicAudio_type, "out"));
}

FIR::~FIR()
{
	if (fir) delete fir;
}

void
FIR::set_state (Graph * g, int state)
{
	if (state == Plugin::Prepared)
	{
		if (fir == 0)
			throw AssertionError() << "no filter kernel set";

		fir->reset();
	}

	Plugin::set_state (g, state);
}

void
FIR::process_audio (int frames)
{
	CyclicAudio * event = ins[0]->pop_mix();
	
	if (event == 0)
		return;

	d_sample * s = event->data;
	for (int i = 0; i < frames; ++i)
		s[i] = fir->process (s[i]);
	
	outs[0]->push (event);
}

/* /////////////////////////////////////////////////////////////////////// */

/* FIR getsets */

static PyObject *
FIR_get_c (FIR * fir)
{
	DSP::FIR * f = fir->fir;

	PyObject * C = PyTuple_New (f->n);

	for (int i = 0; i < f->n; ++i)
		PyTuple_SetItem (C, i, PyFloat (f->c[i]));
	
	return C;
}

static int
FIR_set_c (FIR * fir, PyObject * C)
Py_TRY
{
	if (fir->state != Plugin::Off)
		throw AssertionError() << "cannot do that on a running plugin";

	if (!PySequence_Check (C))
		throw TypeError() << "need a sequence for the filter kernel.";

	int n = PySequence_Size (C);
	if (n < 1)
		throw ValueError() << "filter kernel is too small.";
	
	d_sample c[n];

	for (int i = 0; i < n; ++i)
	{
		PyRef f (PySequence_GetItem (C, i));
		
		if (!PyArg_Parse (f, "f", c + i))
			throw TypeError() << "c[" << i << "] is not a float coefficient.";
	}

	if (fir->fir) delete fir->fir;
	fir->fir = 0;
	
	fir->fir = new DSP::FIR (n, c);

	return 0;
}
Py_CATCH (-1);

static PyGetSetDef 
FIR_getsets [] =
{
	{"c", (getter) FIR_get_c, (setter) FIR_set_c, 
	 "c, tuple containing the filter kernel coefficients."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
FIR_get_kernel (FIR * fir)
Py_TRY
{
	AudioSignal * t = NEW (AudioSignal);
	PyRef ref (t);

	DSP::FIR & f = *(fir->fir);
	t->set_size (f.n);
	for (int i = 0; i < f.n; ++i)
		t->data[i] = f.c[i];

	return ::incref (t);
}
Py_CATCH (0);

static PyObject *
FIR_one_shot (FIR * fir, AudioSignal * e)
Py_TRY
{
	if (!AudioSignal_type->is_derived (e))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (e);

	DSP::FIR f (*(fir->fir));

	d_sample * s = e->data;

	for (int i = 0; i < e->size; ++i)
		s[i] = f.process (s[i]);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
FIR_upsample (FIR * fir, PyObject * args)
Py_TRY
{
	AudioSignal * s;
	int ratio;

	if (!PyArg_ParseTuple (args, "O!i", AudioSignal_type, &s, &ratio))
		return 0;

	if (ratio <= 1)
		throw ValueError() << "ratio must be >= 2";

	DSP::FIRUpsampler up (*(fir->fir), ratio);

	AudioSignal * t = NEW (AudioSignal);
	PyRef ref (t);
	t->set_size (s->size * ratio);

	for (int i = 0; i < s->size; ++i)
	{
		int j = i * ratio;
		
		t->data [j] = up.upsample (s->data [i]);
		
		for (int k = 1; k < ratio; ++k)
			t->data [j + k] = up.pad (k);
	}

	return ::incref (t);
}
Py_CATCH (0);

static PyObject *
FIR_downsample (FIR * fir, PyObject * args)
Py_TRY
{
	AudioSignal * s;
	int ratio;

	if (!PyArg_ParseTuple (args, "O!i", AudioSignal_type, &s, &ratio))
		return 0;

	if (ratio <= 1)
		throw ValueError() << "ratio must be >= 2";

	DSP::FIR down (*(fir->fir));

	AudioSignal * t = NEW (AudioSignal);
	PyRef ref (t);
	t->set_size (s->size / ratio);

	for (int i = 0; i < t->size; ++i)
	{
		int j = i * ratio;
		
		t->data [i] = down.process (s->data [j]);
		
		for (int k = 1; k < ratio; ++k)
			down.process (s->data [j + k]);
	}

	return ::incref (t);
}
Py_CATCH (0);

static PyMethodDef
FIR_methods[] = 
{
	{"get_kernel", (PyCFunction) FIR_get_kernel, METH_NOARGS,
	 "get_kernel() -- return the filter coefficients as an AudioSignal."},
	{"one_shot", (PyCFunction) FIR_one_shot, METH_O,
	 "one_shot (audio signal) -- process a signal, "
	 "starting from a filter history of zeros. (in-place)"},
	{"upsample", (PyCFunction) FIR_upsample, METH_VARARGS,
	 "upsample (audio signal, ratio) -- upsample the signal by 'ratio', "
	 "using this filter for the interpolation."},
	{"downsample", (PyCFunction) FIR_downsample, METH_VARARGS,
	 "downsample (audio signal, ratio) -- downsample the signal by 'ratio', "
	 "using this filter for the decimation."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<FIR>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = FIR_getsets;
	t->tp_methods = FIR_methods;
}

char * Py<FIR>::__doc__ = 
"FIR ([c = (1,)]) -- Finite Impulse Response filter.\n"
"The filter supports kernels up to 128 coefficients; the default "
"being an identity set of size 1. The implementation is brute force; "
"therefore its use is confined to mainly academical exercises.";

void
Py<FIR>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"c", 0};
	PyObject * c = 0;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|O", keys, &c))
		return;
	
	if (c) FIR_set_c (this, c);
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_FIR_types (PyObject * dict)
{
	FIR_type = new PyType<FIR> ("FIR", dict);
}
