/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "Gain.h"
#include "Graph.h"

#include "PyType.h"

/* inlined factor application functions */

inline void
apply_factor (d_sample factor, d_sample * data, int frames)
{
	for (int i = 0; i < frames; ++i)
		data[i] *= factor;
}

inline void
apply_factor_rate (d_sample factor, d_sample * data, int frames, d_sample rate)
{
	for (int i = 0; i < frames; ++i)
		data[i] *= factor,
		factor += rate;
}

inline void
apply_factor_peak (d_sample factor, d_sample * data, int frames, d_sample * peak)
{
	for (int i = 0; i < frames; ++i)
	{
		register d_sample p = data[i] * factor;
		data[i] = p;

		p = fabsf (p);

		if (p > *peak)
			*peak = p;
	}
}

inline void
apply_factor_rate_peak (d_sample factor, d_sample * data, int frames, d_sample rate, d_sample * peak)
{
	for (int i = 0; i < frames; ++i)
	{
		register d_sample p = data[i] * factor;
		data[i] = p;

		factor += rate;

		if (p > *peak)
			*peak = p;
	}
}

/* //////////////////////////////////////////////////////////////////////// */

Gain::Gain (Type * t, d_sample g, bool p)
: Plugin (t, Audio)
{
	set_object_name ("Gain");

	state =
	factor = g;
	peak = 0.0;
	do_peak = p;
}

void
Gain::init (int n_outs)
{
	if (outs)
		throw AssertionError() << "cannot do this after instantiation.";

	add_in (new In (this, CyclicAudio_type, IN_SLOTS, "dry"));
	add_in (new In (this, FloatEvent_type, 1, "factor"));

	for (int i = 0; i < n_outs; ++i)
	{
		Out * out = add_out (new Out (this, CyclicAudio_type));
		out->set_object_name ("wet %d", i);
	}
}

void
Gain::end_cycle (CyclicAudio * mix)
{
	for (int i = 1; i < outs; ++i)
	{
		Out * out = outs[i];
		
		if (!out->connected || !out->is_audio)
			continue;

		CyclicAudio * e = (CyclicAudio *) outs[i]->cache->pop();

		*e = *mix;

		out->push (e);
	}

	outs[0]->push (mix);
}

void
Gain::process_audio (int frames)
{
	CyclicAudio * mix = ins[0]->pop_mix();

	if (mix == 0)
		return;
	
	d_frame f = graph->frame;
	d_tick tick;

	d_sample * data = mix->data;

	while (frames > 0)
	{
		int n = min (64, frames);

		bool is_loop = graph->is_loop_frame (f, f + n);

		tick = graph->frame2tick (f + n);

		ins[1]->pop_value (tick, &state);

		d_sample rate = (state - factor) / n;

		/* apply */
		if (do_peak)
		{
			if (state == factor)
				apply_factor_peak (factor, data, n, &peak);
			else
				apply_factor_rate_peak (factor, data, n, rate, &peak);
		}
		else
		{
			if (state == factor)
				apply_factor (factor, data, n);
			else
				apply_factor_rate (factor, data, n, rate);
		}
		
		factor = state;

		frames -= n;
		data += n;
		f += n;

		if (is_loop)
			++loop_cycle,
			f -= graph->loop.frame[1] - graph->loop.frame[0];
	}

	end_cycle (mix);
}

/* //////////////////////////////////////////////////////////////////////// */

Type * StereoGain_type;

StereoGain::StereoGain()
: Gain (StereoGain_type)
{
}

void
StereoGain::init()
{
	add_in (new In (this, CyclicAudio_type, IN_SLOTS, "in:left"));
	add_in (new In (this, CyclicAudio_type, IN_SLOTS, "in:right"));

	add_out (new Out (this, CyclicAudio_type, "out:left"));
	add_out (new Out (this, CyclicAudio_type, "out:right"));
}

void
StereoGain::process_audio (int frames)
{
	CyclicAudio * l = ins[0]->pop_mix();
	CyclicAudio * r = ins[1]->pop_mix();

	if (l)
	{
		apply_factor (factor, l->data, frames);
		outs[0]->push (l);
	}
	
	if (r)
	{
		apply_factor (factor, r->data, frames);
		outs[1]->push (r);
	}
}

/* //////////////////////////////////////////////////////////////////////// */

#define KFRAMES 64

ADSR::ADSR()
: Gain (ADSR_type, 0.0, 0)
{
	set_object_name ("ADSR");

	phase = Off;
	rate = 0.0;
	state = 0.0;

	a = 0.02;
	d = 0.1;
	s = 0.7;
	sr = 24.0;
	r = 0.1;

	Gain::init (1);

	notes_in = add_in (new In (this, MidiEvent_type, 1, "notes in"));
	notes_out = add_out (new Out (this, MidiEvent_type, "notes out"));
	cv_out = add_out (new Out (this, FloatEvent_type, "cv"));
}

void
ADSR::process_audio (int frames)
{
	CyclicAudio * mix = ins[0]->pop_mix();

	if (mix == 0)
		return;

	d_sample * data = mix->data;

	d_frame f = graph->frame;
	d_tick tick;

	while (frames > 0)
	{
		int n = min (64, frames);

		bool is_loop = graph->is_loop_frame (f, f + n);

		tick = graph->frame2tick (f + n);
		d_time delta = (d_time) n / (d_time) graph->sample_rate;

		one_cycle (data, n, delta, tick);

		frames -= n;
		data += n;
		f += n;

		if (is_loop)
		{
			++loop_cycle;

			f -= graph->loop.frame[1] - graph->loop.frame[0];

			/* all notes off ? */
			if (phase != Off)
				phase = Release;
		}
	}

	Gain::end_cycle (mix);
}

void
ADSR::one_cycle (d_sample * data, int frames, d_time delta, d_tick tick)
{
	bool trigger = 0;

	MidiEvent * e;
	while ((e = (MidiEvent *) notes_in->pop_before (tick)))
	{
		if (e->is_note_off())
		{
			notes.note_off (e->data[1]);
		}
		else if (e->is_note_on())
		{
			trigger = 1;
			notes.note_on (e->data[1]);
			velocity = (d_sample) e->data[2] / 127.;
		}

		// msg(), tick, e->tick, e->data[0], e->data[1] 
		//			<< trigger, notes.sounding;
		notes_out->push (e);
	}

	if (trigger && notes.sounding)
	{
		phase = Attack;
		rate = a ? 1 / a : 1;
	}
	else if (phase >= Attack && phase <= Sustain && notes.sounding == 0)
	{
		phase = Release;
		rate = r ? 1 / r : 1;
	}

	/* */

	if (phase == Attack)
	{
		d_sample st = state + rate * delta;

		if (st < 1)
			state = st;
		else
		{
			delta -= (1 - state) / rate;

			phase = Decay;
			state = 1;
			rate = d ? 1 / d : 1;
		}
	}
			
	if (phase == Decay)
	{
		state -= rate * delta;

		if (state < s)
		{
			phase = Sustain;
			state = s;
			rate = sr ? 1 / sr : 0;
		}
	}
	
	if (phase == Sustain)
	{
		state -= rate * delta;

		if (state < 0)
		{
			phase = Off;
			state = 0;
			rate = 0;
		}
	}
	
	else if (phase == Release)
	{
		state -= rate * delta;
		
		if (state <= 0)
		{
			phase = Off;
			state = 0;
			rate = 0;
		}
	}

	/* apply gain. */
	d_sample next_factor;
	
	if (phase == Attack)
		next_factor = velocity * (1 - (1 - state) * (1 - state));
	else
		next_factor = state * velocity;

	d_sample factor_rate = (next_factor - factor) / frames;

	if (do_peak)
		apply_factor_rate_peak (factor, data, frames, factor_rate, &peak);
	else
		apply_factor_rate (factor, data, frames, factor_rate);
	
	/* send state to the control voltage output. */
	if (factor != state && cv_out->connected)
	{
		FloatEvent * cv = (FloatEvent *) cv_out->cache->pop();

		cv->tick = tick;
		cv->value = state;
		
		cv_out->push (cv);
	}

	factor = next_factor;
}

/* //////////////////////////////////////////////////////////////////////// */

static int
Gain_set_db (Gain * g, PyObject * value)
Py_TRY
{
	d_float db, factor;

	if (!PyArg_Parse (value, "d", &db))
		throw blewit();

	factor = pow (10., db / 20.);

	g->set_factor (factor);

	return 0;
}
Py_CATCH (-1);

static PyObject *
Gain_get_db (Gain * g)
{
	return PyFloat (20 * log10 (g->get_factor()));
}

static int
Gain_set_factor (Gain * g, PyObject * value)
Py_TRY
{
	d_sample factor;

	if (!PyArg_Parse (value, "f", &factor))
		throw blewit();

	g->set_factor (factor);

	return 0;
}
Py_CATCH (-1);

static PyObject *
Gain_get_factor (Gain * g)
{
	return PyFloat (g->get_factor());
}

static PyObject *
Gain_get_peak (Gain * g)
{
	d_sample peak = g->peak;
	g->peak = 0;

	return PyFloat (peak);
}

static int
Gain_set_do_peak (Gain * g, PyObject * do_peak)
Py_TRY
{
	if (!PyArg_Parse (do_peak, "i", &g->do_peak))
		throw blewit();

	return 0;
}
Py_CATCH (-1);

static PyObject *
Gain_get_do_peak (Gain * g)
{
	return PyInt (g->do_peak);
}

static PyGetSetDef 
Gain_getsets [] = {
	{"peak", (getter) Gain_get_peak, 0,
	 "the signal peak value so far. The value is reset to 0 upon reading."},
	{"do_peak", (getter) Gain_get_do_peak, (setter) Gain_set_do_peak,
	 "whether peak values are calculated."},
	{"db", (getter) Gain_get_db, (setter) Gain_set_db,
	 "the amplification factor expressed as gain, in dB."},
	{"factor", (getter) Gain_get_factor, (setter) Gain_set_factor,
	 "the amplification factor; 1.0 is unity. The factor value is "
	 "computed from a dB gain value like this: "
	 "def db2factor (db): return pow (10., db / 20.)"},
	{ }
};

Type * Gain_type = 0;

char * Py<Gain>::__doc__ = 
"Gain ([db = 0.0, n_outs = 1, do_peak = 1]) -- "
"an audio plugin that applies amplification and optionally calculates "
"the peak signal value after applying gain.";

void
Py<Gain>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = Gain_getsets;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
}

void
Py<Gain>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"db", "n_outs", "do_peak", 0};
	double db = 0.;
	int n_outs = 1;
	int _do_peak = 1;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|dii", keys,
				&db, &n_outs, &_do_peak))
		throw blewit();

	if (n_outs < 1)
		throw ValueError() << "too few outs.";
	if (n_outs > 1024)
		throw ValueError() << "too many outs.";
	
	init (n_outs);

	factor = pow (10., db / 20.);

	set_factor (factor);
	do_peak = _do_peak;
}

/* //////////////////////////////////////////////////////////////////////// */

char * Py<StereoGain>::__doc__ = 
"StereoGain ([db = 0, do_peak = 0]) -- "
"applies gain to two channels of audio.";

void
Py<StereoGain>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Gain_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
}

void
Py<StereoGain>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"db", "do_peak", 0};
	double db = 0.;
	int _do_peak = 1;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|di", keys,
				&db, &_do_peak))
		throw blewit();

	init();

	factor = pow (10., db / 20.);

	set_factor (factor);
	do_peak = _do_peak;
}

/* ////////////////////////////////////////////////////////////////////// */

/* A */

static int
ADSR_set_a (ADSR * adsr, PyObject * value)
{
	if (!PyArg_Parse (value, "f", &adsr->a))
		return -1;
	return 0;
}

static PyObject *
ADSR_get_a (ADSR * adsr)
{
	return PyFloat (adsr->a);
}

/* D */

static int
ADSR_set_d (ADSR * adsr, PyObject * value)
{
	if (!PyArg_Parse (value, "f", &adsr->d))
		return -1;
	return 0;
}

static PyObject *
ADSR_get_d (ADSR * adsr)
{
	return PyFloat (adsr->d);
}

/* S */

static int
ADSR_set_s (ADSR * adsr, PyObject * value)
{
	if (!PyArg_Parse (value, "f", &adsr->s))
		return -1;
	return 0;
}

static PyObject *
ADSR_get_s (ADSR * adsr)
{
	return PyFloat (adsr->s);
}

/* SR */

static int
ADSR_set_sr (ADSR * adsr, PyObject * value)
{
	if (!PyArg_Parse (value, "f", &adsr->sr))
		return -1;
	return 0;
}

static PyObject *
ADSR_get_sr (ADSR * adsr)
{
	return PyFloat (adsr->sr);
}

/* R */

static int
ADSR_set_r (ADSR * adsr, PyObject * value)
{
	if (!PyArg_Parse (value, "f", &adsr->r))
		return -1;
	return 0;
}

static PyObject *
ADSR_get_r (ADSR * adsr)
{
	return PyFloat (adsr->r);
}

static PyGetSetDef 
ADSR_getsets [] = {
	{"a", (getter) ADSR_get_a, (setter) ADSR_set_a,
	 "'attack' is the time until full amplification."},
	{"d", (getter) ADSR_get_d, (setter) ADSR_set_d,
	 "'decay' is the time from full gain until sustain level."},
	{"s", (getter) ADSR_get_s, (setter) ADSR_set_s,
	 "'sustain' is the amplification factor at which the decay phase ends."},
	{"sr", (getter) ADSR_get_sr, (setter) ADSR_set_sr,
	 "'sustain rate' is the time until sustain amplification is halved."},
	{"r", (getter) ADSR_get_r, (setter) ADSR_set_r,
	 "'release' is the time from note off to zero amplification."},
	{ }
};

Type * ADSR_type = 0;

char * Py<ADSR>::__doc__ = 
"ADSR ([a = 0.02, d = 0.1, s = 0.7, sr = 24.0, r = 0.1]) -- "
"a simple Attack-Decay-Sustain-Release based audio gain. The ADSR responds "
"to MidiEvents, which are routed through. The amplification factor "
"ramps linearly (= gain decays exponentially) during all phases except attack, "
"where a square mapping combined "
"with linear interpolation is used. In addition to being "
"applied to an audio stream routed through the plugin, "
"the current factor is also sent on a third output port. ";

void
Py<ADSR>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Gain_type;
	t->tp_getset = ADSR_getsets;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
}

void
Py<ADSR>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"a", "d", "s", "sr", "r", 0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|fffff", keys,
				&a, &d, &s, &sr, &r))
		throw blewit();
}

/* ////////////////////////////////////////////////////////////////////// */

void
init_Gain_types (PyObject * dict)
{
	Gain_type = new PyType<Gain> ("Gain", dict);
	StereoGain_type = new PyType<StereoGain> ("StereoGain", dict);
	ADSR_type = new PyType<ADSR> ("ADSR", dict);
}
