/* Copyright 2003 tim goetze <tim@quitte.de> -- see 'COPYING.gz'. */

#include "basics.h"

static inline float
frandom()
{
  return (float) random() / (float) RAND_MAX;
} 
  
#include "dsp/Sine.h"
#include "dsp/Granular.h"

#include "Graph.h"
#include "PyType.h"

class Granular
: public NamedObject,
	public DSP::Granular
{
	public:
		AudioSignal * src;
		
		Granular()
			{
				src = 0;
				fs = SAMPLE_RATE;
			}

		~Granular()
			{
				::decref (src, true);
			}

		void init (double _fs, AudioSignal * s);

		void one_cycle (d_sample * s, int frames);
};

/* /////////////////////////////////////////////////////////////////////// */

void
Granular::init (double _fs, AudioSignal * s)
{
	::incref (src = s);
	((DSP::Granular *) this)->init (_fs, s->data, s->size);
}

void
Granular::one_cycle (d_sample * s, int frames)
{
	memset (s, 0, sizeof (d_sample) * frames);
	((DSP::Granular *) this)->one_cycle (s, frames);
}

/* //////////////////////////////////////////////////////////////////////// */

Type * Granular_type;

static PyObject *
Granular_get_gain (Granular * g)
{
	return PyFloat (g->gain);
}

static int
Granular_set_gain (Granular * g, PyObject * value)
{
	if (!PyArg_Parse (value, "f", &g->gain))
		return -1;
	return 0;
}

static PyObject *
Granular_get_gain_jitter (Granular * g)
{
	return PyFloat (g->gain_jitter);
}

static int
Granular_set_gain_jitter (Granular * g, PyObject * value)
{
	if (!PyArg_Parse (value, "f", &g->gain_jitter))
		return -1;
	return 0;
}

/* envelope */
static PyObject *
Granular_get_fade (Granular * g)
{
	return PyFloat (g->fade / g->fs);
}

static int
Granular_set_fade (Granular * g, PyObject * value)
{
	double v;

	if (!PyArg_Parse (value, "d", &v))
		return -1;

	g->fade = (int) (v * g->fs);
	return 0;
}

static PyObject *
Granular_get_body (Granular * g)
{
	return PyFloat (g->body / g->fs);
}

static int
Granular_set_body (Granular * g, PyObject * value)
{
	double v;

	if (!PyArg_Parse (value, "d", &v))
		return -1;

	g->body = (int) (v * g->fs);
	return 0;
}

static PyObject *
Granular_get_step (Granular * g)
{
	return PyFloat (g->step / g->fs);
}

static int
Granular_set_step (Granular * g, PyObject * value)
{
	double v;

	if (!PyArg_Parse (value, "d", &v))
		return -1;

	g->step = (int) (v * g->fs);
	return 0;
}

static PyObject *
Granular_get_jitter (Granular * g)
{
	return PyFloat (g->jitter);
}

static int
Granular_set_jitter (Granular * g, PyObject * value)
{
	double v;

	if (!PyArg_Parse (value, "d", &v))
		return -1;

	g->jitter = fmod (v, 1);
	return 0;
}

static PyGetSetDef 
Granular_getsets [] =
{
	{"gain", (getter) Granular_get_gain, (setter) Granular_set_gain, 
	 "linear gain control."},
	{"gain_jitter", (getter) Granular_get_gain_jitter, (setter) Granular_set_gain_jitter, 
	 "gain jitter control (max. attenuation)."},
	{"fade", (getter) Granular_get_fade, (setter) Granular_set_fade, 
	 "envelope fadein and ~out length, s."},
	{"body", (getter) Granular_get_body, (setter) Granular_set_body, 
	 "envelope body length, s."},
	{"step", (getter) Granular_get_step, (setter) Granular_set_step, 
	 "time between two consecutiv grains, s."},
	{"jitter", (getter) Granular_get_jitter, (setter) Granular_set_jitter, 
	 "sample positional jitter, [0 .. 1]."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
Granular_one_shot (Granular * g, AudioSignal * s)
{
	if (!AudioSignal_type->is_derived (s))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (s);
	
	g->one_cycle (s->data, s->size);

	return ::incref (s);
}

static PyMethodDef
Granular_methods [] =
{
	{"one_shot", (PyCFunction) Granular_one_shot, METH_O,
	 "one_shot (signal) -- in-place process an AudioSignal."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<Granular>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = NamedObject_type;
	t->tp_getset = Granular_getsets;
	t->tp_methods = Granular_methods;
}

char * Py<Granular>::__doc__ = 
"Granular (audio signal[, gain = 1]) -- granular oscillator.";

void
Py<Granular>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"src", "fs", 0};

	AudioSignal * src = 0;
	double fs = SAMPLE_RATE;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!|d", keys, 
				AudioSignal_type, &src, &fs))
		return;

	init (fs, src);
}

void
init_Granular_types (PyObject * dict)
{
	Granular_type = new PyType<Granular> ("Granular", dict);
}
