/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "Graph.h"
#include "GraphRequests.h"
#include "TempoMap.h"
#include "AudioClock.h"
#include "PythonThread.h"

GraphRequest::GraphRequest (Graph * g, Mutex * m, Type * t)
: NamedObject (t)
{
	graph = g;
	mutex = m;

	blown = 0;
	need_pre_post = 0;
}

/* ////////////////////////////////////////////////////////////////////// */

Graph::Graph()
: Object (Graph_type),
	Thread ("Graph"),
	plugins ("plugins"),
	poll (get_type_name (this)),
	pipe ("pipe", this)
{
	msg(), "this=PyO:", (PyObject *) this;
	msg(), "this=O:", (Object *) this;
	msg(), "this=G:", (Graph *) this;

	state = Plugin::Off;

	requests = new GraphRequestFifo;

	time_map = 0;

	audio_clock = 0;
	rt_clock = 0;
	disk_scheduler = 0;
	script_scheduler = 0;

	quarter = QUARTER;
	sample_rate = SAMPLE_RATE;
	frames_per_cycle = FRAMES_PER_CYCLE;

	audio_cycle_time = (d_time) frames_per_cycle / (d_time) sample_rate;
	cpu = 0;

	/* set all clocks */
	frame = 0;
	
	tick = 0;
	time = 0.0;
	beat_tick = 0;
	beat_time = 0;

	next_clock = 0;

	virtual_tick = 0;
	virtual_time = 0.0;

	memset (&loop, 0, sizeof (loop));

	skew = 1.0;

	silence = new CyclicAudio();

	quitting = 0;

	poll.add (pipe);

	start_thread();
}

Graph::~Graph()
{
#if 1
	if (state >= Plugin::Activated)
	{
		GraphRequest * r = new GraphStateChange (this, Plugin::Off);
		try 
		{
			single_request (r);
		} 
		catch (blewit & b) 
		{
			msg(), "could not stop: ", b.what(), 
					" -- continuing graph shutdown nonetheless.";	
		}
	}
#endif
	quitting = 1;
	pipe.write_trigger (Quit);

	join_thread();

	if (audio_clock)
	{
		ScopedLock lock (((AudioClock *) audio_clock)->mutex);
		try 
		{
			audio_clock->set_state (this, Plugin::Off);
		} 
		catch (blewit & b)
		{
			msg(), "ignoring: ", b.what();
		}
		audio_clock->graph = 0;
		audio_clock = 0;
	}

	/* disconnect all */
	for (int i = 0; i < plugins; ++i)
	{
		OutList & outs = plugins[i]->outs;

		for (int j = 0; j < outs; ++j)
			outs[j]->connect (0);

		plugins[i]->graph = 0;
	}

	plugins.clear();

	::decref (silence, true);
	::decref (time_map, true);

	delete requests;
}

void
Graph::set_state (int s)
{
	if (state == s)
		return;
	
	if (s == Plugin::Off) 
		;
	else if (s == Plugin::Prepared)
	{
		if (frames_per_cycle <= 0 || sample_rate <= 0)
			throw AssertionError() << "invalid 'frames_per_cycle' or 'sample_rate'.";

		if (state != Plugin::Off)
			throw ValueError() << "can only prepare in stopped state.";

		audio_cycle_time = (d_time) frames_per_cycle / (d_time) sample_rate;

		for (int i = 0; i < plugins; ++i)
			plugins[i]->set_frames_per_cycle (frames_per_cycle);

		/* the silent CyclicAudio */
		silence->set_size (frames_per_cycle);
	}
	else if (s == Plugin::Activated)
	{
		if (state < Plugin::Prepared)
			throw ValueError() << "can not activate from stopped state.";

		/* advancing 'virtual' time in Activated state. */
		virtual_tick = tick;
		virtual_time = time;

		tempo_map.recalculate (quarter);
	}
	else if (s == Plugin::Rolling)
	{
		if (state != Plugin::Activated)
			throw ValueError() << "can only roll from activated state.";

		/* Since in Activated state clock signals are emitted too, the current
		 * beat position of devices synced to the clock can be anything. 
		 * Let's make MIDI clock signals match a full quarter when restarting. 
		 */
		set_next_clock();

		msg(), "tick, next_clock: ", tick, next_clock;
	}

	/* for all except audio clock. */
	for (int i = 0; i < plugins; i++)
	{
		Plugin * p = plugins[i];
		
		if (p == audio_clock)
			continue;
			
		p->set_state (this, s);
		if (p->state != s)
			msg(), p, "won't set state to ", s;
	}
	
	if (audio_clock)
		audio_clock->set_state (this, s);

	state = s;
}

void
Graph::set_next_clock()
{
	d_tick clock = quarter / 24;
	d_tick nc = tick - fmod (tick, clock); 

	/* This provokes a burst of clock signals by shifting next_clock backwards
	 * in time, but it's still the better solution if an arpeggiator is synced 
	 * to our output -- the arpeggiator would play in a shifted beat, and that 
	 * sounds plain weird, mostly. 
	 */
	int clocks = 24 - (int) fmod ((next_clock - nc) / clock, 24);
	next_clock = nc - clocks * clock;
}

void
Graph::seek (d_time t)
{
	if (time == t)
		return;

	time = t;
	frame = time2frame (time);
	tick = time2tick (t);
	beat_tick = tick;
	beat_time = t;

	tempo_map.seek_tick (tick);

	set_next_clock();
	
	loop.done.beat = 0;
	loop.done.audio = 0;
	loop.done.time = 0;
	loop.cycles = 0;

	for (int i = 0; i < plugins; i++)
	{
		Plugin * p = plugins[i];

		p->seek();
		p->loop_cycle = 0;

		/* clear all queued events from the plugin outputs. */
		for (int i = 0; i < p->outs; ++i)
		{
			Out * out = p->outs[i];

			Object * o;

			while ((o = out->pop()))
				::decref (o);
		}
	}
	
	/* preroll the beat cycle */
	trigger_beat();
}

void
Graph::loop_done (int flag)
{
	if (flag & Plugin::Audio || !audio_clock)
		loop.done.audio = 1;
	if (flag & Plugin::Time || !rt_clock)
		loop.done.time = 1;
	if (flag & Plugin::Beat || !rt_clock)
		loop.done.beat = 1;
	
	if (!(loop.done.audio && loop.done.time && loop.done.beat))
		return;

	loop.done.audio =
	loop.done.beat = 
	loop.done.time = 0;

#if 0
	msg(), "all loops done. cycles:", loop.cycles;
#endif

	/* in case somebody forgot to switch the box off ... */
	if (++loop.cycles >= 0x7FFFFFFF)
		loop.use = 0;

	/* a good point in time to set a new loop is now. */
	if (loop.queued)
	{
		set_loop (loop.next[0], loop.next[1], true);
		loop.queued = 0;
	}
}

void
Graph::set_loop (d_tick t0, d_tick t1, bool force)
{
	if (!force && cannot_set_loop_now())
	{
		msg(), "enqueuing loop request ", t0, t1;

		loop.next[0] = t0;
		loop.next[1] = t1;
		loop.queued = true;

		return;
	}

	loop.tick[0] = t0;
	loop.tick[1] = t1;
	
	loop.time[0] = tick2time (t0);
	loop.time[1] = tick2time (t1);

	loop.frame[0] = time2frame (loop.time[0]);
	loop.frame[1] = time2frame (loop.time[1]);
}

/* ////////////////////////////////////////////////////////////////////// */

void
Graph::process_audio (int frames)
{
	d_frame next = frame + frames;

	bool is_loop = 
			state >= Plugin::Rolling && is_loop_frame (frame, frame + frames);
			
	if (is_loop)
		next -= loop.frame[1] - loop.frame[0];

	if (state >= Plugin::Rolling && loop.done.time == loop.done.audio)
	{
		/* drift correction */
		d_time dt = frame2time (frame) - time;
		
		/* enforce a lower bound for the skew factor. paranoid safety measure. */
		skew = max (0.5, 1.0 + dt / 2.0); 
			
		/* debug: print drift statistics */
		if (0 && (frame % (2 << 11)) == 0)
			msg(), time, frame, "drift: ", dt, time2frame (time) - frame, 
					"skew: ", skew;
	}

	/* release events not popped from audio outs; necessary because invalid audio 
	 * connections are allowed (= silence), but CyclicAudio Outs have but one
	 * event in the cache. 
	 */
	for (int i = 0; i < plugins; i++)
	{
		Plugin * p = plugins[i];

		if (!p->is_audio())
			continue;

		for (int i = 0; i < p->outs; i++)
		{
			Out * out = p->outs[i];

			if (!out->is_audio)
				continue;

			CyclicAudio * e;
			while ((e = (CyclicAudio *) out->pop()))
			{
				#if 0
				msg(), "audio event left at ", get_name (p), ".", get_name (out);
				#endif
				e->return_to_cache();
			}
		}
	}

	/* iterate all plugins and call those that do audio. */
	double burnt = 0;
	d_time then = gettimeofday();
	for (int i = 0; i < plugins; i++)
	{
		Plugin * p = plugins[i];

		if (p->is_audio() && p->state >= Plugin::Prepared) try
		{
			p->process_audio (frames);
				
			d_time now = gettimeofday();
			d_time dt = now - then;

			burnt += dt;
			
			if (dt > p->cpu)
				p->cpu = dt;

			then = now;
		}
		catch (blewit & b)
		{
			msg(), "frame ", frame, p, "blew it: ", b.what();
			p->set_state (this, Plugin::Off);
		}
	}
	
	if (burnt > cpu)
		cpu = burnt;

	if (state >= Plugin::Rolling)
	{
		frame = next;
		
		if (is_loop)
			loop_done (Plugin::Audio);

		/* synthesize a pulse for non-audio. */
		if (rt_clock)
			rt_clock->wake_up();
	}
}

/* ////////////////////////////////////////////////////////////////////// */

/* called by the audio clock after processing is done to trigger the
 * execution of GraphRequests. 
 */

void
Graph::check_for_requests()
{
	if (!requests->valid)
		return;

	while (requests->next (GraphRequestFifo::TryLock))
		msg(), "request ", requests->peek();
}

/* ////////////////////////////////////////////////////////////////////// */

void
Graph::process_time (d_time dt, int clocks)
{
	/* adjust for audio clock versus system clock drift */
	dt *= skew;

	ScopedLock lock (mutex);

	d_time next_time = time;
	d_tick next_tick = tick;

	bool is_loop = 0;

	if (state >= Plugin::Rolling)
	{
		next_time = time + dt;
		next_tick = time2tick (next_time);
		
		is_loop = is_loop_time (time, next_time);
		if (is_loop)
			next_time -= loop.time[1] - loop.time[0];
	}
	else if (state == Plugin::Activated)
	{
		/* in activated state, we advance 'virtual' time and limit ourselves
		 * to the current tempo zone. */
		virtual_time += dt;
		virtual_tick = virtual_time2tick (time, virtual_time); 

		/* apply modulo quarter looping to virtual time */
		d_tick running = virtual_tick - tick;
		if (running > quarter)
		{
			running -= fmod (running, quarter);
#if 0
			msg(), "looping vtick ", virtual_tick, "by ", running;
#endif
			virtual_tick -= running;
			next_clock -= running;
			virtual_time = virtual_tick2time (tick, virtual_tick);
		}
		
		next_tick = virtual_tick;
	}

	/* determine the number of MIDI clocks for this interval */
	if (clocks < 0)
	{
		if (!finite (next_tick))
			throw AssertionError() << "next tick not finite";

		clocks = 0;
		while (next_clock < next_tick)
			++clocks,
			next_clock += quarter / 24;
	}

	for (int i = 0; i < plugins; i++)
	{
		Plugin * p = plugins[i];

		if (!p->is_time())
			continue;
		
		p->process_time (dt, clocks);
	}
	
	time = next_time;
	tick = time2tick (time);
	
	if (is_loop)
	{
		set_next_clock();
		loop_done (Plugin::Time);
	}

	/* trigger process_beat */
	if (state >= Plugin::Activated)
	{
		if (loop.done.time == loop.done.beat)
		{
			/* allow 1/10th of a second at max */
			if (beat_time - time < .1)
				trigger_beat();
		}
		else /* beat has already looped */
		{
			d_time t = time - (loop.time[1] - loop.time[0]);
			/* allow 1/10th of a second at max */
			if (beat_time - t < .1)
				trigger_beat();
		}
	}
}

bool 
Graph::cannot_set_loop_now()
{
	/* we cannot modify the loop points if any one cycle has already
	 * looped. */
	return 
		loop.done.audio || loop.done.time || loop.done.beat;
}

void
Graph::tempo_changed()
{
	time = tick2time (tick);
	beat_time = tick2time (beat_tick);
	frame = time2frame (time);

	if (state < Plugin::Rolling)
		virtual_time = virtual_tick2time (tick, virtual_tick);
}

inline d_tick 
Graph::get_beat_cycle_length()
{
	return quarter / 8;
}

void
Graph::trigger_beat()
{
	pipe.write_trigger (ProcessBeat);
}

void
Graph::process_beat()
{
	if (state < Plugin::Activated)
		return;

	d_tick dt = get_beat_cycle_length();
	d_tick until = beat_tick + dt;

	/*
	if (loop.done.time == loop.done.beat)
	{
		if (until > tick + dt)
			until = floorf (tick + dt);
	}
	*/
	if (until < beat_tick)
		return;

	bool is_loop = is_loop_tick (beat_tick, until);

#if 0
	msg(), "process_beat @ ", tick, "( ", state, ")", beat_tick, "-> ", until, 
	 		(char *) (is_loop ? "is loop" : ""),
			"delta = ", beat_tick - tick;
#endif

	ScopedLock lock (mutex);

	for (int i = 0; i < plugins; i++)
	{
		Plugin * p = plugins[i];

		if (p->state == Plugin::Off || !p->is_beat())
			continue;
			
		try 
		{
			p->process_beat (until);
		} 
		catch (blewit & b) 
		{
			msg(), p, "process_beat ", until, " blew it: ", b.what();
			p->state = Plugin::Off;
		}
	}

	if (is_loop)
		until -= loop.tick[1] - loop.tick[0];

	beat_tick = until;
	beat_time = tick2time (beat_tick);

	if (is_loop)
		loop_done (Plugin::Beat);
}

/* called by a disk scheduler to have a plugin's process_disk() executed.
 * process_script works identically.
 *
 * the plugin address is verified and the graph mutex is released before
 * the plugin's process_disk() is executed -- we don't want the graph mutex
 * locked while doing blocking IO or waiting for the PL. 
 */
void
Graph::process_disk (Plugin * p)
{
	mutex.lock();

	/* the request for process_disk was posted sometime ago, we now need to 
	 * check whether the plugin is still active, ie is part of this graph. 
	 */
	int i = 0;
	while (i < plugins && p != plugins[i])
		++i;
	
	/* not found. */
	if (i == plugins)
	{
		mutex.unlock();
		msg(), "disk plugin ", (void *) p, "not found.";
		return;
	}

	/* do we have to ::incref ourselves here too? */

	/* now we pin down the plugin, */
	p->mutex.lock();
	::incref (p);
		
	/* release the graph mutex, */
	mutex.unlock();

	/* process the plugin, */
	p->process_disk (p);
	p->mutex.unlock();

	/* and release it. */
	::decref (p);
}

void
Graph::process_script (Plugin * p)
{
	mutex.lock();

	/* the request for process_script was posted sometime ago, we now need to 
	 * check whether the plugin is still active, ie is part of this graph. 
	 */
	int i = 0;
	while (i < plugins && p != plugins[i])
		++i;
	
	/* not found. */
	if (i == plugins)
	{
		mutex.unlock();
		msg(), "script plugin ", (void *) p, "not found.";
		return;
	}

	/* now we pin down the plugin, */
	p->mutex.lock();
	::incref (p);
		
	/* release the graph mutex, */
	mutex.unlock();

	d_time then = gettimeofday();

	/* process the plugin, */
	p->process_script (p);
	p->mutex.unlock();

	d_time cpu = gettimeofday() - then;
	if (cpu > p->cpu)
		p->cpu = cpu;

	/* and release it. */
	::decref (p);
}

/* three plugin list maintenance methods, only called by GraphRequests. */
void
Graph::add_plugin (Plugin * p)
{
	if (AudioClock_type->is_derived (p) && audio_clock)
		throw AssertionError() << "an audio clock is already present.";

	if (p->is_disk() && !disk_scheduler)
		throw AssertionError() << "add a DiskScheduler first.";

	int i;
	for (i = 0; i < plugins; i++)
		if (p->before (plugins[i]))
			break;
	
	plugins.add_object (p, i);
	p->graph = this;

	if (AudioClock_type->is_derived (p))
		audio_clock = p;
	
	if (get_type (p) == DiskScheduler_type)
		disk_scheduler = p;

	if (get_type (p) == ScriptScheduler_type)
		script_scheduler = p;

	if (get_type (p) == RTClock_type)
		rt_clock = p;

	p->set_frames_per_cycle (frames_per_cycle);
}

void
Graph::remove_plugin (Plugin * p)
{
	if (p->graph != this) 
		throw LookupError() << "plugin is not known here.";

	if (p->count_connections()) 
		throw AssertionError() << "plugin is still connected.";

	/* */

	p->pre_remove();

	for (int i = 0; i < plugins; i++)
		plugins[i]->pre_remove_other (p);

	/* */

	if (AudioClock_type->is_derived (p))
		audio_clock = 0;

	if (get_type (p) == RTClock_type)
		rt_clock = 0;

	if (get_type (p) == ScriptScheduler_type)
		script_scheduler = 0;

	plugins.remove (p);
	p->graph = 0;
}

void
Graph::move_plugin (Plugin * p, d_float x, d_float y)
{
	if (p->graph != this) 
		throw LookupError() << "plugin is not known here.";

	/* todo: might add a check for connection validity here. */

	plugins.remove (p);

	p->x = x;
	p->y = y;

	int i;
	for (i = 0; i < plugins; i++)
		if (p->before (plugins[i]))
			break;
	
	plugins.add_object (p, i);
}

/* thread & thread communication */
void
Graph::signal_request()
{
	/* no audio clock or audio not running: execute in process_beat context. */
	if (!audio_clock || state < Plugin::Activated) 
		pipe.write_trigger (ProcessRequest);
}

void
Graph::single_request (GraphRequest * r)
{
	ScopedLock lock (requests->mutex);

	requests->push (r);

	try 
	{
		requests->post();
	}
	catch (blewit & b)
	{
		requests->clear();
		throw;
	}
	requests->clear();
}

void
Graph::beat_thread_request (int todo)
{
	try 
	{
		switch (todo)
		{
			case ProcessBeat:
				process_beat();
				break;

			case ProcessRequest:
				while (!audio_clock && requests->next (GraphRequestFifo::DoLock))
					;
				break;

			case CallPostSeek:
				call_method ("seek_complete", "()");
				break;

			case CallPostState:
				call_method ("state_changed", "()");
				break;
		}
	}
	catch (blewit & b)
	{
		_msg(), "beat_thread_request ", todo, "blew it: ", b.what();
	}
}

void *
Graph::thread()
{
	while (1)
	{
		int r = poll.poll (100);
		
		if (quitting) break;
		
		/* poll error */
		if (r == -1) 
		{
			if (errno == EINTR)
				continue;
			throw IOError (errno) << " in poll(2).";
		}
		
		/* hung up, inval etc: */
		if (poll[0].revents & ~POLLIN)
			throw IOError() << " on signal socket.";

		if (r == 0)
			continue;

		if (poll[0].revents & POLLIN)
			beat_thread_request (pipe.read_trigger());
	}
	return 0;
}


