/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _GRAPH_H_
#define _GRAPH_H_

#include "Thread.h"
#include "Plugin.h"
#include "TempoMap.h"
#include "Pollable.h"
#include "Pipe.h"

/* we don't need to know more here, so declare it a class * only: */
class TimeMap;

#include "AudioSignal.h"

/* encapsulation of graph state changes */
class GraphRequest;
class GraphRequestFifo;

class Graph
: public Object,
	public Thread
{
	public:
		Graph();
		~Graph();
		
		TypedObjectList<Plugin> plugins;

		/* this mutex protects the plugin list and the tempo map. it is held 
		 * during the Time and Beat cycles and while fetching a plugin from
		 * the list for asynchronous Script/Beat/Disk processing. 
		 */
		Mutex mutex;

		/* GraphRequests obtain the lock and then signal they're waiting. if
		 * an AudioClock is present, the request is carried out in audio
		 * thread context; otherwise the Beat thread will step in.
		 */
		GraphRequestFifo * requests;

		/* signal that there are GraphRequests waiting */
		void signal_request();			
		/* execute a single request, called in audio or beat thread, see above. */
		void single_request (GraphRequest *);
		/* check for requests, called by the audio clock. */
		void check_for_requests();
		
		/* one cycle worth of silence. */		
		CyclicAudio * silence; 

		/* current audio parameters, usually set by an audio clock. */
		int frames_per_cycle;
		void set_frames_per_cycle (int);

		int sample_rate;
		void set_sample_rate (int);
		
		/* the (estimated) time one audio cycle takes. This is only valid while
		 * state is >= 1 (Prepared). 
		 */
		d_time audio_cycle_time;
		double cpu; /* max cpu usage in audio cycle */

		/* how many midi ticks in a quarter note. this is the same as a 
		 * MIDI file's 'division' parameter. 
		 */
		d_tick quarter;			

		TempoMap tempo_map;
		TimeMap * time_map;

		d_frame frame;				/* current audio frame */
		d_tick tick;					/* current midi tick */
		d_tick virtual_tick;	/* current virtual midi tick (while Activated) */
		d_tick next_clock;		/* next midi clock signal */
		d_tick beat_tick;			/* about .1 s ahead of tick */
		d_tick beat_time;			/* about .1 s ahead of time */
		d_time time;					/* current song time */
		d_time virtual_time;	/* current song time */
		d_time skew;					/* correction for clock drift */

		/* loop info */
		struct {
			bool use, queued;
			
			/* the next loop points to use, if 'queued'. */
			d_tick next[2];

			d_tick tick[2];
			d_time time[2];
			d_frame frame[2];
			
			struct {
				int beat;
				int time;
				int audio;
			} done;

			/* how many times the loop has been run. */
			int cycles;
		} loop;
		
		bool cannot_set_loop_now();
		
		void set_loop (d_tick t0, d_tick t1, bool force = 0);

		bool is_loop_time (d_time t0, d_time t1)
			{
				return loop.use && t0 < loop.time[1] && t1 >= loop.time[1];
			}

		bool is_loop_tick (d_tick t0, d_tick t1)
			{
				return loop.use && t0 < loop.tick[1] && t1 >= loop.tick[1];
			}

		bool is_loop_frame (d_frame f0, d_frame f1)
			{
				return loop.use && f0 < loop.frame[1] && f1 >= loop.frame[1];
			}

		/* same meaning as Plugin::state, see Plugin.h for the details. */
		int state;				
		void set_state (int _state);
		
		/* these are plugins that provide special functionality; their addresses
		 * are cached here for plugins that need them to operate correctly. 
		 */
		Plugin * rt_clock;
		Plugin * audio_clock;
		Plugin * disk_scheduler;
		Plugin * script_scheduler;

		/* called by GraphRequests. */
		void add_plugin (Plugin *);
		void remove_plugin (Plugin *);
		void move_plugin (Plugin *, d_float x, d_float y);
		
		void seek (d_time time);

		/* the processing loops. */
		void process_audio (int frames);
		void process_time (d_time dt, int clocks);	
		void process_disk (Plugin * p);
		void process_script (Plugin * p);
				
		/* (tick, time) mappings: */
		inline d_tick time2tick (d_time t) 
			{ 
				return tempo_map.time2tick (t); 
			}
		
		inline d_time tick2time (d_tick t) 
			{ 
				return tempo_map.tick2time (t); 
			}
		
		inline d_tick virtual_time2tick (d_time zone, d_time t) 
			{ 
				return tempo_map.virtual_time2tick (zone, t); 
			}
		
		inline d_time virtual_tick2time (d_tick zone, d_tick t) 
			{ 
				return tempo_map.virtual_tick2time (zone, t); 
			}
		
		/* (frame, time) mappings: */
		inline d_time frame2time (d_frame f) 
			{ 
				return (d_time) (f) / (d_time) sample_rate;
			}

		inline d_frame time2frame (d_time t) 
			{ 
				return (d_frame) (t * (d_time) sample_rate);
			}

		/* (tick, frame) mappings: */
		inline d_tick frame2tick (d_frame f)
			{
				return time2tick (frame2time (f));
			}

		inline d_frame tick2frame (d_tick t)
			{
				return time2frame (tick2time (t));
			}

		/* called after tempo map changes */
		void tempo_changed();
		
	protected:
		/* sched_other plugins (ahead of time) */		
		void trigger_beat();		
		void process_beat();	
		virtual void * thread();
		void set_next_clock();

		void loop_done (int flag); /* Beat, Audio, Time */
		d_tick get_beat_cycle_length();
		
		enum {
			Quit = 0,
			ProcessBeat = 1,
			ProcessRequest = 2,
			CallPostSeek = 3,
			CallPostState = 4,
		};
		void beat_thread_request (int todo);
		
	protected:
		bool quitting;
		Poll poll;
		Pipe pipe;
};

/* these requests are constructed in one thread, and passed to the graph
 * to be executed in the audio thread.
 */
class GraphRequest
: public NamedObject
{
	public:
		Graph * graph;

		/* an additional mutex that may protect what we are trying to change,
		 * eg. a plugin mutex. 
		 */
		Mutex * mutex; 

		/* exceptions have to pass the thread border, too */
		blewit blown;	
		
	public:
		GraphRequest (Graph *, Mutex *, Type *);
		
		/* we need a default constructor to make Py<GraphRequest>. */
		GraphRequest()
			{ }

		virtual ~GraphRequest() 
			{ }

		/* if the request is executed within the audio thread, trylock() is
		 * to be called. otherwise, we may block waiting for a mutex, since 
		 * we are running in sched_other.
		 */
		virtual void lock() 
			{ 
				if (mutex) mutex->lock(); 
			}
		
		virtual bool trylock() 
			{ 
				if (mutex)
					return mutex->trylock();
				return true; 
			}

		/* this is supposed to do the real modification. */
		virtual void execute()
			{ }
		
		virtual void unlock() 
			{ 
				if (mutex) mutex->unlock(); 
			}
		
		/* if resources need be allocated for the request to complete (like
		 * inserting a plugin into a running graph), pre_post() should do that.
		 * 
		 * the request protocol will ensure this is done within the calling
		 * thread.
		 */
		atomic need_pre_post;
		virtual void pre_post() 
			{ }
};

#endif /* _GRAPH_H_ */
