/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _GRAPHREQUESTS_H_
#define _GRAPHREQUESTS_H_

#include "TempoMap.h"
#include "PythonThread.h"

/* we won't try locking forever, see GraphRequest::next() below. */
#define MAX_LOCK_TRIES	8

class GraphRequestFifo
: public TypedObjectFifo<GraphRequest>
{
	friend class Graph;

	private:
		atomic valid;
		int fd[2];

	public:
		Mutex mutex;
		int lock_tries;

		GraphRequestFifo() 
			: TypedObjectFifo<GraphRequest> (62, "graph requests")
			{
				/* prepair ;) communication sockets. */
				if (socketpair (AF_LOCAL, SOCK_STREAM, 0, fd) < 0)
					throw OSError (errno) << "cannot get a socketpair(2).";
			}

		~GraphRequestFifo() 
			{
				close (fd[0]);
				close (fd[1]);
			}

		/* executes all requests. if a request throws an exception, the execution
		 * is stopped and the exception is passed on.
		 */
		void post()
			{
				if (have() == 0)
					throw AssertionError() << "nothing to do.";

				valid = 1;
				lock_tries = 0;

				while (valid && peek())
				{
					peek()->graph->signal_request();

					char sig;
					read (fd[1], &sig, sizeof (char));
					
					if (sig == PrePost)
					{
						peek()->pre_post();
						valid = 1;
					}
					else if (sig == UnLock)
					{
						/* there may be a thread having locked the graph (or another
						 * mutex) waiting.
						 */
						sched_yield();
					}
					else 
					{
						if (sig == Blown) 
							throw peek()->blown;

						/* sig == Done, we're through. */
						break;
					}
				}
			}

	private:
		/* inform the waiting thread of completion, exceptions or Plugin::pre_* 
		 * code needed to execute in non-rt context. */
		enum {
			Done = 'd',
			Blown = 'b',
			PrePost = 'p',
			UnLock = 'l'
		};
		
		void pass_signal (char sig)
			{
				write (fd[0], &sig, sizeof (char));
			}
		
		/* the Graph calls this, either from its sched_other thread or
		 * at the end of an audio chunk from check_for_requests.
		 * 
		 * mode is TryLock if we're in audio thread context, meaning that
		 * we cannot block waiting for a mutex. otherwise it's DoLock.
		 *
		 * returns true if there are more requests to be executed. */
		enum 
		{
			TryLock = 't',
			DoLock = 'l' 
		};

		bool next (int mode)
			{
				GraphRequest * r = peek();

				if (!valid || r == 0)
					return 0;

				/* may need code executed outside the audio thread to allocate
				 * resources first. */
				if (r->need_pre_post)
				{
#if TRACE_EXEC
					msg(), "pre_post: ", get_name (r);
#endif

					valid = 0;
					pass_signal (PrePost);

					return 0;
				}
				
				if (mode == DoLock)
					r->lock();
				else if (!r->trylock()) 
				{
					if (++lock_tries < MAX_LOCK_TRIES)
					{
						msg(), "can't lock ", lock_tries;
						pass_signal (UnLock);
						/* try later */
						return 0;
					}
					
					r->blown = RuntimeError() << "can't lock after " 
																		<< lock_tries << " attempts.";
					valid = 0;
					pass_signal (Blown);

					return 0; 
				}

#if TRACE_EXEC
				msg(), "exec: ", get_name (r);
#endif

				try 
				{
					r->execute();
				} 
				catch (blewit & b) 
				{
					r->unlock();
				
					/* r failed, stop processing. */
					r->blown = b;

					valid = 0;
					pass_signal (Blown);
				
					return 0;
				}
				
				r->unlock();

				/* r is done, on to the next. */
				lock_tries = 0;

				::decref (pop());

				if (peek())
					return true;

				/* that was the last one. */
				valid = 0;
				pass_signal (Done);

				return 0;
			}
};

/* ////////////////////////////////////////////////////////////////////// */

class GraphStateChange
: public GraphRequest
{
	public:
		int state;
		
		GraphStateChange()
			{ }

		GraphStateChange (Graph * g, int s)
			: GraphRequest (g, &g->mutex, GraphStateChange_type)
			{
				state = s;

				set_object_name ("set_state: %d", state);
			}
			
		virtual void execute()
			{
				graph->set_state (state);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphStateChange (graph = " << (void *) graph 
					<< ", state = " << state 
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class GraphAddPlugin
: public GraphRequest
{
	public:
		Plugin * plugin;

		GraphAddPlugin()
			{ 
				plugin = 0;
			}

		GraphAddPlugin (Graph * g, Plugin * p)
			: GraphRequest (g, &g->mutex, GraphAddPlugin_type)
			{
				::incref (plugin = p);

				need_pre_post = 1;

				set_object_name ("add_plugin: %s", get_type_name (p));
			}
			
		~GraphAddPlugin()
			{
				::decref (plugin);
			}

		virtual void pre_post()
			{
				plugin->pre_insert (graph);

				/* necessary for the plugin to allocate resources. */
				for (int i = 0; i <= graph->state; i++)
					plugin->set_state (graph, i);

				--need_pre_post;
			}

		virtual void execute()
			{
				graph->add_plugin (plugin);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphAddPlugin (graph = " << (void *) graph 
					<< ", plugin = " << (void *) plugin 
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class GraphRemovePlugin
: public GraphRequest
{
	public:
		Plugin * plugin;

		GraphRemovePlugin()
			{
				plugin = 0;
			}

		GraphRemovePlugin (Graph * g, Plugin * p)
			: GraphRequest (g, &g->mutex, GraphRemovePlugin_type)
			{
				::incref (plugin = p);

				set_object_name ("remove_plugin: %s", get_type_name (p));
			}
			
		~GraphRemovePlugin()
			{
				::decref (plugin);
			}

		virtual void execute()
			{
				graph->remove_plugin (plugin);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphRemovePlugin (graph = " << (void *) graph 
					<< ", plugin = " << (void *) plugin 
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class GraphMovePlugin
: public GraphRequest
{
	public:
		Plugin * plugin;
		d_float x, y;

		GraphMovePlugin()
			{
				plugin = 0;
				x = 0.0;
				y = 0.0;
			}

		GraphMovePlugin (Graph * g, Plugin * p, d_float _x, d_float _y)
			: GraphRequest (g, &g->mutex, GraphMovePlugin_type)
			{
				::incref (plugin = p);
				x = _x;
				y = _y;

				set_object_name ("move_plugin: %s %.2f %.2f", 
						get_type_name (p), x, y);
			}
			
		~GraphMovePlugin()
			{
				::decref (plugin);
			}

		virtual void execute()
			{
				graph->move_plugin (plugin, x, y);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphMovePlugin (graph = " << (void *) graph 
					<< ", plugin = " << (void *) plugin 
					<< ", x = " << x
					<< ", y = " << y
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class GraphConnect
: public GraphRequest
{
	public:
		Out * out;
		In * in; 

		GraphConnect()
			{
				out = 0;
				in = 0;
			}

		GraphConnect (Graph * g, Out * o, In * i)
			: GraphRequest (g, &o->mutex, GraphConnect_type)
			{
				::incref (out = o);
				::incref (in = i);

				if (i)
					set_object_name ("connect: %s:%s -> %s:%s", 
							get_type_name (o->plugin), 
							get_name (o), 
							get_type_name (i->plugin),
							get_name (i)
						);
				else
					set_object_name ("connect: %s:%s -> None", 
							get_type_name (o->plugin), 
							get_name (o)
						);
			}

		~GraphConnect()
			{
				::decref (out);
				::decref (in);
			}

		virtual void execute()
			{
				out->connect (in);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphConnect (graph = " << (void *) graph 
					<< ", out = " << (void *) out 
					<< ", in = " << (void *) in 
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class GraphStoreTempo
: public GraphRequest
{
	public:
		d_float bpm;
		d_tick tick;

		GraphStoreTempo()
			{ }

		GraphStoreTempo (Graph * g, d_float _bpm, d_tick _tick)
			: GraphRequest (g, &g->mutex, GraphStoreTempo_type)
			{
				bpm = _bpm;
				tick = _tick;

				set_object_name ("add_tempo: %.2f bpm at %d", bpm, tick);
			}
			
		virtual void execute()
			{
				graph->tempo_map.add (bpm, tick);
				graph->tempo_map.recalculate (graph->quarter);

				graph->tempo_changed();

				/* unconditionally recalculate the loop points. */
				graph->set_loop (graph->loop.tick[0], graph->loop.tick[1], true);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphStoreTempo (graph = " << (void *) graph 
					<< ", bpm = " << bpm
					<< ", tick = " << tick
					<< ")";
			}
};

class GraphRemoveTempo
: public GraphRequest
{
	public:
		d_tick tick;
		int n;
		
		GraphRemoveTempo()
			{ }

		GraphRemoveTempo (Graph * g, d_tick _tick, int _n = 1)
			: GraphRequest (g, &g->mutex, GraphRemoveTempo_type)
			{
				tick = _tick;
				n = _n;

				set_object_name ("remove_tempo: from %d, n = %d", tick, n);
			}
			
		virtual void execute()
			{
				graph->tempo_map.remove (tick, n);
				graph->tempo_map.recalculate (graph->quarter);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphRemoveTempo (graph = " << (void *) graph 
					<< ", tick = " << tick
					<< ", n = " << n
					<< ")";
			}
};

/* ////////////////////////////////////////////////////////////////////// */

class GraphSeek
: public GraphRequest
{
	public:
		d_time time;
		
		GraphSeek()
			{ }

		GraphSeek (Graph * g, d_time t)
			: GraphRequest (g, &g->mutex, GraphSeek_type)
			{
				time = t;

				set_object_name ("seek: %.3f", time);
			}
			
		virtual void execute()
			{
				graph->seek (time);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphSeek (graph = " << (void *) graph 
					<< ", time = " << time 
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class GraphSetQuarter
: public GraphRequest
{
	public:
		d_tick quarter;

		GraphSetQuarter()
			{ }

		GraphSetQuarter (Graph * g, d_tick q)
			: GraphRequest (g, &g->mutex, GraphSetQuarter_type)
			{
				quarter = q;

				set_object_name ("set_quarter: %.1f", quarter);
			}
			
		virtual void execute()
			{
				graph->quarter = quarter;
				graph->tempo_map.recalculate (graph->quarter);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphSetQuarter (graph = " << (void *) graph 
					<< ", quarter = " << quarter
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class GraphSetLoop
: public GraphRequest
{
	public:
		d_tick t0, t1;

		GraphSetLoop()
			{ }

		GraphSetLoop (Graph * g, d_tick _t0, d_tick _t1)
			: GraphRequest (g, &g->mutex, GraphSetLoop_type)
			{
				t0 = _t0;
				t1 = _t1;

				set_object_name ("set_loop: %.1f %.1f", t0, t1);
			}
			
		virtual void execute()
			{
				graph->set_loop (t0, t1);
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphSetLoop (graph = " << (void *) graph 
					<< ", t0 = " << t0
					<< ", t1 = " << t1
					<< ")";
			}
};

/* /////////////////////////////////////////////////////////////////////// */

class GraphSetUseLoop
: public GraphRequest
{
	public:
		bool use;

		GraphSetUseLoop()
			{ }

		GraphSetUseLoop (Graph * g, bool _use)
			: GraphRequest (g, &g->mutex, GraphSetUseLoop_type)
			{
				use = _use;

				set_object_name ("set_use_loop: %s", use ? "on" : "off");
			}
			
		virtual void execute()
			{
				if (graph->cannot_set_loop_now())
					throw ValueError() << "cannot set loop right now.";

				graph->loop.use = use;
			}

		virtual String & operator >> (String & s)
			{
				return s 
					<< "GraphSetUseLoop (graph = " << (void *) graph 
					<< ", use = " << (int) use
					<< ")";
			}
};

#endif /* _GRAPHREQUESTS_H_ */
