/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "Graph.h"
#include "PyType.h"

#include "dsp/RBJ.h"
#include "dsp/Eq.h"
#include "dsp/Zoelzer.h"
#include "dsp/complex.h"
#include "dsp/ComplexIIR.h"

static Type * IIR_type;

/* needs to be a power of two. */
#define SIZE 32

class IIR
: public Plugin
{
	public:
		/* order and coefficients */
		int n;
		double a[SIZE], b[SIZE];

		/* history */
		int h;
		double x[SIZE], y[SIZE];

		IIR();
		
		/* Plugin:: */
		virtual void set_state (Graph * g, int s);
		virtual void process_audio (int frames);

		inline void one_shot (d_sample * s, int frames);
	
		double dc_gain()
			{
				double A = 0, B = 0;

				for (int i = 0; i < n; ++i)
					A += a[i],
					B += b[i];

				return A / (1 - B);
			}

		double nyquist_gain()
			{
				double A = 0, B = 0;

				for (int i = 0; i < n; i += 2)
					A += a[i],
					B += b[i];
				for (int i = 1; i < n; i += 2)
					A -= a[i],
					B -= b[i];

				return A / (1 - B);
			}

	protected:
		inline int one_cycle (d_sample * s, int frames, int h, double * x, double * y);
		inline int one_cycle_3 (d_sample * s, int frames, int h, double * x, double * y);
};

IIR::IIR()
: Plugin (IIR_type, Audio)
{
	memset (a, 0, sizeof (a));
	memset (b, 0, sizeof (b));

	n = 1;
	a[0] = 1;

	add_in (new In (this, CyclicAudio_type, IN_SLOTS, "in"));
	add_out (new Out (this, CyclicAudio_type, "out"));
}

void
IIR::set_state (Graph * g, int state)
{
	if (state == Plugin::Prepared)
	{
		memset (x, 0, sizeof (x));
		memset (y, 0, sizeof (y));
		h = 0;
	}

	Plugin::set_state (g, state);
}

void
IIR::process_audio (int frames)
{
	CyclicAudio * event = ins[0]->pop_mix();
	
	if (event == 0)
		return;

	d_sample * s = event->data;
	
	if (n == 3)
		h = one_cycle_3 (s, frames, h, x, y);
	else
		h = one_cycle (s, frames, h, x, y);
	
	outs[0]->push (event);
}

int
IIR::one_cycle (d_sample * s, int frames, int h, double * x, double * y)
{
	for (int i = 0; i < frames; ++i)
	{
		register double out = s[i];
		
		x[h] = out;
		
		out *= a[0];

		for (int j = 1; j < n; ++j)
		{
			register int z = (h - j) & (SIZE - 1);
			out += a[j] * x[z];
			out += b[j] * y[z];
		}

		/* tried denormal elimination but it's counterproductive on avg. */
		s[i] = y[h] = out;

		h = (h + 1) & (SIZE - 1);
	}

	return h;
}

/* since this is such a common case, it gets its own implementation. */
int
IIR::one_cycle_3 (d_sample * s, int frames, int h, double * x, double * y)
{
	for (int i = 0; i < frames; ++i)
	{
		register int z = h;
		register double out = s[i];
		
		x[z] = out;
		
		out *= a[0];
		
		z = (z - 1) & (SIZE - 1);
		out += a[1] * x[z];
		out += b[1] * y[z];

		z = (z - 1) & (SIZE - 1);
		out += a[2] * x[z];
		out += b[2] * y[z];

		s[i] = y[h] = out;

		h = (h + 1) & (SIZE - 1);
	}

	return h;
}

void 
IIR::one_shot (d_sample * s, int frames)
{
	double x[SIZE], y[SIZE];

	for (int i = 0; i < SIZE; ++i)
		x[i] = y[i] = 0.;

	if (n == 3)
		one_cycle_3 (s, frames, 0, x, y);
	else
		one_cycle (s, frames, 0, x, y);
}

/* /////////////////////////////////////////////////////////////////////// */

/* IIR getsets */

static PyObject *
IIR_get_ab (IIR * iir)
{
	PyObject * ab = PyTuple_New (2);
	
	PyObject * A = PyTuple_New (iir->n);
	PyObject * B = PyTuple_New (iir->n);

	PyTuple_SetItem (ab, 0, A);
	PyTuple_SetItem (ab, 1, B);

	for (int i = 0; i < iir->n; ++i)
	{
		PyTuple_SetItem (A, i, PyFloat (iir->a[i]));
		PyTuple_SetItem (B, i, PyFloat (iir->b[i]));
	}
	
	return ab;
}

static int
IIR_set_ab (IIR * iir, PyObject * ab)
Py_TRY
{
	PyObject * A, * B;
	
	if (!PyArg_ParseTuple (ab, "OO", &A, &B))
		return -1;

	int n = PySequence_Size (A);
	if (n < 1 || n > SIZE)
		throw ValueError() << "len (a) must be in [1 .. 32].";
	
	if (PySequence_Size (B) != n)
		throw ValueError() << "len (b) must be len (a).";

	iir->n = 1;

	for (int i = 0; i < n; ++i)
	{
		register PyObject * f;
		
		f = PySequence_GetItem (A, i);
		
		if (!PyArg_Parse (f, "d", iir->a + i))
			throw TypeError() << "a[" << i << "] is not a float coefficient.";

		f = PySequence_GetItem (B, i);
		
		if (!PyArg_Parse (f, "d", iir->b + i))
			throw TypeError() << "b[" << i << "] is not a float coefficient.";

		::decref (f, true);
	}

	iir->n = n;

	return 0;
}
Py_CATCH (-1);

static PyGetSetDef 
IIR_getsets [] =
{
	{"ab", (getter) IIR_get_ab, (setter) IIR_set_ab, 
	 "ab, tuple of (a, b) where 'a' and 'b' are tuples themselves, "
	 "containing the recursion coefficients. 'b[0]' is ignored."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
IIR_one_shot (IIR * iir, AudioSignal * e)
Py_TRY
{
	if (!AudioSignal_type->is_derived (e))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (e);

	iir->one_shot (e->data, e->size);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
IIR_cheby_lp (IIR * iir, PyObject * args)
Py_TRY
{
	double f, ripple;
	int order;

	if (!PyArg_ParseTuple (args, "ddi", &f, &ripple, &order))
		return 0;

	if (order < 2 || order > 16)
		throw ValueError() << "order must be >= 2 and <= 16.";

	DSP::ChebyshevIIR c (f, 0, ripple, order);

	iir->n = 0;
	for (int i = 0; i <= order; ++i)
		iir->a[order - i] = c.a[i];

	for (int i = 0; i < order; ++i)
		iir->b[order - i] = c.b[i];

	iir->n = order + 1;

	return PyNone();
}
Py_CATCH (0);

/* parses a tuple of (poles, zeros) into a DSP::PZs plane where 'poles' and
 * 'zeros' are sequences of PyComplex 
 */
static void
parse_poles_zeros (PyObject * args, DSP::PZs & plane)
{
	PyObject * poles, * zeros = 0;

	if (!PyArg_ParseTuple (args, "O|O", &poles, &zeros))
		throw blewit();

	if (!PySequence_Check (poles))
		throw TypeError() << "need a sequence of poles as 1st argument.";

	if (zeros && !PySequence_Check (zeros))
		throw TypeError() << "need a sequence of zeros or None as 2nd argument.";

	/* parse poles */
	int n = PySequence_Length (poles);

	for (int i = 0; i < n; ++i)
	{
		PyObject * co = PySequence_GetItem (poles, i);
		
		if (!PyComplex_Check (co))
		{
			::decref (co, true);
			throw TypeError() << "pole " << i << " is not complex.";
		}

		DSP::complex c;
		c.re = PyComplex_RealAsDouble (co);
		c.im = PyComplex_ImagAsDouble (co);

		plane.add_pole (c);
	}
		
	if (zeros == 0)
		return;

	/* parse zeros */
	n = PySequence_Length (zeros);

	for (int i = 0; i < n; ++i)
	{
		PyObject * co = PySequence_GetItem (zeros, i);
		
		if (!PyComplex_Check (co))
		{
			::decref (co, true);
			throw TypeError() << "zero " << i << " is not complex.";
		}

		DSP::complex c;
		c.re = PyComplex_RealAsDouble (co);
		c.im = PyComplex_ImagAsDouble (co);

		plane.add_zero (c);
	}
}

/* todo: probably broken, untested anyway. */
static PyObject *
IIR_from_splane (IIR * iir, PyObject * args)
Py_TRY
{
	DSP::PZs pzs;

	parse_poles_zeros (args, pzs);

	for (int i = 0; i < pzs.npoles; ++i)
		pzs.poles[i] = DSP::blt (pzs.poles[i]);

	for (int i = 0; i < pzs.nzeros; ++i)
		pzs.zeros[i] = DSP::blt (pzs.zeros[i]);

	iir->n = pzs.compute_ab (iir->a, iir->b, sizeof (iir->a));

	return PyNone();
}
Py_CATCH (0);

static PyObject *
IIR_from_zplane (IIR * iir, PyObject * args)
Py_TRY
{
	DSP::PZs pzs;

	parse_poles_zeros (args, pzs);
	iir->n = pzs.compute_ab (iir->a, iir->b, sizeof (iir->a));

	return PyNone();
}
Py_CATCH (0);

template <class PROTOTYPE>
static PyObject *
_iir_biquad (IIR * iir, PyObject * args)
{
	double fc, Q;

	if (!PyArg_ParseTuple (args, "dd", &fc, &Q))
		return 0;

	msg(), fc, Q;
	iir->n = 0;
	PROTOTYPE (fc, Q, iir->a, iir->b);
	iir->n = 3;

	return PyNone();
};

template <class PROTOTYPE>
static PyObject *
_iir_biquad_gain (IIR * iir, PyObject * args)
{
	double fc, Q, dB;

	if (!PyArg_ParseTuple (args, "ddd", &fc, &Q, &dB))
		return 0;

	msg(), fc, Q, dB;
	iir->n = 0;
	PROTOTYPE (fc, Q, dB, iir->a, iir->b);
	iir->n = 3;

	return PyNone();
};

static PyObject * _bp (IIR * iir, PyObject * args)
{
	return _iir_biquad<DSP::BP> (iir, args);
}

static PyObject * _rbj_lp (IIR * iir, PyObject * args)
{
	return _iir_biquad<DSP::RBJ::LP> (iir, args);
}

static PyObject * _rbj_bp (IIR * iir, PyObject * args)
{
	return _iir_biquad<DSP::RBJ::BP> (iir, args);
}

static PyObject * _rbj_hp (IIR * iir, PyObject * args)
{
	return _iir_biquad<DSP::RBJ::HP> (iir, args);
}

static PyObject * _rbj_notch (IIR * iir, PyObject * args)
{
	return _iir_biquad<DSP::RBJ::Notch> (iir, args);
}

static PyObject * _rbj_lo_shelve (IIR * iir, PyObject * args)
{
	return _iir_biquad_gain<DSP::RBJ::LoShelve> (iir, args);
}

static PyObject * _rbj_peaking (IIR * iir, PyObject * args)
{
	return _iir_biquad_gain<DSP::RBJ::PeakingEQ> (iir, args);
}

static PyObject * _rbj_hi_shelve (IIR * iir, PyObject * args)
{
	return _iir_biquad_gain<DSP::RBJ::HiShelve> (iir, args);
}

static PyObject * _zoelzer_lp (IIR * iir, PyObject * args)
{
	return _iir_biquad<DSP::Zoelzer::LP> (iir, args);
}

/* filter gain */

static PyObject *
IIR_dc_gain (IIR * iir)
{
	return PyFloat (iir->dc_gain());
}

static PyObject *
IIR_nyquist_gain (IIR * iir)
{
	return PyFloat (iir->nyquist_gain());
}

static PyObject *
IIR_scale (IIR * iir, PyObject * args)
{
	double f;

	if (!PyArg_ParseTuple (args, "d", &f))
		return 0;

	for (int i = 0; i < iir->n; ++i)
		iir->a[i] *= f;

	return PyNone();
}

static PyMethodDef
IIR_methods[] = 
{
	{"one_shot", (PyCFunction) IIR_one_shot, METH_O,
	 "one_shot (audio event) -- process the contents of a single audio event, "
	 "starting from a filter history of zeroes. the audio signal is processed "
	 "in-place."},
	{"cheby_lp", (PyCFunction) IIR_cheby_lp, METH_VARARGS,
	 "cheby_lp (f, ripple, order[, alpha = .05]) -- compute coefficients for a "
	 "Chebyshev lo-pass (or a Butterworth if ripple is 0). NB the filter gain "
	 "probably needs scaling before the filter is useful, for example like so:\n"
	 "iir.scale (1 / iir.dc_gain())."},
	{"from_splane", (PyCFunction) IIR_from_splane, METH_VARARGS,
	 "from_splane (poles, zeros) -- compute from poles and zeros in the "
	 "s-plane; 'poles' and 'zeros' are tuples of complex numbers."},
	{"from_zplane", (PyCFunction) IIR_from_zplane, METH_VARARGS,
	 "from_zplane (poles, zeros) -- compute from poles and zeros in the "
	 "z-plane; 'poles' and 'zeros' are tuples of complex numbers."},
	{"bp", (PyCFunction) _bp, METH_VARARGS,
	 "bp (fc, Q) -- initialize filter coefficients for a band-pass."},
	{"rbj_lp", (PyCFunction) _rbj_lp, METH_VARARGS,
	 "rbj_lp (fc, Q) -- initialize filter coefficients for a lo-pass bi-quad "
	 "according to Robert Bristow-Johnson."},
	{"rbj_bp", (PyCFunction) _rbj_bp, METH_VARARGS,
	 "rbj_bp (fc, Q) -- initialize filter coefficients for a band-pass bi-quad "
	 "according to Robert Bristow-Johnson."},
	{"rbj_hp", (PyCFunction) _rbj_hp, METH_VARARGS,
	 "rbj_hp (fc, Q) -- initialize filter coefficients for a hi-pass bi-quad "
	 "according to Robert Bristow-Johnson."},
	{"rbj_notch", (PyCFunction) _rbj_notch, METH_VARARGS,
	 "rbj_notch (fc, Q) -- initialize filter coefficients for a notch bi-quad "
	 "according to Robert Bristow-Johnson."},
	{"rbj_lo_shelve", (PyCFunction) _rbj_lo_shelve, METH_VARARGS,
	 "rbj_lo_shelve (fc, Q, dB) -- initialize filter coefficients for a lo-shelve "
	 "bi-quad according to Robert Bristow-Johnson."},
	{"rbj_peaking", (PyCFunction) _rbj_peaking, METH_VARARGS,
	 "rbj_peaking (fc, Q, dB) -- initialize filter coefficients for a peaking eq "
	 "bi-quad according to Robert Bristow-Johnson."},
	{"rbj_hi_shelve", (PyCFunction) _rbj_hi_shelve, METH_VARARGS,
	 "rbj_hi_shelve (fc, Q, dB) -- initialize filter coefficients for a hi-shelve "
	 "bi-quad according to Robert Bristow-Johnson."},
	{"zoelzer_lp", (PyCFunction) _zoelzer_lp, METH_VARARGS,
	 "zoelzer_lp (fc, Q) -- initialize filter coefficients for a lo-pass bi-quad "
	 "according to Udo Zoelzer."},
	{"dc_gain", (PyCFunction) IIR_dc_gain, METH_NOARGS,
	 "dc_gain() -- return the gain (linear) at DC."},
	{"nyquist_gain", (PyCFunction) IIR_nyquist_gain, METH_NOARGS,
	 "nyquist_gain() -- return the gain (linear) at Nyquist."},
	{"scale", (PyCFunction) IIR_scale, METH_VARARGS,
	 "scale (factor) -- scale the filter gain by a linear 'factor'."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<IIR>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = IIR_getsets;
	t->tp_methods = IIR_methods;
}

char * Py<IIR>::__doc__ = 
"IIR ([ab = ((1,), (0,))]) -- Infinite Impulse Response filter.\n"
"The filter supports up to 16 recursion coefficients; the default "
"being an identity set. The biquad variation "
"(three 'a' coefficients, two 'b') is special-cased for better performance.";

void
Py<IIR>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"ab", 0};
	PyObject * ab = 0;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|O", keys, &ab))
		return;
	
	if (ab) IIR_set_ab (this, ab);
}

/* /////////////////////////////////////////////////////////////////////// */

static Type * Pulse_type;

class Pulse
: public Plugin
{
	public:
		int n, width, period;

		Pulse();

		virtual void set_state (Graph * g, int s);
		virtual void process_audio (int frames);
};

Pulse::Pulse()
: Plugin (Pulse_type, Audio)
{
	n = 0;
	width = 1;
	period = 2048;

	add_out (new Out (this, CyclicAudio_type, "out"));
}

void
Pulse::set_state (Graph * g, int state)
{
	if (state == Plugin::Prepared)
		n = 0;

	Plugin::set_state (g, state);
}

void
Pulse::process_audio (int frames)
{
	CyclicAudio * event = (CyclicAudio *) outs[0]->cache->pop();

	if (event == 0)
		return;

	d_sample * s = event->data;
	
	while (frames)
	{
		int i = 0;
		
		if (n < width)
		{
			i = min (width - n, frames);

			for (int j = 0; j < i; ++j)
				*s++ = 1.0;
		}
		else if (n < period)
		{
			i = min (period - n, frames);

			for (int j = 0; j < i; ++j)
				*s++ = 0.0;
		}
		else
		{
			n = 0;
		}

		n += i;
		frames -= i;
	}

	outs[0]->push (event);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
Pulse_get_period (Pulse * p)
{
	return PyInt (p->period);
}

static int
Pulse_set_period (Pulse * p, PyObject * arg)
{
	if (!PyArg_Parse (arg, "i", &p->period))
		return -1;
	
	return 0;
}

static PyObject *
Pulse_get_width (Pulse * p)
{
	return PyInt (p->width);
}

static int
Pulse_set_width (Pulse * p, PyObject * arg)
{
	if (!PyArg_Parse (arg, "i", &p->width))
		return -1;
	return 0;
}

static PyGetSetDef 
Pulse_getsets [] =
{
	{"period", (getter) Pulse_get_period, (setter) Pulse_set_period, 
	 "the pulse repeat period, in samples."},
	{"width", (getter) Pulse_get_width, (setter) Pulse_set_width, 
	 "the width of the pulse, in samples."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<Pulse>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = Pulse_getsets;
}

char * Py<Pulse>::__doc__ = 
"Pulse ([period = 2048, width = 1]) -- rectangular, amplitude 1 pulse.";

void
Py<Pulse>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"period", "width", 0};
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|ii", keys, &period, &width))
		return;
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_IIR_types (PyObject * dict)
{
	IIR_type = new PyType<IIR> ("IIR", dict);
	Pulse_type = new PyType<Pulse> ("Pulse", dict);
}
