/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#ifdef WITH_IIWU

#include "IIWU.h"
#include "Graph.h"

#include "PyType.h"

/* set to 0 if the effect sends should not be exposed as Outs; needs a
 * pending patch (that will probably not be applied) to fluidsynth. 
 */
#define FX_SEND 0

void
IIWU::log (int level, char * msg, void * self)
{
	static char * prefix [] = 
		{
			"Panic: ", "Error: ", "Warning: ", "Info: ", "Debug: ", "?: "
		};

	if (level < 0 || level > 4) level = 5;

	_msg(), (IIWU *) self, prefix [level], msg;
}

IIWU::IIWU()
: Plugin (IIWU_type, Audio)
{
	synth = 0;
	cmd_handler = 0;

	sample_rate = SAMPLE_RATE;
	voices = 12;
	flags = 0;

	gain = 1.0;
	path = 0;

	set_object_name ("iiwu");

	add_in (new In (this, MidiEvent_type, 16, "midi"));

	add_out (new Out (this, CyclicAudio_type, "left"));
	add_out (new Out (this, CyclicAudio_type, "right"));

	#if FX_SEND
	add_out (new Out (this, CyclicAudio_type, "reverb send"));
	add_out (new Out (this, CyclicAudio_type, "chorus send"));
	#endif

	for (int i = FLUID_PANIC; i < LAST_LOG_LEVEL; i++)
		fluid_set_log_function (i, log, this);
}

IIWU::~IIWU()
{
	if (cmd_handler)
		delete_fluid_cmd_handler (cmd_handler);
	if (synth)
		delete_fluid_synth (synth);
	if (path)
		free (path);
}

void
IIWU::pre_insert (Graph * g)
{
	if (synth)
		return;
	
	sample_rate = g->sample_rate;
	open();
}

void
IIWU::open()
{
	fluid_settings_t * settings = new_fluid_settings();
	
	if (settings == 0)
		throw blewit() << "new_fluid_settings() failed.";

	fluid_settings_setstr (settings, "audio.driver", "");
	fluid_settings_setstr (settings, "audio.sample-format", "float");
	fluid_settings_setstr (settings, "audio.output-channels", "2");

	fluid_settings_setstr (settings, "midi.driver", "");

	synth = new_fluid_synth (settings);

	if (synth == 0)
		throw blewit() << "new_fluid_synth() failed.";
	
	cmd_handler = new_fluid_cmd_handler (synth);

	if (path)
		add_sf2 (path);
}

int
IIWU::add_sf2 (char * path)
{
	if (synth == 0)
		throw blewit() << "not opened yet.";
		
	msg(), "loading sf2: ", path;

	int result = fluid_synth_sfload (synth, path, 0);

	if (result < 0)
		throw blewit() << "fluid_synth_sfload ('" << path << "') failed: " << result;

	return result;
}

void
IIWU::parse_event (MidiEvent * e)
{
	if (e->is_sysex())
	{
		/* this can't be a command. */
		if (e->length > 200)
			return;
			
		/* need to protect fluid from passing in a non-zero-terminated string. */
		char command [e->length + 1];
		
		strncpy (command, (char *) e->data, e->length);
		command [e->length] = 0;
		
		fluid_command (cmd_handler, command, fileno (stderr));
	}

	else if (e->is_note_off())
		fluid_synth_noteoff (synth, e->channel(), e->data[1]);

	else if (e->is_note_on())
		fluid_synth_noteon (synth, e->channel(), e->data[1], e->data[2]);

	else if (e->is_program())
		fluid_synth_program_change (synth, e->channel(), e->data[1]);
	
	else if (e->is_controller())
	{
		if (e->data[1] == 0)
			fluid_synth_bank_select (synth, e->channel(), e->data[2]);
		else
			fluid_synth_cc (synth, e->channel(), e->data[1], e->data[2]);
	}
	
	else if (e->is_pitch())
		fluid_synth_pitch_bend (synth, e->channel(), e->data[1] | (e->data[2] << 7));
	
	::decref (e);
}

void
IIWU::all_notes_off()
{
	for (int i = 0; i < 16; ++i)
		fluid_synth_cc (synth, i, 123, 0);
}

void 
IIWU::seek()
{
	if (state >= Activated)
		all_notes_off();
}

void 
IIWU::set_state (Graph * g, int s)
{
	if (s < Rolling && state == Rolling)
		all_notes_off();

	Plugin::set_state (g, s);
}

void IIWU::process_audio (int frames)
{
	if (synth == 0)
		throw AssertionError() << this << "isn't opened yet.";

	fluid_synth_set_gain (synth, gain);

	CyclicAudio * l = (CyclicAudio *) outs[0]->cache->pop();
	CyclicAudio * r = (CyclicAudio *) outs[1]->cache->pop();

	#if FX_SEND
	CyclicAudio * reverb = (CyclicAudio *) outs[2]->cache->pop();
	CyclicAudio * chorus = (CyclicAudio *) outs[3]->cache->pop();
	#endif

	int written = 0;
	d_frame f = graph->frame + graph->frames_per_cycle;

	while (written < frames)
	{
		int n = min (64, frames);

		bool is_loop = graph->is_loop_frame (f, f + n);

		f += n;
		d_tick tick = graph->frame2tick (f);
		
		MidiEvent * e;

		while ((e = (MidiEvent *) ins[0]->pop_before (tick)))
			parse_event (e);

		if (is_loop)
		{
			if (0)
				/* all-notes-off. */
				for (int i = 0; i < 16; ++i)
					fluid_synth_cc (synth, i, 123, 0);

			++loop_cycle;
			
			f -= graph->loop.frame[1] - graph->loop.frame[0];
		}

		#if FX_SEND
		fluid_synth_write_float_dry_fx (synth,
				n, written, 1,
				l->data, r->data, 
				reverb->data, chorus->data);
		#else
		fluid_synth_write_float (synth, n, 
				l->data, written, 1, 
				r->data, written, 1);
		#endif
		
		written += n;
	}

	outs[0]->push (l);
	outs[1]->push (r);

	#if FX_SEND
	outs[2]->push (reverb);
	outs[3]->push (chorus);
	#endif
}

/* //////////////////////////////////////////////////////////////////////// */

/* Py<IIWU> */

static PyObject *
IIWU_add_sf2 (IIWU * fluid, PyObject * path) 
Py_TRY
{
	if (!PyString_Check (path))
		throw TypeError() << "path must be a string.";

	int id = fluid->add_sf2 (PyString_AsString (path));
	return PyInt (id);
} 
Py_CATCH (0);

static PyObject *
IIWU_cmdline (IIWU * fluid, PyObject * cmd)
Py_TRY
{
	if (fluid->synth == 0)
		throw AssertionError() << "the synth isn't instantiated yet.";

	if (!PyString_Check (cmd))
		throw TypeError() << "command must be a string.";

	fluid_command (fluid->cmd_handler, PyString_AsString (cmd), fileno (stderr));

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef 
IIWU_methods [] = 
{
  {"add_sf2", (PyCFunction) IIWU_add_sf2, METH_O,
	 "add_sf2 (path) -- add a soundfont to the fluid sf2 stack."},
  {"cmdline", (PyCFunction) IIWU_cmdline, METH_O,
	 "cmdline ('fluid command') -- issue an fluid command, returns a "
	 "string reply."},
	{ }
};

static PyObject * 
IIWU_get_gain (IIWU * fluid)
{
	return PyFloat (fluid->gain);
}

static int 
IIWU_set_gain (IIWU * fluid, PyObject * value)
{
	if (!PyArg_Parse (value, "f", &fluid->gain))
		return -1;
	return 0;
}

static PyGetSetDef 
IIWU_getsets [] = 
{
	{"gain", (getter) IIWU_get_gain, (setter) IIWU_set_gain,
	 "the output signal is multiplied by this value."},
	{ }
};

void
Py<IIWU>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = IIWU_getsets;
	t->tp_methods = IIWU_methods;

	PyType_Ready (t);

	/* constructor flags */
	PyDict_SetItemString (t->tp_dict, "Reverb", PyInt (Reverb));
	PyDict_SetItemString (t->tp_dict, "Chorus", PyInt (Chorus));
	PyDict_SetItemString (t->tp_dict, "Verbose", PyInt (Verbose));
}

Type * IIWU_type = 0;

char * Py<IIWU>::__doc__ = 
"IIWU (voices[, path = None, flags = 0]) -- "
"a plugin that wraps the 'If I Were You' wavetable "
"synthesizer.\n"
"The 'voices' setting determines the maximum number of concurrently "
"sounding voices and thereby the maximum CPU usage. 'path' should "
"point to a soundfont file in SF2 format. 'flags' is a combination "
"of the constants found in the type dict.";

void
Py<IIWU>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"voices", "path", "flags", 0};
	char * _path = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "i|si", keys,
				&voices, &_path, &flags))
		throw blewit();

	if (voices < 1)
		voices = 1;
	if (voices > 64)
		voices = 64;

	/* will be opened later. */
	if (_path)
		path = strdup (_path);
}

#endif /* WITH_IIWU */

/* //////////////////////////////////////////////////////////////////////// */

void
init_IIWU_types (PyObject * dict)
{
	#ifdef WITH_IIWU
	IIWU_type = new PyType<IIWU> ("IIWU", dict);
	#endif /* WITH_IIWU */
}
