/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "InOut.h"
#include "Plugin.h"
#include "Graph.h"

/* log event cache exits/entries */
#define DBG_POP_PUSH 0

/* log pushes to Out */
#define DBG_PUSH_OUT 0

In::In (Plugin * p, Type * t, int _n_slots, const char * name)
: NamedObject (In_type, name)
{
	flags = 0;
	plugin = p;
	set_type (t);

	connected = 0;

	/* arbitrary limit. */
	if ((n_slots = _n_slots) > 1022)
		throw OverflowError() << 1022 << " maximum slots for an In.";

	slots = new Out * [n_slots];

	for (int i = 0; i < n_slots; ++i)
		slots[i] = 0;
	
	current = 0;
}

In::~In()
{
	for (int i = 0; i < n_slots; ++i)
		::decref (slots[i], true);

	delete [] slots;
}

void
In::set_type (Type * t)
{
	type = t;
	
	if (CyclicAudio_type->is_derived (t))
		set_flag (Audio);
	if (Event_type->is_derived (t))
		set_flag (Events);
}

/* called by Out::connect. */
void
In::connect (Out * out)
{
	for (int i = 0; i < n_slots; ++i)
		if (slots[i] == out)
			throw AssertionError() << this << " is already connected to " << out;

	int i = -1;
	while (++i < n_slots)
	{
		if (slots[i])
			continue;
		break;
	}

	if (i == n_slots)
		throw OverflowError() << this << " already has " 
													<< n_slots << "connections.";

	++connected;
	slots[i] = out;
}

int
In::have()
{
	int n = 0;

	for (int i = 0; i < n_slots; ++i)
		if (slots[i])
			n += slots[i]->have();
	
	return n;
}

void
In::disconnect (Out * out)
{
	if (current == out)
		current = 0;

	for (int i = 0; i < n_slots; ++i)
	{
		if (slots[i] != out)
			continue;

		--connected;

		slots[i] = 0;
		return;
	}
	
	throw AssertionError() << this << " is not connected to " << out;
}

/* find the slot providing the next event. */
Object *
In::next()
{
	current = 0;

	/* if we don't deal with tick-marked events, we don't have a clue about 
	 * order, so we return the first object we find. */
	for (int i = 0; i < n_slots; ++i)
	{
		if (slots[i]->peek() == 0)
			continue;

		return (current = slots[i])->peek();
	}
	
	return 0;
}

/* find the next event before 'tick' and in 'this->plugin->loop_cycle'*/
Event *
In::peek_before (d_tick tick)
{
	Event * next = 0;
	int loop = plugin->loop_cycle;

	for (int i = 0; i < n_slots; ++i)
	{
		Out * out = slots[i];

		if (out == 0)
			continue;

		Event * e = (Event *) out->peek();

		if (e == 0)
			continue;

		/* hardcore debugging, creates *lots* of output. */
		if (0 && get_type (e) == MidiEvent_type)
		{
			MidiEvent * me = (MidiEvent *) e;
			msg(), loop, tick, ": slot, loop, tick: ", i, e->loop, e->tick, 
					me->data[0], me->data[1];
		}

		if (e->loop > loop)
			continue;

		if (e->loop == loop)
		{
			if (e->tick > tick)
				continue;
		}

		/* an event from a previous loop. */
		else if (e->loop < loop)
		{
			/* this event needs to play right away; we do that by adjusting the
			 * event tick. but first we have to check for a Sysex event; they
			 * do not come from a cache so we wouldn't operate on a temporary
			 * copy but change the original event (probably within a Part). */
			if (get_type (e) == SysexEvent_type)
			{
				::decref (out->pop());
				continue;
			}

			/* with all other events we can simply adjust the event tick for
			 * direct transmission. */
			e->tick = 0;
			current = out;
			return e;
		}

		current = out;
		tick = e->tick;
		next = e;
	}

	return next;
}

/* pop and return the next event before 'tick' */
Event *
In::pop_before (d_tick tick)
{
	assert (Event_type->is_derived (type));

	Event * e = peek_before (tick);

	if (e)
		current->pop();

	return e;
}

void
In::debug_midi_queue_dump()
{
	_msg m;

	m, "queue:";

	for (int i = 0; i < n_slots; ++i)
	{
		if (!slots[i])
			continue;

		MidiEvent * me = (MidiEvent *) slots[i]->peek_object();
		if (me)
			m, "\n\tslot ", i, me->data[0], me->data[1], me->tick, "(", me->loop, ")";
	}
}

/* pop and update a float value. */
int 
In::pop_value (d_tick tick, d_float * value)
{
	assert (FloatEvent_type->is_derived (type));

	int popped = 0;

	FloatEvent * e;

	while ((e = (FloatEvent *) pop_before (tick)))
	{
		*value = e->value;
		::decref (e);
		++popped;
	}

	return popped;
}

/* pop from all connected Outs and mixdown into one event. */
CyclicAudio *
In::pop_mix()
{
	assert (CyclicAudio_type->is_derived (type));

	CyclicAudio * event = 0, * e = 0;

	for (int i = 0; i < n_slots; ++i)
	{
		if (slots[i] == 0)
			continue;
			
		e = (CyclicAudio *) slots[i]->pop();
		if (e == 0)
			continue;

		if (event == 0)
		{
			event = e;
			continue;
		}

		*event += *e;

		e->return_to_cache();
	}
	
	return event;
}

/* ///////////////////////////////////////////////////////////////////////// */

/* number of objects that can be queued in an Out or its cache. */

static inline int
get_fifo_size (Type * t)
{
	if (CyclicAudio_type->is_derived (t))
		return 2;

	if (DecycledAudio_type->is_derived (t))
		return 4;

	/* not an event: don't cache. */
	if (!Event_type->is_derived (t))
		return 0;

	if (MidiEvent_type->is_derived (t))
		return 1022;

	/* arbitrary limit. */
	return 126;
}

EventCache::EventCache (Out * o)
: ObjectFifo (get_fifo_size (o->type)),
	events ("cached", get_fifo_size (o->type))
{
	going_down = 0;

	Type * t = (out = o)->type;

	set_object_type (EventCache_type);
	set_object_name ("%s.%s.cache", get_name (out->plugin), get_type_name (t));
	
	/* allocate events for the cache. simply creating an array of instances
	 * instead of calling new for each seems a tempting option for two reasons:
	 * it is quicker (also in the long run) and it removes the need for the
	 * events array used to uncache the objects. otoh we'd need to wait for all
	 * cached events to be returned before we could free the array, resulting
	 * in a very complicated scenario at object shutdown that would involve
	 * blocking until all events have returned. Not attractive. 
	 */
	while (have() < size - 1)
	{
		Object * e = (Object *) t->__new__();

		e->object_cache = this;
		
		events.add_object (e);
		push_object (e);
	}
}

EventCache::~EventCache()
{
	going_down = 1;
	
	release_events();

	/* Fallthrough to ~ObjectFifo which does the actual free()ing. */
}

void
EventCache::release_events()
{
	/* mark all events as uncached, no matter whether they are in the cache or
	 * in a thread-local disposal fifo. This is the only purpose of the 'events'
	 * list. 
	 */
	for (int i = 0; i < events; ++i)
		events[i]->object_cache = 0;
}

Object * 
EventCache::pop()
{
	#if DBG_POP_PUSH
	msg(), 
			get_name (out->plugin), ".", get_name (out), 
			".cache::pop ", peek_object();
	#endif

	if (have() < 1)
		throw RuntimeError() 
					<< get_type_name (out->plugin) << " " 
					<< get_name (out->plugin) << "."
					<< get_name (out) << ".cache is empty.";
	
	return (Event *) pop_object();
}

void 
EventCache::push (Object * e)
{
	#if DBG_POP_PUSH
	msg(), get_name (out->plugin), ".", get_name (out), ".cache::push ", e;
	#endif

	if (going_down)
	{
		::decref (e);
		return;
	}

	if (is_full())
	{
		::decref (e);
		throw RuntimeError() << this << "is full pushing " << e;
	}

	push_object (e);
}

/* ///////////////////////////////////////////////////////////////////////// */

In *
Plugin::next_event (d_tick before)
{
	In * found = 0;
	
	for (int i = 0; i < ins; ++i)
	{
		In * in = ins[i];
		
		if (!in->is_event())
			continue;
		
		Event * e = in->peek_before (before);

		if (e == 0)
			continue;

		found = in;
		before = e->tick;
	}

	return found;
}

void
Plugin::set_frames_per_cycle (int n)
{
	for (int i = 0; i < outs; ++i)
	{
		Out * out = outs[i];
		
		if (!out->is_audio)
			continue;
		
		EventCache * cache = out->cache;

		int j = cache->have();
		while (j--)
		{
			/* resize all audio events in the cache to fit the new frames_per_cycle
			 * setting. */
			CyclicAudio * e = (CyclicAudio *) cache->pop_object();
			e->set_size (n);
			cache->push_object (e);
		}
	}
}

/* ///////////////////////////////////////////////////////////////////////// */

Out::Out (Plugin * p, Type * t, const char * name)
: ObjectFifo (get_fifo_size (t), name)
{
	set_object_type (Out_type);

	plugin = p;
	connected = 0;
	cache = 0;
	blocked = 0;

	set_type (t);
}

Out::~Out()
{
	connect (0);

	::decref (cache, true);
}

void
Out::set_type (Type * t)
{
	if (connected)
		throw AssertionError() << this << " already connected in set_type().";

	type = t;

	is_event = Event_type->is_derived (t);
	is_audio = CyclicAudio_type->is_derived (t);

	::decref (cache);
	
	cache = new EventCache (this);
}

/* connect to an input. a check is done on whether the event type transmitted
 * is (or is derived from) the type that the input expects. With an atomic
 * pointer data type, this is considered safe against a thread pushing data. 
 */
void
Out::connect (In * in)
{
	if (in)
	{
		if (!plugin || !in->plugin)
			throw AssertionError() << "detached port.";

		if (!plugin->graph)
			throw RuntimeError() << "can not connect outside a graph.";
			
		if (plugin->graph != in->plugin->graph)
			throw RuntimeError() << "can only connect within the same graph.";

		if (is_audio && !plugin->before (in->plugin))
			throw RuntimeError() << "violates the audio processing order.";

		if (!in->type->is_derived (type))
			throw TypeError() << in << "does not dig '" << type->tp_name << "'.";

		if (in->connected == in->n_slots)
			throw OverflowError() << in << "does not accept any more connections.";
	}

	In * i = connected;

	connected = 0;
	
	if (i) 
	{
		i->disconnect (this);
		::decref (i);
	}
	
	if (in)
	{
		::incref (connected = in);
		connected->connect (this);

		/* if we have left-over data (we shouldn't), trigger processing. */
		if (have())
		{
			Plugin * p = connected->plugin;

			if (p && (p->flags & Plugin::Script))
				p->wake_up();
		}
	}
}

/* sends the event to a connected In and returns != 0 if everything went
 * ok, otherwise discards the event and returns 0.
 * NB: steals a reference to 'e'. 
 */
bool
Out::push (Object * e)
{
	#if DBG_POP_PUSH
	msg(), get_name (plugin), ".", get_name (this), "::push ", e;
	#endif

	In * in = connected;

	if (in == 0 || !ObjectFifo::push_object (e))
	{
		::decref (e);
		return 0;
	}
	
	/* trigger script execution. */
	Plugin * p = in->plugin;

	if (p == 0)
		;
	if (p->flags & Plugin::Script)
	{
		Graph * g = p->graph;
		
		if (g && g->script_scheduler)
			p->wake_up();
	}
	else if (in->is_trigger())
	{
		p->process_trigger (in);
	}
	
	return true;
}

/* a separate implementation of push() for CyclicAudio that immediately
 * returns events to their cache if unconnected.
 */
bool
Out::push (CyclicAudio * e)
{
	#if DBG_PUSH_OUT
	msg(), get_name (plugin), ".", get_name (this), "::push ", e;
	#endif

	if (connected && ObjectFifo::push_object (e))
		return true;

	/* don't push if we're not connected. */
	e->return_to_cache();
	return 0;
}
