/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _INOUT_H_
#define _INOUT_H_

#include "List.h"
#include "ObjectFifo.h"
#include "Thread.h"
#include "AudioSignal.h"

class In;
class Out;
class Plugin;

#define IN_SLOTS 64

/* really an object cache, but only used inside the graph, where every signal
 * is called an event, even if it's not an Event. I offer my apologies for any
 * inconveniences this may cause. :)
 */
class EventCache
: public ObjectFifo,
	public Cache
{
	protected:
		atomic going_down;

	public:
		Out * out;
		
		/* Needs to remember all events created to set their cache to 0 when going
		 * down -- the events may still be travelling along the graph, or be held
		 * in a trashfifo. setting the cache member to 0 ensures they are finally
		 * deallocated.
		 */
		PointerList<Object> events;

		EventCache()
			: events ("cached")
			{
				throw NotImplementedError() << "cannot instantiate without an Out.";
			}

		EventCache (Out *);
		~EventCache();

		Object * peek() { return ObjectFifo::peek_object(); }
		Object * pop();
		virtual void push (Object * o);

		/* sets event->cache to 0 for all */
		void release_events();
};

class Out
: public ObjectFifo
{
	public:
		Plugin * plugin;
		Type * type;
		In * connected;
		
		bool is_event;
		bool is_audio;
		bool blocked;

		EventCache * cache;

		Mutex mutex; 
		
	public:
		Out (Plugin * p = 0, Type * t = Object_type, const char * name = 0);
		~Out();

		/* don't use this after connecting. */
		void set_type (Type * t);

		void connect (In * to);

		bool push (Object *);
		bool push (CyclicAudio *);

		Object * pop() 
			{ 
				if (this == 0 || blocked)
					return 0;

				return pop_object(); 
			}

		Object * peek() 
			{ 
				if (this == 0 || blocked)
					return 0;

				return peek_object();
			}

};

class In
: public NamedObject
{
	friend class Out;

	public:
		Plugin * plugin;
		Type * type;	/* base type for accepted events */

		atomic connected; /* number of connected Outs */

		int n_slots; /* number of slots for connections */
		Out ** slots; /* slots[i] = (Out *) or 0 */
		
		Out * current; /* the Out holding the next event. */
		
		enum {
			Audio = 1, /* takes AudioSignal events (mostly CyclicAudio) */
			Events = 2, /* takes Events */
			Trigger = 4 /* flag indicating the plugin is to be called immediately 
										 upon receiving an event -- means (due to threading
										 issues) that only one Out can connect to this In */
		};
		int flags;

	public:
		In (Plugin * p = 0, Type * t = Object_type, int _n_slots = 1, const char * name = 0);
		~In();

		/* call is valid only before connecting. */
		void set_type (Type * t);

		/* return count of pending events */
		int have();
		
		/* returns the current object without removing it from 'current' */
		Object * peek() 
			{ 
				return current ? current->peek() : next();
			}

		/* returns the current object and removes it from 'current' */
		Object * pop() 
			{ 
				if (current == 0)
					next();

				if (current == 0)
					return 0;
				
				Object * o = current->pop();
				current = 0;
				return o;
			}

		/* It is considered that creating more subclasses of In and Out that
		 * are specialized in some sort of Event/Object would result in bloat.
		 *
		 * Following are specialized peek and pop implementations. */
		Event * peek_before (d_tick tick);
		Event * pop_before (d_tick tick);

		/* does pop_before and updates a value. returns the number of events
		 * popped. */
		int pop_value (d_tick tick, d_float * value);

		/* does pop() on all slots, adding (mixing) the audio signals.
		 * return is 0 if no signal (not one event) is present at the input. */
		CyclicAudio * pop_mix();

		/* write info about input queues to stdmsg */
		void debug_midi_queue_dump();

		bool is_audio()
			{ return flags & Audio; }
		bool is_event()
			{ return flags & Events; }
		bool is_trigger()
			{ return flags & Trigger; }

		void set_flag (int f)
			{ flags |= f; }
		
	private:
		/* sets 'current' to the Out holding the next object. */
		Object * next();
		
		/* called by Out::connect() only. */
		void connect (Out * out);
		void disconnect (Out * out);

};

typedef TypedObjectList<In> InList;
typedef TypedObjectList<Out> OutList;

/* how loops are supposed to work:
 *
 * since plugins should not need to differentiate between events meant
 * to play before the loop point and events beyond it (the only hint is
 * the event's timing info, which actually doesn't even suffice for this
 * purpose because plugins play any event that is marked 'before the current
 * time') we set Out::r_loop to the event index in the fifo that holds
 * the first event after the loop. Out::pop() and push() will block this
 * event until r_loop is reset to -1.
 * 
 * this is bullshit, and needs to be replaced by something simpler and more
 * suited.
 */

#endif /* !_INOUT_H_ */
