/* (c) 2001-4 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include <basics.h>

#ifdef WITH_JACK

#include "Jack.h"
#include "Graph.h"
#include "PythonThread.h"

#include "PyType.h"

static int 
_jack_buffer_size (jack_nframes_t n, void * self)
{
	return ((Jack *) self)->jack_buffer_size (n);
}

static int 
_jack_sample_rate (jack_nframes_t n, void * self)
{
	return ((Jack *) self)->jack_sample_rate (n);
}

static void 
_jack_shutdown (void * self)
{
	((Jack *) self)->jack_shutdown();
}

static int 
_jack_process (jack_nframes_t n, void * self)
{
	return ((Jack *) self)->jack_process (n);
}

#if 0
static void 
_jack_verror (const char * fmt, ...)
{
	char * s = 0;
	va_list args;
	
	va_start (args, fmt);
	vasprintf (&s, fmt, args);
	va_end (args);
	
	msg(), "Jack says: ", s;

	free (s);
}
#endif

static void 
_jack_error (const char * error)
{
	msg(), "Jack says: ", error;
}

/* //////////////////////////////////////////////////////////////////////// */

Jack::Jack()
{
	handle = 0;
	py_thread = 0;

	jackd_is_down = true;

	jack_set_error_function (_jack_error);

	if (sizeof (jack_default_audio_sample_t) != sizeof (d_sample))
		throw AssertionError() << "Jack and I do not agree on the "
				"audio sample size.";
}

Jack::~Jack()
{
	if (handle)
		jack_client_close (handle);
	
	if (py_thread)
		py_thread->active = 0;
}

void
Jack::init (char * name)
{
	set_object_name (name);

	handle = jack_client_new (name);
	
	if (handle == 0)
		throw IOError() << "can't register with the Jack server. Is it up?";

	jack_on_shutdown (handle, _jack_shutdown, this);

	sample_rate = jack_get_sample_rate (handle);
	buffer_size = jack_get_buffer_size (handle);
	
	jack_set_buffer_size_callback (handle, _jack_buffer_size, this);
	jack_set_sample_rate_callback (handle, _jack_sample_rate, this);

	jack_set_process_callback (handle, _jack_process, this);

	/* the simplest is to activate now. the graph expects audio clocks to
	 * call check_for_requests() which we can't do without being called back
	 * by jackd. */
	 
	if (jack_activate (handle))
		throw IOError() << "jackd refuses to activate this client.";

	jackd_is_down = false;
}

void 
Jack::connect (const char * source, const char * dest)
{
	int r = jack_connect (handle, source, dest);
	
	if (r)
		throw IOError() << "jack_connect ('" << source << "', '" 
				<< dest << "') fails.";
}

void 
Jack::disconnect (const char * source, const char * dest)
{
	int r = jack_disconnect (handle, source, dest);
	
	if (r)
		throw IOError() << "jack_disconnect ('" << source << "', '" 
				<< dest << "') fails.";
}

void
Jack::restart_graph()
{
	if (graph == 0)
		return;

	try 
	{
		/* ensure exclusive access to the graph. */
		ScopedLock lock (graph->mutex);

		int s = graph->state;
	
		graph->set_state (Off);
		graph->sample_rate = sample_rate;
		graph->frames_per_cycle = buffer_size;

		/* as soon as the graph state goes > 0, it will do the necessary 
		 * resource allocations etc.
		 */
		for (int i = 1; i <= s; ++i)
			graph->set_state (i);
	}
	catch (blewit & b)
	{
		msg(), this, "can't restart the graph: ", b.what();
		state = Off;
	}
}

int
Jack::jack_buffer_size (int n)
{
	buffer_size = n;
	state = Off;

	restart_graph();

	return 0;
}

int
Jack::jack_sample_rate (int n)
{
	sample_rate = n;
	state = Off;

	restart_graph();

	return 0;
}

void
Jack::jack_shutdown()
{
	msg(), "jackd went down.";

	jackd_is_down = true;
	state = Off;

	/* this is no usable audio clock any more. */
	graph->audio_clock = 0;
}

int
Jack::jack_process (int frames)
{
	/* the only way to attach to the jack_process thread is this: */
	if (py_thread == 0)
		py_thread = PythonThread::create ("Jack");
	
	if (graph == 0)
		return 0;

	if (state >= Activated)
	{
		try
		{
			graph->process_audio (frames);
		}
		catch (blewit & b)
		{
			msg(), "audio cycle blew it: ", b.what();
			state = Off;
		}
	}

	graph->check_for_requests();

	return 0;
}

void
Jack::pre_insert (Graph * g)
{
	if (jackd_is_down)
		throw AssertionError() << "jackd is down.";

	if (g->state != Off)
		throw ValueError() << "graph must be in stopped state.";
	
	if (g->frames_per_cycle != buffer_size)
	{
		msg(), "Jack vs graph frames/cycle mismatch. fixing ...";
		g->frames_per_cycle = buffer_size;
	}

	if (g->sample_rate != sample_rate)		
	{
		msg(), "Jack vs graph sample_rate mismatch. fixing ...";
 		g->sample_rate = sample_rate;
	}
}

void
Jack::set_state (Graph * g, int _state)
{
	/* nothing to do. */
	state = _state;
}

/* //////////////////////////////////////////////////////////////////////// */

JackIn::JackIn()
: Plugin (JackIn_type, Audio),
	ports ("JackIn::ports"),
	events ("JackIn::events")
{
}

JackIn::~JackIn()
{
	for (int i = 0; i < events; ++i)
	{
		events[i]->object_cache = 0;
		::decref (events[i], true);
	}

	/* fallthrough, ~events frees the events. */
}

void
JackIn::init (Jack * jack, int n_ports)
{
	char port_name [32];

	for (int i = 0; i < n_ports; ++i)
	{
		sprintf (port_name, "in_%d", i);
		
		jack_port_t * port = jack_port_register (jack->handle, port_name, 
				JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);

		if (port == 0)
			throw IOError() << "registering Jack input " << i << " fails.";

		add_port (port);
	}
}

void
JackIn::add_port (jack_port_t * port)
{
	ports.add_object (port);
	
	CyclicAudio * e = new CyclicAudio();
	e->object_cache = this;
	events.add_object (e);

	add_out (new Out (this, CyclicAudio_type, jack_port_name (port)));
}

/* inherited from Cache: */
void 
JackIn::push (Object * e)
{
	/* 'e->data' points to a Jack shm sample buffer; zap the pointer so
	 * ~CyclicAudio will not try to free it. 
	 */
	 
	((CyclicAudio *) e)->data = 0;
}

void
JackIn::process_audio (int frames)
{
	for (int i = 0; i < ports; ++i)
	{
		d_sample * src = (d_sample *) jack_port_get_buffer (ports[i], frames);
		
		if (src == 0)
		{
			msg(), this, "port ", i, " yields no buffer.";
			continue;
		}

		CyclicAudio * e = events[i];
		e->data = src;
		outs[i]->push (e);
	}
}

/* //////////////////////////////////////////////////////////////////////// */

JackOut::JackOut()
: Plugin (JackOut_type, Audio),
	ports ("JackIn::ports")
{
}

JackOut::~JackOut()
{
}

void
JackOut::init (Jack * jack, int n_ports)
{
	char port_name [32];

	for (int i = 0; i < n_ports; ++i)
	{
		sprintf (port_name, "out_%d", i);
		
		jack_port_t * port = jack_port_register (jack->handle, port_name, 
				JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);

		if (port == 0)
			throw IOError() << "registering Jack output " << i << " fails.";

		add_port (port);
	}
}

void
JackOut::add_port (jack_port_t * port)
{
	ports.add_object (port);

	add_in (new In (this, CyclicAudio_type, IN_SLOTS, jack_port_name (port)));
}

void
JackOut::process_audio (int frames)
{
	for (int i = 0; i < ports; ++i)
	{
		CyclicAudio * e = ins[i]->pop_mix();

		d_sample * dest = (d_sample *) jack_port_get_buffer (ports[i], frames);
		
		if (dest == 0)
		{
			msg(), this, "port ", i, " yields no buffer.";
			state = Off;
		}

		else if (e == 0)
			memset (dest, 0, frames * sizeof (d_sample));

		else if (dest != e->data)
			memcpy (dest, e->data, frames * sizeof (d_sample));

		if (e)
			e->return_to_cache();
	}
}

/* // Py<Jack> //////////////////////////////////////////////////////////// */

static PyObject * 
Jack_get_jackd_is_down (Jack * jack)
{
	return PyInt (jack->jackd_is_down);
}

/* ~/audio/src/jack-audio-connection-kit-0.91.1/libjack/client.c */

static PyObject * 
Jack_get_ports (Jack * jack)
{
	const char ** ports = jack_get_ports (jack->handle, 0, 0, 0);

	if (ports == 0)
		return PyNone();

	int n = 0;
	for (const char ** p = ports; *p; ++p)
		++n;

	PyObject * l = PyList_New (n);
	if (l == 0)
		return 0;

	n = 0;
	for (const char ** p = ports; *p; ++p, ++n)
		PyList_SetItem (l, n, PyString (*p));

	free (ports);

	return l;
}

static PyGetSetDef 
Jack_getsets [] = {
	{"jackd_is_down", (getter) Jack_get_jackd_is_down, (setter) 0,
	 "whether jackd is currently processing."},
	{"ports", (getter) Jack_get_ports, (setter) 0,
	 "all ports on the jack graph."},
	{ }
};

static PyObject *
Jack_connect (Jack * jack, PyObject * args)
Py_TRY
{
	PyObject * o[2];
	const char * s[2];

	if (!PyArg_ParseTuple (args, "OO", o, o + 1))
		return 0;
	
	for (int i = 0; i < 2; ++i)
	{
		/* the In/Out on a JackOut/~In will have a legal name in jack's port
		 * namespace. */
		if (Object_type->is_derived (o[i]))
			s[i] = get_name (o[i]);
		else if (!PyArg_Parse (o[i], "s", s + i))
			return 0;
	}
			
	jack->connect (s[0], s[1]);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
Jack_disconnect (Jack * jack, PyObject * args)
Py_TRY
{
	PyObject * o[2];
	const char * s[2];

	if (!PyArg_ParseTuple (args, "OO", o, o + 1))
		return 0;
	
	for (int i = 0; i < 2; ++i)
	{
		/* the In/Out on a JackOut/~In will have a legal name in jack's port
		 * namespace. */
		if (Object_type->is_derived (o[i]))
			s[i] = get_name (o[i]);
		else if (!PyArg_Parse (o[i], "s", s + i))
			return 0;
	}
			
	jack->disconnect (s[0], s[1]);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef 
Jack_methods [] = 
{
  {"connect", (PyCFunction) Jack_connect, METH_VARARGS,
	 "connect ('source', 'dest') -- asks Jack to connect two Jack ports. "
	 "'source' and 'dest' can be strings or In/Out ports on a JackOut/JackIn, "
	 "as long as their name has not been changed."},
  {"disconnect", (PyCFunction) Jack_disconnect, METH_VARARGS,
	 "disconnect ('source', 'dest') -- asks Jack to disconnect two Jack ports. "
	 "'source' and 'dest' can be strings or In/Out ports on a JackOut/JackIn, "
	 "as long as their name has not been changed."},
	{ }
};

Type * Jack_type = 0;

char * Py<Jack>::__doc__ = 
"Jack (name) -- an AudioClock that connects to "
"the Jack audio server, identifying itself as 'name'. "
"After creating the Jack plugin, you may want to create JackIn and/or JackOut "
"plugins that stream audio signal in from or out into the Jack graph.";

void
Py<Jack>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = AudioClock_type;
	t->tp_getset = Jack_getsets;
	t->tp_methods = Jack_methods;
}

void
Py<Jack>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {"name", 0};
	char * name;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s", keys, &name))
		throw blewit();
	
	init (name);
}

/* ///////////////////////////////////////////////////////////////////// */

Type * JackIn_type = 0;

char * Py<JackIn>::__doc__ = 
"JackIn (jack, n_ports) -- a plugin that streams "
"audio data in from the Jack graph.";

void
Py<JackIn>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
}

void
Py<JackIn>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {"jack", "n_ports", 0};
	Jack * jack = 0;
	int n_ports = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!i", keys,
			Jack_type, &jack, &n_ports))
		throw blewit();
	
	if (n_ports < 1)
		throw ValueError() << "nonsense without at least one port.";

	init (jack, n_ports);
}

/* ///////////////////////////////////////////////////////////////////// */

Type * JackOut_type = 0;

char * Py<JackOut>::__doc__ = 
"JackOut (jack, n_ports) -- a plugin that streams "
"audio signal out into the Jack graph.";

void
Py<JackOut>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
}

void
Py<JackOut>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {"jack", "n_ports", 0};
	Jack * jack = 0;
	int n_ports = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!i", keys,
			Jack_type, &jack, &n_ports))
		throw blewit();
	
	if (n_ports < 1)
		throw ValueError() << "nonsense without at least one port.";

	init (jack, n_ports);
}

#endif /* WITH_JACK */

/* ///////////////////////////////////////////////////////////////////// */

void
init_Jack_types (PyObject * dict)
{
	#ifdef WITH_JACK
	Jack_type = new PyType<Jack> ("Jack", dict);
	JackIn_type = new PyType<JackIn> ("JackIn", dict);
	JackOut_type = new PyType<JackOut> ("JackOut", dict);
	#endif /* WITH_JACK */
}
