/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "LFO.h"

#include "PyType.h"

d_float 
LFO::default_wave[] = {0, 1,};

LFO::LFO()
: Plugin (LFO_type, Audio)
{
	state = 0;
	tick = 0;

	rate = QUARTER * 2 / 3;
	interpolate = 1;

	wave = new d_float [LFO_WAVE_SIZE];
	set_wave (default_wave, 2);

	add_in (new In (this, FloatEvent_type, 1, "rate"));
	add_out (new Out (this, FloatEvent_type, "cv out"));
}

LFO::~LFO()
{
	delete [] wave;
}

void
LFO::set_wave (d_float * new_wave, int new_points)
{
	if (new_points < 2)
		throw OverflowError() << " need more than one point.";
	if (new_points > LFO_WAVE_SIZE)
		throw OverflowError() << LFO_WAVE_SIZE << " maximum points.";

	memcpy (wave, new_wave, points * sizeof (d_float));
	points = new_points;
}

void
LFO::seek()
{
	tick = graph->tick;
	msg(), "seek ", tick;
}

void
LFO::process_audio (int frames)
{
	d_frame f = graph->frame;

	d_tick t = graph->tick;

	while (frames > 0)
	{
		/* subcycle in 64 frame units. */
		int n = min (64, frames);

		bool is_loop = graph->is_loop_frame (f, f + n);

		d_tick next = graph->frame2tick (f + n);

		/* update the parameters from input FloatEvents. */
		ins[0]->pop_value (next, &rate);
		one_cycle (next - t);

		/* update the counters. */
		t = next;
		frames -= n;
		f += n;

		/* and do a possible loop jump. */
		if (is_loop)
			++loop_cycle,
			t = graph->loop.tick[0],
			f -= graph->loop.frame[1] - graph->loop.frame[0];
	}
}

void
LFO::one_cycle (d_tick ticks)
{
	/* anything is better than dividing by 0. */
	d_float r = (rate == 0) ? 0.0001 : rate;

	tick = fmod (tick + ticks, r);
	d_float phase = tick / r;

	d_float x = phase * points;
	int ix = (int) x;
	x -= ix;

	if (interpolate == 0)
		state = wave [ix % points];
	else
	{
		d_float L1 = wave [--ix % points];
		d_float L0 = wave [++ix % points];
		d_float H0 = wave [++ix % points];
		d_float H1 = wave [++ix % points];

		/* this 3rd order spline interpolation is from musicdsp.org;
		 * code credited to a Dave from muon software, 
		 * idea and reasoning by Josh Scholar. 
		 */
		state = 
			L0 + .5 * 
			x * (H0 - L1 + 
			x * (H0 + L0 * -2 + L1 + 
			x * ((H0 - L0) * 9 + (L1 - H1) * 3 + 
			x * ((L0 - H0) * 15 + (H1 -  L1) * 5 + 
			x * ((H0 - L0) * 6 + (L1 - H1) * 2)))));

		// msg(), tick, phase, L1, L0, H0, H1, state;
	}

	if (!outs[0]->connected)
		return;

	/* output. */
	FloatEvent * e = (FloatEvent *) outs[0]->cache->pop();
	e->value = state;
	e->tick = graph->tick;
	outs[0]->push (e);
}

/* //////////////////////////////////////////////////////////////////////// */

static int
LFO_set_rate (LFO * lfo, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &lfo->rate))
		return -1;
	return 0;
}

static PyObject *
LFO_get_rate (LFO * lfo)
{
	return PyFloat (lfo->rate);
}

static int
LFO_set_interpolate (LFO * lfo, PyObject * value)
{
	if (!PyArg_Parse (value, "i", &lfo->interpolate))
		return -1;
	return 0;
}

static PyObject *
LFO_get_interpolate (LFO * lfo)
{
	return PyInt (lfo->interpolate);
}

static PyObject *
LFO_get_wave_size (LFO * lfo)
{
	return PyInt (LFO_WAVE_SIZE);
}

static int
LFO_set_wave (LFO * lfo, PyObject * value)
Py_TRY
{
	if (!PySequence_Check (value))
		throw TypeError() << "need a list of values.";

	int l = PySequence_Size (value);
	
	/* can't just update the values because that would leave us with a garbage
	 * LFO wave if setting it fails somewhere half-way. */
	d_float new_wave [l];

	for (int i = 0; i < l; ++i)
		if (!PyArg_Parse (PySequence_GetItem (value, i), "d", new_wave + i))
			return -1;

	lfo->set_wave (new_wave, l);
		
	return 0;
}
Py_CATCH (-1);

static PyObject *
LFO_get_wave (LFO * lfo)
{
	PyObject * tuple = PyTuple_New (lfo->points);

	for (int i = 0; i < lfo->points; ++i)
		PyTuple_SetItem (tuple, i, PyFloat (lfo->wave[i]));

	return tuple;
}

static PyGetSetDef 
LFO_getsets [] = 
{
	{"rate", (getter) LFO_get_rate, (setter) LFO_set_rate,
	 "the wave cycle repeats every 'rate' ticks."},

	{"interpolate", (getter) LFO_get_interpolate, (setter) LFO_set_interpolate,
	 "whether to apply 3rd order spline interpolation."},

	{"wave", (getter) LFO_get_wave, (setter) LFO_set_wave,
	 "a tuple of numbers describing evenly spaced points on the wave."},

	{"wave_size", (getter) LFO_get_wave_size, (setter) 0,
	 "how many points the wave can contain."},

	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

Type * LFO_type = 0;

char * Py<LFO>::__doc__ = 
"LFO ([rate = 144, wave = (0, 1), interpolate = 1]) -- "
"A low-frequency oscillator, built around a wavepoint array. "
"The repeat 'rate' of the oscillator is expressed in ticks/cycle."
"The 'wave' can be "
"set to an arbitrary collection of numbers up to 'wave_size'; for generating "
"the output stream "
"they are seen as evenly spaced points on the wave. Points inbetween are "
"calculated by a third order spline algorithm, so if you want a square wave "
"LFO you need to turn off 'interpolate' or set a lot of points and accept "
"a certain overshoot.";

void
Py<LFO>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = LFO_getsets;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
}

void
Py<LFO>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"rate", "wave", "interpolate", 0};
	PyObject * _wave = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|dOi", keys,
				&rate, &_wave, &interpolate))
		return;
	
	if (_wave)
		LFO_set_wave (this, _wave);
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_LFO_types (PyObject * dict)
{
	LFO_type = new PyType<LFO> ("LFO", dict);
}
