/* (c) 2001-4 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include <dlfcn.h>
#include "basics.h"

#include "LadspaPlugin.h"
#include "Graph.h"
#include "PyType.h"

inline Type *
ladspa_event_type (const LADSPA_Descriptor * d, int i)
{
	if (d->PortDescriptors[i] & LADSPA_PORT_AUDIO)
		return CyclicAudio_type;
	return FloatEvent_type;
}

LadspaPort::LadspaPort (const LADSPA_Descriptor * d, int i)
{
	sample_rate = SAMPLE_RATE;
	index = i;
	event = 0;

	const LADSPA_PortRangeHint & hint = d->PortRangeHints [i];

	hints = hint.HintDescriptor;

	if (hints & LADSPA_HINT_BOUNDED_BELOW)
		bounds[0] = hint.LowerBound;
	else
		bounds[0] = -1;

	if (hints & LADSPA_HINT_BOUNDED_ABOVE)
		bounds[1] = hint.UpperBound;
	else
		bounds[1] = 1;

	LADSPA_Data def;

	/* if no default hints are provided, default to DEFAULT_LOW */
	def = bounds[0] + (bounds[0] + bounds[1]) / 4;

	double lo = bounds[0], hi = bounds[1];

	/* prevent calling log (0) */
	if (hints & LADSPA_HINT_LOGARITHMIC)
	{
		if (lo == 0) lo = .0000001;
		if (hi == 0) hi = .0000001;
	}

	int masked = hints & LADSPA_HINT_DEFAULT_MASK;

	if (masked == LADSPA_HINT_DEFAULT_MINIMUM)
		def = lo;
	else if (masked == LADSPA_HINT_DEFAULT_LOW)
	{
		if (hints & LADSPA_HINT_LOGARITHMIC)
			def = exp (log (lo) * .75 + log (hi) * .25);
		else
			def = lo * .75 + hi * .25;
	}
	else if (masked == LADSPA_HINT_DEFAULT_MIDDLE)
	{
		if (hints & LADSPA_HINT_LOGARITHMIC)
			def = exp (log (lo) * .5 + log (hi) * .5);
		else
			def = lo * .5 + hi * .5;
	}
	else if (masked == LADSPA_HINT_DEFAULT_HIGH)
	{
		if (hints & LADSPA_HINT_LOGARITHMIC)
			def = exp (log (lo) * .25 + log (hi) * .75);
		else
			def = lo * .25 + hi * .75;
	}
	else if (masked == LADSPA_HINT_DEFAULT_MAXIMUM)
		def = hi;
	else if (masked == LADSPA_HINT_DEFAULT_0)
		def = 0;
	else if (masked == LADSPA_HINT_DEFAULT_1)
		def = 1;
	else if (masked == LADSPA_HINT_DEFAULT_100)
	{
		def = 100;
		if (hints & LADSPA_HINT_SAMPLE_RATE)
			def /= sample_rate;
	}
	else if (masked == LADSPA_HINT_DEFAULT_440)
	{
		def = 440;
		if (hints & LADSPA_HINT_SAMPLE_RATE)
			def /= sample_rate;
	}
	
	if (hints & LADSPA_HINT_SAMPLE_RATE)
		def *= sample_rate;

	set_value (def);
}

void
LadspaPort::set_value (LADSPA_Data v)
{
	LADSPA_Data lower = bounds[0];
	LADSPA_Data upper = bounds[1];

	if (hints & LADSPA_HINT_SAMPLE_RATE)
		lower *= sample_rate,
		upper *= sample_rate;

	/* quietly saturate. */
	if (v < lower) v = lower;
	if (v > upper) v = upper;

	if (hints & LADSPA_HINT_INTEGER)
		v = floor (v + 0.49999);

	value = v;
}

LADSPA_Data
LadspaPort::get_value()
{
	return value;
}

/* //////////////////////////////////////////////////////////////////////// */

LadspaIn::LadspaIn (LadspaPlugin * p, const LADSPA_Descriptor * d, int i)
: In (p, ladspa_event_type (d, i), IN_SLOTS, d->PortNames[i]),
	LadspaPort (d, i)
{
	set_object_type (LadspaIn_type);
}

void
LadspaIn::event_cycle (d_tick tick)
{
	FloatEvent * e;

	while ((e = (FloatEvent *) pop_before (tick)))
	{
		value = e->value;
		::decref (e);
	}
}

LadspaOut::LadspaOut (LadspaPlugin * p, const LADSPA_Descriptor * d, int i)
: Out (p, ladspa_event_type (d, i), d->PortNames[i]),
	LadspaPort (d, i)
{
	set_object_type (LadspaOut_type);
}

void
LadspaOut::event_cycle (d_tick tick)
{
	if (old_value == value || connected == 0)
		return;
	
	FloatEvent * e = (FloatEvent *) cache->pop();
	
	e->tick = tick;
	
	old_value = 
	e->value = value;

	push (e);
}

/* //////////////////////////////////////////////////////////////////////// */

LadspaPlugin::LadspaPlugin()
: Plugin (LadspaPlugin_type, Audio),
	ports ("LadspaPorts")
{
	dlpath = 0;
	dlhandle = 0;

	handle = 0;
	subcycle = 0;
}

LadspaPlugin::~LadspaPlugin()
{
	cleanup();

	if (dlpath) free (dlpath);
	if (dlhandle) dlclose (dlhandle);
}

void
LadspaPlugin::cleanup()
{
	if (handle == 0)
		return;
		
	if (descriptor->deactivate)
		descriptor->deactivate (handle);

	descriptor->cleanup (handle);

	handle = 0;
}

void
LadspaPlugin::open (const char * _path)
{
	if (strchr (_path, '/'))
		dlpath = strdup (_path);
	else
		asprintf (&dlpath, LADSPA_DEFAULT_PATH"%s", _path);
		
	dlhandle = dlopen (dlpath, RTLD_LAZY);

	if (!dlhandle)
		throw OSError() << dlerror();
	
	get_descriptor =
		(LADSPA_Descriptor_Function) dlsym (dlhandle, "ladspa_descriptor");
	
	if (!get_descriptor)
		throw LookupError() << "dlsym ('ladspa_descriptor'): " << dlerror();
}

/* init as next plugin from a library */
void
LadspaPlugin::init (LadspaPlugin * p)
{
	open (p->dlpath);

	descriptor = get_descriptor (index = p->index + 1);

	if (descriptor == 0)
		throw StopIteration() << "no more plugins in " << this;
	
	setup();
}

void
LadspaPlugin::init (int id)
{
	index = 0;
	while ((descriptor = get_descriptor (index)))
	{
		if ((int) descriptor->UniqueID == id || id == -1)
			return setup();
		
		index++;
	}

	throw LookupError() << "plugin descriptor not found by id: " << id;
}

void
LadspaPlugin::init (char * name)
{
	int n = strlen (name);

	index = 0;
	while ((descriptor = get_descriptor (index)))
	{
		if (strncmp (descriptor->Label, name, n) == 0)
			return setup();
		
		index++;
	}

	throw LookupError() << "plugin descriptor not found by name: " << name;
}

void
LadspaPlugin::setup()
{
	set_object_name (descriptor->Label);

	audio_ports = 0;

	for (int i = 0; i < (int) descriptor->PortCount; i++)
	{
		LadspaPort * port = 0;

		int f = descriptor->PortDescriptors[i];

		if (f & LADSPA_PORT_INPUT)
		{
			LadspaIn * in = new LadspaIn (this, descriptor, i);

			add_in (in);
			port = (LadspaPort *) in;

			if (in->is_audio())
				++audio_ports;
		}
		else
		{
			LadspaOut * out = new LadspaOut (this, descriptor, i);
			
			add_out (out);
			port = (LadspaPort *) out;

			if (out->is_audio)
				++audio_ports;
		}

		ports.add_object (port);
	}
	
	if (!(descriptor->Properties & LADSPA_PROPERTY_HARD_RT_CAPABLE))
		msg(), this, " is not HARD_RT_CAPABLE.";

	if (descriptor->Properties & LADSPA_PROPERTY_INPLACE_BROKEN)
		msg(), this, " is INPLACE_BROKEN.";
}

void
LadspaPlugin::instantiate (int fs)
{
	handle = descriptor->instantiate (descriptor, fs);
	if (handle == 0)
		throw AssertionError() << this << " refuses instantiation.";

	/* connect all ports. this also, without much meaning, connects the audio 
	 * ports but it's more readable than looping twice, over ins and outs. 
	 */
	for (int i = 0; i < ports; ++i)
	{
		ports[i]->sample_rate = fs;
		connect (ports[i]);
	}
}

void
LadspaPlugin::set_state (Graph * g, int _state)
{
	if (_state == state)
		return;

	if (_state == Off)
		cleanup();
	else if (_state == Prepared)
	{
		if (state != Off)
			throw AssertionError() << this << " can only prepare in stopped state.";

		if (handle == 0)
			instantiate (g->sample_rate);
		
		if (descriptor->activate)
			descriptor->activate (handle);
	}		

	Plugin::set_state (g, _state);
}

void
LadspaPlugin::connect (LadspaPort * port, LADSPA_Data * location)
{
	if (location == 0)
		location = &port->value;

	descriptor->connect_port (handle, port->index, location);
}

void
LadspaPlugin::process_audio (int frames)
{
	if (handle == 0)
		return;

	CyclicAudio * silence = graph->silence;

	/* assign input events */

	// msg(), get_name (this), ":";

	int have = 0;
	CyclicAudio * in_events [(int) ins];

	for (int i = 0; i < ins; ++i)
	{
		LadspaIn * in = (LadspaIn *) ins[i];

		if (!in->is_audio())
			continue;

		in->event = in->pop_mix();
		
		if (in->event == 0)
			in->event = silence;
		else
			in_events [have++] = in->event;

		// msg(), i, " (in): ", (void *) in->event;
	}

	bool inplace = !(descriptor->Properties & LADSPA_PROPERTY_INPLACE_BROKEN);

	for (int o = 0; o < outs; ++o)
	{
		LadspaOut * out = (LadspaOut *) outs[o];

		if (!out->is_audio)
		{
			/* to detect a change of the value, store it. later we'll compare and
			 * emit an event describing the change. */
			out->old_value = out->value;
			continue;
		}
		
		/* try an inplace event from an input */
		if (inplace && have)
			out->event = in_events [--have];
		else
		{
			/* fetch an event from the cache and set the frame time. */
			out->event = (CyclicAudio *) out->cache->pop();
			out->event->frame = graph->frame;
		}

		// msg(), o, " (out): ", (void *) out->event;
	}

	/* */

	d_frame f = graph->frame;

	int per_cycle = subcycle ? 64 : frames;
	int written = 0;

	while (written < frames)
	{
		int n = min (per_cycle, frames - written);

		bool is_loop = graph->is_loop_frame (f, f + n);
		d_tick tick = graph->frame2tick (f + n);

		for (int i = 0; i < ins; ++i)
		{
			LadspaIn * in = (LadspaIn *) ins[i];

			if (in->is_audio())
				connect (in, in->event->data + written);
			else
				in->event_cycle (tick);
		}

		for (int i = 0; i < outs; ++i)
		{
			LadspaOut * out = (LadspaOut *) outs[i];

			if (out->is_audio)
				connect (out, out->event->data + written);
		}

		descriptor->run (handle, n);
		
		for (int i = 0; i < outs; ++i)
		{
			LadspaOut * out = (LadspaOut *) outs[i];

			if (!out->is_audio)
				out->event_cycle (tick);
		}

		f += n;
		written += n;

		if (is_loop)
			++loop_cycle,
			f -= graph->loop.frame[1] - graph->loop.frame[0];
	}

	while (have)
		in_events[--have]->return_to_cache();
	
	for (int i = 0; i < outs; ++i)
	{
		LadspaOut * out = (LadspaOut *) outs[i];

		if (out->is_audio)
			out->push (out->event);
	}
}

void
LadspaPlugin::one_shot (int n, AudioSignal ** s)
{
	if (n <= 0)
		throw ValueError() << "no signals to play with.";

	int frames = s[0]->size;

	if (handle == 0)
		throw RuntimeError() << "plugin isn't instantiated.";

	d_time then = gettimeofday();

	int m = 0;
	
	for (int i = 0; i < ins; ++i)
	{
		LadspaIn * port = (LadspaIn *) ins[i];

		if (!(port->is_audio()))
			continue;
		if (m == n)
			throw ValueError() << "not enough AudioSignals provided";
		if (s[m]->size != frames)
			throw ValueError() << "signal " << m << " is wrong size.";
		connect (port, s[m]->data);
		m++;
	}

	for (int i = 0; i < outs; ++i)
	{
		LadspaOut * port = (LadspaOut *) outs[i];

		if (!(port->is_audio))
			continue;
		if (m == n)
			throw ValueError() << "not enough AudioSignals provided";
		if (s[m]->size != frames)
			throw ValueError() << "signal " << m << " is wrong size.";

		connect (port, s[m]->data);
		m++;
	}

	if (m < n)
		msg(), "ignoring ", n - m, "extra signals in ", this,"::one_shot()";

	descriptor->run (handle, frames);

	/* tell us how long it took. */
	cpu = gettimeofday() - then;
}

/* //////////////////////////////////////////////////////////////////////// */

static int
LadspaPlugin_set_subcycle (LadspaPlugin * p, PyObject * value)
{
	if (!PyArg_Parse (value, "i", &p->subcycle))
		return -1;
	return 0;
}

static PyObject *
LadspaPlugin_get_subcycle (LadspaPlugin * p)
{
	return PyInt (p->subcycle);
}

static PyObject *
LadspaPlugin_get_id (LadspaPlugin * p)
{
	return PyInt (p->descriptor->UniqueID);
}

static PyObject *
LadspaPlugin_get_label (LadspaPlugin * p)
{
	return PyString (p->descriptor->Name);
}

static PyGetSetDef 
LadspaPlugin_getsets [] = {
	{"subcycle", (getter) LadspaPlugin_get_subcycle, (setter) LadspaPlugin_set_subcycle,
	 "whether to subdivide audio cycles into chunks of 64 samples "
	 "for better input response."},
	{"id", (getter) LadspaPlugin_get_id, (setter) 0,
	 "the plugin's unique ID, according to its LADSPA descriptor."},
	{"label", (getter) LadspaPlugin_get_label, (setter) 0,
	 "the plugin's LADSPA descriptor member 'Name'."},
	{ }
};

static PyObject *
LadspaPlugin_next (LadspaPlugin * p)
{
	LadspaPlugin * next = 0;

	try 
	{
		/* have to use the GC allocator. */
		next = NEW (LadspaPlugin);
		next->init (p);

		return next;
	}
	catch (blewit & b)
	{
		/* we'd like to 'catch (StopIteration &)' but that doesn't work, at least
		 * not with gcc 2.95. 
		 */
		delete next;
		
		if (b.py_exc == PyExc_StopIteration)
			return PyNone();
		
		b.throw_at_python();
	}
	
	return 0;
}

static PyObject *
LadspaPlugin_one_shot (LadspaPlugin * lad, PyObject * args)
Py_TRY
{
	if (lad->graph)
		throw AssertionError() << "this plugin managed by a graph.";

	if (lad->state != Plugin::Off)
		throw RuntimeError() << "busy.";

	int n = PySequence_Size (args);
	
	if (n <= 0)
		throw ValueError() << "need a few audio signals.";
	
	AudioSignal * s[n];

	for (int i = 0; i < n; ++i)
	{
		s[i] = (AudioSignal *) PySequence_GetItem (args, i);
		if (!AudioSignal_type->is_derived (s[i]))
		{
			while (--i >= 0)
				::decref (s[i], true);
			throw TypeError() << "argument " << i << " is not an AudioSignal.";
		}
	}

	try 
	{
		lad->one_shot (n, s);
	} 
	catch (blewit & b) 
	{
		for (int i = 0; i < n; ++i)
			::decref (s[i], true);
		b.throw_at_python();
		return 0;		
	}
	
	for (int i = 0; i < n; ++i)
		::decref (s[i], true);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
LadspaPlugin_instantiate (LadspaPlugin * lad, PyObject * args)
Py_TRY 
{
	if (lad->graph)
		throw AssertionError() << "this plugin managed by a graph.";

	if (lad->handle)
		throw AssertionError() << "already instantiated.";

	int fs = 44100;

	if (!PyArg_ParseTuple (args, "|i", &fs))
		return 0;
	
	lad->instantiate (fs);

	if (lad->descriptor->activate)
		lad->descriptor->activate (lad->handle);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
LadspaPlugin_reset (LadspaPlugin * lad)
Py_TRY 
{
	if (lad->graph)
		throw AssertionError() << "this plugin managed by a graph.";

	if (lad->handle == 0)
		throw AssertionError() << "not instantiated yet.";

	if (lad->descriptor->deactivate)
		lad->descriptor->deactivate (lad->handle);

	if (lad->descriptor->activate)
		lad->descriptor->activate (lad->handle);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef 
LadspaPlugin_methods [] = 
{
  {"instantiate", (PyCFunction) LadspaPlugin_instantiate, METH_VARARGS,
	 "instantiate ([sample rate = 44100]) -- prepare the plugin for off-line "
	 "processing."},
  {"next", (PyCFunction) LadspaPlugin_next, METH_NOARGS,
	 "next() -- returns the next plugin from the shared object, or None."},
  {"one_shot", (PyCFunction) LadspaPlugin_one_shot, METH_VARARGS,
	 "one_shot (audio signals) -- run the plugin on the provided audio data."},
  {"reset", (PyCFunction) LadspaPlugin_reset, METH_NOARGS,
	 "reset() -- resets the plugin's internal state by calling its deactivate() "
	 "and activate() methods."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

/* LadspaPlugin as mapping: strings are compared to {In,Out}->object_name */
static int
LadspaPlugin_length (LadspaPlugin * p)
{
	return p->ins + p->outs;
}

static PyObject *
LadspaPlugin_find_port (LadspaPlugin * p, PyObject * key)
{
	char * s = 0;

	if (!PyArg_Parse (key, "s", &s))
		return 0;
	
	/* is it an In? */
	for (int i = 0; i < p->ins; ++i)
		if (!strcmp (get_name (p->ins[i]), s))
			return p->ins[i];

	/* is it an Out? */
	for (int i = 0; i < p->outs; ++i)
		if (!strcmp (get_name (p->outs[i]), s))
			return p->outs[i];
	
	throw KeyError() << p << " has no port named '" << s << "'.";
}

static PyObject *
LadspaPlugin_subscript (LadspaPlugin * p, PyObject * key)
Py_TRY
{
	return ::incref (LadspaPlugin_find_port (p, key));
}
Py_CATCH (0);

static int
LadspaPlugin_ass_subscript (LadspaPlugin * p, PyObject * key, PyObject * val)
Py_TRY
{
	float f;

	if (!PyArg_Parse (val, "f", &f))
		return -1;
	
	PyObject * port = LadspaPlugin_find_port (p, key);
	
	if (LadspaOut_type->is_derived (port))
		((LadspaPort *) ((LadspaOut *) port))->set_value (f);
	else /* (LadspaIn_type->is_derived (port)) */
		((LadspaPort *) ((LadspaIn *) port))->set_value (f);

	return 0;
}
Py_CATCH (-1);

static PyMappingMethods 
LadspaPlugin_as_mapping = 
{
	(inquiry) LadspaPlugin_length, /* mp_length */
	(binaryfunc) LadspaPlugin_subscript, /* mp_subscript */
	(objobjargproc) LadspaPlugin_ass_subscript, /* mp_ass_subscript */
};

Type * LadspaPlugin_type = 0;

char * Py<LadspaPlugin>::__doc__ = 
"LadspaPlugin ('ladspa.so'[, name = 'name' or id = -1, subcycle = 0]) -- "
"an external audio DSP plugin. The plugin is searched for in "
"'"LADSPA_DEFAULT_PATH"' "
"if the filename does not mention a path (contains no '/').\n"
"If the file opened contains more than one plugin, use 'name' to "
"choose one by its name, or 'id' to choose by its unique LADSPA ID. "
"The default is to instantiate the first plugin from the file.";

static void
ladspa_hint (PyObject * dict, int i, char * s)
{
	PyDict_SetItemString (dict, s, PyInt (i));
}

void
Py<LadspaPlugin>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = LadspaPlugin_getsets;
	t->tp_methods = LadspaPlugin_methods;
	t->tp_as_mapping = &LadspaPlugin_as_mapping;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;

	PyType_Ready (t);

	ladspa_hint (t->tp_dict, LADSPA_HINT_BOUNDED_BELOW, "BOUNDED_BELOW");
	ladspa_hint (t->tp_dict, LADSPA_HINT_BOUNDED_ABOVE, "BOUNDED_ABOVE");
	ladspa_hint (t->tp_dict, LADSPA_HINT_TOGGLED, "TOGGLED");
	ladspa_hint (t->tp_dict, LADSPA_HINT_SAMPLE_RATE, "SAMPLE_RATE");
	ladspa_hint (t->tp_dict, LADSPA_HINT_LOGARITHMIC, "LOGARITHMIC");
	ladspa_hint (t->tp_dict, LADSPA_HINT_INTEGER, "INTEGER");
}

void
Py<LadspaPlugin>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"path", "name", "id", "subcycle", 0};
	char * path;
	char * name = 0;
	int id = -1;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|sii", keys,
				&path, &name, &id, &subcycle))
		throw blewit();

	if (id != -1 && name)
		throw AssertionError() << "cannot specify both 'name' and 'id'.";
	
	open (path);
	if (name)
		init (name);
	else
		init (id);
}

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
LadspaPort_get_bounds (LadspaPort * l)
{
	PyObject * t = PyTuple_New (2);

	for (int i = 0; i < 2; ++i)
	{
		LADSPA_Data b = l->bounds[i];
		
		if (l->hints & LADSPA_HINT_SAMPLE_RATE)
			b *= l->sample_rate;

		PyTuple_SetItem (t, i, PyFloat (b));
	}

	return t;
}

static PyObject *
LadspaIn_get_bounds (LadspaIn * l)
{
	return LadspaPort_get_bounds (l);
}

static PyObject *
LadspaIn_get_value (LadspaIn * l)
{
	return PyFloat (l->get_value());
}

static int
LadspaIn_set_value (LadspaIn * l, PyObject * value)
Py_TRY
{
	LADSPA_Data v;

	if (!PyArg_Parse (value, "f", &v))
		throw blewit();

	l->set_value (v);

	return 0;
}
Py_CATCH (-1);

static PyObject *
LadspaIn_get_hints (LadspaIn * l)
{
	return PyInt (l->hints);
}

static PyGetSetDef 
LadspaIn_getsets [] = {
	{"bounds", (getter) LadspaIn_get_bounds, 0,
	 "a tuple containing the lower and upper bound for the port value."},
	{"value", (getter) LadspaIn_get_value, (setter) LadspaIn_set_value,
	 "the current port value."},
	{"hints", (getter) LadspaIn_get_hints, 0,
	 "the hints define additional information about the port value, "
	 "according to the constants defined in LadspaPlugin."},
	{ }
};

Type * LadspaIn_type = 0;

char * Py<LadspaIn>::__doc__ = 
"LadspaIn() -- "
"an In port on a LadspaPlugin.\n"
"You cannot instantiate this in Python.";

void
Py<LadspaIn>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = In_type;
	t->tp_getset = LadspaIn_getsets;
}

void
Py<LadspaIn>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
LadspaOut_get_bounds (LadspaOut * l)
{
	return LadspaPort_get_bounds (l);
}

static PyObject *
LadspaOut_get_value (LadspaOut * l)
{
	return PyFloat (l->get_value());
}

static PyObject *
LadspaOut_get_hints (LadspaOut * l)
{
	return PyInt (l->hints);
}

static PyGetSetDef 
LadspaOut_getsets [] = {
	{"bounds", (getter) LadspaOut_get_bounds, 0,
	 "a tuple containing the lower and upper bound for the port value."},
	{"value", (getter) LadspaOut_get_value, 0,
	 "the current port value."},
	{"hints", (getter) LadspaOut_get_hints, 0,
	 "the hints define additional information about the port value, "
	 "according to the constants defined in LadspaPlugin."},
	{ }
};

Type * LadspaOut_type = 0;

char * Py<LadspaOut>::__doc__ = 
"LadspaOut() -- "
"an Out port on a LadspaPlugin.\n"
"You cannot instantiate this in Python.";

void
Py<LadspaOut>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Out_type;
	t->tp_getset = LadspaOut_getsets;
}

void
Py<LadspaOut>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* //////////////////////////////////////////////////////////////////////// */

void init_Ladspa_types (PyObject * dict)
{
	LadspaPlugin_type = new PyType<LadspaPlugin> ("LadspaPlugin", dict);
	LadspaIn_type = new PyType<LadspaIn> ("LadspaIn", dict);
	LadspaOut_type = new PyType<LadspaOut> ("LadspaOut", dict);
}
