/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _LIST_H_
#define _LIST_H_

#include "basics.h"

#include "Chain.h"

/* The basic List stores fixed-size objects, allocating memory in steps of
 * a given granularity. it is only capable of returning void pointers to
 * the listed objects. It keeps a chain of iterators that are adjusted when
 * objects are inserted or removed. 
 *
 * the core functionality is in the List. the templates derived basically only 
 * supply type casts for the list items. 
 *
 * ObjectLists however perform reference counting on the pointers to objects
 * they contain. */

class List;
class Slice;

class BasicIterator;
template <class T> class Iterator;
	
template <class T> class InstanceList;
class CharList;
class ByteList;
class SampleList;

template <class T> class PointerList;
class ObjectList;

class BasicIterator
:	public Chained<BasicIterator>
{
	friend class List;
	
	public:
		List * list;
		int index;

	public:
		BasicIterator();
		BasicIterator (BasicIterator &);

		virtual ~BasicIterator();
			
		void init (List * l, int i = 0);

		virtual void pre_remove (int i, int n)
			{
				if (index > i && index < i + n)
					index = i;

				if (this && get_next())
					get_next()->pre_remove (i, n);
			}

		virtual void post_insert (int i, int n)
			{
				if (index > i)
					index += n;

				if (this && get_next())
					get_next()->post_insert (i, n);
			}
		
		void move (int i)
			{
				init (list, i);
			}

		void next() 
			{
				index++;
			}

		int have();

		void * resolve();
		
		operator int();

		bool operator == (BasicIterator & it) { return operator == ((int) it); }
		bool operator <= (BasicIterator & it) { return operator <= ((int) it); }
		bool operator <  (BasicIterator & it) { return operator < ((int) it); }
		bool operator >= (BasicIterator & it) { return operator >= ((int) it); }
		bool operator >  (BasicIterator & it) { return operator > ((int) it); }

		bool operator == (int i) { return index == i; }
		bool operator > (int i) { return index > i; }
		bool operator >= (int i) { return index >= i; }
		bool operator < (int i) { return index < i; }
		bool operator <= (int i) { return index <= i; }

		bool valid();
};

template <class T>
class Iterator
:	public BasicIterator
{
	public:
		Iterator (List * l = 0, int i = 0)
			: BasicIterator()
			{
				init (l, i);
			}
		
		Iterator (BasicIterator & it)
			: BasicIterator (it)
			{}
			
		operator T () 
			{ 
				return get();
			}

		T get()
			{
				return (T &) *((T *) list->item (index)); 
			}
};

class IteratorChain
:	public Chain<BasicIterator>
{
	public:
		IteratorChain()
			: Chain<BasicIterator>()
			{}
								
		void
			pre_remove (int i, int n) 
				{ 
					if (first)
						get_first()->pre_remove (i, n); 
				}
					
		void
			post_insert (int i, int n) 
				{ 
					if (first)
						get_first()->post_insert (i, n); 
				}
};

/* the List */

class List
{
	protected:
		size_t size;
		int allocated, granularity;
		
		IteratorChain iterators;
			
	protected:
		void reallocate (int n);

	public:
		void * items;
		int length;
		const char * name;
		
		enum {
			Start = 0, 
			End = -1
		};

		List (const char * _name, int _size = 0, int g = 16)
			: iterators()
			{
				name = _name;
				length = 0;
				items = 0;
				size = _size;
				allocated = 0;
				granularity	= g;

				/* can allocate later. */
				if (_size)
					init (_size);
			}
		
		virtual ~List()
			{
				clear();

				if (items)
					free (items);
			}

		void init (int _size)
			{
				if (items) 
					throw AssertionError() << "List '" << name << "' at " << this 
																 << "already has an allocation.";
				allocated = granularity;
				items = malloc (allocated * size);
			}

		int	get_item_size() { return size; }
		
		void * item (int i) 
			{
				if (i < 0 || i >= length)
				{
					throw IndexError() << "no item " << i << " in List '" << name 
														 << "' at " << this;
				}

				return (void *) ((byte *) items + i * size);
			}

		operator int () { return length; }

		void add_iterator (BasicIterator * it) { iterators.prepend (it); }
		void remove_iterator (BasicIterator * it) { iterators.remove (it); }
									
	private:
		void basic_post_insert (int i, int n = 1) 
			{	
				if (i < 0 || i >= length || n < 1) return;
					iterators.post_insert (i, n);
					post_insert (i, n);
			}
		
		void basic_pre_remove (int i, int n = 1) 
			{	
				if (i < 0 || i >= length || n < 1) return;
				iterators.pre_remove (i, n);
				pre_remove (i, n);
			}
		
	protected:
		virtual void post_insert (int i, int n = 1)
			{ }
		
		virtual void pre_remove (int i, int n = 1)
			{ }

		
	public:
		int insert (int n = 1, int i = End);
		void copy_items (const void * data, int n);
		
		int insert_items (const void * data, int n = 1, int i = End);
		bool remove_items (int i, int n = 1);
		
		int insert_item (void * data, int i = End)
			{
				return insert_items (data, 1, i);
			}

		/* uses memcpy regardless of sizeof(item). don't use with small items
		 * if you're worried about performance. */
		void swap (int m, int n);

		void move (int from, int to);

		/* this makes the list forget it allocated anything. therefore, after
		 * calling this, the list is not usable anymore. returns the items buffer
		 * that you should later free(). */
		void * forget_allocation();
		
		void clear();
};

/* instance list, and instance lists */

template <class T>
class InstanceList
:	public List
{
	public:
		InstanceList (const char * _name, int g = 8)
			: List (_name, sizeof (T), g)
			{ }

		void insert_item (const T & t, int i = List::End)
			{
				insert_items (&t, 1, i);
			}

		operator T * () 
			{
				return (T *) items;
			}

		T &	operator [] (int i)
			{
				return *((T *) item (i));
			}

		InstanceList<T> & operator << (T & t)
			{
				insert_item (t);

				return *this;
			}
};

class CharList
:	public InstanceList<char>
{
	public:
		CharList (const char * _name)
			: InstanceList<char> (_name, 80)
			{ }
		
		CharList & operator << (char * s)
			{
				int n = strlen (s);
				reallocate (n);
				memcpy (item (length), s, n);
				length += n;
				return *this;
			}

		CharList & operator << (char c)
			{
				reallocate (length + 1);
				(*this) [length++] = c;
				return *this;
			}
};

/* pointer list, and pointer lists */

template <class T>
class PointerList
:	public List
{
	public:
		PointerList (const char * _name, int g = 8)
			: List (_name, sizeof (T *), g)
			{ }

		T *	operator [] (int i)
			{
				return *((T **) item (i));
			}

		int find (const T * item)
			{
				for (int i = 0; i < length; i++)
					if ((*this) [i] == item)
						return i;
				return -1;
			}

		virtual int add_object (T * obj, int i = End)
			{
				i = insert (1, i);
				*((T **) item (i)) = obj;
				post_insert (i);
				return i;
			}
		
		virtual int remove (const T * obj)
			{
				int i = find (obj);
				if (i < 0) throw LookupError() << "object to remove was not found.";
				remove_items (i, 1);
				return i;
			}
};

class ObjectList
:	public PointerList<Object>
{
	public:
		bool going_down;
		
		ObjectList (const char * _name, int g = 12)
			: PointerList<Object> (_name, g)
			{ 
				going_down = 0;
			}

		~ObjectList()
			{
				/* can 'really' decref in a destructor. */
				going_down = true;

				clear();
			}
		
		void clear_and_decref()
			{
				for (int i = 0; i < length; ++i)
					::decref ((*this)[i], true);
				length = 0;
			}
		
		virtual void post_insert (int i, int n) 
			{
				List::post_insert (i, n);
				
				while (n--) 
					::incref ((*this) [i++]);
			}

		virtual void pre_remove (int i, int n) 
			{
				List::pre_remove (i, n);
				
				while (n--) 
					::decref ((*this) [i++], going_down);
			}
};

template <class T>
class TypedObjectList
:	public ObjectList
{
	public:
		TypedObjectList (const char * _name, int g = 12)
			: ObjectList (_name, g)
			{ }

		T * & operator [] (int i)
			{ 
				return *((T **) item (i)); 
			}
};

#endif /* _LIST_H_ */
