/* (c) 2001-4 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "TempoMap.h"
#include "TimeMap.h"

#include "PyType.h"
#include "PythonThread.h"

/* tempo maps are managed by a graph. Access to the tempo map follows the
 * GraphRequest protocol, ie. happens under realtime constraints. The
 * granularity of the EventList used is comparably high, to avoid audio
 * glitches because of a tempo-storing operation. 
 */
TempoMap::TempoMap()
: events (TempoEvent_type, 8192)
{
	quarter = QUARTER;
	current = 0;
}

TempoMap::~TempoMap()
{
}

void
TempoMap::add (d_float bpm, d_tick tick)
{
	int i = events.index_of_tick (tick);

	if (i < 0 || i >= events || events[i]->tick != tick)
	{
		/* NEW() isn't rt-friendly but we hold the PL anyway. */
		TempoEvent * e = NEW (TempoEvent);

		e->bpm = bpm;
		e->tick = tick;

		i = events.add_event (e);

		/* shift current event */
		if (i <= current)
			++current;
			
		::decref (e);
	}
	else
	{
		events[i]->bpm = bpm;
	}
}

void
TempoMap::remove (d_tick tick, int n)
{
	int i = events.index_of_tick (tick);
	
	if (i < 0 || i >= events)
	{
		msg(), "no tempo events to remove after ", tick;
		return;
	}

	if (events[i]->tick < tick)
		i++;

	/* clamp n to the valid range */
	if (n < 0)
		n = (int) events - i;
	else
		n = min (n, (int) events - i);
	
	/* adjust current event */
	if (current > i)
	{
		if (current < i + n)
			current = i;
		else
			current -= n;
	}

	events.remove_items (i, min (n, (int) events - i));
}

void
TempoMap::seek_tick (d_tick tick)
{
	current = events.index_of_tick (tick);
}

void
TempoMap::seek_time (d_time time)
{
	int i;
	
	for (i = 1; i < events; i++)
		if (events[i]->time > time)
			break;

	current = i - 1;
}

/* for every tempo change, calculate time and seconds per tick */
void
TempoMap::recalculate (d_tick _quarter)
{
	quarter = _quarter;

	d_time s_per_tick = 
			default_sec_per_tick = (60. / BPM) / quarter;

	d_tick tick = 0;
	d_time time = 0;

	for (int i = 0; i < events; i++)
	{
		TempoEvent * e = events[i];
		
		d_tick dtick = e->tick - tick;

		tick = e->tick;
		
		time += s_per_tick * (d_time) dtick;

		e->time = time;
	
		e->sec_per_tick = 
				s_per_tick = (60. / e->bpm) / quarter;
	}
}

d_time
TempoMap::virtual_tick2time (d_tick zone, d_tick t)
{
	if ((int) events == 0)
		return default_sec_per_tick * t;

	int i = 1;

	/* find the event just prior to t */
	while (i < events && events[i]->tick < zone)
		++i;
	
	TempoEvent * e = events[i - 1];
	
	d_time dt = (d_time) (t - e->tick);
	return e->time + e->sec_per_tick * dt;
}

d_tick
TempoMap::virtual_time2tick (d_time zone, d_time t)
{
	if ((int) events == 0)
		return t / default_sec_per_tick;

	int i = 1;

	/* find the event just prior to t */
	while (i < events && events[i]->time < zone)
		++i;
	
	TempoEvent * e = events[i - 1];
	
	return e->tick + (d_tick) ((t - e->time) / e->sec_per_tick);
}

/* /////////////////////////////////////////////////////////////////////// */

TimeMap::TimeMap()
: NamedObject (TimeMap_type),
	events (TimeEvent_type),
	default_time (4, 4, QUARTER, 0)
{
}

TimeMap::~TimeMap()
{
}

int
TimeMap::add (TimeEvent * e)
{
	TimeIter iter (this, e->tick);

	if (iter.valid() && ((TimeEvent *) iter)->tick == e->tick)
	{
		msg(), "removing double time event at ", e->tick;
		events.remove_items (iter.index);
	}
	
	return events.add_event (e);
}

void
TimeMap::remove (TimeEvent * e)
{
	events.remove (e);
}

void
TimeMap::tick2bbt (d_tick tick, BBT & bbt)
{
	int i = events.index_of_tick (tick) - 1;
	
	if (i >= 0 && i < events)
		(*this)[i]->map_tick (tick, bbt);
	else
		default_time.map_tick (tick, bbt);
}

d_tick
TimeMap::bbt2tick (BBT & bbt)
{
	int hi = events.length, lo = -1;
	TimeEvent * e;

	while (hi - lo > 1)
	{
		int probe = (hi + lo) / 2;

		e = (TimeEvent *) events [probe];
		if (e->bbt.compare (bbt) < 0)
			lo = probe;
		else
			hi = probe;
	}

	lo = max (lo, 0);
	e = lo < events.length ? (TimeEvent *) events[lo] : &default_time;

	return e->map_bbt (bbt);
}

void
TimeMap::recalculate()
{
	TimeIter iter (this, 0);
	TimeEvent * last = &default_time;
	
	while (iter.valid())
	{
		last->update_length();

		TimeEvent * e = iter;
		last->map_tick (e->tick, e->bbt);
		last = e;

		iter.next();
	}
}

/* // TimeIter ///////////////////////////////////////////////////////////// */

void 
TimeEvent::map_tick (d_tick t, BBT & b)
{
	d_tick delta = t - tick;

	b.bar = (int) (delta / length.bar);
	delta -= b.bar * length.bar;
	
	b.beat = (int) (delta / length.beat);
	b.tick = delta - b.beat * length.beat;

	b.bar += bbt.bar;
	b.beat += bbt.beat;
}

d_tick
TimeEvent::map_bbt (BBT & b)
{
	d_tick t = tick;

	t += (b.bar - bbt.bar) * length.bar;
	t += (b.beat - bbt.beat) * length.beat;
	t += (b.tick - bbt.tick);

	return t;
}

void
TimeEvent::update_length()
{
	length.beat = ticks * (4 / d);
	length.bar = n * length.beat;
}

TimeIter::TimeIter (TimeMap * map, d_tick t)
: Iterator<TimeEvent *> (&map->events, map->events.index_of_tick (t))
{ 
	tick = t;
}

/* // Py<TimeMap> ////////////////////////////////////////////////////////// */

/* sequence methods */

static int 
TimeMap_sq_length (TimeMap * p)
{
	return p->events.length;
}

static PyObject *
TimeMap_sq_item (TimeMap * p, int i) 
Py_TRY
{
	if (i >= 0 && i < p->events)
		return ::incref ((*p)[i]); 
	throw IndexError() << "no item " << i << " in " << p;
} 
Py_CATCH (0);

static PySequenceMethods 
TimeMap_as_sequence = 
{
	(inquiry) TimeMap_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) TimeMap_sq_item, /* sq_item */
	(intintargfunc) 0, /* sq_slice */
	(intobjargproc) 0, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

static PyObject *
TimeMap_add (TimeMap * self, PyObject * args)
Py_TRY
{
	TimeEvent * event = 0;
	d_tick t = D_TICK_MAX;
	
	if (!PyArg_ParseTuple (args, "O!|d", TimeEvent_type, &event, &t))
		return 0;

	if (event == 0)
		throw ValueError() << "needs a TimeEvent";

	/* clone it. */
	TimeEvent * e = NEW (TimeEvent);
	e->clone (event);
	if (t != D_TICK_MAX)
		e->tick = t;

	ScopedLock lock (self->events.mutex);

	try 
	{
		self->add (e);
		self->recalculate();
	} 
	catch (blewit & b) 
	{
		::decref (e);
		throw;
	}

	return e;
} 
Py_CATCH (0);

static PyObject *
TimeMap_remove (TimeMap * self, PyObject * args)
Py_TRY
{
	TimeEvent * e = 0;
	
	if (!PyArg_ParseTuple (args, "O!", TimeEvent_type, &e))
		return 0;

	ScopedLock lock (self->events.mutex);

	self->remove (e);
	self->recalculate();

	return PyNone();
} 
Py_CATCH (0);

static PyObject * 
TimeMap_lock (TimeMap * p)
{
	p->events.mutex.lock();
	return PyNone();
}

static PyObject * 
TimeMap_unlock (TimeMap * p)
{
	p->events.mutex.unlock();
	return PyNone();
}

static PyObject *
TimeMap_bbt (TimeMap * map, PyObject * arg)
{
	d_tick t;
	
	if (!PyArg_Parse (arg, "d", &t))
		return 0;

	BBT bbt;

	/* ensure exclusive access */
	{
		ScopedLock lock (map->events.mutex);
		map->tick2bbt (t, bbt);
	}

	return Py_BuildValue ("iif", bbt.bar, bbt.beat + 1, (double) bbt.tick);
}

static PyObject *
TimeMap_tick (TimeMap * map, PyObject * args)
{
	BBT bbt;

	if (!PyArg_ParseTuple (args, "iid", &bbt.bar, &bbt.beat, &bbt.tick))
		return 0;

	bbt.beat -= 1;

	d_tick tick;

	/* ensure exclusive access */
	{
		ScopedLock lock (map->events.mutex);
		tick = map->bbt2tick (bbt);
	}

	return PyFloat (tick);
}

static PyMethodDef 
TimeMap_methods [] = 
{
  {"add", (PyCFunction) TimeMap_add, METH_VARARGS,
	 "add (event[, t]) -- add 'event' at optional tick 't'."},
	{"remove", (PyCFunction) TimeMap_remove, METH_VARARGS,
	 "remove (event) -- remove 'event'."},
	{"lock", (PyCFunction) TimeMap_lock, METH_NOARGS,
	 "lock() -- by acquiring this lock, the part is protected from other threads' "
	 "attempts at reading or modifying its contents."},
	{"unlock", (PyCFunction) TimeMap_unlock, METH_NOARGS,
	 "unlock() -- releases the protecting lock."},
	{"bbt", (PyCFunction) TimeMap_bbt, METH_O,
	 "bbt (tick) -- returns tick mapped into (bar, beat, tick)."},
	{"tick", (PyCFunction) TimeMap_tick, METH_VARARGS,
	 "tick (bar, beat = 0, tick = 0) -- returns bar, beat, tick mapped to ticks."},
	{}
};

Type * TimeMap_type = 0;

char * Py<TimeMap>::__doc__ = 
"TimeMap (list) -- a Plugin that manages TimeEvents, enabling it to map "
"tick to bar.beat.tick.";

void
Py<TimeMap>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();
}

void
Py<TimeMap>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = NamedObject_type;
	t->tp_methods = TimeMap_methods;
	t->tp_as_sequence = &TimeMap_as_sequence;
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_Map_types (PyObject * dict)
{
	TimeMap_type = new PyType<TimeMap> ("TimeMap", dict);
}
