/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "MidiStream.h"
#include "Graph.h"

#include "PyType.h"

MidiInThread * MidiInThread::the_one = 0;
atomic MidiInThread::refcount = 0;

MidiInThread::MidiInThread()
: Thread ("MidiIn"),
	poll_fds ("midi in poll_fds"),
	midi_ins ("midi ins")
{
	if (the_one) throw AssertionError() << "there should be only one.";

	/* prepair ;) communication sockets (TODO: use Pipe()). */
	if (socketpair (AF_LOCAL, SOCK_STREAM, 0, request_fd) < 0)
		throw OSError (errno) << "cannot get a socketpair(2).";
	if (socketpair (AF_LOCAL, SOCK_STREAM, 0, result_fd) < 0)
		throw OSError (errno) << "cannot get a socketpair(2).";

	/* AlsaClock is 15, RTClock is 5 (if priorities are not modified). */
	start_thread_rt (20);

	refcount = 0;
}

MidiInThread::~MidiInThread()
{
	signal_request (Quit);
	
	join_thread();

	the_one = 0;
}

void
MidiInThread::add (Pollable * in)
{
	if (!the_one) the_one = new MidiInThread();

	++refcount;

	the_one->signal_request (AddPollable, in);
}

void
MidiInThread::remove (Pollable * in)
{
	if (!the_one) throw AssertionError() << "no MidiInThread.";

	the_one->signal_request (RemovePollable, in);

	if (refcount.dec_is_zero())
		delete the_one;
}

void
MidiInThread::signal_request (char action, Pollable * midi_in)
{
	/* it's assumed only one thread is managing Pollables, so we don't need
	 * to protect this. 
	 */
	act_on = midi_in;

	write (request_fd[0], &action, 1);
	read (result_fd[0], &action, 1);
}

void
MidiInThread::rebuild_poll_fds()
{
	poll_fds.clear();

	poll_fds.insert (1);
	poll_fds[0].fd = request_fd[1];
	poll_fds[0].events = POLLIN;

	for (int i = 0; i < midi_ins; i++)
		midi_ins[i]->get_pollfd (poll_fds);
}

void *
MidiInThread::thread()
{
	bool rebuild = 1;
	
	while (1)
	{
		if (rebuild)
		{
			rebuild_poll_fds();
			rebuild = 0;
		}

		for (int i = 0; i < poll_fds; i++)
			poll_fds[i].revents = 0;

		int r = poll (&poll_fds[0], poll_fds.length, 1000);

		if (r < 0)
		{
			if (errno != EINTR)
				throw IOError (errno) << "in poll().";
			continue;
		}

		if (poll_fds[0].revents & POLLHUP)
			throw StopIteration() << "hung up.";
		if (poll_fds[0].revents & POLLERR) /* request fd */
			throw IOError() << "poll error on request socket.";

		/* check for input on midi sockets. */
		for (int i = 1; i < poll_fds; i++)
		{
			if (poll_fds[i].revents & ~POLLIN)
				throw IOError() << "poll error on " << midi_ins[i - 1];
				
			if (poll_fds[i].revents & POLLIN) try
			{
				midi_ins[i - 1]->do_io();
			}
			catch (blewit & b)
			{
				msg(), midi_ins[i - 1], "blew it: ", b.what();
			}
		}

		/* request signalled */
		if (poll_fds[0].revents & POLLIN) 
		{
			char action;
			read (request_fd[1], &action, 1);
			
			if (action == AddPollable)
				midi_ins.add_object (act_on);
			else if (action == RemovePollable)
				midi_ins.remove (act_on);

			rebuild = 1;

			/* give feedback */
			write (result_fd[1], &action, 1);
			
			if (action == Quit)
				break;
		}
	}

	return 0;
}

/* //////////////////////////////////////////////////////////////////////// */

MidiIn::MidiIn()
: MidiStream (MidiIn_type, 0),
  data ("midi in buffer", 16384) 
{
	event = 0;
	running = 0;
	expect = Status;
	log = false;
}

MidiIn::~MidiIn()
{
	::decref (event, true);

	if (fd != -1)
		MidiInThread::remove (this);
}

void
MidiIn::get_pollfd (InstanceList<struct pollfd> & list)
{
	if (fd == -1)
		throw AssertionError() << "can't poll -- MidiIn is not open.";

	int i = list.insert (1);

	list[i].fd = fd;
	list[i].events = POLLIN;
}

void
MidiIn::open (char * path)
{
	fd = ::open (path, O_RDONLY | O_NONBLOCK);

	if (fd == -1) throw IOError (errno) << "opening " << path;

	set_object_name ("%s in", path);
	add_out (new Out (this, MidiEvent_type, "MIDI"));
	add_out (new Out (this, SysexEvent_type, "Sysex"));

	MidiInThread::add (this);
}

void
MidiIn::start_event (byte status)
{
	data.clear();
	data << status;

	if (status == 0xF0)
	{
		/* if we get one of these, a detour to brk(2) is in it. */
		event = (SysexEvent *) outs[1]->cache->pop();
		bytes = -1;
	}
	else
	{
		event = (MidiEvent *) outs[0]->cache->pop();

		if (status < 0x80) /* paranoid safety measure. */
			bytes = 0;
		else if (status < 0xC0)	/* note off/on, note touch & control change */
			bytes = 2,
			running = status;
		else if (status <= 0xDF) /* program change & channel touch */
			bytes = 1,
			running = status;
		else if (status < 0xF0) /* pitch wheel */
			bytes = 2,
			running = status;
		else if (status == 0xF1) /* SMPTE time code */
			bytes = 1;
		else if (status == 0xF2) /* song position */
			bytes = 2;
		else if (status == 0xF3) /* song select */
			bytes = 1;
		else
			bytes = 0;
		/* realtime events 0xF8 - 0xFF are handled in read_and_parse(). */
	}
	
	event->tick = graph ? graph->tick : 0;
	event->loop = graph ? graph->loop.cycles : 0;

	if (bytes)
		expect = Data;
	else
	{
		dispatch(); /* status byte only event */
		expect = Status;
	}
}

void
MidiIn::dispatch()
{
	if (!event) return;

	/* can cause a call to brk(2) if the event is sysex but there's not 
	 * much we can do about that without leaving the bounds of sanity
	 * seeing that sysex can bring any amount of data. 
	 */
	event->set_data (&data[0], data.length);

	if (log)
		_msg(), get_name (this), ": ", event;

	if (event->is_sysex())
		outs[1]->push (event);
	else
		outs[0]->push (event);
	
	event = 0;
	expect = Status;
}

/* read and parse as many bytes as possible, emitting events as they come.
 *
 * untested with MIDI events 0xF1 - 0xF6, 0xF8 - 0xFF.
 *
 * sysex data will be cut off at MAX_UINT - 1 bytes because of the 'bytes' 
 * counter overflowing. don't hold your breath to see this happen. ;)
 */
void
MidiIn::do_io()
{
	byte b;

	while (read (fd, &b, 1) == 1)
	{
		/* active nonsense. */
		if (b == 0xFE) 
			continue;

		/* b is a status byte. */
		if (b & 0x80)
		{
			/* single-byte realtime message. */
			if (b >= 0xF8) 
			{
				MidiEvent * e = (MidiEvent *) outs[0]->cache->pop();

				e->tick = graph ? graph->tick : 0;
				e->set_data (&b, 1);
				
				outs[0]->push (e);
					
				continue;
			}

			/* event terminated by new status. */
			if (expect == Data) 
			{
				/* append the end-of-sysex (EOX = 0xF7) byte. */
				if (data[0] == 0xF0 && b == 0xF7)
				{
					data << b;
					dispatch();

					continue;
				}

				/* event was not complete when new status was received; discard it. */
				dispatch();
			}

			/* new status, new event. */
			start_event (b);
			continue;
		}

		/* b is a data byte. 
		 *
		 * check for running status.
		 */
		if (expect == Status) 
		{
			/* if we haven't caught on to running status yet we have to drop this. */
			if (!running) 
				continue;

			start_event (running);

			/* and proceed to data ... */
		}

		if (expect == Data)
		{
			data << b;

			if (--bytes == 0)
				dispatch();
		}
	}
}

/* ////////////////////////////////////////////////////////////////////// */

MidiOut::MidiOut()
: MidiStream (MidiOut_type, Plugin::Time),
	all_off_cc ("all_off_cc", 16),
	direct (32, "midi out direct")
{
	do_clock = true;
	log = false;

	/* estimated MIDI wire throughput. 
	 * this seems too much for many MIDI interfaces found on PCM/MIDI cards.
	 */
	throughput = 31250. / 10.; 

	/* estimated predelay (for a note-on message). */
	predelay = 3. / throughput;

	/* state */
	running = 0x00; 
	writable = 0.0;
	data_index = 0;
	clocks = 0;
	all_off = -1;

	current.event = 0;
	current.tick = 0;

	/* construct all-notes-off controller messages -- we need to mix them into
	 * the data stream after a seek or loop. */
	byte data [] = {0xB0, 123, 0};
	for (int i = 0; i < 16; i++)
	{
		data[0] = 0xB0 | i;
		
		MidiEvent * e = NEW (MidiEvent);
		e->set_data (data, 3);
		
		all_off_cc.add_object (e);
		
		::decref (e, true);
	}
}

MidiOut::~MidiOut()
{
	if (state != Plugin::Off && fd != -1)
		all_notes_off_now();
	
	::decref (current.event, true);
}

void
MidiOut::open (char * path)
{
	fd = ::open (path, O_WRONLY | O_NONBLOCK);
	if (fd == -1)
		throw IOError (errno) << " opening " << path;

	set_object_name ("%s out", path);

	add_in (new In (this, MidiEvent_type, IN_SLOTS, "send"));
}

void
MidiOut::send (MidiEvent * e)
{
	if (graph && state != Plugin::Off)
	{
		direct.push (e);
		
		/* if push throws, we need not (and do not because we never get here), 
		 * incref(). */ 
		::incref (e);
	}
	else
	{
		int n = write (fd, e->data, e->length);
		if (n != e->length)
			throw IOError() << "wrote " << n << " instead of " << e->length <<
					" bytes";
	}
}

/* Plugin:: */
void
MidiOut::seek()
{
	/* cancel the current event, if any */
	::decref (current.event);
	current.event = 0;
	current.tick = 0;

	/* running status, if any, stays unchanged. */

	/* transmission restarts with a first byte now. */
	data_index = 0;

	/* trigger the all-notes-off sequence if playing. */
	if (state == Plugin::Rolling)
		all_off = 0;
}

void
MidiOut::set_state (Graph * g, int s)
{
	if (s == Plugin::Off)
		all_notes_off_now(); /* stopping processing, kill all notes now. */

	else if (state == Plugin::Rolling)
		all_off = 0; /* trigger all-notes-off */

	Plugin::set_state (g, s);
}

/* returns the next event to send before 'next', if any. */
MidiEvent *
MidiOut::current_event (d_tick next)
{
	/* do we have all-off waiting to be sent */
	if (all_off >= 0)
		return all_off_cc [all_off];
	
	if (current.event)
		return current.event;
	
	/* do we have events in 'direct' queue */
	if (direct.have())
		return current.event = direct.pop();

	/* fetch the next event from inputs. */
	if ((current.event = (MidiEvent *) ins[0]->pop_before (next)))
		current.tick = current.event->tick;

	return current.event;
}

/* called from process_time, fetches the next current_event() */
MidiEvent *
MidiOut::next_event (d_tick next)
{
	data_index = 0;
	current.tick = 0;

	/* busy sending all-notes-off? */
	if (all_off >= 0)
	{
		if (++all_off >= all_off_cc)
			all_off = -1;
	}
	else
	{
		::decref (current.event);
		current.event = 0;
	}
	
	return current_event (next);
}

/* takes a slice of time and sequences events and and clock signals into
 * a byte stream that is fed to the file descriptor.
 */
void
MidiOut::process_time (d_time dt, int _clocks)
{
	/* how many bytes can be written? */
	writable += throughput * dt;
	int to_write = (int) writable;
	writable -= to_write;
	
	/* interleave clock signals with the data stream. */
	int clock_interval = TEN_TO_THE_NINTH;

	if (!do_clock)
		clocks = 0;
	else
	{
		/* writing more than 24 clock signals doesn't make sense. */
		_clocks %= 24;

		if (_clocks > 0)
		{
			clocks += _clocks;
			clock_interval = (to_write > clocks) ? (to_write / clocks) : (to_write);
		}
	}

	/* take care of loop cycling. */
	if (graph->is_loop_time (graph->time, graph->time + dt))
		++loop_cycle;

	if (to_write <= 0)
		return;
		
	/* until which tick do we play events? -- not perfect in the loop case,
	 * but reasonable seeing we rarely write more than a few bytes at a time.
	 */
	d_tick next = graph->time2tick (graph->time + dt + predelay);

	/* create a buffer for the bytestream on the stack. */
	int bytes = 0;
	byte buffer [to_write];

	MidiEvent * e = current_event (next);

	while (bytes < to_write && e)
	{
		if (clocks && (bytes % clock_interval) == 0)
		{
			buffer [bytes++] = 0xF8;
			--clocks;
			continue;
		}

		if (data_index == 0) /* next event */
		{
			if (current.tick > next) /* will do later. */
				break;
				
			byte status = e->data[0];
			
			if (status >= 0x80)
			{
				if (status == 0xF0) /* sysex, turn off running status. */
					running = 0;
				else if (status > 0xF0) /* realtime, no status change */
					;
				else if (status != running) /* new running status. */
					running = status;
				else
				{
					/* same as running status, continue with next byte. */
					++data_index;
					continue;
				}
			}
		}

		buffer [bytes++] = e->data [data_index];

		if (++data_index >= e->length) 
		{
			/* done with this event, fetch the next one. */
			e = next_event (next);
		}
	}

	/* append clock ticks that could not be interleaved with data. */
	while (bytes < to_write && clocks)
		buffer [bytes++] = 0xF8,
		--clocks;

	if (bytes == 0) 
		return;
	
	int written = write (fd, buffer, bytes);
	 
	if (log)
	{
		_msg m;

		m, get_name (this), ": ", do_clock;
		for (int i = 0; i < bytes; ++i)
			m, buffer[i];
	}
	
	if (written == -1)
		msg(), "write error: ", strerror (errno);
	else if (written != bytes)
		msg(), "only ", written, "bytes written instead of ", bytes;
}

/* generates an array containing all-notes-off for all midi channels. */
void
MidiOut::all_notes_off_now()
{
	/* regardless of the state of the communication (we might be in the middle 
	 * of an event right now) the new status byte will be recognized as such.
	 * i don't have the official midi spec, but all devices i have here act
	 * correctly in this case.
	 */
	byte all_notes_off[] = {0, 123, 0};

	for (int channel = 0; channel < 16; channel++)
	{
		all_notes_off[0] = 0xB0 | channel;
		write (fd, all_notes_off, 3);
	}
}

/* ////////////////////////////////////////////////////////////////////// */

Type * MidiIn_type = 0;

static int
MidiIn_set_log (MidiIn * m, PyObject * value)
{
	if (!PyArg_Parse (value, "i", &m->log))
		return -1;
	return 0;
}

static PyObject *
MidiIn_get_log (MidiIn * m)
{
	return PyInt (m->log);
}

static PyGetSetDef 
MidiIn_getsets [] = {
	{"log", (getter) MidiIn_get_log, (setter) MidiIn_set_log,
	 "whether to log incoming MIDI events to the stdmsg file."},
	{ }
};

/* ////////////////////////////////////////////////////////////////////// */

char * Py<MidiIn>::__doc__ = 
"MidiIn ('/path/to/file') -- receives MIDI events from a file you specify. "
"Common choices "
"for files will be /dev/snd/midiCxDy (ALSA card x device y) or /dev/midiN "
"(OSS midi driver), but fifos work, too.\n"
"This plugin needs to reside "
"in a graph holding a RTClock to put proper timestamps on the received "
"events.\n"
"All plugins of this type share a worker thread that wakes them up as soon "
"as incoming data is signalled. This makes their use on regular files "
"dubious. The worker thread is automatically started and stopped; there's "
"no special scheduler plugin you'd need to put into the graph.";

void
Py<MidiIn>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = MidiIn_getsets;
}

void
Py<MidiIn>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"path", 0};

	char * path;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s", keys,
	    &path))
		throw blewit();

	open (path);
}

/* ///////////////////////////////////////////////////////////////// */

static int
MidiOut_set_do_clock (MidiOut * m, PyObject * value)
{
	if (!PyArg_Parse (value, "i", &m->do_clock))
		return -1;
	return 0;
}

static PyObject *
MidiOut_get_do_clock (MidiOut * m)
{
	return PyInt (m->do_clock);
}

static int
MidiOut_set_log (MidiOut * m, PyObject * value)
{
	if (!PyArg_Parse (value, "i", &m->log))
		return -1;
	return 0;
}

static PyObject *
MidiOut_get_log (MidiOut * m)
{
	return PyInt (m->log);
}

static int
MidiOut_set_predelay (MidiOut * m, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &m->predelay))
		return -1;
	return 0;
}

static PyObject *
MidiOut_get_predelay (MidiOut * m)
{
	return PyFloat (m->predelay);
}

static int
MidiOut_set_throughput (MidiOut * m, PyObject * value)
Py_TRY
{
	double t;

	if (!PyArg_Parse (value, "d", &t))
		return -1;
	/* ensure we can write at least 96 clock ticks at 240 bpm plus some */
	if (t < 128)
		throw ValueError() << "throughput must be at least 128 bytes/second.";

	m->throughput = t;

	return 0;
}
Py_CATCH (-1);

static PyObject *
MidiOut_get_throughput (MidiOut * m)
{
	return PyFloat (m->throughput);
}

static PyGetSetDef 
MidiOut_getsets [] = {
	{"do_clock", (getter) MidiOut_get_do_clock, (setter) MidiOut_set_do_clock,
	 "whether to send MIDI clock sync signals."},
	{"log", (getter) MidiOut_get_log, (setter) MidiOut_set_log,
	 "whether to log the outgoing MIDI bytestream to the stdmsg file."},
	{"predelay", (getter) MidiOut_get_predelay, (setter) MidiOut_set_predelay,
	 "to achieve the best timing of events, they should be written to the "
	 "MIDI device early to compensate for the message transmission time. "
	 "This predelay value defaults to about one millisecond (transmission "
	 "time for three bytes, or a note-on message)."},
	{"throughput", (getter) MidiOut_get_throughput, (setter) MidiOut_set_throughput,
	 "assumed bandwidth of the MIDI interface in bytes/second."},
	{ }
};

/* ///////////////////////////////////////////////////////////////// */

static PyObject *
MidiOut_write (MidiOut * m, MidiEvent * event)
Py_TRY
{
	if (!MidiEvent_type->is_derived (event))
		throw TypeError() << "need a MidiEvent, not " << get_type_name (event);

	m->send (event);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
MidiOut_methods [] = {
	{"write", (PyCFunction) MidiOut_write, METH_O,
	 "write (event) -- send a MIDI event right now."},
	{ }
};

Type * MidiOut_type = 0;

char * Py<MidiOut>::__doc__ = 
"MidiOut ('/path/to/file'[, do_clock = 1]) -- "
"transmits MIDI events on a file you specify, and does this "
"mostly according to the MIDI standard. Common choices for files will be "
"/dev/snd/midiCxDy (ALSA card x device y) or /dev/midiN (OSS midi driver), "
"but fifos work, too.\n"
"For proper operation this plugin must reside in a graph that also holds a "
"RTClock. This is because the plugin operation is based on the assumption "
"that it drives a MIDI port at a bandwidth of 31.25 kbit (raw). Therefore it "
"requires a precise idea of time to properly schedule the streaming of "
"outbound MIDI data."
"\nThis plugin sends a continuous stream of "
"MIDI clock signals (bytes of value 0xF8) by default. This can be disabled "
"by setting 'do_clock' to 0.";

void
Py<MidiOut>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = MidiOut_getsets;
	t->tp_methods = MidiOut_methods;
}

void
Py<MidiOut>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"path", "do_clock", 0};

	char * path;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|i", keys,
	    &path, &do_clock))
		throw blewit();

	open (path);
}

/* ///////////////////////////////////////////////////////////////// */

void
init_MidiStream_types (PyObject * dict)
{
	MidiIn_type = new PyType<MidiIn> ("MidiIn", dict);
	MidiOut_type = new PyType<MidiOut> ("MidiOut", dict);
}
