/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _MIDISTREAM_H_
#define _MIDISTREAM_H_

#include "Plugin.h"
#include "Event.h"
#include "Thread.h"
#include "Pollable.h"
#include "Pipe.h"

class MidiIn;

/* MidiInThread keeps a list of MidiIns whose file descriptors are poll(2)ed
 * for data. there shall only be one at a time, accessed through the static
 * functions add() and remove() only which take care of creating and killing
 * the singleton.
 *
 * the list of MidiIns is maintained in thread context only.
 */
class MidiInThread
: public Thread
{
	public:
		virtual void * thread();

		static void add (Pollable *);
		static void remove (Pollable *);
		
		MidiInThread();
		~MidiInThread();

	private:
		int request_fd[2];
		int result_fd[2];

		InstanceList<struct pollfd> poll_fds;
		
		PointerList<Pollable> midi_ins;

		static atomic refcount;
		static MidiInThread * the_one;

		void rebuild_poll_fds();

		Pollable * act_on;
		void signal_request (char action, Pollable * = 0);
		enum /* actions */
		{
			AddPollable = 'a',
			RemovePollable = 'r',
			Quit = 'q'
		};
};

/* base to MidiIn and MidiOut, the MidiStream maintains a file descriptor
 * and is a Plugin. 
 */
class MidiStream
: public Plugin
{
	public:
		int fd;

		MidiStream (Type * t, int _flags)
			: Plugin (t, _flags)
			{
				fd = -1;
			}

		~MidiStream()
			{
				if (fd != -1) close (fd);
			}

		virtual void open (char * path) = 0;
};

/* MidiIn is an inbound MIDI connection via a file descriptor, which is 
 * poll(2)ed for input in a MidiInThread via the Pollable interface.
 */
class MidiIn
: public MidiStream,
	public Pollable
{
	public:
		MidiIn();
		~MidiIn();

		virtual void open (char * path);

		/* Pollable:: */
		virtual void get_pollfd (InstanceList<struct pollfd> & list); 
		virtual void do_io();
		virtual operator Object * () { return (Object *) this; }

		/* dump incoming events to stdmsg */
		bool log;
	
	private:
		MidiEvent * event;
		InstanceList<byte> data;

		void start_event (byte status);

		/* emit the current event */
		void dispatch(); 
		
		/* current running status */
		byte running; 

		int expect;
		enum 
		{
			Status = 's',
			Data = 'd'
		};

		/* how many data bytes left to assign to the current event */
		int bytes; 
};

/* MidiOut is an outbound MIDI connection via a file descriptor. */
class MidiOut
: public MidiStream
{
	private:
		byte running;	/* running status */
		double writable; /* number bytes that can be written */
		int data_index; /* which byte from the next event to write */
		int clocks; /* how many midi clock signals still to send, 0 in normal op */
		
	public:
		d_time predelay; /* how early to send events */
		double throughput; /* bytes per second */
		bool do_clock; /* whether to send F8 clock signals */	
		bool log;

		struct {
			MidiEvent * event; /* the event currently being sent. */
			d_tick tick; /* the tick at which to send the current event. */
		} current;

		MidiOut();
		~MidiOut();

		virtual void open (char * path);

		/* push into 'direct' and play asap. */
		void send (MidiEvent * e);

		/* Plugin:: */
		virtual void process_time (d_time dt, int _clocks);
		virtual void seek();
		virtual void set_state (Graph *, int);

		virtual void pre_remove_other (Plugin * p)
			{
				if (RTClock_type->is_derived (p))
					throw RuntimeError() << this << " needs " << p;
			}

		/* bypasses the queue, takes a lot of time with bad interfaces. */
		void all_notes_off_now();
	
	protected:
		MidiEvent * current_event (d_tick next);
		MidiEvent * next_event (d_tick next);

		int all_off; /* all notes off index */
		TypedObjectList<MidiEvent> all_off_cc;

		TypedObjectFifo<MidiEvent> direct; /* 'send now' fifo */
};

#endif /* !_MIDISTREAM_H_ */
