/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "Plugin.h"
#include "Graph.h"
#include "dsp/Noise.h"
#include "dsp/BiQuad.h"
#include "dsp/Sine.h"
#include "dsp/Lorenz.h"
#include "dsp/Roessler.h"

#include "PyType.h"

Type * Noise_type;

class Noise
: public Plugin
{
	public:
		int method;

		DSP::PinkNoise pink;
		DSP::OneOverFNoisz noisz;

		Noise();

		/* Plugin::* */
		virtual void process_audio (int frames);

		void one_cycle (d_sample *, int frames);
};


Noise::Noise()
: Plugin (Noise_type, Audio)
{
	method = 0;
	add_out (new Out (this, CyclicAudio_type, "noise"));
}

void
Noise::one_cycle (d_sample * s, int frames)
{
	if (method == 0)
	{
		for (int i = 0; i < frames; ++i)
			s[i] = pink.get_white();
	}
	else if (method == 1)
	{
		for (int i = 0; i < frames; ++i)
			s[i] = pink.get_pink();
	}
	else if (method == 2)
	{
		for (int i = 0; i < frames; ++i)
			s[i] = noisz.get();
	}
}

void
Noise::process_audio (int frames)
{
	if (!outs[0]->connected)
		return;

	CyclicAudio * e = (CyclicAudio *) outs[0]->cache->pop();

	e->frame = graph->frame;

	one_cycle (e->data, frames);

	outs[0]->push (e);
}

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
Noise_get_method (Noise * n)
{
	return PyInt (n->method);
}

static int
Noise_set_method (Noise * n, PyObject * method)
Py_TRY
{
	if (!PyArg_Parse (method, "i", &n->method))
		return -1;

	return 0;
}
Py_CATCH (-1);

static PyObject *
Noise_get_rows (Noise * n)
{
	return PyInt (n->pink.n_rows);
}

static int
Noise_set_rows (Noise * n, PyObject * rows)
Py_TRY
{
	int r;
	
	if (!PyArg_Parse (rows, "i", &r))
		return -1;
	
	if (r < 8 || r > PINK_MAX_RANDOM_ROWS)
		throw ValueError() << "'rows' must be >= 8 and <= " << PINK_MAX_RANDOM_ROWS;

	n->pink.init (r);

	return 0;
}
Py_CATCH (-1);

static PyGetSetDef
Noise_getsets [] = 
{
	{"rows", (getter) Noise_get_rows, (setter) Noise_set_rows,
	 "changes the noise character for pink noise."},
	{"method", (getter) Noise_get_method, (setter) Noise_set_method,
	 "0 -- white noise\n"
	 "1 -- pink noisz\n"
	 "2 -- 1/f noisz"},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
Noise_one_shot (Noise * n, AudioSignal * s)
Py_TRY
{
	if (!AudioSignal_type->is_derived (s))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (s);
	
	n->one_cycle (s->data, s->size);

	return ::incref (s);
}
Py_CATCH (0);

static PyMethodDef
Noise_methods [] =
{
	{"one_shot", (PyCFunction) Noise_one_shot, METH_O,
	 "one_shot (signal) -- sample into an AudioSignal."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

char * Py<Noise>::__doc__ = 
"Noise ([method = 0, rows = 12]) -- a plugin that generates noise.";

void
Py<Noise>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = Noise_getsets;
	t->tp_methods = Noise_methods;
}

void
Py<Noise>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"method", "rows", 0};
	PyObject * rows = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|iO", keys, &method, &rows))
		return;
	
	if (rows)
		Noise_set_rows (this, rows);
}

/* BoundNoise //////////////////////////////////////////////////////////////// */

static Type * BoundNoise_type;

class BoundNoise
: public Plugin
{
	public:
		DSP::PinkNoise pink;
		DSP::BiQuad filter;
		DSP::Sine sine;

		double f;

	public:
		BoundNoise();

		void init (double f, double fs);

		virtual void set_state (Graph *, int);
		virtual void process_audio (int frames);

		void one_cycle (d_sample *, int);
};

/* /////////////////////////////////////////////////////////////////////// */

static d_sample 
	BN_A[] = {0.46875377529788653, 0.93750755059577306, 0.46875377529788653},
	BN_B[] = {0, 0.44263102271985172, 0.43238407847169436};
		
BoundNoise::BoundNoise()
: Plugin (BoundNoise_type, Audio),
	filter (BN_A, BN_B)
{
	f = 400;

	add_out (new Out (this, CyclicAudio_type, "out"));
}

void
BoundNoise::init (double _f, double fs)
{
	f = _f;
	sine.set_f (f, fs, 0);
}

void
BoundNoise::set_state (Graph * g, int s)
{
	if (s == Plugin::Prepared)
		init (f, g->sample_rate);

	Plugin::set_state (g, s);
}

void
BoundNoise::one_cycle (d_sample * s, int frames)
{
	for (int i = 0; i < frames; ++i)
	{
		s[i] = filter.process (pink.get_white());
	}
}

void 
BoundNoise::process_audio (int frames)
{
	CyclicAudio * e = (CyclicAudio *) outs[0]->cache->pop();
	
	e->frame = graph->frame;

	d_sample * s = e->data;

	one_cycle (s, frames);

	outs[0]->push (e);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
BoundNoise_one_shot (BoundNoise * l, AudioSignal * e)
Py_TRY
{
	if (!AudioSignal_type->is_derived (e))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (e);

	l->one_cycle (e->data, e->size);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
BoundNoise_methods[] = 
{
	{"one_shot", (PyCFunction) BoundNoise_one_shot, METH_O,
	 "one_shot (signal) -- sample the fractal into 'signal'."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
BoundNoise_get_f (BoundNoise * s)
{
	return PyFloat (s->f);
}

static int
BoundNoise_set_f (BoundNoise * s, PyObject * value)
Py_TRY
{
	double f;

	if (!PyArg_Parse (value, "d", &f))
		return -1;
	
	if (f <= .0)
		throw ValueError() << "h must not be <= 0.";
	if (f > .015)
		throw ValueError() << "h must not be > 0.15";

	s->f = f;

	return 0;
}
Py_CATCH (-1);

static PyGetSetDef 
BoundNoise_getsets [] =
{
	{"f", (getter) BoundNoise_get_f, (setter) BoundNoise_set_f, 
		"frequency, > 0 and < 0.026 are stable."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<BoundNoise>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_methods = BoundNoise_methods;
	t->tp_getset = BoundNoise_getsets;
}

char * Py<BoundNoise>::__doc__ = 
"BoundNoise ([f = .01]) -- filtered and ring-modulated noisz.";

void
Py<BoundNoise>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"f", 0};
	PyObject * f = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|O", keys, &f))
		return;

	if (f) BoundNoise_set_f (this, f);
}

/* // Lorenz /////////////////////////////////////////////////////////////// */

static Type * Lorenz_type;

class Lorenz
: public Plugin
{
	public:
		DSP::Lorenz lorenz;

	public:
		Lorenz();

		void init (double _h = .001, double seed = .0);

		virtual void set_state (Graph *, int);
		virtual void process_audio (int frames);

		void step();
		void one_cycle (d_sample *, int);
};

/* /////////////////////////////////////////////////////////////////////// */

Lorenz::Lorenz()
: Plugin (Lorenz_type, Audio)
{
	add_out (new Out (this, CyclicAudio_type, "out"));
}

void
Lorenz::init (double h, double seed)
{
	lorenz.init (h, seed);
}

void
Lorenz::set_state (Graph * g, int s)
{
	Plugin::set_state (g, s);
}

void
Lorenz::one_cycle (d_sample * s, int frames)
{
	for (int i = 0; i < frames; ++i)
		s[i] = lorenz.get();
}

void 
Lorenz::process_audio (int frames)
{
	CyclicAudio * e = (CyclicAudio *) outs[0]->cache->pop();
	
	e->frame = graph->frame;

	d_sample * s = e->data;

	one_cycle (s, frames);

	outs[0]->push (e);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
Lorenz_one_shot (Lorenz * l, AudioSignal * e)
Py_TRY
{
	if (!AudioSignal_type->is_derived (e))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (e);

	l->one_cycle (e->data, e->size);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
Lorenz_step (Lorenz * L, PyObject * args)
{
	int n = 1;
	
	if (!PyArg_ParseTuple (args, "|i", &n))
		return 0;

	DSP::Lorenz & l = L->lorenz;

	for (int i = 1; i < n; ++i)
		l.step();
	
	l.step();

	return Py_BuildValue ("ddd", l.get_x(), l.get_y(), l.get_z());
}

static PyMethodDef
Lorenz_methods[] = 
{
	{"one_shot", (PyCFunction) Lorenz_one_shot, METH_O,
	 "one_shot (signal) -- sample the fractal into 'signal'."},
	{"step", (PyCFunction) Lorenz_step, METH_VARARGS,
	 "step ([n=1]) -- progress the fractal by n steps and return (x, y, z)."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
Lorenz_get_h (Lorenz * s)
{
	return PyFloat (s->lorenz.h);
}

static int
Lorenz_set_h (Lorenz * s, PyObject * value)
Py_TRY
{
	double h;

	if (!PyArg_Parse (value, "d", &h))
		return -1;
	
	if (h <= .0)
		throw ValueError() << "h must not be <= 0.";
	if (h > .015)
		throw ValueError() << "h must not be > 0.015";

	s->lorenz.h = h;

	return 0;
}
Py_CATCH (-1);

static PyGetSetDef 
Lorenz_getsets [] =
{
	{"h", (getter) Lorenz_get_h, (setter) Lorenz_set_h, 
		"hrequency, > 0 and < 0.026 are stable."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<Lorenz>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_methods = Lorenz_methods;
	t->tp_getset = Lorenz_getsets;
}

char * Py<Lorenz>::__doc__ = 
"Lorenz ([h = .01]) -- a Lorenz attractor fractal. only the x coordinate is "
"sampled. for more info "
"go to <http://astronomy.swin.edu.au/~pbourke/fractals/lorenz/>.";

void
Py<Lorenz>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"h", "seed", 0};
	PyObject * h = 0;
	double seed = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|Od", keys, &h, &seed))
		return;

	init (.001, seed);

	if (h) Lorenz_set_h (this, h);
}

/* // Roessler ///////////////////////////////////////////////////////////// */

static Type * Roessler_type;

class Roessler
: public Plugin
{
	public:
		DSP::Roessler roessler;

	public:
		Roessler();

		void init (double _h = .001);

		virtual void set_state (Graph *, int);
		virtual void process_audio (int frames);

		void step();
		void one_cycle (d_sample *, int);
};

/* /////////////////////////////////////////////////////////////////////// */

Roessler::Roessler()
: Plugin (Roessler_type, Audio)
{
	add_out (new Out (this, CyclicAudio_type, "out"));
}

void
Roessler::init (double _h)
{
	roessler.init (_h);
}

void
Roessler::set_state (Graph * g, int s)
{
	Plugin::set_state (g, s);
}

void
Roessler::one_cycle (d_sample * s, int frames)
{
	for (int i = 0; i < frames; ++i)
		s[i] = roessler.get();
}

void 
Roessler::process_audio (int frames)
{
	CyclicAudio * e = (CyclicAudio *) outs[0]->cache->pop();
	
	e->frame = graph->frame;

	d_sample * s = e->data;

	one_cycle (s, frames);

	outs[0]->push (e);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
Roessler_one_shot (Roessler * l, AudioSignal * e)
Py_TRY
{
	if (!AudioSignal_type->is_derived (e))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (e);

	l->one_cycle (e->data, e->size);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
Roessler_step (Roessler * R)
{
	DSP::Roessler & l = R->roessler;
	l.get();
	return Py_BuildValue ("ddd", l.get_x(), l.get_y(), l.get_z());
}

static PyMethodDef
Roessler_methods[] = 
{
	{"one_shot", (PyCFunction) Roessler_one_shot, METH_O,
	 "one_shot (signal) -- sample the fractal into 'signal'."},
	{"step", (PyCFunction) Roessler_step, METH_NOARGS,
	 "step() -- progress the fractal by one step and return (x, y, z)."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
Roessler_get_h (Roessler * s)
{
	return PyFloat (s->roessler.h);
}

static int
Roessler_set_h (Roessler * s, PyObject * value)
Py_TRY
{
	double h;

	if (!PyArg_Parse (value, "d", &h))
		return -1;
	
	if (h <= .0)
		throw ValueError() << "h must not be <= 0.";
	if (h > .096)
		throw ValueError() << "h must not be > 0.096";

	s->roessler.h = h;

	return 0;
}
Py_CATCH (-1);

static PyGetSetDef 
Roessler_getsets [] =
{
	{"h", (getter) Roessler_get_h, (setter) Roessler_set_h, 
		"hrequency, > 0 and < 0.096 are stable."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<Roessler>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_methods = Roessler_methods;
	t->tp_getset = Roessler_getsets;
}

char * Py<Roessler>::__doc__ = 
"Lorenz ([h = .01]) -- a Roessler attractor fractal. the x and z coordinates "
"are sampled. for more info "
"go to <http://astronomy.swin.edu.au/~pbourke/fractals/rossler/>.";

void
Py<Roessler>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"h", 0};
	PyObject * h = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|O", keys, &h))
		return;

	init();

	if (h) Roessler_set_h (this, h);
}

/* //////////////////////////////////////////////////////////////////////// */

void
init_Noise_types (PyObject * dict)
{
	Noise_type = new PyType<Noise> ("Noise", dict);
	Lorenz_type = new PyType<Lorenz> ("Lorenz", dict);
	BoundNoise_type = new PyType<BoundNoise> ("BoundNoise", dict);
	Roessler_type = new PyType<Roessler> ("Roessler", dict);
}
