/* (c) 2004 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include <basics.h>
#include <Thread.h>
#include <PythonThread.h>
#include <AudioSignal.h>
#include <ObjectFifo.h>
#include <List.h>
#include <Pipe.h>
#include <PyType.h>

#include <sys/ioctl.h>
#include <sys/soundcard.h>

Type * OSS_type;

class OSS
: public NamedObject,
	public Thread
{
	protected:
		Poll pfds;

		AudioSignal * signal[2];
		d_sample * read_ptr[2];

		int fd;

		int16 * buffer;

		int fetch_signals();

	public:
		int have; /* frames in queue */
		int blocksize;  /* in frames */
		int channels, sample_rate;
		double gain;

		/* possible trigger values for communication via the pipe */
		enum {
			Sync = 0,
			Quit = 1
		};
		Pipe pipe;
		Pipe sync;

		TypedObjectFifo<AudioSignal> fifo;
		Mutex mutex;

		OSS();
		~OSS();

		void init (char * path, int sample_rate, int channels, int blocksize);
		
		virtual void * thread();
};

OSS::OSS()
: NamedObject (OSS_type),
	Thread ("OSS"),
	pfds ("oss poll"),
	pipe ("oss pipe", this, true),
	sync ("oss sync", this, true),
	fifo (16, "oss fifo")
{
	fd = -1;
	buffer = 0;
	gain = 1.;
	signal[0] = signal[1] = 0;

	have = 0;
	sync.set_nonblocking (Pipe::Write);
}

OSS::~OSS()
{
	if (fd) 
	{
		pipe.write_trigger (OSS::Quit);
		join_thread(); 
		fcntl (fd, F_SETFL, O_NONBLOCK);
		close (fd);
	}

	AudioSignal * a;
	do ::decref ((a = fifo.pop()), true);
	while (a);

	::decref (signal[0], true);
	::decref (signal[1], true);

	if (buffer) free (buffer);
}

void
OSS::init (char * path, int _sample_rate, int _channels, int _blocksize)
{
	if (fd != -1)
		throw AssertionError() << "already initialized";

	blocksize = _blocksize;
	sample_rate = _sample_rate;
	channels = _channels;

	if (channels < 1 || channels > 2)
		throw ValueError() << "only 1 or 2 channel operation";

	fd = open (path, O_WRONLY);
	if (fd == -1)
		throw OSError (errno);

	set_object_name (path);

	/* OSS is broken by design (tm) */
	int frag = 1, b = blocksize;
	while (b > 1)
		++frag, b >>= 1;
	if (b != 1)
		throw ValueError() << "blocksize not a power of 2";

	frag |= 0x7FFF0000;

	int gotfrag = frag;
	if (ioctl (fd, SNDCTL_DSP_SETFRAGMENT, &gotfrag) == -1)
		throw OSError() << "cannot set fragment size";
	if (gotfrag != frag)
		throw OSError() << "want " << blocksize << " fragments, got " << gotfrag;

	int format = AFMT_S16_LE;
	if (ioctl (fd, SNDCTL_DSP_SETFMT, &format) == -1)
		throw OSError (errno) << "setting AFMT_S16_NE";
	if (format != AFMT_S16_LE)
		throw OSError() << "cannot set AFMT_S16_NE";

	int c = channels;
	if (ioctl (fd, SNDCTL_DSP_CHANNELS, &c) == -1)
		throw OSError (errno) << "setting " << c << " channels";
	if (c != channels)
		throw OSError() << "want " << channels << " channels, got " << c;

	int f = sample_rate;
	if (ioctl (fd, SNDCTL_DSP_SPEED, &f) == -1)
		throw OSError (errno) << "setting sample rate " << f;
	if (f != sample_rate)
		throw OSError() << "want " << sample_rate << " Hz, got " << f;
	
	blocksize = 4096;
	if (ioctl (fd, SNDCTL_DSP_GETBLKSIZE, &blocksize) == -1)
		msg(), "cannot determine blocksize, assume ", blocksize;
		
	msg(), "blocksize ", blocksize;

	if (blocksize % (sizeof (int16) * channels))
		throw AssertionError() << "blocksize not a multiple of frame size";

	buffer = (int16 *) malloc (blocksize);
	if (buffer == 0)
		throw MemoryError() << "out of bytes, can you believe it.";

	start_thread();
}

void *
OSS::thread()
{
	int have_frames = 0;
	int sync_requests = 0;
	int silent = 1;

	pfds.add (pipe);
	pfds.add (fd, POLLOUT);
	
	while (1)
	{
		if (pfds.poll (1000000) == 0)
			continue;
			
		if (pfds[0].revents & POLLIN)
		{
			int t = pipe.read_trigger();

			if (t == OSS::Quit)
				break;
			else if (t == OSS::Sync)
				++sync_requests;

			continue;
		}
		
		/* getting here means poll element 1, our pcm fd, can write now */
		audio_buf_info info;
		ioctl (fd, SNDCTL_DSP_GETOSPACE, &info);

		if (info.fragsize != blocksize)
			throw AssertionError() << "oss says fragsize " << info.fragsize <<
					" and we assume " << blocksize;

		/* always write full fragments, otherwise we come here every 32 frames 
		 * (or so) */
		while (info.fragments >= 0)
		{
			int frames = blocksize / (sizeof (int16) * channels);

			int16 * out_0 = buffer;
			int16 * out_1 = buffer + 1;
			
			double scale = gain * 32767.;

			while (frames)
			{
				if (have_frames <= 0)
					have_frames = fetch_signals();

				if (have_frames <= 0)
				{
					if (!silent) msg(), "silent", (silent = 1);
					memset (out_0, 0, frames * channels * sizeof (int16));
					break;
				}

				int n = min (have_frames, frames);
				silent = 0;

				if (channels == 1) 
					for (int i = 0; i < n; ++i)
						out_0[i] = (int16) (scale * read_ptr[0][i]);
				else
					for (int i = 0; i < n; ++i)
					{
						out_0[2 * i] = (int16) (scale * read_ptr[0][i]);
						out_1[2 * i] = (int16) (scale * read_ptr[1][i]);
					}

				out_0 += n;
				out_1 += n;

				read_ptr[0] += n;
				read_ptr[1] += n;

				have_frames -= n;
				frames -= n;
			}

			int written = write (fd, buffer, blocksize);
			if (written == -1)
				throw OSError (errno) << "writing to the dsp device";

			--info.fragments;
		}

		while (sync_requests > 0)
		{
			sync.write_trigger (0);
			--sync_requests;
		}
	}

	return 0;
}

int
OSS::fetch_signals()
{
	ScopedLock lock (mutex);

	if (fifo.have() < channels)
		return 0;

	for (int i = 0; i < channels; ++i)
	{
		::decref (signal[i]);

		signal[i] = fifo.pop();
		read_ptr[i] = signal[i]->data;
	}
	
	for (int i = 1; i < channels; ++i)
		if (signal[i]->size != signal[0]->size)
			throw AssertionError() << "signal size mismatch on channel " << i;

	have -= signal[0]->size;

	return signal[0]->size;
}

/* // Py<OSS> ///////////////////////////////////////////////////////////// */

static void
oss_sync (OSS * oss)
{
	PythonUnlock unlock;

	oss->pipe.write_trigger (OSS::Sync);
	oss->sync.read_trigger();
}

static PyObject *
OSS_play (OSS * oss, PyObject * args)
Py_TRY
{
	AudioSignal * s1 = 0, * s2 = 0;
	
	char * fmt = "O!O!";
	if (oss->channels == 1) fmt += 2;

	if (!PyArg_ParseTuple (args, fmt, 
				AudioSignal_type, &s1, 
				AudioSignal_type, &s2))
		return 0;

	if (s2 && s2->size != s1->size)
		throw ValueError() << "2-channel signal data must be same length";

	ScopedLock lock (oss->mutex);

	while (oss->fifo.have() == oss->fifo.get_size())
	{
		/* perform the equivalent of a blocking write */
		ScopedUnlock unlock (oss->mutex);

		oss_sync (oss);
	}

	::incref (s1);
	oss->fifo.push (s1);
	
	if (s2)
	{
		::incref (s2);
		oss->fifo.push (s2);
	}
	
	oss->have += s1->size;

	return PyNone();
}
Py_CATCH (0);

static PyObject *
OSS_sync (OSS * oss)
Py_TRY
{
	oss_sync (oss);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
OSS_have (OSS * oss)
Py_TRY
{
	ScopedLock lock (oss->mutex);

	return PyInt (oss->have);
}
Py_CATCH (0);

static PyMethodDef
OSS_methods [] =
{
	{"play", (PyCFunction) OSS_play, METH_VARARGS,
	 "play (1 or 2 audio signals) -- queue the AudioSignal(s) for playing. "
	 "the signal should be discarded and not modified after this call "
	 "because it is not copied."},
	{"write", (PyCFunction) OSS_play, METH_VARARGS,
	 "equivalent to play()."},
	{"sync", (PyCFunction) OSS_sync, METH_NOARGS,
	 "sync() -- sleep until the driver signals a write possible condition."},
	{"have", (PyCFunction) OSS_have, METH_NOARGS,
	 "have() -- return the total number of frames queued (comprises all "
	 "AudioSignals in the queue)."},
	{ }
};

/* ///////////////////////////////////////////////////////////////////////// */

static PyObject *
OSS_get_gain (OSS * oss)
{
	return PyFloat (oss->gain);
}

static int
OSS_set_gain (OSS * oss, PyObject * arg)
{
	if (!PyArg_Parse (arg, "d", &oss->gain))
		return -1;
	return 0;
}

static PyObject *
OSS_get_sample_rate (OSS * oss)
{
	return PyInt (oss->sample_rate);
}

static PyObject *
OSS_get_blocksize (OSS * oss)
{
	return PyInt (oss->blocksize / (oss->channels * sizeof (int16)));
}

static PyGetSetDef 
OSS_getsets [] =
{
	{"gain", (getter) OSS_get_gain, (setter) OSS_set_gain, "1 = unity"},
	{"sample_rate", (getter) OSS_get_sample_rate, (setter) 0, "."},
	{"blocksize", (getter) OSS_get_blocksize, (setter) 0, 
	 "number of frames written per cycle and channel."},
	{ }
};

void
Py<OSS>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = NamedObject_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = OSS_getsets;
	t->tp_methods = OSS_methods;
}

char * Py<OSS>::__doc__ = 
"OSS ([path = '/dev/dsp', channels = 1, sample_rate = 44100, blocksize = 4096]) -- "
"OpenSoundSystem audio output. the output is done in a separate thread and "
"uses blocking write(2) calls.\n"
"to play audio, pass AudioSignals to the "
"play() method. the signals are written to a fifo, from which they are "
"fetched by the writer thread; you can queue up to 128 AudioSignals for "
"seamless playback. the individual length of the queued signals can be "
"arbitrary, the have() method will return the collected total "
"number of frames currently "
"waiting for playback in the queue, and you can sync to the end of a "
"block write cycle by using the sync() method.\n"
"only 1 and 2 channel operations are supported. when in 2-channel more, "
"play() will expect 2 AudioSignals instead of just one.";

void
Py<OSS>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {
			"path", "channels", "sample_rate", "blocksize", 0};

	char * path = "/dev/dsp";
	int sample_rate = SAMPLE_RATE, channels = 1, blocksize = 4096;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|siii", keys, 
				&path, &channels, &sample_rate, &blocksize))
		return;

	init (path, sample_rate, channels, blocksize);
}

/* //////////////////////////////////////////////////////////////////// */

void
init_OSS_types (PyObject * dict)
{
	OSS_type = new PyType<OSS> ("OSS", dict);
}
