/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "Py.h"
#include "PythonThread.h"

/* turn on constructor/destructor tracing. */
#define DBG_LIFE			1

/* trace Object::incref and ::decref. */
#define DBG_REFCOUNT	0

atomic Object::objects_alive = 0;

/* when allocating from C++, use the standard python allocator. this means
 * that types that are meant to be python baseclasses will cause a core
 * dump sooner or later if they are allocated through this allocator.
 */

void *
Object::operator new (size_t size)
{
	return PyObject_MALLOC (size);
}

/* this new operator doesn't allocate anything. it's used to make c++
 * constructors work with derived classes when tp_new() is called, after
 * PyType has called the correct python allocator (GC or generic) to get
 * the memory we need. 
 */
 
void *
Object::operator new (size_t size, PyObject * self)
{
 	return self;
}

/* the delete operator checks whether the object's type provides tp_free()
 * itself, which PyType does, otherwise it uses the standard python 
 * deallocator. 
 */
 
void
Object::operator delete (void * self)
{
	Type * type = get_type (self);

	if (type && type->tp_free)
		type->tp_free ((PyObject *) self);
	else
		PyObject_Del (self);
}

/* ////////////////////////////////////////////////////////////////////// */

Object::Object (Type * type)
{
	ob_refcnt = 1; 
	ob_type = type;

	object_cache = 0;

	#if DBG_LIFE
	_msg(), this, ".new()";
	#endif

	++objects_alive;
}

Object::~Object()
{
	#if DBG_LIFE
	_msg(), this, ".delete()";
	#endif

	--objects_alive;
}

void
Object::set_object_type (Type * type)
{ 
	ob_type = (PyTypeObject *) type; 

	#if DBG_LIFE
	_msg(), this, ".set_object_type()";
	#endif
}

Object *
Object::incref()
{
	if (this == 0) 
		return 0;
	
	Py_INCREF (this);

	#if DBG_REFCOUNT
	_msg(), this, ".incref() = ", ob_refcnt;
	#endif

	return this;
}

void
Object::decref (bool really) 
{
	if (this == 0) 
		return;

	if (really)
	{
		#if DBG_REFCOUNT
		_msg(), this, ".decref() = ", ob_refcnt - 1;
		#endif
		
		Py_DECREF (this);
	}
	else
	{
		PythonThread * t = PythonThread::get();

		if (t == 0)
		{
			msg(), "no PythonThread for ", this;
			Py_DECREF (this);
			return;
		}
		
		t->push (this);
	}
}

String &
Object::operator >> (String & s)
{
	s << "<" << ob_type->tp_name << " Object '" 
		<< get_object_name() << "' at " 
		<< (void *) this << ">";

	return s;
}

/* /////////////////////////////////////////////////////////////////////// */

bool 
Object::call_method (char * name, char * format, ...)
{
	PythonLock _pl;

	/* it certainly is a bit lame to check for overloads every time we call,
	 * but there's no way around it since dynamic overloading is just too
	 * handy to preclude. 
	 */
	PyObject * method = PyObject_GetAttrString (this, name);

	if (method == 0)
	{
		PyErr_Clear();
		return false;
	}
					  
	/* build the parameter list. */
	va_list vargs;

	va_start (vargs, format);
	PyObject * args = Py_VaBuildValue (format, vargs);
	va_end (vargs);

	/* args == 0 is caught later. */

	/* call. */
	if (args)
		::decref (PyEval_CallObject (method, args), true);

	::decref (method, true);
	::decref (args, true);

	if (PyErr_Occurred() == 0)
		return true;

	/* ///////////////////////////////////////////////////////////////////// */

	/* send error info to stdmsg. */
	PythonThread::trouble (this);

	return false;
}

/* ////////////////////////////////////////////////////////////////////// */

PyObject *
Type::__new__()
{
	PyObject * o = this->tp_new (this, 0, 0);
	
	if (o == 0) throw blewit();

	return o;
}

