/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _OBJECT_H_
#define _OBJECT_H_

/* gets defined in python2.3's pyconfig.h, but also through -D_GNU_SOURCE */
#undef _POSIX_C_SOURCE
#include <Python.h>

class Type;

#define TYPE(O)  O ## _type
#define NEW(O) (O *) O ## _type->__new__()

/* the base. */
extern Type * Object_type;
extern Type * NamedObject_type;

extern Type * ObjectFifo_type;

/* events. */
extern Type * Event_type;
extern Type * FloatEvent_type;
extern Type * MidiEvent_type;
extern Type * SysexEvent_type;
extern Type * Note_type;
extern Type * TextEvent_type;
extern Type * ClickEvent_type;
extern Type * TimeEvent_type;
extern Type * TempoEvent_type;

/* event lists. */
extern Type * EventList_type;
extern Type * Part_type;

/* . */
extern Type * TimeMap_type;

/* audio events. */
extern Type * AudioSignal_type;
extern Type * CyclicAudio_type;
extern Type * DecycledAudio_type;

/* plugin. */
extern Type * In_type; 
extern Type * Out_type;
extern Type * Plugin_type;
extern Type * EventCache_type;

/* graph. */
extern Type * Graph_type;

extern Type * GraphRequest_type;
extern Type * GraphStateChange_type;
extern Type * GraphAddPlugin_type;
extern Type * GraphRemovePlugin_type;
extern Type * GraphMovePlugin_type;
extern Type * GraphConnect_type;
extern Type * GraphStoreTempo_type;
extern Type * GraphRemoveTempo_type;
extern Type * GraphSeek_type;
extern Type * GraphSetQuarter_type;
extern Type * GraphSetLoop_type;
extern Type * GraphSetUseLoop_type;

/* rt clock. */
extern Type * RTClock_type;

/* midi ports. */
extern Type * MidiStream_type;
extern Type * MidiIn_type;
extern Type * MidiOut_type;

/* audio plugins. */
extern Type * AudioStream_type;
extern Type * AudioClock_type;

/* disk file types / plugins. */
extern Type * SoundFile_type;

extern Type * DiskStream_type;
extern Type * DiskScheduler_type;

/* script plugins. */
extern Type * ScriptPlugin_type;
extern Type * ScriptScheduler_type;

extern Type * DecycledAudio_type;

/* configurables */
#ifdef WITH_JACK
extern Type * Jack_type;
extern Type * JackIn_type;
extern Type * JackOut_type;
#endif

#ifdef WITH_IIWU
extern Type * IIWU_type;
#endif

/* /////////////////////////////////////////////////////////////////////// */

class Object;

class Cache
{
	public:
		virtual void push (Object *) = 0;
};

/* everything that wants to be visible to Python must inherit from 'Object',
 * and it must inherit from 'Object' first. 
 *
 * Caution: not every PyObject is also an Object because they lack the 
 * 'object_cache' property and the virtual table.
 */
class Object
:	public PyObject
{
	public:
		static atomic objects_alive;

		Object (Type * type = Object_type);
		virtual ~Object(); 

		/* ////////////////////////////////////////////////////////////////// */
		
		/* generic caching facility. */

		Cache * object_cache;

		void return_to_cache()
			{
				if (object_cache)
					object_cache->push (this);
				else
					decref();
			}

		/* ////////////////////////////////////////////////////////////////// */
		
		/* Redirection to Python refcounting: */
		
		/* incref returns 'this' for convenience. */
		
		Object * incref(); 
		
		/* If not 'really', this only pushes the Object on the current thread's 
		 * local trash fifo, where it rests until PythonThread::housekeep does 
		 * its duty. Actually destroying an object is not viable under realtime 
		 * constraints, neither with free(3) nor (worse) with the not-thread-
		 * safe Python GC allocator. */
		
		void decref (bool really = 0);
		
		/* ////////////////////////////////////////////////////////////////// */
		
		/* an object can override its type. */

		void set_object_type (Type * type);

		/* ////////////////////////////////////////////////////////////////// */
		
		virtual const char * get_object_name() 
			{
				return "nameless";
			}

		virtual String & operator >> (String & s);

		/* ////////////////////////////////////////////////////////////////// */

		/* checks for the existence of the method 'name' in this::__dict__()
		 * and calls it with a python 'format' string. */

		bool call_method (char * name, char * format, ...);
		
		/* ////////////////////////////////////////////////////////////////// */
		
		/* Object instantiation and passing-away are a little special because
		 * il faut danser avec l'allocateur GC de Python. This is discussed in 
		 * more depth in "PyType.h". */

		/* The C++ allocator relegates to the standard Python allocator. */

		void * operator new (size_t size);
		
		/* The tp_init allocator does *not* allocate, to make C++ constructors 
		 * work on memory allocated by the Python allocator. */
		
		void * operator new (size_t size, PyObject * self);

		/* Checks for tp_free before deciding to free() via the standard Python
		 * deallocator. */
		
		void operator delete (void * self);
};

class Type
:	public PyTypeObject
{
	public:
		Type (const char * name, size_t size)
			{
				memset (this, 0, sizeof (*this));
				
				ob_refcnt = 1;
				ob_size = sizeof (*this);
				ob_type = &PyType_Type;
				
				tp_basicsize = size;
				
				tp_name = const_cast<char *> (name);
			}
		
		~Type()
			{ }

		/* the factory. */
		PyObject * __new__();

		inline int is_derived (PyTypeObject * t)
			{
				int steps = 1;
				
				while (t)
				{
					if (t == this) 
						return steps;
					
					t = t->tp_base;
					
					++steps;
				}

				return 0;
			}

		inline int is_derived (PyObject * o)
			{
				return is_derived (o->ob_type);
			}

};

class NamedObject
: public Object
{
	public:
		char * object_name;
		
		NamedObject (Type * type = NamedObject_type, const char * name = 0)
			: Object (type)
			{
				if (name) object_name = strdup (name);
				else object_name = 0;
			}
			
		~NamedObject()
			{
				if (object_name) free (object_name);
			}
			
		virtual const char * get_object_name() 
			{
				return object_name ? object_name : "unnamed";
			}

		void set_object_name (const char * fmt, ...)
			{
				if (object_name) free (object_name);
				
				va_list args;
				
				va_start (args, fmt);
				vasprintf (&object_name, fmt, args);
				va_end (args);
			}
};

/* ::incref (o) and ::decref (o) are easier to spot. */
inline Object *
incref (void * o)
{
	return ((Object *) o)->incref();
}

/* Set 'really' to true only if you are positively sure the big python lock is 
 * currently held by the thread executing. If in doubt, leave it 0 and have the 
 * actual actual decref be delegated to PythonThread::housekeep(). 
 *
 * Destructors can and must ::decref (child object, true). 
 */
inline void
decref (void * o, bool really = 0)
{
	((Object *) o)->decref (really);
}

/* */

inline Type * 
get_type (void * o)
{
	return (Type *) ((Object *) o)->ob_type;
}

inline const char * 
get_type_name (void * o)
{
	if (!o) 
		return "(null)";
	
	if (!((Object *) o)->ob_type) 
		return "untyped";

	return ((Object *) o)->ob_type->tp_name;
}

inline const char *
get_name (void * o)
{
	if (!o) 
		return "(null)";
	
	/* get_object_name is virtual, so protect PyObjects that don't have it. */
	if (Object_type->is_derived (get_type (o)))
		return ((Object *) o)->get_object_name();

	return "anonymous";
}

#endif /* _OBJECT_H_ */
