/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _OBJECT_FIFO_H_
#define _OBJECT_FIFO_H_

/* a fifo for Object pointers. a reference to queued objects is stolen.
 *
 * objects that are still in the queue will be ::decref()d when the fifo
 * is destroyed. A zero-sized ObjectFifo is legal. */
class ObjectFifo 
: public NamedObject
{
	protected:
		Object ** stack;
		int r, w;
		atomic used; 

	public:
		int size;

		ObjectFifo (int _size = 0, const char * name = 0)
			: NamedObject (ObjectFifo_type, name)
			{
				stack = 0;
				size = _size;
				used = 0;
				r = w = 0;
				
				if (size) 
					alloc (size);
			}
		
		~ObjectFifo()
			{
				Object * o;

				while ((o = pop_object()))
					::decref (o, true);

				if (stack)
					free (stack);
			}

		void alloc (int _size)
			{
				if (stack) 
					throw AssertionError() << "already allocated.";

				size = _size;
				stack = (Object **) calloc (size, sizeof (Object *));
				
				if (!stack) 
					throw MemoryError() << "allocating a stack for " << this;
			}
			
		int get_size() { return size; }
		int have() { return used; }
		bool is_full() { return used == size; }

		Object * pop_object()
			{
				if (used == 0) return 0;
				
				Object * o = stack[r];
				r = (r + 1) % size;
				--used;

				return o;
			}

		Object * peek_object()
			{
				if (used == 0) return 0;
				return stack[r];
			}

		bool push_object (Object * o)
			{
				if (this == 0) return 0;
				if (used == size) return 0;

				++used;
				stack[w] = o;
				w = (w + 1) % size;

				return true;
			}
	
		void clear (bool decref_now = 0)
			{
				Object * o;
				
				while ((o = pop_object()))
					::decref (o, decref_now);
			}
};

template <class T>
class TypedObjectFifo
: public ObjectFifo
{
	public:
		TypedObjectFifo (int _size, const char * name)
			: ObjectFifo (_size, name)
			{ }

		T * pop() 
			{ 
				return (T *) pop_object(); 
			}
		
		T * peek() 
			{ 
				return (T *) peek_object(); 
			}
		
		void push (Object * o) 
			{ 
				push_object (o); 
			}
};

#endif /* !_OBJECT_FIFO_H_ */
