/* (c) 2001-4 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "Part.h"
#include "Graph.h"

#include "PyType.h"
#include "PythonThread.h"

/* /////////////////////////////////////////////////////////////////////// */

Part::Part()
: Plugin (Part_type, Beat)
{
	channel = -1;
	list = 0;
	tick = 0;
	mute = 0;
	transpose = 0;
}

Part::~Part()
{
	note_offs.clear();
	::decref (list, true);
}

void
Part::set_list (EventList * l)
{
	if (list)
		throw AssertionError() << "list already set.";

	::incref (list = l);

	out = add_out (new Out (this, l->type, "out"));
}

void
Part::set_state (Graph * g, int s)
{
	if (s == Plugin::Activated)
		; /* any initializing work */
	if (s == Plugin::Rolling)
	{
		tick = g->tick;

		out->clear (true);
		note_offs.clear();

		enqueue (tick + g->quarter);
	}

	Plugin::set_state (g, s);
}

void
Part::seek()
{
	tick = graph->tick;

	out->clear (true);
	note_offs.clear();

	if (state == Rolling)
		enqueue (tick + graph->quarter / 2);
}

void
Part::process_beat (d_tick until)
{
	if (state != Rolling)
		return;

	/* this happens occasionally. */
	if (until <= tick)
		return;

	ScopedLock lock (mutex);

	bool is_loop = graph->is_loop_tick (tick, until);

#if 0
	if (!strcmp (get_object_name(), "Drums"))
		msg(), graph->tick, "process_beat until: ", until, is_loop ? "loop" : "";
#endif

	if (is_loop)
	{
		enqueue (graph->loop.tick[1]);

		note_offs.clear();

		tick = graph->loop.tick[0];

		until -= graph->loop.tick[1] - graph->loop.tick[0];

		++loop_cycle;
	}

	if (until > tick)
		enqueue (until);

	tick = until;
}

void
Part::enqueue (d_tick until)
{
	ScopedLock lock (list->mutex);

	MidiEvent * off;

	if (!mute)
	{
		int i = list->index_of_tick (tick);

		while (i < *list)
		{
			Event * e = (*list) [i];
			
			if (e->tick >= until)
				break;

			/* send pending note offs before this one: */
			while ((off = note_offs.peek()) && off->tick <= e->tick)
			{
				out->push (off);
				note_offs.pop();
			}

			if (!(get_type (e) == Note_type && mute))
				enqueue_one (e);

			++i;
		}
	}

	/* send all pending note offs left before 'until' now: */
	while ((off = note_offs.peek()) && off->tick <= until)
	{
		out->push (off);
		note_offs.pop();
	}

	tick = until;
}

inline void
Part::enqueue_one (Event * e)
{
#if 0
	if (!strcmp (get_object_name(), "Drums"))
		msg(), loop_cycle, "+++ enqueuing ", (MidiEvent *) e;
#endif

	if (get_type (e) == Note_type)
	{
		queue_note_off ((Note *) e);

		MidiEvent * o = (MidiEvent *) out->cache->pop();

		o->clone (e);
		o->loop = loop_cycle;

		/* transpose and channel */
		o->data[1] = (((Note *) e)->data[1] + transpose) % 120;
		if (channel >= 0)
			o->set_channel (channel);
	
		out->push (o);
		return;
	}
	else if (get_type (e) != out->type)
	{
		/* derived type, won't construct a clone. this might be a sysex of any
		 * size. */
		msg(), this, ": ", get_type_name (e), "s aren't sent from cache.";
		out->push (::incref (e));

		return;
	}

	Event * o = (Event *) out->cache->pop();

	o->clone (e);
	o->loop = loop_cycle;

	if (channel >= 0 && MidiEvent_type->is_derived (o))
		if (((MidiEvent *) o)->has_channel())
			((MidiEvent *) o)->set_channel (channel);
	
	out->push (o);
}

void
Part::queue_note_off (Note * n)
{
	MidiEvent * off = (MidiEvent *) out->cache->pop();

	/* construct a note-off from e: */
	off->clone (n);
	/* transpose */
	off->data[1] = (n->data[1] + transpose) % 120;
	off->tick += n->l;

	/* velocity = 0 marks note-on as note-off. */
	off->data[2] = 0; 

	if (channel >= 0)
		off->set_channel (channel);

	off->loop = loop_cycle;

	note_offs.push (off);
}

/* // Py<Part> //////////////////////////////////////////////////////////// */

Type * Part_type = 0;

char * Py<Part>::__doc__ = 
"Part (list) -- a Plugin that streams out events from an event list.";

/* getsets */
static PyObject * 
Part_get_list (Part * p)
{
	return ::incref (p->list);
}

static PyObject * 
Part_get_mute (Part * p)
{
	return PyInt (p->mute);
}

static int 
Part_set_list (Part * p, PyObject * el) 
Py_TRY
{
	if (!EventList_type->is_derived (el))
		throw TypeError() << "an EventList is required.";
		
	p->set_list ((EventList *) el);

	return 0;
} 
Py_CATCH (-1);

static int
Part_set_mute (Part * p, PyObject * mute)
{
	if (!PyArg_Parse (mute, "i", &p->mute))
		return -1;
	return 0;
}

static PyObject *
Part_get_channel (Part * p)
{
	return p->channel >= 0 ? PyInt (p->channel) : PyNone();
}

static int
Part_set_channel (Part * p, PyObject * channel)
Py_TRY
{
	if (channel == Py_None)
		p->channel = -1;
	else
	{
		int c;

		if (!PyArg_Parse (channel, "i", &c))
			return -1;

		if (c < 0 || c > 15)
			throw ValueError() << "channel must be in (0 .. 15)";
		
		p->channel = c;
	}

	return 0;
}
Py_CATCH (-1);

static PyObject *
Part_get_transpose (Part * p)
{
	return PyInt (p->transpose);
}

static int
Part_set_transpose (Part * p, PyObject * transpose)
{
	if (!PyArg_Parse (transpose, "i", &p->transpose))
		return -1;
	return 0;
}

static PyGetSetDef 
Part_getsets [] = 
{
	{"list", (getter) Part_get_list, (setter) Part_set_list, "The EventList."},
	{"mute", (getter) Part_get_mute, (setter) Part_set_mute, 
	 "whether to send Note events (all others are sent regardless)."},
	{"channel", (getter) Part_get_channel, (setter) Part_set_channel, 
	 "the channel to impose on (MIDI) events traded, or None."},
	{"transpose", (getter) Part_get_transpose, (setter) Part_set_transpose, 
	 "transpose played Notes this many semitones."},
	{ }
};

void
Py<Part>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"list", "channel", "name", 0};

	EventList * e = 0;
	char * name = 0;
	int c = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!|is", keys,
	    EventList_type, &e, &c, &name))
		throw blewit();
	
	channel = c ? ((c - 1) & 0xF) : -1;
	
	set_list (e);
	if (name)
		NamedObject::set_object_name (name);
	else if (e)
		NamedObject::set_object_name (e->get_object_name());
}

void
Py<Part>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = Part_getsets;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
}

/* /////////////////////////////////////////////////////////////////////// */

MidiPlexer::MidiPlexer()
: Plugin (MidiPlexer_type, 0 /* no cycles, just trigger */)
{
	in = add_in (new In (this, MidiEvent_type, 2, "in"));
	in->set_flag (In::Trigger);

	thru = add_out (new Out (this, MidiEvent_type, "thru"));
	copy = add_out (new Out (this, MidiEvent_type, "copy"));

	/* pushing the same event on that was passed in: 'thru' needs no cache */
	thru->cache->release_events();
	thru->cache->clear (true);

	channel = -1;
	block_mask = 0; /* no events blocked initially */

	/* set up controller mapping to identity */
	for (int i = 0; i < 128; ++i)
		cc_map[i] = (byte) i;
}

void
MidiPlexer::process_trigger (In * in)
{
	ScopedLock lock (mutex);

	MidiEvent * e;
	
	while ((e = (MidiEvent *) in->pop()))
	{
		bool fw = true;

		if (get_type (e) == MidiEvent_type)
		{
			if (channel >= 0 && e->has_channel())
				e->set_channel (channel);

			fw = !is_blocked (((MidiEvent *) e)->status());

			/* rewrite controller number */
			if (e->is_controller())
				(*e)[1] = cc_map[(*e)[1]];
			
			/* copy and push */
			MidiEvent * c = (MidiEvent *) copy->cache->pop();
			if (c)
			{
				c->clone (e);
				copy->push (c);
			}
		}

		if (fw)
			thru->push (e);
		else
			::decref (e);
	}
}

/* // Py<MidiPlexer> /////////////////////////////////////////////////////// */

static PyObject *
MidiPlexer_get_channel (MidiPlexer * p)
{
	return p->channel >= 0 ? PyInt (p->channel + 1) : PyNone();
}

static int
MidiPlexer_set_channel (MidiPlexer * p, PyObject * channel)
Py_TRY
{
	if (channel == Py_None)
	{
		p->channel = -1;
		return 0;
	}

	int c;
	if (!PyArg_Parse (channel, "i", &c))
		return -1;
	
	if (c < 1 || c > 16)
		throw ValueError() << "channel is None or from [1 .. 16]";

	p->channel = c - 1;
	return 0;
}
Py_CATCH (-1);

static PyObject *
MidiPlexer_get_block_mask (MidiPlexer * p)
{
	return PyInt (p->block_mask);
}

static int
MidiPlexer_set_block_mask (MidiPlexer * p, PyObject * block_mask)
{
	if (!PyArg_Parse (block_mask, "i", &p->block_mask))
		return -1;
	return 0;
}

static PyGetSetDef 
MidiPlexer_getsets [] = 
{
	{"channel", (getter) MidiPlexer_get_channel, (setter) MidiPlexer_set_channel, 
	 "the channel to impose on MidiEvents routed, or -1."},
	{"block_mask", (getter) MidiPlexer_get_block_mask, (setter) MidiPlexer_set_block_mask, 
	 "bit mask indicating, per MIDI status nibble, whether to forward events:\n"
	 "bit 0 -- 0x80 = NoteOff\n"
	 "bit 1 -- 0x90 = NoteOn\n"
	 "bit 2 -- 0xA0 = NotePressure\n"
	 "bit 3 -- 0xB0 = Controller\n"
	 "bit 4 -- 0xC0 = Program\n"
	 "bit 5 -- 0xD0 = ChannelPressure\n"
	 "bit 6 -- 0xE0 = Pitch\n"
	 "bit 7 -- 0xF0 = Sysex and realtime events"},
	{ }
};

static PyObject *
MidiPlexer_is_blocked (MidiPlexer * self, PyObject * arg)
{
	byte status;
	
	if (!PyArg_Parse (arg, "B", &status))
		return 0;

	return PyInt (self->is_blocked (status));
}

static PyObject *
MidiPlexer_block (MidiPlexer * self, PyObject * arg)
{
	byte status;
	
	if (!PyArg_Parse (arg, "B", &status))
		return 0;

	self->block (status);
	return PyNone();
}

static PyObject *
MidiPlexer_unblock (MidiPlexer * self, PyObject * arg)
{
	byte status;
	
	if (!PyArg_Parse (arg, "B", &status))
		return 0;

	self->unblock (status);
	return PyNone();
}

static PyObject *
MidiPlexer_send (MidiPlexer * self, MidiEvent * arg)
{
	if (!MidiEvent_type->is_derived (arg))
		throw TypeError() << "need a MidiEvent.";

	if (self->channel >= 0 && arg->has_channel())
		arg->set_channel (self->channel);

	ScopedLock lock (self->mutex);
	::incref (arg);
	self->thru->push (arg);
	return PyNone();
}

static PyMethodDef 
MidiPlexer_methods [] = {
	{"is_blocked", (PyCFunction) MidiPlexer_is_blocked, METH_O,
		"is_blocked (MIDI status) -- whether MIDI events with 'status' are "
		"forwarded to the 'thru' output or not."},
	{"block", (PyCFunction) MidiPlexer_block, METH_O,
		"block (MIDI status) -- don't forward MIDI events with 'status' "
		"to the 'thru' output."},
	{"unblock", (PyCFunction) MidiPlexer_unblock, METH_O,
		"unblock (MIDI status) -- do forward MIDI events with 'status' "
		"to the 'thru' output."},
	{"send", (PyCFunction) MidiPlexer_send, METH_O,
		"send (MIDI event) -- send a MIDI event right now. Acquires the plexer "
	  "mutex, which can, depending on the plexer's input connection, "
		"block realtime threads, so use with care."},
	{ }
};

Type * MidiPlexer_type = 0;

char * Py<MidiPlexer>::__doc__ = 
"MidiPlexer() -- a Plugin that copies one inbound stream of MidiEvents "
"to two outbound streams. SysexEvents are only routed through once.";

void
Py<MidiPlexer>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();
}

void
Py<MidiPlexer>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = MidiPlexer_getsets;
	t->tp_methods = MidiPlexer_methods;
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_Part_types (PyObject * dict)
{
	Part_type = new PyType<Part> ("Part", dict);
	MidiPlexer_type = new PyType<MidiPlexer> ("MidiPlexer", dict);
}
