#ifdef WITH_FFTW

#include <PyType.h>

#include "PhaseVocoder.h"

Type * PVocFrame_type;
Type * PhaseVocoder_type;

/* PVocFrame getsets */

int 
PVocFrame_set_frame (PVocFrame * e, PyObject * value)
{
	if (!PyArg_Parse (value, "L", &e->frame))
		return -1;
	return 0;
}

PyObject * 
PVocFrame_get_frame (PVocFrame * e)
{
	return PyFrame (e->frame);
}

static PyGetSetDef 
PVocFrame_getsets [] = 
{
	{"frame", (getter) PVocFrame_get_frame, (setter) PVocFrame_set_frame, 
	 "the frame (point in time, measured in samples)."},
	{ }
};

/* ///////////////////////////////////////////////////////////////////// */

/* sequence methods */

static int 
PVocFrame_sq_length (PVocFrame * e)
{
	return e->size;
}

static PyObject *
PVocFrame_sq_item (PVocFrame * e, int i) 
Py_TRY
{
	if (i < 0 || i >= e->size)
		throw IndexError() << "no sample " << i << " in " << e;
	return PyFloat (e->data [i]); 
} 
Py_CATCH (0);

static int
PVocFrame_sq_ass_item (PVocFrame * e, int i, PyObject * value) 
Py_TRY
{
	if (i < 0 || i >= e->size)
		throw IndexError() << "no sample " << i << " in " << e;
	
	if (!PyArg_Parse (value, "f", e->data + i))
		return -1;

	return 0;
} 
Py_CATCH (-1);

static PySequenceMethods 
PVocFrame_as_sequence = 
{
	(inquiry) PVocFrame_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) PVocFrame_sq_item, /* sq_item */
	(intintargfunc) 0, /* sq_slice */
	(intobjargproc) PVocFrame_sq_ass_item, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

/* Py<PVocFrame> */

char * Py<PVocFrame>::__doc__ = 
"PVocFrame (size) -- a phase vocoder frame as generated and digested by "
"the PhaseVocoder plugin.";

void
Py<PVocFrame>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"size", 0};
	int n;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "i", keys, &n))
		return;
	
	if (n < 2)
		throw ValueError() << "size must be >= 2.";

	set_size (n);
}

void
Py<PVocFrame>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags &= ~Py_TPFLAGS_CHECKTYPES;
	t->tp_base = Object_type;
	t->tp_getset = PVocFrame_getsets;
	t->tp_as_sequence = &PVocFrame_as_sequence;
}

/* // PhaseVocoder ///////////////////////////////////////////////////// */

PhaseVocoder::PhaseVocoder()
: Plugin (PhaseVocoder_type, Audio)
{
	fill = 0;
	frametype = PVOC_AMP_FREQ;
}

PhaseVocoder::~PhaseVocoder()
{
}

void
PhaseVocoder::init (int dir, int fs, int fft, int df)
{
	direction = dir;
	sample_rate = fs;
	fft_length = fft;
	decfac = df;

	if (!vocoder.init (sample_rate, fft_length, decfac, PVPP_STREAMING))
		throw RuntimeError() << "vocoder fails to initialize.";
}

/* ///////////////////////////////////////////////////////////////////// */

void
PhaseVocoder::set_state (Graph * g, int state)
{
}

void
PhaseVocoder::process_audio (int frames)
{
}

/* enter: samples, number of samples, frame to fill;
 * return: number of samples taken, sets frame->frametype if a frame was
 * produced.
 * assumes: frame->size is sufficient (fft_length + 2)
 */
int
PhaseVocoder::produce_frame (d_sample * s, int n, PVocFrame * frame)
{
	frame->frametype = PVOC_INVALID;

	n = min (decfac - fill, n);

	fill += n;
	if (fill >= decfac)
	{
		vocoder.generate_frame (s, frame->data, decfac, frametype);
		frame->frametype = frametype;
		fill -= decfac;
	}

	return n;
}

/* assumes: enough room for 'decfac' samples in 's'
 */
int
PhaseVocoder::produce_audio (PVocFrame * frame, d_sample * s)
{
	if (frame->frametype == PVOC_INVALID)
		throw AssertionError() << "frame is marked invalid.";

	return (int) vocoder.process_frame (frame->data, s, frame->frametype);
}

/* ///////////////////////////////////////////////////////////////////// */

static PyObject *
pvoc_produce_frames (PhaseVocoder * v, AudioSignal * a)
{
	int maxframes = a->size / v->decfac + 2;
	
	PVocFrame * frames [maxframes];
	memset (frames, 0, sizeof (frames));

	int frame = 0;
	int n = a->size;
	d_sample * s = a->data;
	
	while (n)
	{
		if (frames[frame] == 0)
		{
			frames[frame] = NEW (PVocFrame);
			frames[frame]->set_size (v->fft_length + 2);
		}
		
		int taken = v->produce_frame (s, n, frames [frame]);

		/* frame got filled, on to the next */
		if (frames [frame]->frametype != PVOC_INVALID)
			++frame;

		n -= taken;
		s += taken;
	}
	
	msg(), "maxframes ", maxframes, "frame: ", frame;

	PyObject * list = PyList_New (frame);
	for (int i = 0; i < frame; ++i)
		PyList_SetItem (list, i, frames[i]);
	
	if (frames[frame]) ::decref (frames[frame], true);

	return list;
}

static PyObject *
pvoc_produce_audio (PhaseVocoder * v, PVocFrame * frame)
{
	AudioSignal * a = NEW (AudioSignal);
	a->set_size (v->decfac);

	v->produce_audio (frame, a->data);

	return a;
}

static PyObject *
PhaseVocoder_process (PhaseVocoder * v, PyObject * arg)
Py_TRY
{
	if (v->graph && v->state)
		throw RuntimeError() << "can't use this offline in realtime operations.";
	
	if (v->direction > 0)
	{
		if (!AudioSignal_type->is_derived (arg))
			throw TypeError() << "needs an AudioSignal.";

		return pvoc_produce_frames (v, (AudioSignal *) arg);
	}

	if (!PVocFrame_type->is_derived (arg))
		throw TypeError() << "needs a PVocFrame.";
	
	return pvoc_produce_audio (v, (PVocFrame *) arg);
}
Py_CATCH (0);

static PyMethodDef
PhaseVocoder_methods [] =
{
	{"process", (PyCFunction) PhaseVocoder_process, METH_O,
	 "process (audio | frame) -- depending on the vocoders 'direction', "
	 "either produces a sequence of PVocFrames from an AudioSignal, or one "
	 "AudioSignal from one PVocFrame."},
	{ }
};

/* ///////////////////////////////////////////////////////////////////// */

static PyObject *
PhaseVocoder_get_direction (PhaseVocoder * r)
{
	return PyInt (r->direction);
}

static PyObject *
PhaseVocoder_get_fft_length (PhaseVocoder * r)
{
	return PyInt (r->fft_length);
}

static PyObject *
PhaseVocoder_get_decfac (PhaseVocoder * r)
{
	return PyInt (r->decfac);
}

static PyObject *
PhaseVocoder_get_sample_rate (PhaseVocoder * r)
{
	return PyInt (r->sample_rate);
}

static PyGetSetDef 
PhaseVocoder_getsets [] =
{
	{"direction", (getter) PhaseVocoder_get_direction, (setter) 0,
	 ">= 0 if the vocoder produces PVocFrames from audio, < 0 otherwise."},
	{"fft_length", (getter) PhaseVocoder_get_fft_length, (setter) 0,
	 "length of the Fourier Transform. powers of two work better; start "
	 "with 1024."},
	{"sample_rate", (getter) PhaseVocoder_get_sample_rate, (setter) 0,
	 "."},
	{"decfac", (getter) PhaseVocoder_get_decfac, (setter) 0,
	 "the decimation factor (step size between taking the FT)."},
	{ }
};

void
Py<PhaseVocoder>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = PhaseVocoder_getsets;
	t->tp_methods = PhaseVocoder_methods;
}

char * Py<PhaseVocoder>::__doc__ = 
"PhaseVocoder() -- generates PVocFrames from an audio signal or vice versa.";

void
Py<PhaseVocoder>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {
			"direction", "sample_rate", "fft_length", "decfac", 0};

	int dir, fft = 1024, sr = SAMPLE_RATE, df = 256;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "i|iii", keys, 
				&dir, &sr, &fft, &df))
		return;

	init (dir, sr, fft, df);
}

/* //////////////////////////////////////////////////////////////////// */

void
init_PhaseVocoder_types (PyObject * dict)
{
	PVocFrame_type = new PyType<PVocFrame> ("PVocFrame", dict);
	PhaseVocoder_type = new PyType<PhaseVocoder> ("PhaseVocoder", dict);
}

#endif /* WITH_FFTW */
