/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _PLUGIN_H_
#define _PLUGIN_H_

#include "InOut.h"

class Graph;

class Plugin
: public NamedObject
{
	public:
		/* the plugin is usually part of a processing graph. */
		Graph * graph;
	
		/* events are received and transmitted at the In/Out connection points. */
		InList ins;
		OutList outs;

		/* the graph processing order is determined by 'y' first, then 'x'. */
		d_float x, y;
		
		bool before (Plugin * p)
			{
				return y < p->y || (y == p->y && x < p->x);
			}

		/* The loop cycle the plugin is currently in. The plugin is responsible
		 * for the correctness of this value. It is used by In::pop_before() to
		 * correctly evaluate which event comes next; cf. Event::loop 
		 */
		int loop_cycle;

		/* 'flags' tells the graph in which processing cycles/threads the plugin 
		 * takes part. 
		 *
		 * Audio preempts all others if adjustment of scheduling parameters is 
		 * granted. Time runs at sched_fifo, too, if possible, yet at a lower 
		 * priority. All other threads run at sched_other. 
		 */
		enum 
		{
			Beat = 1,	
			Audio = 2,
			Time = 4,
			Disk = 8,
			Script = 16,
		};		
		int flags;

		bool is_audio() { return flags & Audio; }
		bool is_time() { return flags & Time; }
		bool is_disk() { return flags & Disk; }
		bool is_beat() { return flags & Beat; }
		bool is_script() { return flags & Script; }

		/* for every flag set, a plugin should implement the corresponding
		 * processor function. this process_*() function should fetch all input 
		 * events due in this cycle (denoted by the graph clock and the parameter 
		 * given) and process them. 
		 */
		  
 		/* cycling at about every eighth note: */
		virtual void process_beat (d_tick until) 
			{ msg(), "empty process_beat(): ", this; }
		
	 	/* every audio cycle; this cycle is triggered by an AudioClock plugin. 
		 * all audio connection points must be serviced in this cycle. 
		 */
		virtual void process_audio (int frames)
			{ msg(), "empty process_audio(): ", this; }

		/* every clock pulse; these cycles are triggered by a RTClock.  
		 * parameters are the time elapsed since the last call (dt) and the 
		 * number of midi clocks coinciding with the interval. one midi clock 
		 * is the 24th part of a quarter note. 
		 */
		virtual void process_time (d_time dt, int clocks)
			{ msg(), "empty process_time(): ", this; }
		
		/* initiated by a DiskScheduler */
		virtual void process_disk (Plugin * p) 
			{ msg(), "empty process_disk(): ", this; }
		
		/* initiated by a ScriptScheduler */ 
		virtual void process_script (Plugin * p) 
			{ msg(), "empty process_script(): ", this; }

		/* initiated by pushing an event into an In flagged In::Trigger */
		virtual void process_trigger (In * in)
			{ msg(), "empty process_trigger(): ", this; }

		/* wakes up scheduler, clock and script plugins. */
		virtual void wake_up() { }
		
		
		/* The plugin starts in Off state. When switching up to Prepared, the 
		 * plugin shall allocate all resources it needs for processing. Two audio
		 * cycles are run by AlsaClock after a successful switch to Prepared to
		 * prime the playback stream.
		 * 
		 * the state is then raised to Activated, in which state all processing
		 * cycles in the graph are running, although the graph's clocks aren't
		 * advanced.
		 * 
		 * the state may then become Rolling, indicating that all clocks are 
		 * running and that playback and recording are active.
		 * 
		 * the state will then drop back to either Off or Activated, and the
		 * state cycle begins again (or the plugin is removed from the graph). 
		 */
		enum 
		{
			Off = 0,				/* full stop. */
			Prepared = 1,		/* prepared to roll, still stopped. */
			Activated = 2,	/* processing runs, clocks, playback & capture do not. */
			Rolling = 3			/* full steam ahead. */
		};
		
		int state;
		
		/* should be implemented by plugins that need to do initialization
		 * tasks prior to processing (malloc, open, read etc.). 
		 */
		virtual void set_state (Graph * g, int _state) {state = _state;}
		
		/* disk and script scheduling run asynchronously. This mutex is held by 
		 * the graph while disk or script is processed, and should also be held 
		 * in the deconstructor, so that the deconstructor will never run 
		 * concurrently with processing. 
		 */
		Mutex mutex;

		/* set by the graph, this is the maximum time (in seconds) the plugin 
		 * spent processing audio. 
		 */
		d_time cpu;
		
	public:
		Plugin (Type * t = Plugin_type, int _flags = 0)
			: NamedObject (t),
				ins ("ins"), outs ("outs")
			{
				graph = 0;
				x = y = 0;
				loop_cycle = 0;
				flags = _flags;
				state = Off;
				cpu = 0.0;
			}

		~Plugin()
			{
				ScopedLock lock (mutex);

				ins.clear();
				outs.clear();
			}
	
		/* append (i = -1) an In and steal a reference. */
		In * add_in (In * in, int i = -1)
			{
				ins.add_object (in, i);
				::decref (in, true);
				
				return in;
			}

		/* append (i = -1) an Out and steal a reference. */
		Out * add_out (Out * out, int i = -1)
			{
				outs.add_object (out, i);
				::decref (out, true);
				
				return out;
			}

		/* the returned value is only reliable if the graph lock is held, which
		 * means it can't run concurrently with Out::connect(). 
		 */
		int count_connections()
			{
				int n = 0;

				for (int i = 0; i < ins; i++)
					n += ins[i]->connected;

				for (int i = 0; i < outs; i++)
					if (outs[i]->connected)
						++n;

				return n;
			}
		
		/* return the In that has the next event before a tick. Implementation
		 * in InOut.cc
		 */
		In * next_event (d_tick before);
		
	public:
		/* Following is a bunch of virtual methods that are called by either
		 * the graph or GraphRequests to notify the plugin of state changes,
		 * insertion into or removal from a graph.
		 */
		
		/* move to the current graph position. This method is run under realtime
		 * constraints. 
		 */
		virtual void seek() 
			{ }
		
		/* these four are called before / after the plugin is inserted into / 
		 * removed from a graph. the pre_* methods may throw an exception, the
		 * post_* should not.
		 */
		virtual void pre_insert (Graph * g) 
			{
				if (graph)
					throw AssertionError() << "already in a graph.";
			}
		virtual void post_insert (Graph * g) { }

		virtual void pre_remove() { }
		virtual void post_remove (Graph * g) { }

		/* these two are called before another plugin is inserted into/removed
		 * from the same graph. Plugins that rely on other plugins to operate
		 * correctly can throw an exception to prevent the insertion/removal. 
		 */
		virtual void pre_insert_other (Plugin *) { }
		virtual void pre_remove_other (Plugin *) { }

		/* if a plugin has CyclicAudio Outs, this allocates the necessary
		 * sample buffers in the object cache. The default implementation is in 
		 * Out.cc near Out::Out(). 
		 */
		virtual void set_frames_per_cycle (int frames_per_cycle);
};

#endif /* !_PLUGIN_H_ */
