/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "AlsaAudio.h"

#include "PyType.h"

static int 
AlsaClock_set_sample_rate (AlsaClock * c, PyObject * value) 
Py_TRY
{
	if (c->state > Plugin::Off)
		throw AssertionError() << "already rolling.";

	int sr;
	
	if (!PyArg_Parse (value, "i", &sr))
		return -1;

	if (sr < 1) 
		throw ValueError() << "invalid sample_rate.";

	c->sample_rate = sr;
	return 0;
} 
Py_CATCH (-1);

static PyObject *
AlsaClock_get_sample_rate (AlsaClock * c)
{
	return PyInt (c->sample_rate);
}

/* frames_per_cycle */

static int 
AlsaClock_set_frames_per_cycle (AlsaClock * c, PyObject * value) 
Py_TRY
{
	if (c->state > Plugin::Off)
		throw AssertionError() << "only in stopped state.";

	int fpc;
	
	if (!PyArg_Parse (value, "i", &fpc))
		return -1;

	if (fpc <= 2)
		throw ValueError() << "frames_per_cycle must be > 2.";

	c->frames_per_cycle = fpc;
	return 0;
} 
Py_CATCH (-1);

static PyObject *
AlsaClock_get_frames_per_cycle (AlsaClock * c)
{
	return PyInt (c->frames_per_cycle);
}

/* show_stats */

static int 
AlsaClock_set_show_stats (AlsaClock * c, PyObject * value) 
Py_TRY
{
	int s;
	
	if (!PyArg_Parse (value, "i", &s))
		return -1;

	if (s < 0)
		throw ValueError() << "show_stats must be 0 or positive.";

	c->show_stats = s;
	return 0;
} 
Py_CATCH (-1);

static PyObject * 
AlsaClock_get_show_stats (AlsaClock * c)
{
	return PyInt (c->show_stats);
}

/* max_xrun_cycles */

static int 
AlsaClock_set_max_xrun_cycles (AlsaClock * c, PyObject * value) 
Py_TRY
{
	int v;
	
	if (!PyArg_Parse (value, "i", &v))
		return -1;

	if (v < 0)
		throw ValueError() << "invalid max_xrun_cycles value.";

	c->max_xrun_cycles = v;
	return 0;
} 
Py_CATCH (-1);

static PyObject * 
AlsaClock_get_max_xrun_cycles (AlsaClock * c)
{
	return PyInt (c->max_xrun_cycles);
}

/* priority */

static int 
AlsaClock_set_priority (AlsaClock * c, PyObject * value) 
Py_TRY
{
	int p;
	
	if (!PyArg_Parse (value, "i", &p))
		return -1;

	c->set_thread_priority (p);
	return 0;
} 
Py_CATCH (-1);

static PyObject *
AlsaClock_get_priority (AlsaClock * c)
{
	return PyInt (c->thread_priority);
}

static PyGetSetDef AlsaClock_getsets [] = {
	{"show_stats", 
	 (getter) AlsaClock_get_show_stats, 
	 (setter) AlsaClock_set_show_stats, 
	 "show statistics every 'show_stats' processing cycles. They are printed "
	 "on the 'stdmsg' logfile."},
	{"sample_rate", 
		(getter) AlsaClock_get_sample_rate, 
		(setter) AlsaClock_set_sample_rate,
	 "the sample rate to use. Set this before starting the graph."},
	{"frames_per_cycle", 
		(getter) AlsaClock_get_frames_per_cycle, 
		(setter) AlsaClock_set_frames_per_cycle,
	 "the number of audio frames to process per cycle. Set this before "
	 "starting the graph."},
	{"max_xrun_cycles", 
	 (getter) AlsaClock_get_max_xrun_cycles, 
	 (setter) AlsaClock_set_max_xrun_cycles, 
	 "the clock stops audio processing when an xrun happens twice within "
	 "'max_xrun_cycles' cycles of processing, default is 4. "
	 "Set this to -1 to disable this "
	 "feature and have the clock run on no matter what happens "
	 "(not recommended)."},
	{"priority", (getter) AlsaClock_get_priority, (setter) AlsaClock_set_priority,
	 "the scheduling priority the clock thread has."},
	{}
};

void
Py<AlsaClock>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = AudioClock_type;
	t->tp_getset = AlsaClock_getsets;
}

Type * AlsaClock_type = 0;

char * Py<AlsaClock>::__doc__ = 
"AlsaClock ([sample_rate = 44100, frames_per_cycle = 2048, buffer_cycles = 2]) -- "
"an AudioClock that drives cyclic audio "
"processing using AlsaStreams. the 'frames_per_cycle' parameter should "
"be a power of two for most sound cards; USB 1.x devices will need a "
"multiple of sample_rate / 1000, rounded up to the nearest integer "
"instead afaik (untested with USB).\n"
"Add this to the graph, then the ALSA stream(s), "
"then set the graph state to 1 to configure the streams and make them "
"instantiate their ports.\n"
"Xruns are reported on the 'stdmsg' logfile.\n"
"If more than one AlsaStream is present in the graph, "
"they are linked to run "
"in sync. Although ALSA allows it, it is usually not a good idea to try "
"and run different hardware audio interfaces on one AlsaClock; they will "
"soon start "
"to drift because their hardware sample word clocks are not in sync if "
"you haven't taken special precautions (SP/DIF or ADAT sync, or a soldering "
"iron). "
"The Midiman Delta series audio cards are claimed to work on a common "
"sample clock.";

void
Py<AlsaClock>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"sample_rate", "frames_per_cycle", "buffer_cycles", 0};
	int sr = SAMPLE_RATE;
	int fpc = FRAMES_PER_CYCLE;
	int bc = 2;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|iii", keys, &sr, &fpc, &bc))
		throw blewit();
	
	if (sr < 1 || sr > TEN_TO_THE_SIXTH)
		throw ValueError() << "invalid sample rate value.";
	if (fpc < 1 || fpc > TEN_TO_THE_SIXTH)
		throw ValueError() << "invalid frames per cycle value.";
	if (bc < 2 || bc > 1024)
		throw ValueError() << "invalid buffer_cycles value.";

	sample_rate = sr;
	frames_per_cycle = fpc;
	buffer_cycles = bc;
}

/* ///////////////////////////////////////////////////////////////////// */

static PyObject *
AlsaStream_set_gain (AlsaStream * s, PyObject * args)
Py_TRY
{
	int i;
	float g;
	
	if (!PyArg_ParseTuple (args, "if", &i, &g))
		return 0;

	if (i < 0 || i > s->channels)
		throw IndexError() << "no such channel";

	s->gain[i] = g;

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AlsaStream_get_gain (AlsaStream * s, PyObject * args)
Py_TRY
{
	int i;
	
	if (!PyArg_ParseTuple (args, "i", &i))
		return 0;

	if (i < 0 || i > s->channels)
		throw IndexError() << "no such channel";

	return PyFloat (s->gain[i]);
}
Py_CATCH (0);

static PyMethodDef
AlsaStream_methods [] =
{
	{"get_gain", (PyCFunction) AlsaStream_get_gain, METH_VARARGS,
	 "get_gain (channel) -- returns the gain on 'channel' (1 is unity)."},
	{"set_gain", (PyCFunction) AlsaStream_set_gain, METH_VARARGS,
	 "set_gain (channel, gain) -- sets the gain on 'channel' (1 is unity)."},
	{ }
};

static int 
AlsaStream_set_relaxed (AlsaStream * s, PyObject * value) 
{
	if (!PyArg_Parse (value, "i", &s->relaxed))
		return -1;
	return 0;
} 

static PyObject * 
AlsaStream_get_relaxed (AlsaStream * s)
{
	return PyInt (s->relaxed);
}

static PyGetSetDef AlsaStream_getsets [] = {
	{"relaxed", 
	 (getter) AlsaStream_get_relaxed, (setter) AlsaStream_set_relaxed, 
	 "If not relaxed, the driver will stop if an 'xrun' occurs. "
	 "Is only evaluated when the stream starts, ie. set this first if you "
	 "can't afford the slightest audio glitch."},
	{ }
};

Type * AlsaStream_type = 0;

char * Py<AlsaStream>::__doc__ = 
"AlsaStream() -- abstract base type for audio input and output streams "
"managed by ALSA sound drivers.\n"
"AlsaStreams (~In, ~Out) are configured when a Graph is set to 'Prepared' "
"state. They always ask the device for the maximum number of channels "
"available. After configuring, every channel is assigned an In or Out "
"depending on signal flow direction.\n"
"Plugins of this type don't work without an AlsaClock in the same graph.\n"
"You cannot instantiate this in Python, use AlsaIn and ~Out instead.";

void
Py<AlsaStream>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = AlsaStream_getsets;
	t->tp_methods = AlsaStream_methods;
}

void
Py<AlsaStream>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* ///////////////////////////////////////////////////////////////////// */

Type * AlsaIn_type = 0;

char * Py<AlsaIn>::__doc__ = 
"AlsaIn (device[, relaxed = 1]) -- "
"receives multi-channel audio signal from an ALSA "
"device. "
"Disable 'relaxed' to have the driver stop if an xrun happens, ie. if "
"you cannot afford even a single sample overrun.\n"
"Plugins of this type don't work without an AlsaClock in the same graph.\n"
"Alsa streams have only been tested with hardware devices (device = 'hw:0,0') "
"so far.";

void
Py<AlsaIn>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = AlsaStream_type;
}

void
Py<AlsaIn>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {"device", "relaxed", 0};
	char * device;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|i", keys,
			&device, &relaxed))
		throw blewit();
	
	/* does *not* configure it. */
	open (device);
}

/* ///////////////////////////////////////////////////////////////////// */

Type * AlsaOut_type = 0;

char * Py<AlsaOut>::__doc__ = 
"AlsaOut (device[, relaxed = 1]) -- "
"sends multi-channel audio signal to an ALSA device. "
"Disable 'relaxed' to have the driver stop if an xrun happens, ie. if "
"you cannot afford even a single sample underrun.\n"
"Plugins of this type don't work without an AlsaClock in the same graph.\n"
"Alsa streams have only been tested with hardware devices (device = 'hw:0,0') "
"so far.";

void
Py<AlsaOut>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = AlsaStream_type;
}

void
Py<AlsaOut>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {"device", "relaxed", 0};
	char * device;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|i", keys,
			&device, &relaxed))
		throw blewit();
	
	/* does *not* configure it. */
	open (device);
}

/* ///////////////////////////////////////////////////////////////////// */

void
init_AlsaAudio_types (PyObject * dict)
{
	AlsaStream_type = new PyType<AlsaStream> ("AlsaStream", dict);
	AlsaIn_type = new PyType<AlsaIn> ("AlsaIn", dict);
	AlsaOut_type = new PyType<AlsaOut> ("AlsaOut", dict);
	AlsaClock_type = new PyType<AlsaClock> ("AlsaClock", dict);
}
