/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "AudioSignal.h"
#include "AudioClock.h"

#include "PyType.h"

#ifdef WITH_FFTW
#include <srfftw.h>
#endif /* WITH_FFTW */

#include "dsp/Sine.h"
#include "dsp/Phase.h"
#include "dsp/pisarenko.h"
#include "dsp/sinc.h"
#include "dsp/windows.h"

/* AudioSignal getsets */

int 
AudioSignal_set_frame (AudioSignal * e, PyObject * value)
{
	if (!PyArg_Parse (value, "L", &e->frame))
		return -1;
	return 0;
}

PyObject * 
AudioSignal_get_frame (AudioSignal * e)
{
	return PyFrame (e->frame);
}

static PyGetSetDef 
AudioSignal_getsets [] = 
{
	{"frame", (getter) AudioSignal_get_frame, (setter) AudioSignal_set_frame, 
	 "the frame (point in time, measured in samples)."},
	{ }
};

/* ///////////////////////////////////////////////////////////////////// */

static PyObject *
AudioSignal_inplace_add (AudioSignal * a, PyObject * arg)
Py_TRY
{
	if (!AudioSignal_type->is_derived (a))
		throw TypeError() << __FUNCTION__;

	if (AudioSignal_type->is_derived (arg))
	{
		AudioSignal * b = (AudioSignal *) arg;
		int n = min (a->size, b->size);

		for (int i = 0; i < n; ++i)
			a->data[i] += b->data[i];
	}
	else
	{
		d_sample f;
		
		if (!PyArg_Parse (arg, "f", &f))
			return 0;

		for (int i = 0; i < a->size; ++i)
			a->data[i] += f;
	}

	return ::incref (a);
}
Py_CATCH (0);

static PyObject *
AudioSignal_inplace_subtract (AudioSignal * a, PyObject * arg)
Py_TRY
{
	if (!AudioSignal_type->is_derived (a))
		throw TypeError() << __FUNCTION__;

	if (AudioSignal_type->is_derived (arg))
	{
		AudioSignal * b = (AudioSignal *) arg;
		int n = min (a->size, b->size);

		for (int i = 0; i < n; ++i)
			a->data[i] -= b->data[i];
	}
	else
	{
		d_sample f;
		
		if (!PyArg_Parse (arg, "f", &f))
			return 0;

		for (int i = 0; i < a->size; ++i)
			a->data[i] -= f;
	}

	return ::incref (a);
}
Py_CATCH (0);

static PyObject *
AudioSignal_inplace_multiply (AudioSignal * a, PyObject * arg)
Py_TRY
{
	if (!AudioSignal_type->is_derived (a))
		throw TypeError() << __FUNCTION__;

	if (AudioSignal_type->is_derived (arg))
	{
		AudioSignal * b = (AudioSignal *) arg;
		int n = min (a->size, b->size);

		for (int i = 0; i < n; ++i)
			a->data[i] *= b->data[i];
	}
	else
	{
		d_sample f;
		
		if (!PyArg_Parse (arg, "f", &f))
			return 0;

		for (int i = 0; i < a->size; ++i)
			a->data[i] *= f;
	}

	return ::incref (a);
}
Py_CATCH (0);

static PyObject *
AudioSignal_inplace_divide (AudioSignal * a, PyObject * arg)
Py_TRY
{
	if (!AudioSignal_type->is_derived (a))
		throw TypeError() << __FUNCTION__;

	d_sample f;
	
	if (!PyArg_Parse (arg, "f", &f))
		return 0;

	if (f == 0)
		throw ZeroDivisionError() << a;

	for (int i = 0; i < a->size; ++i)
		a->data[i] /= f;

	return ::incref (a);
}
Py_CATCH (0);

static void
shift (AudioSignal * a, int n)
{
	if (n == 0)
		return;

	int size = abs (n);
	int move = a->size - size;

	d_sample * t = new d_sample [size];
	
	if (n < 0)
	{
		memcpy (t, a->data, size * sizeof (d_sample));
		memmove (a->data, a->data + size, move * sizeof (d_sample));
		memcpy (a->data + move, t, size * sizeof (d_sample));
	}
	else
	{
		memcpy (t, a->data + move, size * sizeof (d_sample));
		memmove (a->data + size, a->data, move * sizeof (d_sample));
		memcpy (a->data, t, size * sizeof (d_sample));
	}

	delete [] t;
}

static PyObject *
AudioSignal_inplace_lshift (AudioSignal * a, PyObject * arg)
Py_TRY
{
	if (!AudioSignal_type->is_derived (a))
		throw TypeError() << __FUNCTION__;

	if (a->size == 0)
		return ::incref (a);

	int n;
	
	if (!PyArg_Parse (arg, "i", &n))
		return 0;

	n = -n;
	n %= a->size;
	
	shift (a, n);

	return ::incref (a);
}
Py_CATCH (0);

static PyObject *
AudioSignal_inplace_rshift (AudioSignal * a, PyObject * arg)
Py_TRY
{
	if (!AudioSignal_type->is_derived (a))
		throw TypeError() << __FUNCTION__;

	if (a->size == 0)
		return ::incref (a);

	int n;
	
	if (!PyArg_Parse (arg, "i", &n))
		return 0;
	
	n %= a->size;
	
	shift (a, n);

	return ::incref (a);
}
Py_CATCH (0);

static PyObject * 
AudioSignal_add (AudioSignal * a, AudioSignal * b)
Py_TRY
{
	if (!AudioSignal_type->is_derived (b))
		throw TypeError() << "can only add a second AudioSignal.";

	AudioSignal * r = NEW (AudioSignal);

	if (r == 0)
		throw MemoryError() << "no new audio signal, sorry.";

	r->set_size (a->size + b->size);
	
	for (int i = 0; i < a->size; ++i)
		r->data[i] = a->data[i];
	
	d_sample * d = r->data + a->size;
	for (int i = 0; i < b->size; ++i)
		d[i] = b->data[i];

	return r;
}
Py_CATCH (0);

static PyNumberMethods 
AudioSignal_as_number = {
	(binaryfunc) AudioSignal_add,	/*nb_add*/
	0,	/*nb_subtract*/
	0,	/*nb_multiply*/
	0, /*nb_divide*/
	0,	/*nb_remainder*/
	0,	/*nb_divmod*/
	0,	/*nb_power*/
	0,	/*nb_negative*/
	0,	/*nb_positive*/
	0,	/*nb_absolute*/
	0,	/*nb_nonzero*/
	0,	/*nb_invert*/
	0,	/*nb_lshift*/
	0,	/*nb_rshift*/
	0,	/*nb_and*/
	0,	/*nb_xor*/
	0,	/*nb_or*/
	0,		/*nb_coerce*/
	0,	/*nb_int*/
	0,	/*nb_long*/
	0,	/*nb_float*/
	0,	/*nb_oct*/
	0, 	/*nb_hex*/
	(binaryfunc) AudioSignal_inplace_add,			/*nb_inplace_add*/
	(binaryfunc) AudioSignal_inplace_subtract,			/*nb_inplace_subtract*/
	(binaryfunc) AudioSignal_inplace_multiply,			/*nb_inplace_multiply*/
	(binaryfunc) AudioSignal_inplace_divide,			/*nb_inplace_divide*/
	0,			/*nb_inplace_remainder*/
	0,			/*nb_inplace_power*/
	(binaryfunc) AudioSignal_inplace_lshift,			/*nb_inplace_lshift*/
	(binaryfunc) AudioSignal_inplace_rshift,			/*nb_inplace_rshift*/
	0,			/*nb_inplace_and*/
	0,			/*nb_inplace_xor*/
	0,			/*nb_inplace_or*/
	0,	/* nb_floor_divide */
	0,	/* nb_true_divide */
	0,			/* nb_inplace_floor_divide */
	0,			/* nb_inplace_true_divide */
};

/* ///////////////////////////////////////////////////////////////////// */

/* AudioSignal methods */

static PyObject *
AudioSignal_sine (AudioSignal * a, PyObject * arg)
Py_TRY
{
	double cycles, phase = 0.0;

	if (!PyArg_ParseTuple (arg, "d|d", &cycles, &phase))
		return 0;

#if 1 /* quicker but less accurate in the (very) long run */
	DSP::Sine sine (2 * cycles, a->size, phase);

	for (int i = 0; i < a->size; ++i)
		a->data[i] = sine.get();
#else
	double theta = cycles * 2 * M_PI / ((double) a->size);

	for (int i = 0; i < a->size; ++i)
		a->data[i] = sin (phase),
		phase += theta;
#endif
		
	return PyNone();	
}
Py_CATCH (0);

static PyObject *
AudioSignal_rectify (AudioSignal * a)
{
	for (int i = 0; i < a->size; ++i)
		a->data[i] = sgn (a->data[i]);
	return PyNone();
}

static PyObject *
AudioSignal_copy (AudioSignal * e, PyObject * args)
Py_TRY
{
	AudioSignal * s = 0;

	if (!PyArg_ParseTuple (args, "|O!", AudioSignal_type, &s))
		return 0;

	if (s)
	{
		*e = *s;
		return PyNone();
	}

	s = NEW (AudioSignal);
	
	try 
	{ 
		s->set_size (e->size);
		*s = *e; 
	} 
	catch (std::exception & e) 
	{
		::decref (s, true);
		throw;
	}

	return s;
}
Py_CATCH (0);

static PyObject *
AudioSignal_diff (AudioSignal * e, AudioSignal * d)
Py_TRY
{
	if (!AudioSignal_type->is_derived (d))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (d);
	if (d->size != e->size)
		throw ValueError() << "need an event of the same size.";

	register double diff = 0;
	register d_sample * s = e->data, * t = d->data;
	
	for (int i = 0; i < e->size; ++i)
	{
		register double d = s[i] - t[i];
		diff += d * d;
	}

	return PyFloat (sqrt (diff));
}
Py_CATCH (0);

static PyObject *
AudioSignal_peak (AudioSignal * e)
{
	register d_sample peak = 0.0;
	register d_sample * s = e->data;
	
	for (int i = 0; i < e->size; ++i)
		if (fabs (s[i]) > fabs (peak)) peak = s[i];

	return PyFloat (peak);
}

static PyObject *
AudioSignal_normalize (AudioSignal * e)
{
	double peak = 0.0;
	register d_sample * s = e->data;
	
	for (int i = 0; i < e->size; ++i)
	{
		register double d = fabs (s[i]);
		if (d > peak) peak = d;
	}

	if (peak)
	{
		peak = .9996 / peak;
		
		for (int i = 0; i < e->size; ++i)
			s[i] *= peak;
	}

	return PyNone();
}

static PyObject *
AudioSignal_silence (AudioSignal * e)
{
	e->silence();
	return PyNone();
}

/* windows */

static PyObject *
AudioSignal_blackman (AudioSignal * e)
{
	DSP::blackman<DSP::apply_window> (e->data, e->size);
	return PyNone();
}

static PyObject *
AudioSignal_blackman_harris (AudioSignal * e)
{
	DSP::blackman_harris<DSP::apply_window> (e->data, e->size);
	return PyNone();
}

static PyObject *
AudioSignal_hanning (AudioSignal * e)
{
	DSP::hanning<DSP::apply_window> (e->data, e->size);
	return PyNone();
}

static PyObject *
AudioSignal_kaiser (AudioSignal * e, PyObject * args)
{
	double beta = 1.0;
	
	if (!PyArg_ParseTuple (args, "|d", &beta))
		return 0;

	DSP::kaiser<DSP::apply_window> (e->data, e->size, beta);

	return PyNone();
}

static PyObject *
AudioSignal_rms (AudioSignal * e)
{
	register double rms = 0.f;
	register d_sample * s = e->data;
	
	for (int i = 0; i < e->size; ++i)
		rms += s[i] * s[i];

	return PyFloat (sqrt (rms) / (double) e->size);
}

static PyObject *
AudioSignal_ramp (AudioSignal * s, PyObject * args)
{
	d_sample s0, s1;
	
	if (!PyArg_ParseTuple (args, "ff", &s0, &s1))
		return 0;

	double x = s0;
	double step = (s1 - s0) / (double) s->size;

	for (int i = 0; i < s->size; ++i, x += step)
		s->data[i] = x;

	return PyNone();
}

static PyObject *
AudioSignal_paste (AudioSignal * a, PyObject * args)
Py_TRY
{
	int frame;
	AudioSignal * b;
	double gain = 1.;

	if (!PyArg_ParseTuple (args, "O!i|d", 
				AudioSignal_type, &b, &frame, &gain))
		return 0;
	
	if (frame >= a->size)
		throw IndexError() << "frame " << frame << " is beyond " << a;
	
	int n; /* number of frames to mix */
	d_sample * s = b->data; /* src */
	d_sample * d = a->data; /* dest */
	
	if (frame < 0) /* mix before start of signal */
	{
		n = b->size + frame;
		if (n <= 0) /* nothing to do */
			return PyNone();
		
		n = min (a->size, n);
		s -= frame;
	}
	else
	{
		n = min (a->size - frame, b->size);
		d += frame;
	}

	for (int i = 0; i < n; ++i)
		d[i] += gain * s[i];

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AudioSignal_grain (AudioSignal * a, PyObject * args) 
Py_TRY
{
	int frame, length, fade;

	if (!PyArg_ParseTuple (args, "iii", &frame, &length, &fade))
		return 0;
	
	if (frame < 0)
		throw IndexError() << "frame " << frame << " is beyond " << a;
	if (frame + length >= a->size)
		throw IndexError() << "frame " << frame + length << " is beyond " << a;
	if (fade > length / 2)
		throw ValueError() << "fade must be less than half the grain length.";

	AudioSignal * b = NEW (AudioSignal);
	
	b->set_size (length);
	
	d_sample g [fade];

#if 0
	/* fade envelope using gaussian distribution, doesn't sound so good */
	double step = 5. / (double) fade;
	
	double x = -5.;
	for (int i = 0; i < fade; ++i, x += step)
		g[i] = exp (-x * x * .5);
#else
	/* raised-cosine fade, much smoother */
	double x = M_PI / 2;
	double step = x / (double) fade;
	
	for (int i = 0; i < fade; ++i, x += step)
	{
		g[i] = cos (x);
		g[i] *= g[i];
	}
#endif
	
	d_sample * s = a->data + frame;
	d_sample * d = b->data;

	int i = 0;
	for (  ; i < fade; ++i)
		d[i] = s[i] * g[i];

	int n = length - fade;
	for (  ; i < n; ++i)
		d[i] = s[i];

	for (int j = fade - 1; i < length; ++i, --j)
		d[i] = s[i] * g[j];

	return b;
} 
Py_CATCH (0);

/* i'm not sure if we should allow python code to set the data size. some
 * plugins rely on the data size remaining constant (SamplePlayer for 
 * instance).
 */
static PyObject *
AudioSignal_set_data (AudioSignal * s, PyObject * data)
Py_TRY
{
	signed short * src;
	int n;

	if (PyString_AsStringAndSize (data, (char **) &src, &n))
		return 0;

	n /= 2;
	s->set_size (n);

	for (int i = 0; i < n; ++i)
		s->data[i] = (d_sample) (src[i]) * 3.0517578125e-05;

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AudioSignal_pisarenko1 (AudioSignal * s, PyObject * args)
Py_TRY
{
	int f0 = 0, f1 = s->size;

	if (!PyArg_ParseTuple (args, "|ii", &f0, &f1))
		return 0;
	
	if (f0 < 0 || f1 < f0 + 8 || f1 > s->size)
		throw ValueError() << "window is out of bounds.";

	double p = DSP::pisarenko1 (s->data + f0, f1 - f0);

	return PyFloat (p);
}
Py_CATCH (0);

template <int M>
static PyObject *
AudioSignal_xfade (PyObject * self, PyObject * args)
Py_TRY
{
	AudioSignal * s = (AudioSignal *) self;
	AudioSignal * t;
	
	if (!PyArg_ParseTuple (args, "O!", AudioSignal_type, &t))
		return 0;
	
	if (t->size != s->size)
		throw ValueError() << "xfade signals must be same size";

	if (M == 0) /* linear */
	{
		for (int i = 0; i < s->size; ++i)
		{
			register double fi = (double) i / (double) s->size;
			register double fo = 1. - fi;
			s->data[i] = fi * t->data[i] + fo * s->data[i];
		}
	}
	else if (M == 1) /* sin */
	{
		DSP::Sine gain (.25, s->size, 0);
		
		for (int i = 0; i < s->size; ++i)
		{
			double g = gain.get();
			g *= g;
			s->data[i] = g * s->data[i] + (1. - g) * t->data[i];
		}
	}
	else if (M == 2) /* sqrt */
	{
		for (int i = 0; i < s->size; ++i)
		{
			register double fi = (double) i / (double) s->size;
			register double fo = 1. - fi;
			s->data[i] = sqrt (fi) * t->data[i] + sqrt (fo) * s->data[i];
		}
	}

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AudioSignal_sample (AudioSignal * s, PyObject * method)
Py_TRY
{
	if (!PyCallable_Check (method))
		throw TypeError() << "needs a callable";

	PyRef args (PyTuple_New (1));
	PyRef x (PyFloat (0.));

	PyTuple_SetItem (args, 0, x);

	for (int i = 0; i < s->size; ++i)
	{
		((PyFloatObject *) x.o)->ob_fval = (double) i / (double) s->size;
		PyRef result (PyEval_CallObject (method, args));

		if (result == 0)
			throw blewit();
		
		if (!PyArg_Parse (result, "f", s->data + i))
			throw blewit();
	}

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AudioSignal_filter (AudioSignal * s, PyObject * method)
Py_TRY
{
	if (!PyCallable_Check (method))
		throw TypeError() << "needs a callable";

	PyRef args (PyTuple_New (1));
	PyRef x (PyFloat (0.));

	PyTuple_SetItem (args, 0, x);

	for (int i = 0; i < s->size; ++i)
	{
		((PyFloatObject *) x.o)->ob_fval = s->data[i];
		PyRef result (PyEval_CallObject (method, args));

		if (result == 0)
			throw blewit();
		
		if (!PyArg_Parse (result, "f", s->data + i))
			throw blewit();
	}

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AudioSignal_xfade_loop (AudioSignal * s, PyObject * args)
Py_TRY
{
	int f0, f1;

	if (!PyArg_ParseTuple (args, "ii", &f0, &f1))
		return 0;
	
	int length = f1 - f0;
	
	if (f0 < length)
		throw ValueError() << "not enough samples before loop start.";

	if (f0 < 0 || length < 8 || f1 > s->size)
		throw ValueError() << "window is out of bounds.";

	d_sample * dst = s->data + f0 - length;
	d_sample * src = s->data + f0;
	
	/* linear x-fade */
	for (int i = 0; i < length; ++i)
	{
		register double fi = (double) i / (double) length;
		register double fo = 1. - fi;
		dst[i] = fi * src[i] + fo * dst[i];
	}

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AudioSignal_find_0_crossing (AudioSignal * s, PyObject * args)
Py_TRY
{
	int f0 = 0;
	int sign = -1;

	if (!PyArg_ParseTuple (args, "|ii", &f0, &sign))
		return 0;
	
	sign = sgn (sign);

	if (f0 < 0 || f0 >= s->size)
		throw ValueError() << "window is out of bounds.";

	d_sample last = s->data[f0];
	for (int i = f0 + 1; i < s->size; ++i)
	{
		if (sgn (last) != sgn (s->data[i]) && sgn (last) == sign)
		{
			double fract = - last / (s->data[i] - last);

			return PyFloat (i + fract);
		}

		last = s->data[i];
	}

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AudioSignal_at (AudioSignal * s, PyObject * arg)
Py_TRY
{
	double d;

	if (!PyArg_Parse (arg, "d", &d))
		return 0;

	if (d < 0 || d >= s->size)
		throw IndexError() << "no such sample";
	
	DSP::Phase phi (d);

	return PyFloat (phi.cubic_get (s->data));
}
Py_CATCH (0);

static PyObject *
AudioSignal_xfade_into (AudioSignal * s, PyObject * args)
Py_TRY
{
	AudioSignal * t;
	int in_s, in_t;

	if (!PyArg_ParseTuple (args, "O!ii", AudioSignal_type, &t, &in_s, &in_t))
		return 0;

	if (in_s < 0 || in_s >= s->size)
		throw IndexError() << "insertion index is out of bounds.";

	if (in_t < 0 || in_t >= t->size)
		throw IndexError() << "insertion index is out of bounds.";

	AudioSignal * r = NEW (AudioSignal);
	if (r == 0) throw MemoryError();
	
	r->set_size (in_s + t->size - in_t);
	
	d_sample * srcs = s->data;
	d_sample * dst = r->data;
	
	int start_mix = in_s - in_t;

	/* straight copy if in_s is > in_t */
	for (int i = 0; i < start_mix; ++i)
		dst[i] = srcs [i];
	
	d_sample * srct = t->data;

	if (start_mix < 0)
		srct -= start_mix,
		start_mix = 0;
	else
		dst += start_mix,
		srcs += start_mix;

	/* fade length */
	int fade = in_s - start_mix;

	/* fade s in, t out */
	for (int i = 0; i < fade; ++i)
	{
		double f = (double) i / (double) fade;
		dst[i] = srcs[i] * (1 - f) + srct[i] * f;
	}

	/* copy t */
	srct += fade;
	dst += fade;
	
	int copy = t->size - in_t;
	for (int i = 0; i < copy; ++i)
		dst[i] = srct[i];

	return r;
}
Py_CATCH (0);

static PyObject *
AudioSignal_envelope (AudioSignal * a, PyObject * args)
Py_TRY
{
	if (!PySequence_Check (args))
		throw TypeError() << "needs a sequence";

	int n = PySequence_Length (args);

	int index[2];
	double gain[2];
	int z = 0;
	
	for (int i = 0; i < n; ++i)
	{
		PyObject * item = PySequence_GetItem (args, i);
		if (item == 0)
			return 0;

		if (!PyArg_ParseTuple (item, "id", index + z, gain + z))
		{
			::decref (item, true);
			return 0;
		}
		
		::decref (item, true);
		
		if (index[z] < 0) 
			index[z] = a->size + index[z];
		if (index[z] < 0 || index[z] > a->size)
			throw IndexError() << index[z];

		if (i == 0)
		{
			++z;
			continue;
		}

		int di = index[1] - index[0];
		if (di < 0)
			throw ValueError() << "indices must rise monotonically";

		double dg = gain[1] - gain[0];

		d_sample * samples = a->data + index[0];
		for (int s = 0; s < di; ++s)
			samples[s] *= gain[0] + (dg * s) / di;
		
		index[0] = index[1];
		gain[0] = gain[1];
	}

	return PyNone();
}
Py_CATCH (0);

static PyObject *
AudioSignal_upsample (AudioSignal * s, PyObject * arg)
Py_TRY
{
	int n;

	if (!PyArg_Parse (arg, "i", &n))
		return 0;
	if (n <= 1)
		throw ValueError() << "oversampling ratio must be > 1";

	AudioSignal * t = NEW (AudioSignal);
	PyRef ref (t);

	t->set_size (n * s->size);
	
	for (int i = 0; i < s->size; ++i)
		t->data[i * n] = s->data[i];

	return ::incref (t);
}
Py_CATCH (0);

static PyObject *
AudioSignal_downsample (AudioSignal * s, PyObject * arg)
Py_TRY
{
	int n;

	if (!PyArg_Parse (arg, "i", &n))
		return 0;
	if (n <= 1)
		throw ValueError() << "oversampling ratio must be > 1";
	if (s->size % n)
		throw ValueError() << 
				"signal length must be a multiple of oversampling ratio";

	AudioSignal * t = NEW (AudioSignal);
	PyRef ref (t);

	t->set_size (s->size / n);
	
	for (int i = 0; i < t->size; ++i)
		t->data[i] = s->data[i * n];

	return ::incref (t);
}
Py_CATCH (0);

static PyObject *
AudioSignal_sinc (AudioSignal * s, PyObject * arg)
Py_TRY
{
	double omega;

	if (!PyArg_Parse (arg, "d", &omega))
		return 0;

	DSP::sinc (omega, s->data, s->size);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
AudioSignal_methods [] =
{
	{"at", (PyCFunction) AudioSignal_at, METH_O,
	 "at (index) -- get the sample at index, which can be a floating "
	 "point number. uses cubic interpolation."},
	{"sine", (PyCFunction) AudioSignal_sine, METH_VARARGS,
	 "sine (cycles[, phase = 0]) -- sample a variable number of cycles of sin(), "
	 "starting at 'phase'."},
	{"rectify", (PyCFunction) AudioSignal_rectify, METH_NOARGS,
	 "rectify() -- replace all samples with -1 or 1 depending on their sign "
	 "(inplace)."},
	{"set_data", (PyCFunction) AudioSignal_set_data, METH_O,
	 "set_data (data) -- set signal to a copy of native-endian, 16-bit signed "
	 "string 'data'. this is a hack you should not use."},
	{"copy", (PyCFunction) AudioSignal_copy, METH_VARARGS,
	 "copy ([signal]) -- copy the signal data from 'signal', or, if "
	 "no 'signal' is provided, return a copy."},
	{"diff", (PyCFunction) AudioSignal_diff, METH_O,
	 "diff (signal) -- sample-wise compare, returns the square root of "
	 "the sum of all squared per-sample differences."},
	{"silence", (PyCFunction) AudioSignal_silence, METH_NOARGS,
	 "silence() -- set all samples to zero."},
	{"ramp", (PyCFunction) AudioSignal_ramp, METH_VARARGS,
	 "ramp (s0, s1) -- sample a linear ramp from 's0' to 's1'."},
	{"peak", (PyCFunction) AudioSignal_peak, METH_NOARGS,
	 "peak() -- return the peak amplitude of the signal."},
	{"normalize", (PyCFunction) AudioSignal_normalize, METH_NOARGS,
	 "normalize() -- scale the signal so that the peak sample value is +-1."},
	{"hanning", (PyCFunction) AudioSignal_hanning, METH_NOARGS,
	 "hanning() -- apply a Hanning window to the signal samples (inplace)."},
	{"blackman", (PyCFunction) AudioSignal_blackman, METH_NOARGS,
	 "blackman() -- apply a Blackman window to the signal samples (inplace)."},
	{"blackman_harris", (PyCFunction) AudioSignal_blackman_harris, METH_NOARGS,
	 "blackman_harris() -- apply a Blackman-Harris window to the signal "
	 "samples (inplace)."},
	{"kaiser", (PyCFunction) AudioSignal_kaiser, METH_VARARGS,
	 "kaiser ([beta = 1]) -- apply a Kaiser window to the signal samples (inplace)."},
	{"rms", (PyCFunction) AudioSignal_rms, METH_NOARGS,
	 "rms() -- calculate root-mean-square."},
	{"paste", (PyCFunction) AudioSignal_paste, METH_VARARGS,
	 "paste (signal, frame[, gain = 1.]) -- add 'signal' at 'frame' with 'gain'. "
	 "'frame' can be a negative number."},
	{"grain", (PyCFunction) AudioSignal_grain, METH_VARARGS,
	 "grain (frame, length, fade) -- return the sub-signal from 'frame' to "
	 "'frame' + 'length', faded in and out over 'fade' samples."},
	{"pisarenko1", (PyCFunction) AudioSignal_pisarenko1, METH_VARARGS,
	 "pisarenko1 ([start, end]) -- performs Pisarenko Harmonic Decomposition "
	 "for a single sinusoid. multiply the returned value with the sample "
	 "rate to get the frequency in Hz. Estimates get more precise with "
	 "larger sample sizes and better S/N ratios."},
	{"sample", (PyCFunction) AudioSignal_sample, METH_O,
	 "sample (method) -- sample the return value of 'method (x)' over a range "
	 "[0 .. 1[."},
	{"filter", (PyCFunction) AudioSignal_filter, METH_O,
	 "filter (method) -- iterate over all samples, replacing them with the "
	 "return value of 'method (self[i])'."},
	{"xfade", (PyCFunction) AudioSignal_xfade<0>, METH_VARARGS,
	 "xfade (other signal) -- merge with another signal, crossfading for "
	 "constant signal energy."},
	{"xfade_loop", (PyCFunction) AudioSignal_xfade_loop, METH_VARARGS,
	 "xfade_loop (start, end) -- blends the slice [start:end] with "
	 "the slice [start - (end - start):start] so that start:end form "
	 "a smooth loop (inplace)."},
	{"find_0_crossing", (PyCFunction) AudioSignal_find_0_crossing, METH_VARARGS,
	 "find_0_crossing ([start = 0, sign = -1]) -- returns a (fractional) sample index "
	 "where the waveform crosses 0. the search starts at sample index 'start' "
	 "and matches only 0 crossings coming from 'sign', which defaults to -1, "
	 "meaning that transitions from the lower to the upper lobe are detected."},
	{"xfade_into", (PyCFunction) AudioSignal_xfade_into, METH_VARARGS,
	 "xfade_into (other, my index, his index ) -- returns a new "
	 "audio signal composed of 'self' and 'other'. the source signals "
	 "are cross-faded so that at 'my index', 'self' is inaudible, and 'other' "
	 "is at full gain."},
	{"envelope", (PyCFunction) AudioSignal_envelope, METH_VARARGS,
	 "envelope ((index, gain), (index, gain), ...) -- applies a volume "
	 "envelope consisting of linear amplitude fades (inplace)."},
	{"upsample", (PyCFunction) AudioSignal_upsample, METH_O,
	 "upsample (n) -- returns a copy of the signal, upsampled by factor "
	 "'n', which must be an integer. the signal is simply interleaved with 0s."},
	{"downsample", (PyCFunction) AudioSignal_downsample, METH_O,
	 "downsample (n) -- returns a copy of the signal, downsampled by factor "
	 "'n', which must be an integer. the signal is simply decimated."},
	{"sinc", (PyCFunction) AudioSignal_sinc, METH_O,
	 "sinc (omega) -- samples the sinc function <sin (x * pi) / (x * pi)> "
	 "with dx = omega / pi."},
	{ }
};

/* sequence methods //////////////////////////////////////////////////////// */

static int 
AudioSignal_sq_length (AudioSignal * e)
{
	return e->size;
}

static PyObject *
AudioSignal_sq_item (AudioSignal * e, int i) 
Py_TRY
{
	if (i < 0 || i >= e->size)
		throw IndexError() << "no sample " << i << " in " << e;
	return PyFloat (e->data [i]); 
} 
Py_CATCH (0);

static PyObject *
AudioSignal_slice (AudioSignal * s, int n, int m) 
Py_TRY
{
	if (n > s->size)
		n = s->size;
	else
		n %= s->size;
	
	if (m >= s->size)
		m = s->size;
	else
		m %= s->size;
	
	if (n > m)
	{
		int h = n;
		n = m;
		m = h;
	}
	
	m -= n;

	if (m == 0)
		return PyNone();

	AudioSignal * t = NEW (AudioSignal);
	
	t->set_size (m);
	
	for (int i = 0; i < m; ++i)
		t->data[i] = s->data[n + i];

	return t;
} 
Py_CATCH (0);

static int
AudioSignal_sq_ass_item (AudioSignal * e, int i, PyObject * value) 
Py_TRY
{
	if (i < 0 || i >= e->size)
		throw IndexError() << "no sample " << i << " in " << e;
	
	if (!PyArg_Parse (value, "f", e->data + i))
		return -1;

	return 0;
} 
Py_CATCH (-1);

static PySequenceMethods 
AudioSignal_as_sequence = 
{
	(inquiry) AudioSignal_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) AudioSignal_sq_item, /* sq_item */
	(intintargfunc) AudioSignal_slice, /* sq_slice */
	(intobjargproc) AudioSignal_sq_ass_item, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

/* Py<AudioSignal> ///////////////////////////////////////////////////////// */

Type * AudioSignal_type = 0;

char * Py<AudioSignal>::__doc__ = 
"AudioSignal (size) -- a sequence of sampled audio signal. Supports basic "
"principles of the sequence protocol [] as well as some inplace arithmetic.";

void
Py<AudioSignal>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"size", 0};
	int n;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "i", keys, &n))
		return;
	
	if (n < 2)
		throw ValueError() << "size must be >= 2.";

	set_size (n);
}

void
Py<AudioSignal>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags &= ~Py_TPFLAGS_CHECKTYPES;
	t->tp_base = Object_type;
	t->tp_getset = AudioSignal_getsets;
	t->tp_methods = AudioSignal_methods;
	t->tp_as_sequence = &AudioSignal_as_sequence;
	t->tp_as_number = &AudioSignal_as_number;
}

/* // CyclicAudio //////////////////////////////////////////////////////// */

char * Py<CyclicAudio>::__doc__ = 
"CyclicAudio() -- sampled audio signal as used in the audio cycle.\n"
"You cannot instantiate this in Python.";

Type * CyclicAudio_type = 0;

void
Py<CyclicAudio>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<CyclicAudio>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = AudioSignal_type;
	/* t->tp_as_sequence = &AudioSignal_as_sequence; */
}

/* // AudioClock ////////////////////////////////////////////////////////// */

Type * AudioClock_type = 0;

char * Py<AudioClock>::__doc__ = 
"AudioClock() -- abstract base type for audio clocks. These plugins schedule "
"audio processing cycles in a graph. Most likely, an AudioClock will run or "
"depend on a realtime thread that communicates with a sound device.";

void
Py<AudioClock>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
}

void
Py<AudioClock>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* // FFT //////////////////////////////////////////////////////////////// */

#ifdef WITH_FFTW

static Type * FFT_type = 0;

class FFT
: public NamedObject
{
	public:
		int size;
		rfftw_plan plan, nalp;

		FFT();
		~FFT();

		void init (int s);
		void one (d_sample * s, d_sample * t);
		void eno (d_sample * s, d_sample * t);
};

FFT::FFT()
: NamedObject (FFT_type, "fft")
{
	size = 0;
	plan = 0;

	if (sizeof (fftw_real) != sizeof (d_sample))
		throw AssertionError() << 
				"sizeof (fftw_real) != sizeof (d_sample), please recompile using sfftw.";
}

FFT::~FFT()
{
	if (plan)
	{
		rfftw_destroy_plan (plan);
		rfftw_destroy_plan (nalp);
	}
}

void
FFT::init (int s)
{
	size = s;
	plan = rfftw_create_plan (size, FFTW_FORWARD, FFTW_ESTIMATE);
	nalp = rfftw_create_plan (size, FFTW_BACKWARD, FFTW_ESTIMATE);
}

void
FFT::one (d_sample * src, d_sample * dst)
{
	rfftw_one (plan, src, dst);
}

void
FFT::eno (d_sample * src, d_sample * dst)
{
	rfftw_one (nalp, src, dst);
}

/* //////////////////////////////////////////////////////////////////// */

static PyObject *
FFT_get_size (FFT * fft)
{
	return PyInt (fft->size);
}

static PyGetSetDef
FFT_getsets[] = 
{
	{"size", (getter) FFT_get_size, (setter) 0,
	 "size, in samples, of the FFT."},
	{ }
};

static PyObject *
FFT_one (FFT * fft, PyObject * args)
Py_TRY
{
	AudioSignal * e = 0;
	int start = 0;

	if (!PyArg_ParseTuple (args, "O!|i", AudioSignal_type, &e, &start))
		return 0;

	int size = fft->size;

	if (e->size - start < size)
		throw IndexError() << "not enough data.";
	
	AudioSignal * f = NEW (AudioSignal);
	if (f == 0) throw MemoryError();
	
	f->set_size (size);
	
	fft->one (e->data + start, f->data);

	/* scale to [-1, 1] */
	double root = 1. / sqrt (size);
	for (int i = 0; i < f->size; ++i)
		f->data[i] *= root;

	return f;
}
Py_CATCH (0);

static PyObject *
FFT_eno (FFT * fft, AudioSignal * e)
Py_TRY
{
	if (!AudioSignal_type->is_derived (e))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (e);

	int size = fft->size;

	/* output size */
	AudioSignal * f = NEW (AudioSignal);
	if (f == 0) throw MemoryError();
	
	f->set_size (size);
	
	/* make a temp array to carry the f-domain data for the transform.
	 * we need this if padding/truncating, and it saves the input signal
	 * from being modified by fftw.
	 */
	d_sample * temp = (d_sample *) malloc (sizeof (d_sample) * size);
	if (temp == 0)
	{
		::decref (f, true);
		throw MemoryError();
	}
	
	int n = size / 2;
	int n1 = size - 1;
	int m = e->size / 2;
	int m1 = e->size - 1;

	/* copy data from 'e' into temp */
	int mn = min (n, m); 
	for (int i = 0; i < mn; ++i)
		temp[i] = e->data[i], /* +f */
		temp[n1 - i] = e->data[m1 - i]; /* -f */
	
	/* zero-pad if necessary */
	for (int i = mn; i < n; ++i)
		temp[i] = temp[n1 - i] = 0;

	fft->eno (temp, f->data);

	free (temp);

	/* scale to [- sqrt (size) / 2, sqrt (size) / 2] so that the inverse
	 * fft returns the original signal.
	 */
	double root = 1. / sqrt (size);
	for (int i = 0; i < f->size; ++i)
		f->data[i] *= root;

	return f;
}
Py_CATCH (0);

/* FIXME: does this compute correctly scaled results? */
static PyObject *
FFT_power (FFT * fft, PyObject * args)
Py_TRY
{
	AudioSignal * e = 0;
	int logit = 0;

	if (!PyArg_ParseTuple (args, "O!|i", AudioSignal_type, &e, &logit))
		return 0;

	int size = fft->size;

	if (e->size != size)
		throw ValueError() << "size must be " << fft->size;
	
	AudioSignal * f = NEW (AudioSignal);
	if (f == 0) throw MemoryError();
	
	f->set_size ((size + 2) / 2);
			
	d_sample * t = e->data;
	d_sample * s = f->data;

	float k = sqrt (size) / 2;
	k *= k;
	k = 1. / k;

	int i = 0;
	s[i] = sqrt (t[i] * t[i] * k);
	for (  ; i < size / 2; ++i)
		s[i] = sqrt (t[i] * t[i] * k + t[size - i] * t[size - i] * k);
	s[i] = sqrt (t[i] * t[i] * k);
	
	if (logit)
	{
		for (i = 0; i < size / 2; ++i)
			s[i] = s[i] ? 20 * log10 (s[i]) : -144.f;
	}

	return f;
}
Py_CATCH (0);

static PyMethodDef
FFT_methods[] = 
{
	{"one", (PyCFunction) FFT_one, METH_VARARGS,
	 "one (signal[, start = 0]) -- compute the FT of an audio signal. returns "
	 "a new signal, scaled to sqrt (FFT size) / 2; for the layout of the "
	 "returned data please consult the FFTW documentation."},
	{"eno", (PyCFunction) FFT_eno, METH_O,
	 "eno (fft'd signal) -- apply the inverse FT to an audio signal. if "
	 "the f-domain data doesn't match the FFT's size, it is truncated or "
	 "zero-padded to fit, meaning the signal is resampled."},
	{"power", (PyCFunction) FFT_power, METH_VARARGS,
	 "power (fft'd signal[, logit = 0]) -- return the power spectrum of "
	 "frequency domain data, normalized to [0 .. 1] or in log scale."},
	{ }
};

/* //////////////////////////////////////////////////////////////////// */

char * Py<FFT>::__doc__ = 
"FFT (size) -- Jean-Baptiste Joseph Fourier's transform. Code "
"by fftw.org. Only does real to real transforms.";

void
Py<FFT>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = NamedObject_type;
	t->tp_getset = FFT_getsets;
	t->tp_methods = FFT_methods;
}

void
Py<FFT>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {"size", 0};
	int n;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "i", keys, &n))
		return;

	if (n < 8)
		throw ValueError() << "a FFT of " << n << " samples is ridiculous.";

	init (n);
}

#endif /* WITH_FFTW */

/* ///////////////////////////////////////////////////////////////////// */

static Type * Inverter_type;

class Inverter
: public Plugin
{
	public:
		Inverter();

		virtual void process_audio (int frames);
};

Inverter::Inverter()
: Plugin (Inverter_type, Audio)
{
	add_in (new In (this, CyclicAudio_type, IN_SLOTS, "in"));
	add_out (new Out (this, CyclicAudio_type, "out"));
}

void
Inverter::process_audio (int frames)
{
	CyclicAudio * event = ins[0]->pop_mix();
	
	if (event == 0)
		return;

	d_sample * s = event->data;
	
	for (int i = 0; i < frames; ++i)
		s[i] = -s[i];
	
	outs[0]->push (event);
}

/* ///////////////////////////////////////////////////////////////////// */

char * Py<Inverter>::__doc__ = 
"Inverter() -- Audio plugin that inverts the signal processed.";

void
Py<Inverter>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
}

void
Py<Inverter>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {0};
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		return;
}

/* ///////////////////////////////////////////////////////////////////// */

static PyObject *
catwaves (PyObject * self, PyObject * args)
Py_TRY
{
	AudioSignal * a, * b;

	if (!PyArg_ParseTuple (args, "O!O!", 
				AudioSignal_type, &a, AudioSignal_type, &b))
		return 0;
	
	if (a->size != b->size)
		throw ValueError(), "signals must be same length.";

	byte buf[4];
	int fd = fileno (stdout);

	for (int i = 0; i < a->size; ++i)
	{
		int l = (int) (a->data[i] * 32767.);
		int r = (int) (b->data[i] * 32767.);

		buf[0] = (byte) (l & 0xFF);
		buf[1] = (byte) ((l >> 8) & 0xFF);
		buf[2] = (byte) (r & 0xFF);
		buf[3] = (byte) ((r >> 8) & 0xFF);

		write (fd, buf, 4);
	}
	
	return PyNone();
}
Py_CATCH (0);

static PyMethodDef  
module_methods [] = 
{
	{"catwaves", (PyCFunction) catwaves, METH_VARARGS,
	 "catwaves (left signal, right signal) -- dump audio signals to stdout "
	 "as a 16-bit little-endian signed stereo sample stream."},
	{ }
};

/* ///////////////////////////////////////////////////////////////////// */

void
init_Audio_types (PyObject * dict)
{
	AudioSignal_type = new PyType<AudioSignal> ("AudioSignal", dict);
	CyclicAudio_type = new PyType<CyclicAudio> ("CyclicAudio", dict);

	AudioClock_type = new PyType<AudioClock> ("AudioClock", dict);

#ifdef WITH_FFTW
	FFT_type = new PyType<FFT> ("FFT", dict);
#endif /* WITH_FFTW */

	Inverter_type = new PyType<Inverter> ("Inverter", dict);

	for (PyMethodDef * md = module_methods; md->ml_name; ++md)
	{
		PyObject * m = PyCFunction_New (md, 0);

		if (m == 0)
		{
			msg(), "cannot create method: ", md->ml_name;
			continue;
		}

		if (PyDict_SetItemString (dict, md->ml_name, m))
			msg(), "cannot register method: ", md->ml_name;
	}
}
