/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "Graph.h"
#include "GraphRequests.h"
#include "TimeMap.h"

#include "PyType.h"

Type * GraphRequest_type = 0;

char * Py<GraphRequest>::__doc__ = 
"GraphRequest() -- abstract base type for modifications to a graph.";

static PyObject *
GraphRequest_get_graph (GraphRequest * r)
{
	return ::incref (r->graph);
}

static PyGetSetDef
GraphRequest_getsets [] = 
{
	{"graph", (getter) GraphRequest_get_graph, (setter) 0,
	 "the graph this request is directed at."},
	{ }
};

static void
exec_request (GraphRequest * r)
{
	Graph * g = r->graph;

	ScopedLock lock (g->requests->mutex);

	g->requests->push (r);

	try 
	{
		PythonUnlock unlock;
		g->requests->post();
	} 
	catch (blewit & b) 
	{
		g->requests->clear();

		throw;
	}
	g->requests->clear();
}

static PyObject *
GraphRequest_exec (GraphRequest * r)
Py_TRY
{
	::incref (r);
	exec_request (r);
}
Py_CATCH (0);

static PyMethodDef 
GraphRequest_methods [] = 
{
	{"exec", (PyCFunction) GraphRequest_exec, METH_NOARGS,
	 "execute this request."},
	{ }
};

void
Py<GraphRequest>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphRequest>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = Object_type;
	type->tp_methods = GraphRequest_methods;
	type->tp_getset = GraphRequest_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphStateChange_type = 0;

char * Py<GraphStateChange>::__doc__ = 
"GraphStateChange() -- represents a request for a graph state change.";

static PyObject *
GraphStateChange_get_state (GraphStateChange * r)
{
	return PyInt (r->state);
}

static PyGetSetDef
GraphStateChange_getsets [] = 
{
	{"state", (getter) GraphStateChange_get_state, (setter) 0,
	 "the state to change to."},
	{ }
};

void
Py<GraphStateChange>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphStateChange>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphStateChange_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphAddPlugin_type = 0;

char * Py<GraphAddPlugin>::__doc__ = 
"GraphAddPlugin() -- a request to a graph to add a plugin.";

static PyObject *
GraphAddPlugin_get_plugin (GraphAddPlugin * r)
{
	return ::incref (r->plugin);
}

static PyGetSetDef
GraphAddPlugin_getsets [] = 
{
	{"plugin", (getter) GraphAddPlugin_get_plugin, (setter) 0,
	 "the plugin to add."},
	{ }
};

void
Py<GraphAddPlugin>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphAddPlugin>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphAddPlugin_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphRemovePlugin_type = 0;

char * Py<GraphRemovePlugin>::__doc__ = 
"GraphRemovePlugin() -- a request to a graph to remove a plugin.";

static PyObject *
GraphRemovePlugin_get_plugin (GraphRemovePlugin * r)
{
	return ::incref (r->plugin);
}

static PyGetSetDef
GraphRemovePlugin_getsets [] = 
{
	{"plugin", (getter) GraphRemovePlugin_get_plugin, (setter) 0,
	 "the plugin to remove."},
	{ }
};

void
Py<GraphRemovePlugin>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphRemovePlugin>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphRemovePlugin_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphMovePlugin_type = 0;

char * Py<GraphMovePlugin>::__doc__ = 
"GraphMovePlugin() -- a request to a graph to move a plugin.";

static PyObject *
GraphMovePlugin_get_plugin (GraphMovePlugin * r)
{
	return ::incref (r->plugin);
}

static PyObject *
GraphMovePlugin_get_x (GraphMovePlugin * r)
{
	return PyFloat (r->x);
}

static PyObject *
GraphMovePlugin_get_y (GraphMovePlugin * r)
{
	return PyFloat (r->y);
}

static PyGetSetDef
GraphMovePlugin_getsets [] = 
{
	{"plugin", (getter) GraphMovePlugin_get_plugin, (setter) 0,
	 "the plugin to remove."},
	{"x", (getter) GraphMovePlugin_get_x, (setter) 0,
	 "the 'x' coordinate to move the plugin to."},
	{"y", (getter) GraphMovePlugin_get_y, (setter) 0,
	 "the 'y' coordinate to move the plugin to."},
	{ }
};

void
Py<GraphMovePlugin>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphMovePlugin>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphMovePlugin_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphConnect_type = 0;

char * Py<GraphConnect>::__doc__ = 
"GraphConnect() -- a request to a graph to (dis)connect an output point.";

static PyObject *
GraphConnect_get_out (GraphConnect * r)
{
	return r->out ? ::incref (r->out) : PyNone();
}

static PyObject *
GraphConnect_get_in (GraphConnect * r)
{
	return r->in ? ::incref (r->in) : PyNone();
}

static PyGetSetDef
GraphConnect_getsets [] = 
{
	{"out", (getter) GraphConnect_get_out, (setter) 0,
	 "the output connection point."},
	{"in", (getter) GraphConnect_get_in, (setter) 0,
	 "the input connection point."},
	{ }
};

void
Py<GraphConnect>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphConnect>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphConnect_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphStoreTempo_type = 0;

char * Py<GraphStoreTempo>::__doc__ = 
"GraphStoreTempo() -- a request to a graph to store a change of tempo.";

static PyObject *
GraphStoreTempo_get_bpm (GraphStoreTempo * r)
{
	return PyFloat (r->bpm);
}

static PyObject *
GraphStoreTempo_get_tick (GraphStoreTempo * r)
{
	return PyFloat (r->tick);
}

static PyGetSetDef
GraphStoreTempo_getsets [] = 
{
	{"bpm", (getter) GraphStoreTempo_get_bpm, (setter) 0,
	 "the tempo in beats per minute."},
	{"tick", (getter) GraphStoreTempo_get_tick, (setter) 0,
	 "the tick at which to change tempo."},
	{ }
};

void
Py<GraphStoreTempo>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphStoreTempo>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphStoreTempo_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphRemoveTempo_type = 0;

char * Py<GraphRemoveTempo>::__doc__ = 
"GraphRemoveTempo() -- a request to a graph to remove a change of tempo.";

static PyObject *
GraphRemoveTempo_get_n (GraphRemoveTempo * r)
{
	return PyFloat (r->n);
}

static PyObject *
GraphRemoveTempo_get_tick (GraphRemoveTempo * r)
{
	return PyFloat (r->tick);
}

static PyGetSetDef
GraphRemoveTempo_getsets [] = 
{
	{"tick", (getter) GraphRemoveTempo_get_tick, (setter) 0,
	 "the tick at which to remove the tempo change(s)."},
	{"n", (getter) GraphRemoveTempo_get_n, (setter) 0,
	 "the number of tempo changes to remove."},
	{ }
};

void
Py<GraphRemoveTempo>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphRemoveTempo>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphRemoveTempo_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphSeek_type = 0;

char * Py<GraphSeek>::__doc__ = 
"GraphSeek() -- a request to a graph to change the current position in time.";

static PyObject *
GraphSeek_get_time (GraphSeek * r)
{
	return PyFloat (r->time);
}

static PyGetSetDef
GraphSeek_getsets [] = 
{
	{"time", (getter) GraphSeek_get_time, (setter) 0,
	 "the position to move to."},
	{ }
};

void
Py<GraphSeek>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphSeek>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphSeek_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphSetQuarter_type = 0;

char * Py<GraphSetQuarter>::__doc__ = 
"GraphSetQuarter() -- a request to a graph to change the number of MIDI "
"ticks per quarter note.";

static PyObject *
GraphSetQuarter_get_quarter (GraphSetQuarter * r)
{
	return PyFloat (r->quarter);
}

static PyGetSetDef
GraphSetQuarter_getsets [] = 
{
	{"quarter", (getter) GraphSetQuarter_get_quarter, (setter) 0,
	 "the new number of ticks per quarter note."},
	{ }
};

void
Py<GraphSetQuarter>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphSetQuarter>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphSetQuarter_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphSetLoop_type = 0;

char * Py<GraphSetLoop>::__doc__ = 
"GraphSetLoop() -- a request to a graph to set the loop points.";

static PyObject *
GraphSetLoop_get_t0 (GraphSetLoop * r)
{
	return PyFloat (r->t0);
}

static PyObject *
GraphSetLoop_get_t1 (GraphSetLoop * r)
{
	return PyFloat (r->t1);
}

static PyGetSetDef
GraphSetLoop_getsets [] = 
{
	{"t0", (getter) GraphSetLoop_get_t0, (setter) 0,
	 "the loop start point."},
	{"t1", (getter) GraphSetLoop_get_t1, (setter) 0,
	 "the loop start point."},
	{ }
};

void
Py<GraphSetLoop>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphSetLoop>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphSetLoop_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GraphSetUseLoop_type = 0;

char * Py<GraphSetUseLoop>::__doc__ = 
"GraphSetUseLoop() -- a request to a graph to en-/disable the loop points.";

static PyObject *
GraphSetUseLoop_get_use (GraphSetUseLoop * r)
{
	return PyInt (r->use);
}

static PyGetSetDef
GraphSetUseLoop_getsets [] = 
{
	{"use", (getter) GraphSetUseLoop_get_use, (setter) 0,
	 "whether to use the loop points."},
	{ }
};

void
Py<GraphSetUseLoop>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GraphSetUseLoop>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = GraphRequest_type;
	type->tp_getset = GraphSetUseLoop_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */



/* /////////////////////////////////////////////////////////////////////// */

/* getsets */
 
static PyObject * 
Graph_get_plugins (Graph * g)
{
	ScopedLock lock (g->mutex);

	return PyTuple (&g->plugins);
}

static PyObject * 
Graph_get_tempo_map (Graph * g)
{
	ScopedLock lock (g->mutex);

	return PyTuple (&g->tempo_map.events);
}

static PyObject *
Graph_get_time_map (Graph * g)
{
	return g->time_map ? ::incref (g->time_map) : PyNone();
}

static int
Graph_set_time_map (Graph * g, TimeMap * map)
Py_TRY
{
	if (!TimeMap_type->is_derived (map))
		throw TypeError() << "need a TimeMap.";

	::decref (g->time_map, true);
	g->time_map = map;

	return 0;
}
Py_CATCH (-1);

static PyObject * 
Graph_get_state (Graph * g)
{
	return PyInt (g->state);
}

static int 
Graph_set_state (Graph * g, PyObject * value) 
Py_TRY
{
	int state;
	
	if (!PyArg_Parse (value, "i", &state))
		return -1;
	if (state < 0 || state > 3) 
		throw ValueError() << "state must be in (0, 1, 2, 3).";

	/* quietly run through all intermediate states. */
	while (g->state < state)
		exec_request (new GraphStateChange (g, g->state + 1));
	if (g->state > state)
		exec_request (new GraphStateChange (g, state));
	return 0;
} 
Py_CATCH (-1);

static PyObject * Graph_get_sample_rate (Graph * g)
{
	return PyInt (g->sample_rate);
}

static PyObject * Graph_get_frames_per_cycle (Graph * g)
{
	return PyInt (g->frames_per_cycle);
}

static PyObject * Graph_get_frame (Graph * g)
{
	return PyFrame (g->frame);
}

static PyObject * Graph_get_time (Graph * g)
{
	return PyFloat (g->time);
}

static PyObject * Graph_get_tick (Graph * g)
{
	return PyFloat (g->tick);
}

static PyObject * Graph_get_virtual_time (Graph * g)
{
	return PyFloat (g->virtual_time);
}

static PyObject * Graph_get_virtual_tick (Graph * g)
{
	return PyFloat (g->virtual_tick);
}

static PyObject * 
Graph_get_cpu (Graph * g)
{
	d_time cpu = g->cpu / g->audio_cycle_time;
		
	g->cpu = 0;

	return PyFloat (cpu);
}

/* set_{tick,time,frame} == seek. */
static inline int
Graph_set_seek (Graph * g, d_time t)
Py_TRY
{
	exec_request (new GraphSeek (g, t));
	return 0;
} 
Py_CATCH (-1);

static int
Graph_set_frame (Graph * g, PyObject * frame)
{
	d_frame f;
	
	if (!PyArg_Parse (frame, "L", &f))
		return -1;
	
	return Graph_set_seek (g, g->frame2time (f));
}

static int
Graph_set_time (Graph * g, PyObject * time)
{
	d_time t;
	
	if (!PyArg_Parse (time, "d", &t))
		return -1;
	
	return Graph_set_seek (g, t);
}

static int
Graph_set_tick (Graph * g, PyObject * tick)
{
	d_tick t;
	
	if (!PyArg_Parse (tick, "d", &t))
		return -1;
	
	return Graph_set_seek (g, g->tick2time (t));
}

/* more getsets. */

static PyObject * Graph_get_quarter (Graph * g)
{
	return PyInt ((int) g->quarter);
}

static int 
Graph_set_quarter (Graph * g, PyObject * value) 
Py_TRY
{
	int quarter;
	
	if (!PyArg_Parse (value, "i", &quarter))
		return -1;

	if (quarter < 24) 
		throw ValueError() << "quarter must be >= 24.";

	if (quarter % 3 || quarter % 4)
		msg(), "quarter value should be a multiple of both 3 and 4.";

	exec_request (new GraphSetQuarter (g, quarter));
	return 0;
} 
Py_CATCH (-1);

/* loop */
static PyObject *
Graph_get_loop (Graph * g)
{
	return Py_BuildValue ("dd", g->loop.tick[0], g->loop.tick[1]);
}

static int 
Graph_set_loop (Graph * g, PyObject * value) 
Py_TRY
{
	d_tick t0, t1;
	
	if (!PyArg_ParseTuple (value, "dd", &t0, &t1))
		return -1;

	/* swap if needed. */
	if (t1 < t0)
		swap (t0, t1);

	if (t1 - t0 < 2 * g->quarter) 
		throw ValueError() 
				<< "loop points must be at least two quarter notes apart.";

	exec_request (new GraphSetLoop (g, t0, t1));
	return 0;
} 
Py_CATCH (-1);

static PyObject *
Graph_get_use_loop (Graph * g)
{
	return PyInt (g->loop.use);
}

static int 
Graph_set_use_loop (Graph * g, PyObject * value) 
Py_TRY
{
	int use;
	
	if (!PyArg_Parse (value, "i", &use))
		return -1;

	if (use == g->loop.use)
		return 0;

	if (use && g->loop.tick[1] - g->loop.tick[0] < 2 * g->quarter)
		throw ValueError() << "need valid loop points first.";

	exec_request (new GraphSetUseLoop (g, use));
	return 0;
} 
Py_CATCH (-1);

static PyGetSetDef 
Graph_getsets [] = 
{
	{"plugins", (getter) Graph_get_plugins, (setter) 0, 
	 "a tuple containing all plugins registered with this graph, sorted by "
	 "processing order."},
	{"tempo_map", (getter) Graph_get_tempo_map, (setter) 0, 
	 "a tuple containing all events from the graph's tempo map."},
	{"time_map", (getter) Graph_get_time_map, (setter) Graph_set_time_map, 
	 "the TimeMap."},
	{"state", (getter) Graph_get_state, (setter) Graph_set_state, 
	 "The state of the Graph can be one of the following:\n"
	 "0 -- all clocks are stopped. No processing of any kind takes place.\n"
	 "1 -- everything is prepared for rolling, but otherwise the same as 0.\n"
	 "2 -- processing takes place but no playback or recording, clocks are not advanced.\n"
	 "3 -- clocks are running, and all processing is enabled.\n"
	 "These states correspond directly to the four constants defined in Plugin."},
	{"sample_rate", 
	 (getter) Graph_get_sample_rate, (setter) 0,
	 "the sample rate to use. Usually set by the audio clock in the graph."},
	{"frames_per_cycle", 
	 (getter) Graph_get_frames_per_cycle, (setter) 0,
	 "the number of audio "
	 "frames to process per cycle. "
	 "Usually set by the audio clock in the graph."},
	{"cpu",
	 (getter) Graph_get_cpu, (setter) 0,
	 "maximum relative amount of cpu usage for handling an audio cycle."},
	{"frame",
	 (getter) Graph_get_frame, (setter) Graph_set_frame,
	 "the current audio frame."},
	{"time",
	 (getter) Graph_get_time, (setter) Graph_set_time,
	 "the current song time."},
	{"tick",
	 (getter) Graph_get_tick, (setter) Graph_set_tick,
	 "the current tick."},
	{"virtual_time",
	 (getter) Graph_get_virtual_time, (setter) 0,
	 "the current virtual song time (counted when state is Activated)."},
	{"virtual_tick",
	 (getter) Graph_get_virtual_tick, (setter) 0,
	 "the current virtual tick (counted when state is Activated)."},
	{"quarter",
	 (getter) Graph_get_quarter, (setter) Graph_set_quarter,
	 "the (integer) number of ticks per quarter note."},
	{"loop",
	 (getter) Graph_get_loop, (setter) Graph_set_loop,
	 "a tuple of the start and end points of the loop, in ticks."},
	{"use_loop",
	 (getter) Graph_get_use_loop, (setter) Graph_set_use_loop,
	 "whether the loop mechanism is active or not."},
	{ }
};

/* methods */

static PyObject * 
Graph_add_plugin (Graph * g, PyObject * args)
Py_TRY
{
	Plugin * p = 0;
	d_float x = NAN, y = NAN;
	
	if (!PyArg_ParseTuple (args, "O!|dd", Plugin_type, &p, &x, &y))
		return 0;

	if (p->graph)
		throw AssertionError() << p << " is already part of a graph.";
	
	if (finite (x) && finite (y))
		p->x = x,
		p->y = y;
	else if (finite (x))
		p->y = x;

	exec_request (new GraphAddPlugin (g, p));
	return ::incref (p);
} 
Py_CATCH (0);

static PyObject *
Graph_remove_plugin (Graph * g, PyObject * args, PyObject * kwds) 
Py_TRY
{
	static char * keys [] = {"plugin", 0};

	Plugin * p = 0;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!", keys,
	    Plugin_type, &p))
		return 0;

	exec_request (new GraphRemovePlugin (g, p));
	return PyNone();
} 
Py_CATCH (0);

static PyObject * 
Graph_move_plugin (Graph * g, PyObject * args, PyObject * kwds) 
Py_TRY
{
	static char * keys [] = {"plugin", "x", "y", 0};

	Plugin * p = 0;
	d_float x = NAN, y = NAN;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!dd", keys,
	    Plugin_type, &p, &x, &y))
		return 0;

	if (p->graph != g)
		throw AssertionError() << p << " is not part of this graph.";

	exec_request (new GraphMovePlugin (g, p, x, y));
	return PyNone();
} 
Py_CATCH (0);

static PyObject *
Graph_connect (Graph * g, PyObject * args, PyObject * kwds) 
Py_TRY
{
	static char * keys [] = {"out", "in", 0};

	Out * out;
	In * in;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!O!", keys,
	    Out_type, &out, In_type, &in))
		return 0;

	exec_request (new GraphConnect (g, out, in));
	return PyNone();
} 
Py_CATCH (0);

static PyObject *
Graph_disconnect (Graph * g, PyObject * args, PyObject * kwds) 
Py_TRY
{
	static char * keys [] = {"out", 0};

	Out * out;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!", keys,
	    Out_type, &out))
		return 0;

	exec_request (new GraphConnect (g, out, 0));
	return PyNone();
} 
Py_CATCH (0);

static PyObject * 
Graph_store_tempo (Graph * g, PyObject * args, PyObject * kwds) 
Py_TRY
{
	static char * keys [] = {"bpm", "tick", 0};

	d_float bpm;
	d_tick tick;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "dd", keys,
	    &bpm, &tick))
		return 0;

	exec_request (new GraphStoreTempo (g, bpm, tick));
	return PyNone();
} 
Py_CATCH (0);

static PyObject *
Graph_remove_tempo (Graph * g, PyObject * args, PyObject * kwds) 
Py_TRY
{
	static char * keys [] = {"tick", "n", 0};

	d_tick tick = 0;
	int n = 1;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "di", keys,
	    &tick, &n))
		return 0;

	exec_request (new GraphRemoveTempo (g, tick, n));
	return PyNone();
} 
Py_CATCH (0);

static PyObject *
Graph_tick2time (Graph * g, PyObject * arg) 
{
	d_tick tick = 0;

	if (!PyArg_Parse (arg, "d", &tick))
		return 0;

	return PyFloat (g->tick2time (tick));
} 

static PyObject *
Graph_time2tick (Graph * g, PyObject * arg) 
{
	d_time time;

	if (!PyArg_Parse (arg, "d", &time))
		return 0;

	return PyFloat (g->time2tick (time));
} 

static PyObject *
Graph_frame2time (Graph * g, PyObject * arg) 
{
	d_frame frame = 0;

	if (!PyArg_Parse (arg, "L", &frame))
		return 0;

	return PyFloat (g->frame2time (frame));
} 

static PyObject *
Graph_time2frame (Graph * g, PyObject * arg) 
{
	d_time time;

	if (!PyArg_Parse (arg, "d", &time))
		return 0;

	return PyFrame (g->time2frame (time));
} 

static PyObject *
Graph_frame2tick (Graph * g, PyObject * arg) 
{
	d_frame frame = 0;

	if (!PyArg_Parse (arg, "L", &frame))
		return 0;

	return PyFloat (g->frame2tick (frame));
} 

static PyObject *
Graph_tick2frame (Graph * g, PyObject * arg) 
{
	d_tick tick;

	if (!PyArg_Parse (arg, "d", &tick))
		return 0;

	return PyFrame (g->tick2frame (tick));
} 

static PyObject *
Graph_time_as_string (Graph * g)
{
	d_time time = g->time;
	int s = (int) time;
	int c = (int) ((time - s) * 100);
	int h = s / 3600;
	s -= h * 3600;
	int m = s / 60;
	s -= m * 60;
	
	char hmsc [16];
	sprintf (hmsc, "%2d:%02d:%02d.%02d", h, m, s, c);

	return PyString (hmsc);
}

static PyMethodDef 
Graph_methods [] = 
{
  {"add", (PyCFunction) Graph_add_plugin, METH_VARARGS,
	 "add (plugin[, y]|[, x, y]) -- add 'plugin' to the graph "
	 "at the position denoted by ('x', 'y'). returns the plugin for "
	 "convenience."},
  {"remove", (PyCFunction) Graph_remove_plugin, METH_VARARGS | METH_KEYWORDS,
	 "remove (plugin) -- remove 'plugin' from the graph."},
  {"move", (PyCFunction) Graph_move_plugin, METH_VARARGS | METH_KEYWORDS,
	 "move (plugin, x, y) -- move 'plugin' to ('x', 'y'). Because this "
	 "operation changes the processing order, it may produce unexpected "
	 "silence."},
  {"connect", (PyCFunction) Graph_connect, METH_VARARGS | METH_KEYWORDS,
	 "connect (out, in) -- connect a plugin output to a plugin input."},
  {"disconnect", (PyCFunction) Graph_disconnect, METH_VARARGS | METH_KEYWORDS,
	 "disconnect (out) -- disconnect a plugin output."},
  {"store_tempo", (PyCFunction) Graph_store_tempo, METH_VARARGS | METH_KEYWORDS,
	 "store_tempo (bpm, tick) -- store a change of tempo to 'bpm' at 'tick'. "},
  {"remove_tempo", (PyCFunction) Graph_remove_tempo, METH_VARARGS | METH_KEYWORDS,
	 "remove_tempo (tick[, n = 1]) -- remove 'n' tempo changes starting at "
	 "'tick'."},
  {"tick2time", (PyCFunction) Graph_tick2time, METH_O,
	 "tick2time (tick) -- translate a tick into time (in seconds)."},
  {"frame2time", (PyCFunction) Graph_frame2time, METH_O,
	 "frame2time (frame) -- translate an audio frame position into time (in seconds)."},
	{"time2tick", (PyCFunction) Graph_time2tick, METH_O,
	 "time2tick (time) -- translate 'time' into ticks."},
	{"time2frame", (PyCFunction) Graph_time2frame, METH_O,
	 "time2frame (time) -- translate 'time' into an audio frame position."},
  {"tick2frame", (PyCFunction) Graph_tick2frame, METH_O,
	 "tick2frame (tick) -- translate a tick into a frame position."},
	{"frame2tick", (PyCFunction) Graph_frame2tick, METH_O,
	 "frame2tick (frame) -- translate 'frame' into ticks."},
	{"time_as_string", (PyCFunction) Graph_time_as_string, METH_NOARGS,
	 "time_as_string() -- return the current position as 'h:mm:ss.cc'."},
	{ }
};

void
Py<Graph>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = NamedObject_type;
	t->tp_getset = Graph_getsets;
	t->tp_methods = Graph_methods;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
}

Type * Graph_type = 0;

char * Py<Graph>::__doc__ = 
"Graph (quarter = 192) "
"-- central to all signal processing. Maintains a "
"list of plugins. "
"It also keeps account of the passage of time in three ways: "
"ticks (musical time), seconds ('real time') and audio frames (samples).";

void
Py<Graph>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"quarter", 0};
	d_tick q = QUARTER;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|d", keys, &q))
		throw blewit();
	
	if (q < 1)
		throw ValueError() << "invalid quarter value " << q;
	quarter = q;
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject * 
Plugin_get_state (Plugin * p)
{
	return PyInt (p->state);
}

static PyObject * 
Plugin_get_flags (Plugin * p)
{
	return PyInt (p->flags);
}

/* */

static PyObject * 
Plugin_get_x (Plugin * p)
{
	return PyFloat (p->x);
}

static PyObject * 
Plugin_get_y (Plugin * p)
{
	return PyFloat (p->y);
}

/* get/set x and y should be done via a GraphMovePlugin request if the plugin
 * is part of a graph. */
static int
Plugin_set_x (Plugin * p, PyObject * value)
Py_TRY
{
	d_float x;
	
	if (!PyArg_Parse (value, "d", &x))
		return -1;
	
	if (p->graph == 0)
		p->x = x;
	else
		exec_request (new GraphMovePlugin (p->graph, p, x, p->y));

	return 0;
} 
Py_CATCH (-1);

static int
Plugin_set_y (Plugin * p, PyObject * value)
Py_TRY
{
	d_float y;
	
	if (!PyArg_Parse (value, "d", &y))
		return -1;
	
	if (p->graph == 0)
		p->y = y;
	else
		exec_request (new GraphMovePlugin (p->graph, p, p->x, y));

	return 0;
} 
Py_CATCH (-1);

static PyObject * 
Plugin_get_ins (Plugin * p)
{
	/* todo: this is cheap; the List object should be used directly instead. */
	return PyTuple (&p->ins);
}

static PyObject * 
Plugin_get_outs (Plugin * p)
{
	/* todo: this is cheap; the List object should be used directly instead. */
	return PyTuple (&p->outs);
}

static PyObject * 
Plugin_get_graph (Plugin * p)
{
	return p->graph ? ::incref (p->graph) : PyNone();
}

static PyObject * 
Plugin_get_cpu (Plugin * p)
{
	if (p->graph == 0 || p->state < Plugin::Activated)
		return PyFloat (p->cpu);

	d_time cpu = p->cpu / p->graph->audio_cycle_time;
		
	p->cpu = 0;

	return PyFloat (cpu);
}

static PyObject * 
Plugin_get_loop_cycle (Plugin * p)
{
	return PyInt (p->loop_cycle);
}

static int
Plugin_set_loop_cycle (Plugin * p, PyObject * value)
Py_TRY
{
	#if 0 
	if (!p->is_script())
		throw TypeError() << "can only do that with ScriptPlugins.";
	#endif

	if (!PyArg_Parse (value, "i", &p->loop_cycle))
		return -1;
	
	return 0;
}
Py_CATCH (-1);

static PyGetSetDef 
Plugin_getsets [] = 
{
	{"state", (getter) Plugin_get_state, (setter) 0, 
	 "Off = 0, Prepared = 1, Activated = 2, Rolling = 3."},
	{"flags", (getter) Plugin_get_flags, (setter) 0, 
	 "Beat = 1, Audio = 2, Time = 4, Disk = 8, Script = 16,\n"
	 "the processing cycles the plugin takes part in."},
	{"cpu", (getter) Plugin_get_cpu, 0, 
	 "The maximum time the plugin spent in an audio cycle so far, expressed "
	 "as a fraction of the time available for a complete audio cycle. "
	 "The value is reset to zero upon reading.\n"
	 "Note that this figure can only serve as an indicator if the audio "
	 "cycle is not run at system-wide top priority."},
	{"x", (getter) Plugin_get_x, (setter) Plugin_set_x, 
	 "the 'x' coordinate, see the documentation on 'y' for what it means."},
	{"y", (getter) Plugin_get_y, (setter) Plugin_set_y, 
	 "the higher the 'y' coordinate, the later a plugin is processed in the "
	 "various cycles. If two plugins are at the same 'y', 'x' decides over "
	 "the resulting order."},
	{"ins", (getter) Plugin_get_ins, 0, "all input connection points."},
	{"outs", (getter) Plugin_get_outs, 0, "all output connection points."},
	{"graph", (getter) Plugin_get_graph, 0, "the graph the plugin resides in."},
	{"loop_cycle", 
	 (getter) Plugin_get_loop_cycle, (setter) Plugin_set_loop_cycle, 
	 "the sequential number of the loop cycle the plugin is in."},
	{0},
};

/* Plugin as mapping: strings are compared to {In,Out}->object_name */
static int
Plugin_length (Plugin * p)
{
	return p->ins + p->outs;
}

static PyObject *
Plugin_subscript (Plugin * p, PyObject * key)
Py_TRY
{
	char * s = 0;

	if (!PyArg_Parse (key, "s", &s))
		return 0;
	
	/* is it an In? */
	for (int i = 0; i < p->ins; ++i)
		if (!strcmp (get_name (p->ins[i]), s))
			return ::incref (p->ins[i]);

	/* is it an Out? */
	for (int i = 0; i < p->outs; ++i)
		if (!strcmp (get_name (p->outs[i]), s))
			return ::incref (p->outs[i]);
	
	throw KeyError() << p << " has no port named '" << s << "'.";
}
Py_CATCH (0);

static PyMappingMethods 
Plugin_as_mapping = 
	{
		(inquiry) Plugin_length, /* mp_length */
		(binaryfunc) Plugin_subscript, /* mp_subscript */
		(objobjargproc) 0, /* mp_ass_subscript */
	};

void
Py<Plugin>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = NamedObject_type;
	t->tp_getset = Plugin_getsets;
	t->tp_as_mapping = &Plugin_as_mapping;

	PyType_Ready (t);

	PyDict_SetItemString (t->tp_dict, "Off", PyInt (Plugin::Off));
	PyDict_SetItemString (t->tp_dict, "Prepared", PyInt (Plugin::Prepared));
	PyDict_SetItemString (t->tp_dict, "Activated", PyInt (Plugin::Activated));
	PyDict_SetItemString (t->tp_dict, "Rolling", PyInt (Plugin::Rolling));
}

Type * Plugin_type = 0;

char * Py<Plugin>::__doc__ = 
"Plugin() -- abstract base type to all functional units in the processing "
"graph. "
"Plugins can have any number of input and output "
"connection points (In and Out respectively) through which they communicate "
"signal data. These ports are listed in the 'ins' and 'outs' tuples.\n"
"Most plugins will not "
"work correctly, or even not at all, if the graph they are in  "
"does not also contain the respective scheduler or clock plugin, "
"for example a DiskStream will not work without a DiskScheduler."
"\nYou cannot instantiate this in Python; use a ScriptPlugin instead.";

void
Py<Plugin>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject * In_get_type (In * in)
{
	return ::incref (in->type);
}

static PyObject * In_get_plugin (In * in)
{
	return ::incref (in->plugin);
}

static PyObject * In_get_n_slots (In * in)
{
	return PyInt (in->n_slots);
}

static PyObject * In_get_connected (In * in)
{
	PyObject * l = PyList_New (0);

	for (int i = 0; i < in->n_slots; ++i)
	{
		Out * o = in->slots[i];
		if (o)
			PyList_Append (l, o);
	}

	return l;
}

static PyGetSetDef 
In_getsets [] = 
{
	{"type", (getter) In_get_type, 0, 
	 "the type of events accepted."},
	{"plugin", (getter) In_get_plugin, 0, 
	 "the Plugin the port is part of."},
	{"n_slots", (getter) In_get_n_slots, 0, 
	 "the max. number of connections supported."},
	{"connected", (getter) In_get_connected, 0, 
	 "a tuple of all Out ports connected to this port."},
	{0},
};

static PyObject *
In_peek (In * in) 
Py_TRY
{
	if (in->plugin && !(in->plugin->flags & Plugin::Script))
		throw TypeError() << "you can only do this on ScriptPlugins.";
		
	Object * e = in->peek();

	return e ? ::incref (e) : PyNone();
} 
Py_CATCH (0);

static PyObject *
In_pop (In * in) 
Py_TRY
{
	if (in->plugin && !(in->plugin->flags & Plugin::Script))
		throw TypeError() << "you can only do this on ScriptPlugins.";
		
	Object * e = in->pop();

	/* no ::incref (e) -- the semantics of an objectfifo demand that we 
	 * inherit the reference to 'e'. */
	return e ? e : PyNone();
} 
Py_CATCH (0);

static PyMethodDef 
In_methods [] = 
{
  {"peek", (PyCFunction) In_peek, METH_NOARGS,
	 "peek() -- return the next event at this input point, without removing "
	 "it from the queue."},
  {"pop", (PyCFunction) In_pop, METH_NOARGS,
	 "pop() -- return the next event at this input point, and remove "
	 "it from the queue."},
	{}
};

void
Py<In>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = NamedObject_type;
	t->tp_getset = In_getsets;
	t->tp_methods = In_methods;
}

Type * In_type = 0;

char * Py<In>::__doc__ = 
"In() -- an input connection point on a plugin, used to receive signal data."
"\nYou cannot instantiate this in Python; "
"only ScriptPlugins have a new_in() method.";

void
Py<In>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject * 
Out_get_plugin (Out * out)
{
	return out->plugin ? ::incref (out->plugin) : PyNone();
}

static PyObject * 
Out_get_type (Out * out)
{
	return ::incref (out->type);
}

static PyObject * 
Out_get_cache (Out * out)
{
	return ::incref (out->cache);
}

static PyObject * 
Out_get_connected (Out * out)
{
	return out->connected ? ::incref (out->connected) : PyNone();
}

static PyGetSetDef 
Out_getsets [] = 
{
	{"plugin", (getter) Out_get_plugin, 0, 
	 "The Plugin this belongs to."},
	{"type", (getter) Out_get_type, 0, 
	 "The type of events transmitted."},
	{"cache", (getter) Out_get_cache, 0, 
	 "An EventCache storing events of the type transmitted."},
	{"connected", (getter) Out_get_connected, 0, 
	 "The In that receives the signal transmitted, or None."},
	{ },
};

static PyObject *
Out_connect (Out * self, PyObject * _in) 
Py_TRY
{
	In * in;
	if (_in == Py_None)
		in = 0;
	else if (!In_type->is_derived (_in))
		throw TypeError() << "an In connection point is required.";
	else
		in = (In *) _in;

	Graph * g = 0;
	if (self->plugin && self->plugin->graph)
		g = self->plugin->graph;
	else if (in->plugin && in->plugin->graph)
		g = in->plugin->graph;

	if (g)
		exec_request (new GraphConnect (g, self, in));
	else
	{
		ScopedLock lock (self->mutex);
		self->connect (in);
	}

	return PyNone();
} 
Py_CATCH (0);

static PyObject *
Out_push (Out * out, PyObject * arg) 
Py_TRY
{
	if (out->plugin && !ScriptPlugin_type->is_derived (out->plugin))
		throw TypeError() << "you can only do this on ScriptPlugins.";
		
	if (!out->type->is_derived (arg))
		throw TypeError() << "unsupported object type for pushing.";

	ScopedLock lock (out->mutex);

	out->push (::incref (arg));
	return PyNone();
} 
Py_CATCH (0);

/* ObjectFifo::pop needs an overload in Out that prevents Python code from 
 * accessing the reader side of the fifo. 
 */
static PyObject *
Out_pop_throw (ObjectFifo *)
{
	return PyError (PyExc_NotImplementedError, "you can't do that.");
}

static PyMethodDef 
Out_methods [] = 
{
  {"connect", (PyCFunction) Out_connect, METH_O,
	 "connect (in) -- connect to an In (an input port on a plugin). Connecting "
	 "to None disconnects. The connection can not be established if the In "
	 "does not dig the type of events the output produces, or if an audio "
	 "connection is asked for that would not work with the processing order "
	 "of the plugins to be connected."},
  {"push", (PyCFunction) Out_push, METH_O,
	 "push (event) -- transmit 'event' through this output point."},
  {"pop", (PyCFunction) Out_pop_throw, METH_NOARGS,
	 "pop() -- throws an exception."},
	{ }
};

void
Py<Out>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = ObjectFifo_type;
	t->tp_getset = Out_getsets;
	t->tp_methods = Out_methods;
}

Type * Out_type = 0;

char * Py<Out>::__doc__ = 
"Out() -- a connection point on a plugin that transmits events.\n"
"You cannot instantiate this in Python; "
"only ScriptPlugins have a new_out() method.";

void
Py<Out>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
EventCache_push (EventCache *, PyObject *)
{
	return PyError (PyExc_NotImplementedError, "not on a cache.");
}

static PyObject *
EventCache_pop (EventCache * cache)
Py_TRY
{
	Plugin * p = cache->out->plugin;

	if (!ScriptPlugin_type->is_derived (p))
		throw TypeError() << "you can only do this on ScriptPlugins.";

	/* Python now holds the only reference to the event. */
	return cache->pop();
} 
Py_CATCH (0);

static PyMethodDef 
EventCache_methods [] = 
{
  {"push", (PyCFunction) EventCache_push, METH_O,
	 "push (event) -- not implemented."},
  {"pop", (PyCFunction) EventCache_pop, METH_NOARGS,
	 "pop() -- fetch an event from the cache. The state of the event is "
	 "undefined."},
	{ }
};

void
Py<EventCache>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = ObjectFifo_type;
	t->tp_methods = EventCache_methods;
}

Type * EventCache_type = 0;

char * Py<EventCache>::__doc__ = 
"EventCache() -- a cache holding preallocated Events. Every 'Out' has "
"a cache holding events of the type the output is registered with. "
"Events fetched from an event cache should be fed to the 'Out' the cache "
"belongs to. For every event you want to transmit, you should fetch a "
"new event from the cache, ie. don't push the same event twice.\n"
"NB: Failing to observe this can cause a core dump.\n"
"You cannot instantiate this in Python.";

void
Py<EventCache>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_Graph_types (PyObject * dict)
{
	In_type = new PyType<In> ("In", dict);
	Out_type = new PyType<Out> ("Out", dict);
	Plugin_type = new PyType<Plugin> ("Plugin", dict);
	EventCache_type = new PyType<EventCache> ("EventCache", dict);

	Graph_type = new PyType<Graph> ("Graph", dict);

	GraphRequest_type = new PyType<GraphRequest> ("GraphRequest", dict);
	GraphStateChange_type = new PyType<GraphStateChange> ("GraphStateChange", dict);
	GraphAddPlugin_type = new PyType<GraphAddPlugin> ("GraphAddPlugin", dict);
	GraphRemovePlugin_type = new PyType<GraphRemovePlugin> ("GraphRemovePlugin", dict);
	GraphMovePlugin_type = new PyType<GraphMovePlugin> ("GraphMovePlugin", dict);
	GraphConnect_type = new PyType<GraphConnect> ("GraphConnect", dict);
	GraphStoreTempo_type = new PyType<GraphStoreTempo> ("GraphStoreTempo", dict);
	GraphRemoveTempo_type = new PyType<GraphRemoveTempo> ("GraphRemoveTempo", dict);
	GraphSeek_type = new PyType<GraphSeek> ("GraphSeek", dict);
	GraphSetQuarter_type = new PyType<GraphSetQuarter> ("GraphSetQuarter", dict);
	GraphSetLoop_type = new PyType<GraphSetLoop> ("GraphSetLoop", dict);
	GraphSetUseLoop_type = new PyType<GraphSetUseLoop> ("GraphSetUseLoop", dict);
}

