/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _PYTYPE_H_
#define _PYTYPE_H_

#include "Py.h"

/* PyType<O> takes care of setting up a PyTypeObject for Py<O>. 
 *
 * things are a little complicated if the type is to serve as a basetype
 * for Python classes; because these types need to follow the Python 
 * GC protocol.
 *
 * Python/Objects/dictobject.c has provided some clarity on how to set up
 * basetypes for Python classes.
 */

template <class O>
class PyType
:	public Type
{
	public:
		void * operator new (size_t size)
			{
				return PyObject_MALLOC (size);
			}

		PyType (const char * name, PyObject * dict)
			: Type (name, sizeof (Py<O>))
			{
				tp_doc = Py<O>::__doc__;
				
				tp_flags = Py_TPFLAGS_DEFAULT;

				tp_getattro = PyObject_GenericGetAttr;
				tp_setattro = PyObject_GenericSetAttr;
						
				tp_new = (newfunc) _new;
				tp_init = (initproc) _init;

				tp_dealloc = (destructor) _dealloc;
				tp_free = (freefunc) _free;

				tp_print = (printfunc) _print;
				tp_repr = (reprfunc) _repr;

				/* fit the type with methods, getsets and inheritance information. */
				Py<O>::type_init (this, dict);

				if (tp_flags & Py_TPFLAGS_BASETYPE)
				{
					tp_flags |= Py_TPFLAGS_HAVE_GC;

					/* if type_init does not set the GC support methods, we provide 
					 * do-nothing default handlers, assuming that the type knows what
					 * it's doing.
					 */
					if (!tp_traverse)
						tp_traverse = (traverseproc) _traverse;

					if (!tp_clear)
						tp_clear = (inquiry) _clear;
				}

				/* finish and register. */
				if ((tp_flags & Py_TPFLAGS_READY) == 0)
					PyType_Ready (this);

				PyDict_SetItemString (dict, tp_name, (PyObject *) this);
			}
		
		~PyType()
			{
				/* never executed. this object is silently free()d by Python. */
			}

		/* default traverse handler does nothing. */
		static int _traverse (O * self, visitproc visit, void * arg)
			{
				return 0;
			}
		
		/* default clear handler needs to do nothing because _traverse doesn't
		 * give anything to the visitproc. 
		 */
		static int _clear (O * self)
			{
				return 0;
			}
		
		static void * _new (PyTypeObject * type, PyObject * args, PyObject * kwds)
			{
				O * self = (O *) PyType_GenericAlloc (type, 1);
				
				if (self == 0)
					return 0;

				try 
				{
					/* call the default constructor for an already allocated self. still 
					 * trying to catch the beauty of this line of code. */
					new (self) O(); 
					
					/* the constructor has reset the type: if we are instantiating
					 * a derived type, we have to override. */
					self->ob_type = type;

					return self;
				}
				catch (blewit & b)
				{
					b.throw_at_python();
				}

				return 0;
			}

		/* the class __init__ wrapper. */
		static int _init (Py<O> * o, PyObject * args, PyObject * kwds)
			{
				try 
				{
					o->__init__ (args, kwds);
					
					/* an exception thrown by a plugin from another module can not be
					 * handled, check if a Python exception has been set. */
					if (PyErr_Occurred())
						return -1;

					return 0;
				}
				catch (blewit & b) 
				{
					b.throw_at_python();
				}

				return -1;
			}

		static void _dealloc (O * o)
			{
				/* If 'o' is from a cache, take care of returning it. This happens
				 * if a ScriptPlugin pops events (from an In or an EventCache) but 
				 * does not push them further. They then go out of scope under Python 
				 * control and end up here.
				 * Because all this happens with the Python lock held and 
				 * PythonThread::housekeep will only do its work holding the Python
				 * lock, it is safe to write the object to its cache fifo.
				 */
				if (o->object_cache)
				{
					/* return_to_cache equivalent */
					o->ob_refcnt = 1;
					o->object_cache->push (o);
					return;
				}

				if (o->ob_type->tp_flags & Py_TPFLAGS_HAVE_GC)
					PyObject_GC_UnTrack (o);	

				/* Object::operator delete will check for tp_free and call it. */
				delete o;
			}
		
		/* called by Object::operator delete */
		static void _free (void * o)
			{
				if (((Object *) o)->ob_type->tp_flags & Py_TPFLAGS_HAVE_GC)
					PyObject_GC_Del (o);
				else /* TODO: check, is anyone using non-GC allocation at all? */
					PyObject_Del (o);
			}
		
		static int _print (O * o, FILE * f, int flags)
			{
				String s;
				*o >> s;

				_msg (f).operator, (s);
				return 0;
			}
			
		static PyObject * _repr (O * o)
			{
				String s;
				*o >> s;
				
				return PyString ((char *) s);
			}
};

#endif /* _PYTYPE_H_ */
