/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _TRASHFIFO_H_
#define _TRASHFIFO_H_

#include "Fifo.h"
#include "List.h"

class Mutex;

/* Each thread is fitted with a fifo that serves as intermediate storage 
 * for object references that are to be released. They are kept in this 
 * fifo until we can safely acquire the python lock ('PL') and do the 
 * actual reference release. 
 * 
 * This housekeeping function is called periodically by the "Trash" thread.
 *
 * Events that originate from a cache are returned to it. 
 *
 * NB: Of course, all this requires increment operations on object refcounts
 * to be atomic. Fortunately, they are on i386, in case they were not, all
 * we need do is change the Py_INCREF macro to treat obj->ob_rfcnt as
 * atomic and recompile python and its libs. Don't you just love open
 * source.
 *
 */
class PythonThread
{
	public:
		/* set to 0 when this thread exits */
		Fifo<Object *> fifo;

		atomic active;
		const char * name;

		/* local PyThreadState for acquiring the PL, and one for swapping. */
		PyThreadState * tstate, * tstate_h;
		
		PythonThread (const char * _name);
		~PythonThread();

		/* returns the PythonThread of the current thread. */
		static PythonThread * get();

		/* the fifo protocol. */
		int have()
			{
				return fifo.have();
			}

		void push (Object *);

		Object * pop()
			{
				return fifo.pop();
			}

	public:
		static PointerList<PythonThread> * all;
		static Mutex * mutex;

		/* the thread that has lock()ed and not yet unlock()ed, for tracing
		 * and debugging. */
		static PythonThread * executing;

		static pthread_key_t key;

		/* the initial interpreter state, used to create thread states. */
		static PyInterpreterState * interp;
		
		/* prepares for running Python code or using the Python allocator
		 * by acquiring the PL and setting 'tstate'. */
		static void lock();
		static void unlock();

		/* prints exception info to stdmsg. */
		static void trouble (Object * o);
		
		static void init();
		static void exit();

		static PythonThread * create (const char * name);
		static void deactivate();

		/* releases all object references held and destroys inactive fifos. 
		 * acquires the PL if told so. */
		static void housekeep (bool do_lock = 0);
};

/* /////////////////////////////////////////////////////////////////////// */

/* scoped handling of a python lock release/reacquire. */

class PythonLock
{
	public:
		PythonLock()
			{
				PythonThread::lock();
			}
		
		~PythonLock()
			{
				PythonThread::unlock();
			}
};

class PythonUnlock
{
	public:
		PyThreadState * tstate;
		
		PythonUnlock()
			{
				tstate = 0;

				/* safeguard, can be true in case SIGINT was sent during a lengthy
				 * operation, or in case our code has ignored python api error 
				 * returns (would we ever do that consciously?)
				 */
				if (PyErr_Occurred())
					throw blewit();

				tstate = PyEval_SaveThread();
			}
	
		~PythonUnlock()
			{
				if (tstate) lock();
			}

		void lock()
			{
				PyEval_RestoreThread (tstate);
				tstate = 0;
			}
};

#endif
