/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include <sys/ioctl.h>
#include <linux/rtc.h>
#include <sys/poll.h>

#include "basics.h"
#include "PyType.h"

#include "Thread.h"
#include "Graph.h"
#include "Pollable.h"
#include "Pipe.h"

#include "RTClock.h"

#define RTC_FREQUENCY	1024

/* poll_fd indices */
enum 
{
	Pulse = 0,
	RTC = 1
};

RTClock::RTClock()
: Plugin (RTClock_type, 0),
	Thread ("RTClock"),
	poll ("rtc"),
	pulse ("rtc pulse", this)
{
	frequency = RTC_FREQUENCY;

	poll.add (pulse);
	
	if ((rtc_fd = open ("/dev/rtc", O_RDONLY)) == -1)
		msg(), "open ('/dev/rtc'): ", strerror (errno);
	else
		poll.add (rtc_fd, POLLIN);

	set_object_name ("rt clock%s", rtc_fd == -1 ? "" : " with /dev/rtc");

	sentinel = 0;
	if (start_thread_rt (5))
		sentinel = new RTSentinel ("rt clock sentinel", 6);
}

RTClock::~RTClock()
{
	pulse.write_trigger (0);

	join_thread();

	if (sentinel)
		delete sentinel;
}

void
RTClock::set_frequency (int f)
{
	if (rtc_fd == -1)
		throw AssertionError() << "rtc is not available.";

	/* yes, we can do this from a different thread. */
	if (ioctl (rtc_fd, RTC_IRQP_SET, f) == -1)
		throw OSError (errno);

	frequency = f;
}

void
RTClock::wake_up()
{ 
	pulse.write_trigger (1);
}

void
RTClock::no_rtc()
{
	msg(), "rtc frequency ", frequency, "not available, doing without.";
	poll.remove_items (1, 1);
	close (rtc_fd);
	rtc_fd = -1;
}

void *
RTClock::thread()
{
	/* determine how many fds to poll */
	if (rtc_fd == -1)
		;
	else if (ioctl (rtc_fd, RTC_IRQP_SET, frequency) < 0)
		no_rtc();
	else if (ioctl (rtc_fd, RTC_PIE_ON) < 0)
		no_rtc();

	d_time then = gettimeofday();
	d_time last_stats = then;

	int cycles = 0;

	while (1) 
	{
		if (sentinel)
			sentinel->give_livesign();

		/* try 1 ms timeout, in case HZ >= frequency. */
		int r = poll.poll (1);
		
		if (r < 0)
			continue;
		
		if (poll[Pulse].revents & POLLIN) 
		{
			int trigger = pulse.read_trigger();
			
			if (trigger == 0)
				break;
		}

		if (rtc_fd && (poll[RTC].revents & POLLIN))
		{
			unsigned long irqs;

			read (rtc_fd, &irqs, sizeof (irqs));
			irqs = (irqs >> 8) - 1;

			if (irqs >= (unsigned long) frequency / 20)
				msg(), irqs, "irqs dropped.";
		}

		++cycles;

		d_time now = gettimeofday();
		d_time dt = now - then;

		then = now;
		
		if (state >= Activated && graph)
			graph->process_time (dt, -1);

		if (rtc_fd == -1)
			continue;

		/* the jitter calculation is incorrect if the rtc is not used, because
		 * it will fall back to ~ the kernel HZ, which is not known here.
		 * nothing tragic implied though.
		 */
		d_time j = fabs (dt - 1.0 / (d_time) frequency);
		if (j > jitter)
			jitter = j;

		if (0 && now - last_stats >= 15.0)
		{
			msg(), "jitter = ", jitter, "seconds max";
			last_stats = now;
		}
	}

	return 0;
}

/* //////////////////////////////////////////////////////////////////////// */

Type * RTClock_type = 0;

char * Py<RTClock>::__doc__ = 
"RTClock ([frequency = 1024]) -- a scheduler plugin that enables a "
"graph to maintain an account of musical and 'real' time. You should add this "
"to any graph that is not limited to purely processing audio.\n"
"If possible, the linux device file '/dev/rtc' is "
"used to generate a periodic pulse of up to 8192 Hz, but you need root "
"privileges to activate frequencies above the default 64 Hz with an unpatched "
"linux kernel. (Note that 2.4.x kernels allow setting the maximum f "
"available to non-privileged users via /proc/sys/dev/rtc/max-user-freq.) "
"No error is raised if the desired frequency is not available. "
"The actual resolution will never drop below the kernel "
"scheduler frequency, which on unpatched i386 production kernels before 2.6 "
"is 100 Hz.\n"
"If an audio clock is registered with the same graph, it will be acting as an "
"additional pulse source for this clock.";

static int 
RTClock_set_frequency (RTClock * c, PyObject * value) 
Py_TRY
{
	int f;
	
	if (!PyArg_Parse (value, "i", &f))
		return -1;

	c->set_frequency (f);
	return 0;
} 
Py_CATCH (-1);

static PyObject * 
RTClock_get_frequency (RTClock * c)
{
	return PyInt (c->frequency);
}

static int 
RTClock_set_priority (RTClock * c, PyObject * value) 
Py_TRY
{
	int p;
	
	if (!PyArg_Parse (value, "i", &p))
		return -1;

	c->set_thread_priority (p);
	return 0;
} 
Py_CATCH (-1);

static PyObject *
RTClock_get_priority (RTClock * c)
{
	return PyInt (c->thread_priority);
}

static PyObject *
RTClock_get_jitter (RTClock * c)
{
	d_time j = c->jitter;
	c->jitter = 0;
	return PyFloat (j);
}

static PyGetSetDef 
RTClock_getsets [] = 
{
	{"frequency", (getter) RTClock_get_frequency, (setter) RTClock_set_frequency, 
	 "the frequency at which the system real time clock is generating pulses, "
	 "should be a power of 2."},
	{"priority", (getter) RTClock_get_priority, (setter) RTClock_set_priority,
	 "the scheduling priority the clock thread enjoys."},
	{"jitter", (getter) RTClock_get_jitter, (setter) 0,
	 "the maximum time a pulse has been delayed. the value is reset to zero "
	 "upon reading."},
	{ }
};

void
Py<RTClock>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"frequency", 0};
	PyObject * f = 0;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|O", keys, &f))
		throw blewit();
	
	if (f)
		RTClock_set_frequency (this, f);
}

void
Py<RTClock>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = RTClock_getsets;
}

/* //////////////////////////////////////////////////////////////////////// */

void
init_RTClock_types (PyObject * dict)
{
	RTClock_type = new PyType<RTClock> ("RTClock", dict);
}
