/* Copyright 2003 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "SamplePlayer.h"
#include "Graph.h"
#include "GUSVoice.h"

#include "PyType.h"

#include "dsp/FPTruncateMode.h"
#include "dsp/Phase.h"

#define NOISE_FLOOR 0.00005 /* -86 dB */

static Type * SampledVoice_type;
static Type * SamplePlayer_type;

/* // SampledVoice ///////////////////////////////////////////////////////// */

SampledVoice::SampledVoice()
: Object (SampledVoice_type)
{
	sample = 0;
	loop.s0 = loop.s1 = 0;
	note = 0;
	range.n0 = range.n1 = 0;
	group = 0;
	v2f[0] = 0.;
	v2f[1] = 1.;
	pan = 0.;
	decay = 3.;
	release = .2;
	rnd2f = 0.;
	env2f = 0.;
	n2f[0] = 1.;
	n2f[1] = 1.;
	tune = 0.;
}

SampledVoice::~SampledVoice()
{
	::decref (sample, true);
}

void
SampledVoice::init (AudioSignal * s)
{
	::incref (sample = s);
	loop.s0 = loop.s1 = s->size;
}

/* // PlayingVoice /////////////////////////////////////////////////////// */

PlayingVoice::PlayingVoice()
{
	state = Off;
	voice = 0;

	/* all else set at note-on, so no need to initialize here */
}

PlayingVoice::~PlayingVoice()
{
	::decref (voice, true);
}

void
PlayingVoice::noteon (SampledVoice * _voice, float _fs, int _note, double v)
{
	if (state)
		off();

	fs = _fs;

	state = Attack;

	::incref (voice = _voice);
	note = _note;
	pitch = n2f (note + voice->tune) / n2f (voice->note);

	release = voice->release;

	/* sample index */
	phi = 0;

	/* envelope */
	gain.env = 1.;
	gain.factor = 1.;

	/* pan */
	double phi = (voice->pan + 1) * M_PI / 4;

	gain.left = (float) (v * cos (phi));
	gain.right = (float) (v * sin (phi));

	/* initialize filter */
	filter.reset();

	double f = 1 - voice->rnd2f * (double) random() / ((double) RAND_MAX);
	f *= voice->v2f[0] + v * (voice->v2f[1] - voice->v2f[0]);

	/* n2f */
	if (voice->range.n1 == voice->range.n0)
		f *= voice->n2f[0];
	else
		f *= voice->n2f[0] + (voice->n2f[1] - voice->n2f[0]) * 
				(double) note / ((double) (voice->range.n1 - voice->range.n0));

	filter.set (f);
	
	filter_factor = 1.;
}

inline void
PlayingVoice::enter_decay()
{
	state = Decay;

	gain.factor = pow (.5, 1. / (voice->decay * fs));
	filter_factor = 1. - voice->env2f * (1. - gain.factor);
}

inline void
PlayingVoice::enter_release()
{
	state = Release;

	gain.factor = pow (.5, 1. / (release * fs));
	filter_factor = 1. - voice->env2f * (1. - gain.factor);
}

void
PlayingVoice::play (CycleContext & cycle)
{
	int looped = voice->is_looped();

	register d_sample * src = voice->sample->data;

	register int written = 0;

	while (written < cycle.frames)
	{
		/* compute possible frames within this loop cycle */
		int can_play = (int) (phi.frames_until (voice->loop.s1) / pitch);

		/* at loop end, adjust things for rolling over it. */
		if (can_play <= 0)
		{
			if (!looped)
			{
				off();
				return;
			}

			if (state < Decay)
				enter_decay();

			phi.loop (voice->loop.s0, voice->loop.s1);

			can_play = (int) (phi.frames_until (voice->loop.s1) / pitch);
		}

		/* stop at loop end or cycle end? */
		int n = written + min (cycle.frames - written, can_play);

		while (written < n)
		{
			/* change to phi.get for linear interpolation */
			register d_sample x = phi.cubic_get (src);

			x = gain.env * filter.process (x);

			cycle.out.l[written] += cycle.gain * gain.left * x;
			cycle.out.r[written] += cycle.gain * gain.right * x;

			gain.env *= gain.factor;
			filter.decay (filter_factor);

			phi += pitch;

			++written;
		}

		if (gain.env * filter.a0 < NOISE_FLOOR)
		{
			off();
			return;
		}
	}
}

/* just play a one-shot sample at original pitch. no loops, no filters.
 */
void
PlayingVoice::play_simple (CycleContext & cycle)
{
	/* FIXME: stops prematurely. */
	int n = voice->sample->size - phi.frame;
	if (n > cycle.frames)
		n = cycle.frames;
	else
	{
		off();
		return;
	}

	d_sample * src = voice->sample->data + phi.frame;
	
	for (int i = 0; i < n; ++i)
	{
		register d_sample x = src[i];
		x = filter.process (x);

		cycle.out.l[i] += cycle.gain * gain.left * x;
		cycle.out.r[i] += cycle.gain * gain.right * x;
	}

	phi.frame += n;

	if (phi.frame >= voice->sample->size)
		off();
}

void
PlayingVoice::noteoff()
{
	if (state == Off) 
		return;
	enter_release();
}

void
PlayingVoice::muffle()
{
	/* drop to -6 dB in 1/10th second */
	release = .1;
	noteoff();
}

void
PlayingVoice::off()
{
	state = Off;
	::decref (voice);
	voice = 0;
}

double
PlayingVoice::get_weight()
{
	if (state == Off)
		return -1;

	double weight = gain.env * filter.a0;

	return weight;
}

/* SamplePlayer ////////////////////////////////////////////////////////// */

SamplePlayer::SamplePlayer()
: Plugin (SamplePlayer_type, Audio),
	voices ("sample player voices", 64),
	to_add (64, "sample player add voice requests"),
	to_remove (64, "sample player remove voice requests")
{
	/* -12 dB default attenuation. */
	gain = pow (10., -12. / 20.);
	transpose = 0;
	layered = 0;

	add_in (new In (this, MidiEvent_type, 8, "notes in"));
}

SamplePlayer::~SamplePlayer()
{
}

void
SamplePlayer::init (int _channels)
{
	channels = _channels;
	
	if (channels == 1)
		add_out (new Out (this, CyclicAudio_type, "out"));
	else
	{
		add_out (new Out (this, CyclicAudio_type, "out:l"));
		add_out (new Out (this, CyclicAudio_type, "out:r"));
	}
}

void
SamplePlayer::process_audio (int frames)
{
	CyclicAudio * l, * r;
	
	l = (CyclicAudio *) outs[0]->cache->pop();
	l->silence();
	
	if (channels == 1)
		r = l;
	else
	{
		r = (CyclicAudio *) outs[1]->cache->pop();
		r->silence();
	}

	d_frame f =
	l->frame = 
	r->frame = graph->frame;

	int written = 0;

	d_tick end_tick = graph->frame2tick (f + frames);
	
	CycleContext cycle;

	while (written < frames)
	{
		int n = frames - written;
		
		/* check when the next event is scheduled so we can process a 
		 * (hopefully large) continuous block until that point. */
		Event * e = ins[0]->pop_before (end_tick);
		d_tick tick;

		if (!e)
			tick = end_tick;
		else
		{
			tick = e->tick + 0.000001;
			d_frame ff = graph->tick2frame (e->tick);
			if (ff > f)
				n = (int) (ff - f);
		}

		if (n)
		{
			cycle.out.l = l->data + written;
			cycle.out.r = r->data + written;
			cycle.frames = n;
			cycle.pitch = 0.;
			cycle.gain = gain;

			one_cycle (cycle);
		}

		while (e) {
			parse_event ((MidiEvent *) e);
			e = ins[0]->pop_before (tick);
		}

		if (graph->is_loop_frame (f, f + n))
		{
			++loop_cycle;
			f -= graph->loop.frame[1] - graph->loop.frame[0];
			end_tick -= graph->loop.tick[1] - graph->loop.tick[0];
		}

		f += n;
		written += n;
	}

	outs[0]->push (l);
	if (channels == 2)
		outs[1]->push (r);

	check_for_requests();
}

inline void
SamplePlayer::one_cycle (CycleContext & cycle)
{
	/* set the FP unit to truncating mode (i386) -- need to truncate a lot
	 * of fp numbers when mapping (float) phase to (int) sample index, and
	 * saving/loading the i387 status word is expensive. 
	 */
	DSP::FPTruncateMode _t;

	for (int i = 0; i < POLYPHONY; ++i)
	{
		PlayingVoice & v = playing[i];

		if (v.state == PlayingVoice::Off)
			continue;
		
		if (v.voice->is_looped())
			v.play (cycle);
		else
			v.play_simple (cycle);
	}
}

inline void
SamplePlayer::parse_event (MidiEvent * e)
{
	/* sanest is to apply transpose right here and not care about it elsewhere. */
	if (e->is_note_on())
		noteon (e->data[1] + transpose, e->data[2]);

	else if (e->is_note_off())
		noteoff (e->data[1] + transpose);

	::decref (e);
}

inline void
SamplePlayer::noteon (int note, int velocity)
{
	for (int i = 0; i < voices; ++i)
	{
		SampledVoice * voice = voices[i];

		if (!voice->has_note (note))
			continue;
			
		start_voice (voice, note, ((double) velocity) / 127.);

		if (!layered)
			break;
	}
}

inline void
SamplePlayer::start_voice (SampledVoice * voice, int note, double velocity)
{
	double lightest = TEN_TO_THE_SIXTH;
	int index = -1;

	/* find a PlayingVoice to handle the new note */
	for (int i = 0; i < POLYPHONY; ++i)
	{
		PlayingVoice & v = playing[i];

		/* muffle voices from the same group */
		if (voice->group > 0 && v.voice && v.voice->group == voice->group)
			v.muffle();
		
		double weight = v.get_weight();

		if (weight < lightest)
			index = i,
			lightest = weight;
	}

	msg(), "assigning ", index, playing[index].get_weight(), "to note ", note, "in voice ", voice;
	playing[index].noteon (voice, graph->sample_rate, note, velocity);
}

inline void
SamplePlayer::noteoff (int note)
{
	/* cannot turn off a note we don't have. */
	for (int i = 0; i < POLYPHONY; ++i)
	{
		if (playing[i].note != note)
			continue;

		playing[i].noteoff();
	}
}

void
SamplePlayer::all_notes_off()
{
	for (int i = 0; i < POLYPHONY; ++i)
		playing[i].off();
}

/* see SamplePlayer_add and friends below, which push voices on the to_add
 * and to_remove fifos.
 */
void
SamplePlayer::check_for_requests()
{
	SampledVoice * voice;
	
	while ((voice = to_add.pop()))
	{
		voices.add_object (voice);
		::decref (voice);
	}

	while ((voice = to_remove.pop()))
	{
		/* SamplePlayer_clear() pushes in reverse order to prevent the voice
		 * list from memmoving a lot. so we compare last first.
		 */
		for (int j = voices - 1; j >= 0; --j)
		{
			if (voices[j] != voice)
				continue;
			voices.remove_items (j, 1);
			break;
		}

		::decref (voice);
	}

	request_lock.unlock();
}

/* // Py<SampledVoice> ///////////////////////////////////////////////////// */

static int
SampledVoice_set_loop (SampledVoice * v, PyObject * loop)
Py_TRY
{
	/* no loop = both (start, end) at sample->size */
	if (!loop || loop == Py_None)
	{
		v->loop.s0 = v->loop.s1 = v->sample->size;
		return 0;
	}
	
	/* else, check first. */
	float l0, l1;

	if (!PyArg_ParseTuple (loop, "ff", &l0, &l1))
		return -1;

	if (l0 < 0 || l0 >= v->sample->size)
		throw ValueError() << "loop start out of bounds.";
	if (l1 <= l0 + 8 || l1 > v->sample->size)
		throw ValueError() << "loop end out of bounds.";

	v->loop.s0 = l0;
	v->loop.s1 = l1;

	return 0;
}
Py_CATCH (-1);

static PyObject *
SampledVoice_get_loop (SampledVoice * v)
{
	return Py_BuildValue ("ff", v->loop.s0, v->loop.s1);
}

/* range */

static int
SampledVoice_set_range (SampledVoice * v, PyObject * range)
Py_TRY
{
	/* else, check first. */
	int n0, n1;

	if (!PyArg_ParseTuple (range, "ii", &n0, &n1))
		return -1;

	if (n1 < n0)
		throw ValueError() << "range must have at least one note.";

	v->range.n0 = n0;
	v->range.n1 = n1;

	return 0;
}
Py_CATCH (-1);

static PyObject *
SampledVoice_get_range (SampledVoice * v)
{
	return Py_BuildValue ("ii", v->range.n0, v->range.n1);
}

/* group */

static int 
SampledVoice_set_group (SampledVoice * v, PyObject * g)
{
	return !PyArg_Parse (g, "i", &v->group);
}

static PyObject *
SampledVoice_get_group (SampledVoice * v)
{
	return PyInt (v->group);
}

/* v2f */
static int
SampledVoice_set_v2f (SampledVoice * v, PyObject * s)
{
	return !PyArg_ParseTuple (s, "ff", v->v2f, v->v2f + 1);
}

static PyObject *
SampledVoice_get_v2f (SampledVoice * v)
{
	return Py_BuildValue ("ff", v->v2f[0], v->v2f[1]);
}

/* n2f */
static int
SampledVoice_set_n2f (SampledVoice * v, PyObject * s)
{
	return !PyArg_ParseTuple (s, "ff", v->n2f, v->n2f + 1);
}

static PyObject *
SampledVoice_get_n2f (SampledVoice * v)
{
	return Py_BuildValue ("ff", v->n2f[0], v->n2f[1]);
}

/* rnd2f */
static int 
SampledVoice_set_rnd2f (SampledVoice * v, PyObject * a)
{
	return !PyArg_Parse (a, "f", &v->rnd2f);
}

static PyObject *
SampledVoice_get_rnd2f (SampledVoice * v)
{
	return PyFloat (v->rnd2f);
}

/* env2f */
static int 
SampledVoice_set_env2f (SampledVoice * v, PyObject * a)
{
	return !PyArg_Parse (a, "f", &v->env2f);
}

static PyObject *
SampledVoice_get_env2f (SampledVoice * v)
{
	return PyFloat (v->env2f);
}

/* pan */
static int 
SampledVoice_set_pan (SampledVoice * v, PyObject * a)
{
	return !PyArg_Parse (a, "f", &v->pan);
}

static PyObject *
SampledVoice_get_pan (SampledVoice * v)
{
	return PyFloat (v->pan);
}

/* note */
static int 
SampledVoice_set_note (SampledVoice * v, PyObject * a)
Py_TRY
{
	int note;
	
	if (!PyArg_Parse (a, "i", &note))
		return -1;
	
	if (note <= 0)
		throw ValueError() << "note must be > 0.";

	v->note = note;
	return 0;
}
Py_CATCH (-1);

static PyObject *
SampledVoice_get_note (SampledVoice * v)
{
	return PyInt (v->note);
}

/* decay */
static int 
SampledVoice_set_decay (SampledVoice * v, PyObject * a)
Py_TRY
{
	float decay;
	
	if (!PyArg_Parse (a, "f", &decay))
		return -1;
	
	if (decay <= 0.0)
		throw ValueError() << "decay must be > 0.";

	v->decay = decay;
	return 0;
}
Py_CATCH (-1);

static PyObject *
SampledVoice_get_decay (SampledVoice * v)
{
	return PyFloat (v->decay);
}

/* release */
static int 
SampledVoice_set_release (SampledVoice * v, PyObject * a)
Py_TRY
{
	float release;
	
	if (!PyArg_Parse (a, "f", &release))
		return -1;
	
	if (release <= 0.0)
		throw ValueError() << "release must be > 0.";

	v->release = release;
	return 0;
}
Py_CATCH (-1);

static PyObject *
SampledVoice_get_release (SampledVoice * v)
{
	return PyFloat (v->release);
}

/* tune */
static int 
SampledVoice_set_tune (SampledVoice * v, PyObject * a)
Py_TRY
{
	if (!PyArg_Parse (a, "f", &v->tune))
		return -1;
	return 0;
}
Py_CATCH (-1);

static PyObject *
SampledVoice_get_tune (SampledVoice * v)
{
	return PyFloat (v->tune);
}

static PyGetSetDef 
SampledVoice_getsets [] =
{
	{"loop", (getter) SampledVoice_get_loop, (setter) SampledVoice_set_loop,
	 "tuple of (loop start, loop end) in frames, or None for no loop."},
	{"range", (getter) SampledVoice_get_range, (setter) SampledVoice_set_range,
	 "(lowest, highest) note this voice can play."},
	{"group", (getter) SampledVoice_get_group, (setter) SampledVoice_set_group,
	 "integer identifier for the voice group (terminates other voices of "
	 "the same group at note-on). Group IDs <= 0 are ignored."},
	{"v2f", (getter) SampledVoice_get_v2f, (setter) SampledVoice_set_v2f,
	 "velocity to filter."},
	{"n2f", (getter) SampledVoice_get_n2f, (setter) SampledVoice_set_n2f,
	 "note to filter."},
	{"pan", (getter) SampledVoice_get_pan, (setter) SampledVoice_set_pan,
	 "pan position, -1 = left, 0 = center, 1 = right."},
	{"rnd2f", (getter) SampledVoice_get_rnd2f, (setter) SampledVoice_set_rnd2f,
	 "randomize voice filter, from -1 to 1."},
	{"env2f", (getter) SampledVoice_get_env2f, (setter) SampledVoice_set_env2f,
	 "randomize voice filter, from -1 to 1."},
	{"note", (getter) SampledVoice_get_note, (setter) SampledVoice_set_note,
	 "natural note of the sample."},
	{"decay", (getter) SampledVoice_get_decay, (setter) SampledVoice_set_decay,
	 "decay ratio, seconds / 6 dB."},
	{"release", (getter) SampledVoice_get_release, (setter) SampledVoice_set_release,
	 "release ratio, seconds / 6 dB."},
	{"tune", (getter) SampledVoice_get_tune, (setter) SampledVoice_set_tune,
	 "tune, 0. = no tuning, 1. = semitone up."},
	{ }
};

void
Py<SampledVoice>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Object_type;
	t->tp_getset = SampledVoice_getsets;
}

char * Py<SampledVoice>::__doc__ = 
"SampledVoice (sample[, property = value, ...]) -- "
"A Sample for use by a SamplePlayer.\n";

void
Py<SampledVoice>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {
			"sample", "loop", "group", "v2f", "pan", "rnd2f", "note", 
			"decay", "release", 0
		};
	AudioSignal * s = 0;
	PyObject * loop = 0;
	PyObject * group = 0;
	PyObject * v2f = 0;
	PyObject * pan = 0;
	PyObject * rnd2f = 0;
	PyObject * note = 0;
	PyObject * decay = 0;
	PyObject * release = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "O!|OOOOOOOO", keys, 
				AudioSignal_type, &s, &loop, 
				&group, &v2f, &pan, &rnd2f, &note, &decay, &release))
		return;
	
	init (s);

	if (loop) SampledVoice_set_loop (this, loop);
	if (group) SampledVoice_set_group (this, group);
	if (v2f) SampledVoice_set_v2f (this, v2f);
	if (pan) SampledVoice_set_pan (this, pan);
	if (rnd2f) SampledVoice_set_rnd2f (this, rnd2f);
	if (note) SampledVoice_set_note (this, note);
	if (decay) SampledVoice_set_decay (this, decay);
	if (release) SampledVoice_set_release (this, release);
}

/* // Py<SamplePlayer> ///////////////////////////////////////////////////// */

/* sequence methods */

static int 
SamplePlayer_sq_length (SamplePlayer * s)
{
	ScopedLock lock (s->request_lock);
	return s->voices;
}

static PyObject *
SamplePlayer_sq_item (SamplePlayer * s, int i) 
Py_TRY
{
	ScopedLock lock (s->request_lock);

	if (i < 0 || i >= s->voices)
		throw IndexError() << "no item " << i << " in " << s;
	
	return ::incref (s->voices[i]);
} 
Py_CATCH (0);

static PySequenceMethods 
SamplePlayer_as_sequence = 
{
	(inquiry) SamplePlayer_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) SamplePlayer_sq_item, /* sq_item */
	(intintargfunc) 0, /* sq_slice */
	(intobjargproc) 0, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
SamplePlayer_add (SamplePlayer * sp, SampledVoice * v)
Py_TRY
{
	if (!SampledVoice_type->is_derived (v))
		throw TypeError() << "need a SampledVoice.";

	/* unlocked by check_for_requests */
	sp->request_lock.lock();
	
	::incref (v);
	sp->to_add.push (v);

	if (sp->state < Plugin::Activated)
		sp->check_for_requests();
	
	return ::incref (v);
}
Py_CATCH (0);

static PyObject *
SamplePlayer_remove (SamplePlayer * sp, SampledVoice * v)
Py_TRY
{
	if (!SampledVoice_type->is_derived (v))
		throw TypeError() << "need a SampledVoice.";

	/* unlocked by check_for_requests */
	sp->request_lock.lock();
	
	::incref (v);
	sp->to_remove.push (v);

	if (sp->state < Plugin::Activated)
		sp->check_for_requests();
	
	return PyNone();
}
Py_CATCH (0);

static PyObject *
SamplePlayer_clear (SamplePlayer * sp)
{
	/* unlocked by check_for_requests */
	sp->request_lock.lock();
	
	int n = sp->voices;
	SampledVoice * voices [n];

	/* copy first because items from the list can be removed while we're still
	 * posting remove requests. */
	for (int i = 0; i < n; ++i)
		voices[i] = sp->voices[i];
	
	/* removing in reverse order saves us some memmoving. */
	for (int i = n - 1; i >= 0; --i)
	{
		::incref (voices[i]);
		sp->to_remove.push (voices[i]);
	}

	if (sp->state < Plugin::Activated)
		sp->check_for_requests();
	
	return PyNone();
}

static PyMethodDef
SamplePlayer_methods [] = 
{
	{"add", (PyCFunction) SamplePlayer_add, METH_O,
	 "add (sampled voice) -- add a SampledVoice to the voice list."},
	{"remove", (PyCFunction) SamplePlayer_remove, METH_O,
	 "remove (sampled voice) -- remove a SampledVoice from the voice list."},
	{"clear", (PyCFunction) SamplePlayer_clear, METH_NOARGS,
	 "clear() -- remove all voices from the voice list."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
SamplePlayer_get_db (SamplePlayer * p)
{
	return PyFloat (20 * log10 (p->gain));
}

static int
SamplePlayer_set_db (SamplePlayer * p, PyObject * a)
{
	double db;

	if (!PyArg_Parse (a, "d", &db))
		return -1;

	p->gain = pow (10., db / 20.);
	return 0;
}

static PyObject *
SamplePlayer_get_transpose (SamplePlayer * p)
{
	return PyInt (p->transpose);
}

static int
SamplePlayer_set_transpose (SamplePlayer * p, PyObject * a)
{
	if (!PyArg_Parse (a, "i", &p->transpose))
		return -1;

	return 0;
}

static PyObject *
SamplePlayer_get_layered (SamplePlayer * p)
{
	return PyInt (p->layered);
}

static int
SamplePlayer_set_layered (SamplePlayer * p, PyObject * a)
{
	if (!PyArg_Parse (a, "i", &p->layered))
		return -1;

	return 0;
}

static PyObject *
SamplePlayer_get_playing (SamplePlayer * p)
{
	int n = 0;
	
	for (int i = 0; i < POLYPHONY; ++i)
		if (p->playing[i].state)
			++n;

	return PyInt (n);
}

static PyObject *
SamplePlayer_get_channels (SamplePlayer * p)
{
	return PyInt (p->channels);
}

static PyGetSetDef 
SamplePlayer_getsets[] =
{
	{"db", (getter) SamplePlayer_get_db, (setter) SamplePlayer_set_db,
	 "volume control."},
	{"transpose", (getter) SamplePlayer_get_transpose, (setter) SamplePlayer_set_transpose,
	 "semitones."},
	{"layered", (getter) SamplePlayer_get_layered, (setter) SamplePlayer_set_layered,
	 "if True, overlapping voice ranges cause multiple oscillators to kick off. "
	 "default is False, ie. just one oscillator per note."},
	{"playing", (getter) SamplePlayer_get_playing, (setter) 0,
	 "the currently sounding number of samples."},
	{"channels", (getter) SamplePlayer_get_channels, (setter) 0,
	 "output channels (1 or 2)."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<SamplePlayer>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = SamplePlayer_getsets;
	t->tp_methods = SamplePlayer_methods;
	t->tp_as_sequence = &SamplePlayer_as_sequence;
}

char * Py<SamplePlayer>::__doc__ = 
"SamplePlayer ([channels = 2, db = -12]) -- a Plugin containing 128 SampledVoices.";

void
Py<SamplePlayer>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"channels", "db", 0};
	int c = 2;
	PyObject * db = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|iO", keys, &c, &db))
		return;

	if (c < 1 || c > 2)
		throw ValueError() << "1 or 2 channels.";

	init (c);
	if (db) SamplePlayer_set_db (this, db);
}

/* /////////////////////////////////////////////////////////////////////// */

Type * GUSVoice_type;

GUSVoice::GUSVoice()
: NamedObject (GUSVoice_type)
{
	sample = 0;
}

GUSVoice::~GUSVoice()
{
	::decref (sample, true);
}

/* ANSI C doesn't want float template parameters, sigh. let's hope gcc is
 * smart enough to recognize the optimization potential in case SHIFT == 0.
 *
 * this one copies patch sample data into an audio signal.
 */
template<class T, int SHIFT> 
void
copy_samples (AudioSignal * _dest, void * _src, double scale)
{
	T * src = (T *) _src;
	d_sample * dest = _dest->data;
	
	scale = 1. / scale;
	d_sample shift = (d_sample) SHIFT;

	for (int i = 0; i < _dest->size; ++i)
		dest[i] = scale * ((d_sample) src[i]) + shift;
}

static int
env_rate (byte b)
{
	int shift = 3 * (3 ^ ((b >> 6) & 3));
	int i = b & 0x3F;

	return i << shift;
}

static int
env_offset (byte b)
{
	return b << (7 + 15);
}

void
GUSPatchFile::read_patch()
{
	GUSVoice * patch = NEW (GUSVoice);
	patches.add_object (patch);
	::decref (patch, true);

	byte name[8];
	read (name, 7);

	patch->set_object_name ("%-7s", name);

	byte fractions = read_8();

	int size = (int) read_32();

	patch->loop.start = read_32();
	patch->loop.end = read_32();

	patch->sample_rate = read_16();

	patch->freq.low = (int) read_32();
	patch->freq.high = (int) read_32();
	patch->freq.root = (int) read_32();

	skip (2); /* tune */
	
	patch->pan = read_8();
	
	for (int i = 0; i < 6; ++i)
		patch->envelope.rates[i] = env_rate (read_8());

	for (int i = 0; i < 6; ++i)
		patch->envelope.offsets[i] = env_offset (read_8());

	skip (6); /* tremolo, vibrato */

	patch->mode = read_8();

	/* adjust loop */
	if (patch->mode & GUSVoice::TwoByteSamples)
		patch->loop.start /= 2,
		patch->loop.end /= 2;

	patch->loop.start += (fractions & 0xF) / 16.;
	patch->loop.end += (fractions & 0xF0) / 256.;

	skip (40); /* scale, reserved */

	patch->sample = NEW (AudioSignal);

	if (patch->mode & GUSVoice::TwoByteSamples)
		patch->sample->set_size (size / 2);
	else
		patch->sample->set_size (size);
	
	void * data = malloc (size);
	if (data == 0)
		throw MemoryError();

	read (data, size);

	if (patch->mode & GUSVoice::TwoByteSamples)
	{
		if (patch->mode & GUSVoice::Unsigned)
			copy_samples<uint16, -1> (patch->sample, data, 32768.);
		else
			copy_samples<int16, 0> (patch->sample, data, 32768.);
	}
	else
	{
		if (patch->mode & GUSVoice::Unsigned)
			copy_samples<uint8, -1> (patch->sample, data, 128.);
		else
			copy_samples<int8, 0> (patch->sample, data, 128.);
	}

	free (data);
}

/* qsort invoked */
static int
compare_voices (GUSVoice ** v0, GUSVoice ** v1)
{
	return (*v0)->freq.root - (*v1)->freq.root;
}

static PyObject *
load_gus_patch (PyObject * self, PyObject * arg)
Py_TRY
{
	char * s;

	if (!PyArg_Parse (arg, "s", &s))
		return 0;

	GUSPatchFile fd (s);

	/* read & parse the header */
	byte head[239];

	fd.read (head, 239);

	if (memcmp (head, "GF1PATCH110\0ID#000002", 22) &&
			memcmp (head, "GF1PATCH100\0ID#000002", 22))
		throw IOError() << "not a GUS patch file: " << s;

	if (head[82] > 2)
		throw IOError() << "cannot handle more than 1 instrument.";
	
	int samples = head[198];

	for (int i = 0; i < samples; ++i)
		fd.read_patch();

	/* sort the voices by root frequency (this one's on the house) */
	qsort (fd.patches.items, fd.patches.length, sizeof (GUSVoice *), 
			(int (*) (const void *, const void *)) compare_voices);

	return PyTuple (&fd.patches);
}
Py_CATCH (0);

static PyMethodDef
_methods [] = 
{
	{"load_gus_patch", (PyCFunction) load_gus_patch, METH_O,
	 "load_gus_patch (path) -- loads a .pat file, returning a tuple of "
	 "GUSVoice objects."},
	{ }
};
	
/* /////////////////////////////////////////////////////////////////////// */

char * Py<GUSVoice>::__doc__ = 
"GUSVoice() -- one voice from a GUS patch file, returned by load_gus_patch().";

static PyObject *
GUSVoice_get_mode (GUSVoice * p)
{
	return PyInt (p->mode);
}

static PyObject *
GUSVoice_get_sample_rate (GUSVoice * p)
{
	return PyInt (p->sample_rate);
}

static PyObject *
GUSVoice_get_freq (GUSVoice * p)
{
	return Py_BuildValue ("iii", p->freq.low, p->freq.high, p->freq.root);
}

static PyObject *
GUSVoice_get_pan (GUSVoice * p)
{
	return PyInt (p->pan);
}

static PyObject *
GUSVoice_get_sample (GUSVoice * p)
{
	return ::incref (p->sample);
}

static PyObject *
GUSVoice_get_loop (GUSVoice * p)
{
	return Py_BuildValue ("ff", p->loop.start, p->loop.end);
}

static PyObject *
GUSVoice_get_env_rates (GUSVoice * p)
{
	int * b = p->envelope.rates;
	return Py_BuildValue ("iiiiii", b[0], b[1], b[2], b[3], b[4], b[5]);
}

static PyObject *
GUSVoice_get_env_offsets (GUSVoice * p)
{
	int * b = p->envelope.offsets;
	return Py_BuildValue ("iiiiii", b[0], b[1], b[2], b[3], b[4], b[5]);
}

static PyGetSetDef
GUSVoice_getsets [] = 
{
	{"mode", (getter) GUSVoice_get_mode, (setter) 0,
	 "mode bits."},
	{"sample", (getter) GUSVoice_get_sample, (setter) 0,
	 "the sample data, an AudioSignal."},
	{"loop", (getter) GUSVoice_get_loop, (setter) 0,
	 "(start, end) sample offsets."},
	{"freq", (getter) GUSVoice_get_freq, (setter) 0,
	 "(low, high, root) frequencies."},
	{"pan", (getter) GUSVoice_get_pan, (setter) 0,
	 "pan position, from [0 .. 127]."},
	{"env_rates", (getter) GUSVoice_get_env_rates, (setter) 0,
	 "envelope rates, 6 bytes: attack, decay, sustain, 3 * release."},
	{"env_offsets", (getter) GUSVoice_get_env_offsets, (setter) 0,
	 "envelope offsets, 6 bytes: attack, decay, sustain, 3 * release."},
	{"sample_rate", (getter) GUSVoice_get_sample_rate, (setter) 0,
	 "the sample rate, int."},
	{ }
};

void
Py<GUSVoice>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<GUSVoice>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = NamedObject_type;
	type->tp_getset = GUSVoice_getsets;
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_SamplePlayer_types (PyObject * dict)
{
	SampledVoice_type = new PyType<SampledVoice> ("SampledVoice", dict);
	SamplePlayer_type = new PyType<SamplePlayer> ("SamplePlayer", dict);
	GUSVoice_type = new PyType<GUSVoice> ("GUSVoice", dict);

	for (PyMethodDef * md = _methods; md->ml_name; ++md)
	{
		PyObject * m = PyCFunction_New (md, 0);

		if (m == 0)
		{
			msg(), "cannot create method: ", md->ml_name;
			continue;
		}

		if (PyDict_SetItemString (dict, md->ml_name, m))
			msg(), "cannot register method: ", md->ml_name;
	}
}
