/* Copyright 2003 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _SAMPLE_PLAYER_H_
#define _SAMPLE_PLAYER_H_

#include "Plugin.h"
#include "AudioSignal.h"
#include "dsp/OnePole.h"
#include "dsp/Phase.h"

#define POLYPHONY 12

class SampledVoice
: public Object
{
	public:
		AudioSignal * sample;
		
		/* loop start, end frame */
		struct {
			float s0, s1;
		} loop;
		
		/* notes comprised */
		struct {
			int n0, n1;
		} range;
		
		int note;

		int group; /* note-on cancels all voices from this group */
		
		float v2f[2]; /* velocity to filter */
		float n2f[2]; /* note to filter */
		float pan;
		float decay; /* only affects filter */
		float release; /* affects volume */
		float rnd2f; /* random to filter */
		float env2f; /* envelope to filter */
		float tune;  /* 0 = straight, 1 = semitone */

		/* //////////////////////////////////////////////////////////////////// */
		
		SampledVoice();
		~SampledVoice();

		void init (AudioSignal *); /* need only set this at object creation */

		inline int is_looped() { return loop.s1 != loop.s0; }
		inline bool has_note (int n) { return n >= range.n0 && n <= range.n1; }
};

/* //////////////////////////////////////////////////////////////////// */

/* wraps the info necessary for a PlayingVoice to do one cycle */
class CycleContext
{
	public:
		int frames;
		
		struct {
			d_sample * l, * r;
		} out, rev;

		d_sample gain;

		float pitch;
};

class PlayingVoice
{
	public:
		SampledVoice * voice;
		
		int state;
		enum {
			Off = 0,
			Attack = 1,
			Decay = 2,
			Sustain = 3,
			Release = 4
		};

		int note;
		float fs;

		DSP::Phase phi; /* keeps frame.fract sample index */

		float release; /* copied from SampledVoice, modified by muffle() */
		float pitch; /* initialized from SampledVoice */

		struct {
			d_sample left, right; /* set according to pan position */
			double env; /* envelope */
			double factor; /* per-sample */
		} gain;
		
		double filter_factor; /* per-sample */

		DSP::OnePoleLP filter;

		PlayingVoice();
		~PlayingVoice();
		
		inline void noteon (SampledVoice *, float _fs, int note, double velocity);

		inline void play (CycleContext &);
		inline void play_simple (CycleContext &);

		inline void noteoff();
		inline void muffle(); /* same as noteoff but sets release to be quick */
		inline void off();

		inline double get_weight();
	
	protected:
		inline void enter_decay();
		inline void enter_release();
};

/* //////////////////////////////////////////////////////////////////////// */

class SamplePlayer
: public Plugin
{
	public:
		TypedObjectList<SampledVoice> voices;
		PlayingVoice playing [POLYPHONY];
		int channels;
		int transpose;
		double gain;

		bool layered;

		Mutex request_lock;
		TypedObjectFifo<SampledVoice> to_add;
		TypedObjectFifo<SampledVoice> to_remove;

		/* //////////////////////////////////////////////////////////////////// */

		SamplePlayer();
		~SamplePlayer();

		void init (int _channels);

		virtual void process_audio (int frames);

		inline void parse_event (MidiEvent *);
		inline void one_cycle (CycleContext &);

		inline void noteon (int note, int velocity);
		inline void noteoff (int note);

		inline void start_voice (SampledVoice * voice, int note, double velocity);

		void all_notes_off();

		void check_for_requests();
};

#endif /* _SAMPLE_PLAYER_H_ */
