/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "Graph.h"
#include "Pipe.h"
#include "Scheduler.h"
#include "PyType.h"
#include "PythonThread.h"

static char *
get_thread_name (int cycle)
{
	if (cycle == Plugin::Disk)
		return "Disk";

	if (cycle == Plugin::Script)
		return "Script";

	return "Cycle ?";
}

Scheduler::Scheduler (Type * t, int _cycle)
: Plugin (t, 0),
	Thread (get_thread_name (_cycle)),
	poll (get_type_name (this)),
	plugin_pipe ("plugin", this),
	trigger_pipe ("trigger", this)
{
	cycle = _cycle;
	need_trigger = 0;

	if (cycle != Plugin::Script && cycle != Plugin::Disk)
		throw AssertionError() << "can only schedule Disk and Script.";

	poll.add (trigger_pipe);

	start_thread();
}

Scheduler::~Scheduler()
{
	PythonUnlock unlock;

	trigger_pipe.write_trigger (0);

	join_thread();
}

void
Scheduler::wake_up()
{
	if (need_trigger == 0)
		return;

	trigger_pipe.write_trigger (1);

	need_trigger = 0;
}

void
Scheduler::process_disk (Plugin * p)
{
	if (cycle != Plugin::Disk)
		throw AssertionError() << "not a Disk scheduler.";

	plugin_pipe.write (&p, sizeof (Plugin *));

	++need_trigger;

	wake_up();
}

void 
Scheduler::process_script (Plugin * p)
{
	if (cycle != Plugin::Script)
		throw AssertionError() << "not a Script scheduler.";

	plugin_pipe.write (&p, sizeof (Plugin *));

	++need_trigger;

	wake_up();
}

void *
Scheduler::thread()
{
	while (1)
	{
		int r = poll.poll (1000);

		if (r < 0)
			continue;

		if (poll[0].revents & POLLHUP)
			throw StopIteration() << this << "hung up.";

		if (r == 0)
			continue;

		r = trigger_pipe.read_trigger();

		if (r == 0)
			break;

		Plugin * p;

		while (1)
		{
			r = plugin_pipe.read (&p, sizeof (Plugin *));

			if (r <= 0)
				break;
			
			if (r != sizeof (Plugin *))
				throw IOError() << "read an incomplete plugin address.";
			
			if (graph == 0)
				/* happens when going down. */
				continue;

			if (cycle == Plugin::Disk)
				graph->process_disk (p);
			else
				graph->process_script (p);
		}
	}

	return 0;
}

/* /////////////////////////////////////////////////////////////////////// */

static int 
Scheduler_set_priority (Scheduler * c, PyObject * value) 
Py_TRY
{
	int p;
	
	if (!PyArg_Parse (value, "i", &p))
		return -1;

	c->set_thread_priority (p);
	return 0;
} 
Py_CATCH (-1);

static PyObject *
Scheduler_get_priority (Scheduler * c)
{
	return PyInt (c->thread_priority);
}

static PyGetSetDef 
Scheduler_getsets [] = 
{
	{"priority", (getter) Scheduler_get_priority, (setter) Scheduler_set_priority,
	 "the scheduling priority the clock thread enjoys."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<Scheduler>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_getset = Scheduler_getsets;
}

Type * Scheduler_type = 0;

char * Py<Scheduler>::__doc__ = 
"Scheduler() -- a plugin that is needed for some plugins (those that need "
"to execute code without realtime constraints, eg Script and Disk) to operate. "
"The Scheduler runs a sleeping worker thread that is woken when needed. "
"It then asks the graph it is registered with to run the plugin in question.";

void
Py<Scheduler>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* /////////////////////////////////////////////////////////////////////// */

void
Py<ScriptScheduler>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Scheduler_type;
}

Type * ScriptScheduler_type = 0;

char * Py<ScriptScheduler>::__doc__ = 
"ScriptScheduler() -- a plugin that is needed for ScriptPlugins to operate, "
"see the Scheduler base type for the details.";

void
Py<ScriptScheduler>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();
}

/* /////////////////////////////////////////////////////////////////////// */

void
Py<DiskScheduler>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Scheduler_type;
}

Type * DiskScheduler_type = 0;

char * Py<DiskScheduler>::__doc__ = 
"DiskScheduler() -- a plugin that is needed for DiskPlugins to operate, "
"see the Scheduler base type for the details.";

void
Py<DiskScheduler>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_Scheduler_types (PyObject * dict)
{
	Scheduler_type = new PyType<Scheduler> ("Scheduler", dict);
	DiskScheduler_type = new PyType<DiskScheduler> ("DiskScheduler", dict);
	ScriptScheduler_type = new PyType<ScriptScheduler> ("ScriptScheduler", dict);
}

