/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"

#include "Graph.h"
#include "ScriptPlugin.h"

#include "PyType.h"
#include "PythonThread.h"

#include "profiler.h"

ScriptPlugin::ScriptPlugin()
: Plugin (ScriptPlugin_type, 0)
{
	woken = -1;

	/* boost internal python priority. don't set this too high or python will
	 * get very sluggish when executing a cpu-heavy script plugin. 
	 * FIXME: broken with python 2.3
	 */
	checkinterval = 1000;
}

ScriptPlugin::~ScriptPlugin()
{
	/* wait for process_* to finish if currently active. */
	ScopedLock lock (mutex);
}

/* called by Out::push() if applicable. */
void
ScriptPlugin::wake_up()
{
	/* graph and graph->script_scheduler are checked in InOut.cc */
	
	/* if already woken, only signal every 32 events to keep things cool. */
	if (!woken.inc_is_zero() && (woken & 0x1F))
		return;
	
	graph->script_scheduler->process_script (this);
}

void 
ScriptPlugin::process_script (Plugin *)
{
	woken = -1;

	if (state == Off)
	{
		for (int i = 0; i < ins; ++i)
			while (ins[i]->pop());
		return;
	}

	for (int i = 0; i < ins; ++i)
	{
		int n = ins[i]->have();
		if (n == 0)
		{
			msg(), this, "::process_script() has nothing to handle";
			continue;
		}

		PyObject * l;

		/* hold while allocating */
		{
			PythonLock _plock;
		
			l = PyList_New (n);
			if (l == 0) throw blewit();
			
			for (int j = 0; j < n; ++j)
			{
				Event * e = (Event *) ins[i]->pop();
				PyList_SetItem (l, j, e);
			}
		}
		
		if (!call_method ("process", "OO", ins[i], l))
			state = Off;

		::decref (l);
	}
}

/* /////////////////////////////////////////////////////////////////////// */

/* helper for get_methods() */
static PyObject *
get_callback (PyObject * o, char * name)
{
	PyObject * method = PyObject_GetAttrString (o, name);
	
	/* the method will go down when we go down. */
	::decref (method, true);

	if (method && PyCallable_Check (method))
		return method;

	return 0;
}

void
ScriptPlugin::get_methods()
{
	if (get_callback (this, "process"))
		flags |= Script;

	if (get_callback (this, "process_beat"))
		flags |= Beat;
	
	PyErr_Clear();
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
ScriptPlugin_new_in (ScriptPlugin * p, PyObject * args, PyObject * kwds)
Py_TRY
{
	static char * keys [] = {"name", "type", "slots", 0};
	char * name;
	Type * t;
	int slots = 1;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "sO!|i", keys,
			&name, &PyType_Type, &t, &slots))
		return 0;
	
	if (!Object_type->is_derived (t))
		throw TypeError() << "only Object and derived types can be received.";

	if (CyclicAudio_type->is_derived (t))
		throw NotImplementedError() 
				<< "you can't do realtime audio DSP with Python (yet).";
	
	ScopedLock lock (p->mutex);
	return ::incref (p->add_in (new In (p, t, slots, name)));
} 
Py_CATCH (0);

static PyObject *
ScriptPlugin_new_out (ScriptPlugin * p, PyObject * args, PyObject * kwds)
Py_TRY
{
	static char * keys [] = {"name", "type", 0};
	char * name;
	Type * t;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "sO!", keys,
			&name, &PyType_Type, &t))
		return 0;
	
	if (!Object_type->is_derived (t))
		throw TypeError() << "only Objects and derived types can be transmitted.";

	if (CyclicAudio_type->is_derived (t))
		throw NotImplementedError() 
			<< "you can't do realtime audio DSP with Python (yet).";

	ScopedLock lock (p->mutex);
	return ::incref (p->add_out (new Out (p, t, name)));
} 
Py_CATCH (0);

static PyMethodDef ScriptPlugin_methods [] = {
  {"new_in", (PyCFunction) ScriptPlugin_new_in, METH_VARARGS | METH_KEYWORDS,
	 "new_in (name, type[, slots = 1]) -- add an input point for 'type' events. "
	 "the input will allow up to 'slots' connecting outputs."},
  {"new_out", (PyCFunction) ScriptPlugin_new_out, METH_VARARGS | METH_KEYWORDS,
	 "new_out (name, type) -- add an output point 'name' for 'type' events."},
	{ }
};

void
Py<ScriptPlugin>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Plugin_type;
	t->tp_methods = ScriptPlugin_methods;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
}

Type * ScriptPlugin_type = 0;

char * Py<ScriptPlugin>::__doc__ = 
"ScriptPlugin() -- a Plugin that is meant to be subclassed, "
"declaring a 'process' and/or a 'process_beat' method. The 'process' "
"method is called, without arguments, whenever the plugin receives data. "
"The plugin should pop all events from its inputs.\n"
"The 'process_beat' method, if defined, is called at about every "
"eighth beat. Arguments passed "
"are two ticks describing an interval, usually an eighth note "
"ahead of the current time and spanning the length of an eighth. "
"The plugin "
"is supposed to generate events for the interval in question.\n"
"NB: Proper event timing around loop points requires plugins to put "
"a correct 'loop' stamp on the events they transmit.\n"
"A ScriptScheduler is needed in the same graph for the 'process' method "
"to work, the 'process_beat' method needs a RTClock.";

void
Py<ScriptPlugin>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();
	
	/* check which methods have been declared in the plugin. */
	get_methods();
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_ScriptPlugin_types (PyObject * dict)
{
	ScriptPlugin_type = new PyType<ScriptPlugin> ("ScriptPlugin", dict);
}
