/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "AudioSignal.h"

#include "SoundFile.h"
#include "WaveFile.h"
#include "OggFile.h"

#include "PyType.h"

static PyObject * 
SoundFile_get_channels (SoundFile * s)
{
	return PyInt (s->channels);
}

static PyObject * 
SoundFile_get_sample_rate (SoundFile * s)
{
	return PyInt (s->sample_rate);
}

static PyObject * 
SoundFile_get_mode (SoundFile * s)
{
	return PyString ((s->mode == O_RDONLY) ? "r" : "w");
}

static PyObject * 
SoundFile_get_path (SoundFile * s)
{
	return PyString (s->get_path());
}

static PyObject * 
SoundFile_get_frames (SoundFile * s)
{
	return PyFrame (s->frames);
}

static PyObject * 
SoundFile_get_frame (SoundFile * s)
{
	return PyFrame (s->tell());
}

static PyGetSetDef 
SoundFile_getsets [] = 
{
	{"channels", (getter) SoundFile_get_channels, 0, "the number of channels."},
	{"sample_rate", (getter) SoundFile_get_sample_rate, 0, "the sample rate."},
	{"mode", (getter) SoundFile_get_mode, 0, "the file access mode."},
	{"path", (getter) SoundFile_get_path, 0, "where the file is."},
	{"frames", (getter) SoundFile_get_frames, 0, "the length, in audio frames."},
	{"frame", (getter) SoundFile_get_frame, 0, "the current file pointer in audio frames."},
	{ },
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
SoundFile_seek (SoundFile * f, PyObject * pos)
Py_TRY
{
	d_frame p;

	if (!PyArg_Parse (pos, "L", &p))
		return 0;

	ScopedLock lock (f->mutex);
	f->seek (p);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
SoundFile_read (SoundFile * f, PyObject * _frames)
Py_TRY
{
	int frames = -1;

	if (!PyArg_ParseTuple (_frames, "|i", &frames))
		return 0;

	if (frames < 0)
		frames = (int) (f->frames - f->tell());

	PyObject * t = PyTuple_New (f->channels);

	d_sample * data[f->channels];

	try
	{
		for (int i = 0; i < f->channels; ++i)
		{
			AudioSignal * s = NEW (AudioSignal);
			s->set_size (frames);
			data[i] = s->data;
			PyTuple_SetItem (t, i, s);
		}

		ScopedLock lock (f->mutex);

		f->read (data, frames);
	}
	catch (std::exception & e)
	{
		::decref (t, true);
		throw;
	}

	return t;
}
Py_CATCH (0);

static PyObject *
SoundFile_write (SoundFile * f, PyObject * args)
Py_TRY
{
	if (f->mode == O_RDONLY)
		throw RuntimeError() << f << " is read-only.";

	if (!PySequence_Check (args))
		throw TypeError() << "need a sequence of AudioSignals.";

	if (PySequence_Size (args) != f->channels)
		throw ValueError() << "need exactly " << f->channels << " signals.";

	int frames_to_write = TEN_TO_THE_NINTH;
	d_sample * data[f->channels];

	for (int i = 0; i < f->channels; ++i)
	{
		AudioSignal * s = (AudioSignal *) PySequence_GetItem (args, i);
		
		if (!AudioSignal_type->is_derived (s))
		{
			::decref (s, true);
			throw TypeError() << "need AudioSignal in [" << i << "].";
		}
		
		frames_to_write = min (s->size, frames_to_write);

		data[i] = s->data;

		/* the signal is still referenced by the args sequence, so this is
		 * safe, we wont lose the data buffers. */
		::decref (s, true);
	}

	ScopedLock lock (f->mutex);

	f->write (data, frames_to_write);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
SoundFile_methods[] =
{
	{"seek", (PyCFunction) SoundFile_seek, METH_O,
	 "seek (frames) -- moves the file pointer."},
	{"read", (PyCFunction) SoundFile_read, METH_VARARGS,
	 "read (frames) -- returns a tuple of AudioSignals, one per channel."},
	{"write", (PyCFunction) SoundFile_write, METH_VARARGS,
	 "write (sequence of AudioSignals[, frames]) -- writes 'frames' "
	 "from a sequence of signals ('channels'), or as many frames as are "
	 "found in the shortest signal in the sequence."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

Type * SoundFile_type = 0;

char * Py<SoundFile>::__doc__ = 
"SoundFile() -- an abstract base type. Derived objects from the extension "
"module can be used by DiskStreams for playback and recording.";

void
Py<SoundFile>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = NamedObject_type;
	t->tp_getset = SoundFile_getsets;
	t->tp_methods = SoundFile_methods;
}

void
Py<SoundFile>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

/* //////////////////////////////////////////////////////////////////////// */

void
Py<WaveFile>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = SoundFile_type;
}

Type * WaveFile_type = 0;

char * Py<WaveFile>::__doc__ = 
"WaveFile ('file.wav'"
"[, mode = 'r', channels = 1, sample_rate = 44100, bits = 0]) "
"-- an uncompressed PCM audio file in .wav format, accessed via "
"memory-mapped I/O. Use with a DiskStream. If 'mode' is 'w', the file is "
"opened for recording. If bits is an integer in (32, 16, 8) the file will "
"contain integer samples of this bit depth; 'bits' = 0 is for 32 bit "
"IEEE floats. Any other 'bits' value will raise an exception.\n"
"No sample rate conversions are done.";

void
Py<WaveFile>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"path", "mode", "channels", "sample_rate", "bits", 0};
	char * path;
	char mode = 'r';
	int channels = -1;
	int sr = SAMPLE_RATE;
	int bits = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|ciii", keys, 
			&path, &mode, &channels, &sr, &bits))
		throw blewit();
	
	if (mode != 'r' && mode != 'w')
		throw ValueError() << "bad file mode.";
	
	if (mode == 'r')
	{
		if (channels != -1)
			throw SyntaxError() << "number of channels is predetermined in read mode.";
		channels = 1;
	}

	if (mode == 'w' && channels == -1)
		channels = 1;

	if (channels < 1 || channels > 255)
		throw ValueError() << "bad number of channels.";

	this->open (path, mode == 'w' ? O_WRONLY : O_RDONLY, channels, sr, bits);
}

/* OggFile //////////////////////////////////////////////////////////////// */

#ifdef WITH_OGG

void
OggFile::open (char * _path)
{
	path = strdup (_path);
	
	f = fopen (path, "r");
	if (f == 0)
		throw OSError (errno) << ":" << path;

	int r = ov_open (f, &vf, 0, 0);
	if (r)
		throw RuntimeError() << "ov_open ('" << path << "'): " << r;
	
	f = 0;

	vorbis_info * vi = ov_info (&vf, -1);
	if (vi == 0)
		throw RuntimeError() << "ov_info ('" << path << "'): failed";

	sample_rate = vi->rate;
	channels = vi->channels;
	frames = 0x7fffFFFF;
}

void
OggFile::seek (d_frame to)
{
	int r = ov_pcm_seek (&vf, start + to);
	
	if (r == 0)
	{
		at = to;
		return;
	}

	throw RuntimeError() << "ov_pcm_seek (" << to << "): " << r;
}

void
OggFile::read (d_sample ** data, int frames)
{
	/* the decoder handles full packets only, therefore we have to readjust
	 * the file position every time. */
	seek (at);

	float ** ogg_data = 0;

	int i = 0;

	while (frames > 0)
	{
		int r = ov_read_float (&vf, &ogg_data, frames, &section);

		if (r == OV_HOLE)
			throw RuntimeError() << "ogg stream interruption";
		if (r == OV_EBADLINK)
			throw RuntimeError() << "bad ogg stream";
		if (r <= 0)
			throw RuntimeError() << "read less than or zero frames";
		
		int n = min (r, frames);
	
		for (int c = 0; c < channels; ++c)
		{
			d_sample * s = data[c] + i;
			d_sample * d = ogg_data[c];
			
			for (int j = 0; j < n; ++j)
				s[j] = d[j];
		}

		frames -= n;
		i += n;
		at += n;
	}
}

void
Py<OggFile>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = SoundFile_type;
}

Type * OggFile_type = 0;

char * Py<OggFile>::__doc__ = 
"OggFile ('file.ogg') -- ogg audio file, read only, no sample rate conversion.";

void
Py<OggFile>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"path", 0};
	char * path;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s", keys, &path))
		throw blewit();
	
	this->open (path);
}

#endif /* WITH_OGG */

/* //////////////////////////////////////////////////////////////////////// */

void
init_SoundFile_types (PyObject * dict)
{
	SoundFile_type = new PyType<SoundFile> ("SoundFile", dict);
	WaveFile_type = new PyType<WaveFile> ("WaveFile", dict);
#ifdef WITH_OGG
	OggFile_type = new PyType<OggFile> ("OggFile", dict);
#endif
}
