/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <sched.h>
#include <signal.h>

#include "basics.h"

#include "Thread.h"
#include "PythonThread.h"

#include "Event.h"

#define DBG_MUTEX		0
#define DBG_LOCK		0
#define DEBUG_PY_THREADS 0

#if (DBG_MUTEX || DBG_LOCK) && !defined (DEBUG)
#error make up your mind!
#endif

/* run housekeep() every * milliseconds: */
#define HOUSEKEEP_INTERVAL	20

atomic Thread::threads_started;

/* PythonThread statics */
pthread_key_t 
	PythonThread::key = 0;
	
PointerList<PythonThread> * 
	PythonThread::all = 0;
	
Mutex * 
	PythonThread::mutex = 0;

PyInterpreterState *
	PythonThread::interp = 0;

/* 0 if initial thread. */
PythonThread *
	PythonThread::executing = 0;

/* this thread's job is to release references to objects that have been 
 * disposed of in other (realtime) threads. we need to hold the PL to do
 * this in a safe fashion, but we cannot acquire (and possibly wait for)
 * the PL in for example the audio thread.
 * 
 * Events that are in the release queue (implemented as a fifo, see 
 * PythonThread) and have 'cache' set are returned to the cache.
 */
class TrashThread
: public Thread
{
	public:
		static TrashThread * the_one;
		int state_fd[2];

		TrashThread()
			: Thread ("Trash")
			{
				if (socketpair (AF_LOCAL, SOCK_STREAM, 0, state_fd) < 0)
					throw OSError (errno) << "cannot get a socketpair(2).";

				start_thread();
			}
			
		~TrashThread()
			{
				write (state_fd[0], this, 1);

				Py_BEGIN_ALLOW_THREADS
				join_thread();
				Py_END_ALLOW_THREADS

				close (state_fd[0]);
				close (state_fd[1]);
			}

		virtual void * thread()
			{
				struct pollfd pfd [] = {{state_fd[1], POLLIN}};

				while (1)
				{
					int r = poll (pfd, 1, HOUSEKEEP_INTERVAL);
					
					if (r < 0 && errno != EINTR)
						throw OSError (errno) << "in poll(2).";
						
					if (r == 1)
						break;
					
					PythonThread::housekeep (true);
				}

				return 0;
			}
};

TrashThread * TrashThread::the_one = 0;

/* /////////////////////////////////////////////////////////////////////// */

/* initializes thread local storage and the list of threads */
void
PythonThread::init()
{	
	if (pthread_key_create (&key, 0))
		throw RuntimeError (errno) << "pthread_key_create failed.";
	
	mutex = new Mutex;
	all = new PointerList<PythonThread> ("trash fifos");
	
	/* we need a thread state so we can get the big python lock. */
	interp = PyThreadState_Get()->interp;

	/* the initial thread needs one, too. */
	create ("initial");
		
	/* this is the thread that will be disposing of released references. */
	TrashThread::the_one = new TrashThread();
}

/* cleans up after all threads have exited. */
void
PythonThread::exit()
{
	delete TrashThread::the_one;
	TrashThread::the_one = 0;
}

/* attaches a PythonThread to the currently running thread in tls. */
PythonThread *
PythonThread::create (const char * name)
{
	PythonThread * t = new PythonThread (name);
	
#if DEBUG_PY_THREADS
	msg(), "PyT::create:", name, t;
#endif

	if (pthread_setspecific (key, t))
		throw RuntimeError (errno) << "pthread_setspecific failed.";
	
	return t;
}

/* returns the PythonThread belonging to the current thread. */
PythonThread *
PythonThread::get()
{
	return (PythonThread *) pthread_getspecific (key);
}

/* called when a thread exits to prepare cleaning up the PythonThread. */
void
PythonThread::deactivate()
{
	PythonThread * fifo = get();
	pthread_setspecific (key, 0);

	fifo->active = 0;
}

/* static PL acquiring */
void
PythonThread::lock()
{
	PythonThread * current = get();

#if DEBUG_PY_THREADS
	msg(), "PyT::lock:", current;
#endif
	
	/* if a new thread was started in python, we need to attach: */
	if (current == 0)
		current = PythonThread::create ("extern");

	PyEval_AcquireLock();

	executing = current;

	current->tstate_h = PyThreadState_Swap (current->tstate);
}

void
PythonThread::unlock()
{
	PythonThread * current = get();

	executing = 0;
	
	PyThreadState_Swap (current->tstate_h);
	PyEval_ReleaseLock();
}

/* print exception info >> stdmsg. */

void
PythonThread::trouble (Object * o)
{
	PyObject * type, * value, * trace;
	PyErr_Fetch (&type, &value, &trace);
	
	if (type == 0)
		_msg(), get_name (o), " ", get_type_name (o), " blew an untyped exception.";
	else
	{
		PyObject * s = PyObject_Str (type);
		
		_msg(), get_name (o), " ", get_type_name (o), " blew it: ", 
				PyString_AsString (s);

		::decref (s, true);
	}

	if (trace)
		PyTraceBack_Print (trace, _msg::stdmsg);

	if (value)
	{
		PyObject * s = PyObject_Str (value);

		FILE * f = PyFile_AsFile (_msg::stdmsg);

		PyObject_Print (s, f, 0);
		fputc ('\n', f);
		fflush (f);

		::decref (s, true);
	}

	::decref (type, true);
	::decref (value, true);
	::decref (trace, true);
}

/* //////////////////////////////////////////////////////////////////////// */

/* loop over all PythonThreads and release disposed-of references. */

void
PythonThread::housekeep (bool do_lock)
{
	if (do_lock)
		lock();

	try /* ... to make sure we never blow up holding the PL. */
	{
		ScopedLock lock (*mutex);

		for (int i = 0; i < *all; )
		{
			PythonThread * fifo = (*all) [i];
			
			while (fifo->have())
				::decref (fifo->pop(), true);
			
			if (fifo->active == 0)
				all->remove (fifo);
			else
				++i;
		}
	}
	catch (blewit & b)
	{
		msg(), "housekeeping blew it: ", b.what();
	}

	if (do_lock)
		unlock();
}

/* //////////////////////////////////////////////////////////////////////// */

PythonThread::PythonThread (const char * _name)
: fifo (1022) /* we have room for 1022 objects, this is arbitrary. */
{
	name = _name;
	active = 1;

	tstate = PyThreadState_New (interp);

	all->add_object (this);
}

PythonThread::~PythonThread()
{
	if (have())
		msg(), "warning: ~PythonThread<", name, "> still has ", have(), " objects.";

	all->remove (this);

	PyThreadState_Clear (tstate);
	PyThreadState_Delete (tstate);
}

void
PythonThread::push (Object * o)
{
	/* initial thread always has the PL */
	if (this == (*all)[0])
	{
		::decref (o, true);
		return;
	}

	if (fifo.is_full())
	{
		_msg(), name, "::trash is full: doing unsafe housekeeping.";

		#ifdef DEBUG
		TRAP;
		#endif

		housekeep();
	}

	if (active)
		fifo.push (o);
	else 
	{
		/* this only happens in the initial thread; all other threads are setting
		 * active to 0 only just before they exit, so no more objects can be pushed
		 * then. 
		 */
		::decref (o, true);
	}
	
}

/* /////////////////////////////////////////////////////////////////////// */

class UID
{
	public:
		int pushed_uid;
		
		UID (int uid, const char * what_for)
			{
				pushed_uid = geteuid();
				if (seteuid (uid))
					msg(), "cannot set uid to ", uid, "for ", what_for;
			}

		~UID()
			{
				seteuid (pushed_uid);
			}
};

void
Thread::init()
{
	threads_started = 0;

	PythonThread::init();
}

Thread::Thread (const char * _thread_name)
{
	thread_name = _thread_name;
	thread_id = 0;
	thread_priority = 0;
}

Thread::~Thread()
{
	/* derived objects should call join_thread() themselves. */
}

void
Thread::start_thread (int priority)
{
	struct sched_param sched;
	sched.sched_priority = priority;
	
	pthread_attr_t attr;
	
	int e;
	if ((e = pthread_attr_init (&attr)))
		throw OSError (e) << "in pthread_attr_init.";
	
	if ((e = pthread_attr_setschedpolicy (&attr, SCHED_OTHER)))
		throw OSError (e) << "in pthread_attr_setschedpolicy.";
				
	if ((e = pthread_attr_setschedparam (&attr, &sched)))
		throw OSError (e) << "in pthread_attr_setschedparam." << priority;
	
	pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);
	
	int nope = pthread_create 
		(
			&thread_id, 
			&attr, 
			(void * (*) (void *)) _thread, 
			this
		);

	if (nope)
		throw OSError (errno) << "pthread_create() failed.";

	++threads_started;
}

/* Starts the thread() SCHED_FIFO, if it's not available do without.
 */
bool
Thread::start_thread_rt (int priority)
{
	if (seteuid (0))
	{
		msg(), "cannot become root for RT scheduling.";
		start_thread();
		return 0;
	}

	struct sched_param sched;
	sched.sched_priority = priority;
	
	pthread_attr_t attr;
	
	int e;
	if ((e = pthread_attr_init (&attr)))
		throw OSError (e) << "in pthread_attr_init.";
	
	if ((e = pthread_attr_setschedpolicy (&attr, SCHED_FIFO)))
		throw OSError (e) << "in pthread_attr_setschedpolicy.";
				
	if ((e = pthread_attr_setschedparam (&attr, &sched)))
		throw OSError (e) << "in pthread_attr_setschedparam." << priority;
	
	pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);
	
	int nope = pthread_create (
		&thread_id, &attr, (void * (*) (void *)) _thread, this);
		
	seteuid (getuid());

	if (nope)
	{
		msg(), thread_name, "cannot start RT thread: ", strerror (errno);

		/* oh well. do a normal thread then */
		start_thread (priority);
		return 0;
	}
	
	thread_priority = priority;

	msg(), thread_name, " ", thread_id, "started RT.";

	++threads_started;

	return 1;
}

/* only SCHED_FIFO. */
void 
Thread::set_thread_priority (int priority)
{
	if (priority > 0)
	{
		UID uid (0, thread_name);

		struct sched_param sched;
		sched.sched_priority = priority;
		
		int error = pthread_setschedparam (thread_id, SCHED_FIFO, &sched);
		
		if (error)
			throw OSError (error) << "(setting scheduling parameters.)";

		thread_priority = priority;
	}
	else
	{
		/* seems this is the only way to lower a thread's priority. */
		if (setpriority (PRIO_PROCESS, getpid(), abs (priority)))
			throw OSError (errno) << "(setpriority)";
		msg(), thread_name, " priority changed to ", 
				getpriority (PRIO_PROCESS, getpid());
	}
	
	thread_priority = priority;
}

/* having some pages allocated for the stack makes a RT thread kick off
 * quicker. 
 */ 
 
static void
allocate_stack_pages (int n)
{
	n *= getpagesize() / sizeof (int);

	int dummy [n];

	for (int i = 0; i < n; i += 1024)
		dummy [i] = i;
}

void *
Thread::_thread (Thread * self)
{
	allocate_stack_pages (64);

	_msg(), "pid ", getpid(), "is thread ", self->thread_name;

	/* need to block (at least) SIGINT and rely on python to shut us down. */
	SigMask(SIGINT).block();

	try 
	{
		PythonThread::create (self->thread_name);
	}
	catch (blewit & b)
	{
		_msg(), self->thread_name, " could not create a PythonThread: ", b.what();
	
		self->thread_id = 0;
		--threads_started;

		return 0;
	}
	
#if 0 
	/* don't catch thread exceptions (only makes sense if running under gdb) */
	self->thread();
#else
	try 
	{
		self->thread();
		_msg(), self->thread_name, " is gone.";
	} 
	catch (blewit & b) 
	{
		_msg(), self->thread_name, " blew it: ", b.what();
	}
#endif

	PythonThread::deactivate();

	self->thread_id = 0;
	--threads_started;

	return 0;
}

/* making signal masks a little more readable. */

SigMask::SigMask (int signo)
{
	sigemptyset (&mask);

	/* there's no signal 0, is there? */
	if (signo >= 0)
		*this << signo;
}

SigMask &
SigMask::operator << (int signo)
{
	sigaddset (&mask, signo);
	return *this;
}

SigMask &
SigMask::operator >> (int signo)
{
	sigdelset (&mask, signo);
	return *this;
}

void
SigMask::set()
{
	sigset_t oldmask;

	int err = pthread_sigmask (SIG_SETMASK, &mask, &oldmask);

	if (err)
		throw blewit (err) << "pthread_sigmask failed.";
}

void
SigMask::block()
{
	sigset_t oldmask;

	int err = pthread_sigmask (SIG_BLOCK, &mask, &oldmask);

	if (err == 0) return;
	
	throw blewit (err) << "pthread_sigmask (SIG_BLOCK) failed.";
}

void
SigMask::unblock()
{
	sigset_t oldmask;

	int err = pthread_sigmask (SIG_UNBLOCK, &mask, &oldmask);

	if (err == 0) return;
	
	throw blewit (err) << "pthread_sigmask (SIG_UNBLOCK) failed.";
}

void
SigMask::get()
{
	sigemptyset (&mask);
	int err = pthread_sigmask (SIG_BLOCK, &mask, &mask);

	if (err == 0) return;
	
	throw blewit (err) << "pthread_sigmask failed.";
}

/* Just a pthread_mutex class wrapper. */
Mutex::Mutex (int kind)
{
	#ifdef DEBUG
	count = 0;
	#endif
	#if DBG_MUTEX
	msg(), "<mutex ", this, ">";
	#endif

	pthread_mutexattr_t attr;
	
	pthread_mutexattr_init (&attr);
	
	switch (kind)
	{
		case Fast :
			pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_DEFAULT);
			break;
		case Checking :
			pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_ERRORCHECK);
			break;
		case Recursive :
		default :
			pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_RECURSIVE);
			break;
	}

	pthread_mutex_init (&mutex, &attr);
}

Mutex::~Mutex()
{
	#if DBG_MUTEX
	msg(), "<~mutex ", this, ">";
	#endif
	
	pthread_mutex_destroy (&mutex);
}

void
Mutex::lock()
{
	#if DBG_MUTEX
	if (++count >= 0)
		msg(), "<++lock", count, " ", this, ">";
	#endif
	
	pthread_mutex_lock (&mutex);
}

bool
Mutex::trylock()
{
	if (pthread_mutex_trylock (&mutex) == EBUSY)
		return false;
	
	#if DBG_MUTEX
	if (++count >= 0)
		msg(), "<++lock", count, " ", this, ">";
	#endif
	
	return true;
}

void
Mutex::unlock()
{
	#if DBG_MUTEX
	if (count-- >= 0)
		msg(), "<--lock", count, " ", this, ">";
	#endif
	
	pthread_mutex_unlock (&mutex);
}

/* Condition */

Condition::Condition()
: Mutex (Fast)
{
	pthread_cond_init (&cond, NULL);
}

Condition::~Condition()
{
	pthread_cond_destroy (&cond);
}

void
Condition::signal()
{
	pthread_cond_signal (&cond);
}

void
Condition::wait()
{
	pthread_cond_wait (&cond, &mutex);
}

bool
Condition::timed_wait (d_time seconds)
{
	struct timeval tv;
	gettimeofday (&tv, NULL);
	
	seconds += (d_time) tv.tv_sec + (d_time) tv.tv_usec / TEN_TO_THE_SIXTH;
		
	struct timespec ts;
	ts.tv_sec = (long) seconds;
	ts.tv_nsec = (long) ((seconds - (d_time) ts.tv_sec) * TEN_TO_THE_NINTH);

	int r;
	do {
		r = pthread_cond_timedwait (&cond, &mutex, &ts);
	} while (r == EINTR);

	return (r != ETIMEDOUT);
}

/*
 * this code is adapted to c++ from the pthread-faq file:
 * 
 * $Id: rwlock.c,v 1.9 1999/02/27 14:19:35 lk Exp $
 *
 * Routines to implement a read-write lock. Multiple readers or one writer
 * can hold the lock at once. Writers are given priority over readers.
 *
 * Copyright (C) 1998-99 Lee Kindness <leekindness@iname.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

RWLock::RWLock()
{
	pthread_mutex_init (&lock, NULL);
	pthread_cond_init (&wcond, NULL);
	pthread_cond_init (&rcond, NULL);
	
	lock_count = 0;
	waiting_writers = 0;
}

RWLock::~RWLock()
{
	pthread_mutex_destroy (&lock);
	pthread_cond_destroy (&wcond);
	pthread_cond_destroy (&rcond);
}

/*
 * rwlock_ReadLock()
 *
 * Obtain a read lock.
 */

void 
RWLock::lock_read()
{
    pthread_mutex_lock (&lock);
    pthread_cleanup_push ((void (*)(void*)) waiting_reader_cleanup, this);
    while ((lock_count < 0) && (waiting_writers))
			pthread_cond_wait (&rcond, &lock);
    lock_count++;
    /* Note that the pthread_cleanup_pop subroutine will
     * execute the rwlocki_WaitingReaderCleanup routine */
    pthread_cleanup_pop (1);
}

void 
RWLock::unlock_read()
{
	pthread_mutex_lock (&lock);
	lock_count--;
	if (!lock_count)
		pthread_cond_signal (&wcond);
	pthread_mutex_unlock(&lock);
}

void
RWLock::lock_write()
{
	pthread_mutex_lock (&lock);
	waiting_writers++;
	pthread_cleanup_push ((void (*)(void*)) waiting_writer_cleanup, this);
	while (lock_count)
		pthread_cond_wait (&wcond, &lock);
	lock_count = -1;
	/* Note that the pthread_cleanup_pop subroutine will
	 * execute the rwlocki_WaitingWriterCleanup routine */
	pthread_cleanup_pop(1);
}
	
void 
RWLock::unlock_write()
{
	pthread_mutex_lock (&lock);
	lock_count = 0;
	if (!waiting_writers)
		pthread_cond_broadcast (&rcond);
	else
		pthread_cond_signal (&wcond);
	pthread_mutex_unlock (&lock);
}

void 
RWLock::waiting_reader_cleanup (RWLock * self)
{
	pthread_mutex_unlock (&self->lock);
}

void 
RWLock::waiting_writer_cleanup (RWLock * self)
{
	self->waiting_writers--;

	if ((!self->waiting_writers) && (self->lock_count >= 0))
 		/*  This only happens if we have been cancelled */
		pthread_cond_broadcast (&self->wcond);
	
	pthread_mutex_unlock (&self->lock);
}


