/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#ifndef _THREAD_H_
#define _THREAD_H_

#include <pthread.h>
#include <signal.h>

class PythonThread;

class Mutex
{
	public:
		enum {
			Fast = 'f',
			Checking = 'c',
			Recursive = 'r'
		};
		
		pthread_mutex_t	mutex;

		#ifdef DEBUG
		int count;
		#endif

		Mutex (int kind = Fast);
		virtual ~Mutex();
						
		bool trylock();
		void lock();
		void unlock();
};

class Thread
{
	private:
		static void	* _thread (Thread * self); /* calls thread() */
		static pthread_key_t trash_key;

	protected:
		virtual void * thread() = 0;

	public:
		static atomic	threads_started;
		const char * thread_name;
		int thread_priority;

		pthread_t thread_id;
		
		/* one-time initialisation. */
		static void init();

		Thread (const char * thread_name);
		virtual ~Thread();
									
		/* will start a SCHED_OTHER thread if uid 0 privileges are not available. 
		 * in this case, 0 is returned; 1 if realtime. */
		bool start_thread_rt (int priority);
		void start_thread (int priority = 0);
		
		int join_thread() 
			{
				void * r;
				
				return pthread_join (thread_id, &r);
			}

		/* commodity helpers. */
		inline static int max_priority() 
			{ 
				return sched_get_priority_max (SCHED_FIFO); 
			}

		inline static pthread_t	self() 
			{
				return pthread_self();
			}

		void set_thread_priority (int priority);
};

class SigMask
{
	private:
		sigset_t mask;

	public:
		SigMask (int signo = -1);
		
		SigMask & operator << (int signo);
		SigMask & operator >> (int signo);

		void set();
		void block();
		void unblock();
		void get();
};

/* handy, takes care of unlocking on all code paths including exceptions. */
class ScopedLock
{
	public:
		Mutex * mutex;
		bool locked;
		
		ScopedLock (Mutex & m)
			{
				mutex = &m;
				mutex->lock();
				locked = 1;
			}

		~ScopedLock()
			{
				if (locked) 
					unlock();
			}

		void unlock()
			{
				mutex->unlock();
				locked = 0;
			}
};

class ScopedUnlock
{
	public:
		Mutex * mutex;
		bool unlocked;
		
		ScopedUnlock (Mutex & m)
			{
				mutex = &m;
				mutex->unlock();
				unlocked = 1;
			}

		~ScopedUnlock()
			{
				if (unlocked) 
					lock();
			}

		void lock()
			{
				mutex->lock();
				unlocked = 0;
			}
};

class ScopedTryLock
{
	public:
		Mutex * mutex;
		bool locked;
		
		ScopedTryLock (Mutex & m)
			{
				mutex = &m;
				locked = mutex->trylock();
			}

		~ScopedTryLock()
			{
				if (locked) 
					unlock();
			}

		void unlock()
			{
				mutex->unlock();
				locked = 0;
			}
};

class Condition
: public Mutex
{
	public:
		pthread_cond_t	cond;

		Condition();
		~Condition();
						
		void signal();
		void wait();
		
		bool timed_wait (d_time seconds);
};

/*
 * this code is adapted to c++ from the pthread-faq file:
 * 
 * $Id: rwlock.c,v 1.9 1999/02/27 14:19:35 lk Exp $
 *
 * Routines to implement a read-write lock. Multiple readers or one writer
 * can hold the lock at once. Writers are given priority over readers.
 * When compiled with RWLOCK_DEBUG defined verbose debugging output is
 * produced which can help track problems such as mismatches and
 * recursive locks.
 *
 * Copyright (C) 1998-99 Lee Kindness <leekindness@iname.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

class RWLock
{
	public:
    pthread_mutex_t lock;
    pthread_cond_t  rcond;
    pthread_cond_t  wcond;
    int             lock_count;
    int             waiting_writers;

										RWLock();
									 ~RWLock();

		void						lock_read();
		void						unlock_read();

		void						lock_write();
		void						unlock_write();

		static void			waiting_reader_cleanup (RWLock *);
		static void			waiting_writer_cleanup (RWLock *);
};

#endif /* _THREAD_H_ */
