/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "Graph.h"
#include "PyType.h"

#include "dsp/SVF.h"
#include "dsp/Sine.h"
#include "dsp/Lorenz.h"

typedef DSP::SVF<2>	SVF;

/*
	source: musicdsp, a lovely filter by paul kellett.
 */
/*
	Moog VCF, variation 1

	Type : 24db resonant lowpass
	References :CSound source code, Stilson/Smith CCRMA paper., Paul Kellett
	version

	Notes : 
	The second "q =" line previously used exp() - I'm not sure if what I've done 
	is any faster, but this line needs playing with anyway as it controls which
	frequencies will self-oscillate. I think it could be tweaked to sound better 
	than it currently does. 

	Highpass / Bandpass : 

	They are only 6dB/oct, but still seem musically useful - the 'fruity' sound 
	of the 24dB/oct lowpass is retained.

	Code : 
	// Moog 24 dB/oct resonant lowpass VCF 
	// References: CSound source code, Stilson/Smith CCRMA paper. 
	// Modified by paul.kellett@maxim.abel.co.uk July 2000 

	float f, p, q;             //filter coefficients 
	float b0, b1, b2, b3, b4;  //filter buffers (beware denormals!) 
	float t1, t2;              //temporary buffers 

	// Set coefficients given f & resonance [0.0...1.0] 

	q = 1.0f - f; 
	p = f + 0.8f * f * q; 
	f = p + p - 1.0f; 
	q = resonance * (1.0f + 0.5f * q * (1.0f - q + 5.6f * q * q)); 

	// Filter (in [-1.0...+1.0]) 

	in -= q * b4;                          //feedback 
	t1 = b1;  b1 = (in + b0) * p - b1 * f; 
	t2 = b2;  b2 = (b1 + t1) * p - b2 * f; 
	t1 = b3;  b3 = (b2 + t2) * p - b3 * f; 
	b4 = (b3 + t1) * p - b4 * f; 
	b4 = b4 - b4 * b4 * b4 * 0.166667f;    //clipping 
	b0 = in; 

	// Lowpass  output:  b4 
	// Highpass output:  in - b4; 
	// Bandpass output:  3.0f * (b3 - b4);
 */
 
static Type * VCF_type;

class VCF
: public Plugin
{
	public:
		d_float f, resonance;
	
		/* history */
		d_sample b[5];
	
		VCF();

		virtual void process_audio (int frames);
	
	private:
		inline void one_cycle (d_sample * s, int frames);
};

/* /////////////////////////////////////////////////////////////////////// */

VCF::VCF()
: Plugin (VCF_type, Audio)
{
	/* initial f is the f of NOTE (= 'Eb4') */
	f = n2f (NOTE);
	resonance = 0.2;

	/* clean the IIR history. */
	memset (b, 0, sizeof (b));

	/* allow up to 16 incoming audio streams; they are mixed. */
	add_in (new In (this, CyclicAudio_type, 16, "input"));

	add_in (new In (this, FloatEvent_type, 1, "f"));
	add_in (new In (this, FloatEvent_type, 1, "resonance"));

	add_out (new Out (this, CyclicAudio_type, "output"));
}

void 
VCF::process_audio (int frames)
{
	CyclicAudio * event = ins[0]->pop_mix();
	
	if (!event)
		return;

	d_sample * s = event->data;

	d_frame ff = graph->frame;

	while (frames > 0)
	{
		/* subcycle in 64 frame units. */
		int n = min (64, frames);

		/* the graph must be queried again at every subcycle because ticks do not 
		 * advance steadily, since the tick rate can be altered through a tempo
		 * change right in the middle of an audio cycle. */
		d_tick tick = graph->frame2tick (ff + n);

		/* update the parameters. */
		ins[1]->pop_value (tick, &f);
		ins[2]->pop_value (tick, &resonance);

		/* do the filter work. */
		one_cycle (s, n);

		/* update the counters. */
		frames -= n;
		s += n;

		/* loop jump. */
		if (graph->is_loop_frame (ff, ff + n))
			++loop_cycle,
			ff -= graph->loop.frame[1] - graph->loop.frame[0];

		/* update the sample clock. */
		ff += n;
	}

	outs[0]->push (event);
}

inline void 
VCF::one_cycle (d_sample * s, int frames)
{
	double fn, p, q;

	fn = f / (d_float) graph->sample_rate;

	q = 1.0f - fn;
	p = fn + 0.8f * fn * q; 
	fn = p + p - 1.0f; 
	q = resonance * (1.0f + 0.5f * q * (1.0f - q + 5.6f * q * q)); 

	d_float t1, t2;

	for (int i = 0; i < frames; ++i)
	{
		register d_sample in = s[i];

		in -= q * b[4];

		t1 = b[1];  b[1] = (in + b[0]) * p - b[1] * fn; 
		t2 = b[2];  b[2] = (b[1] + t1) * p - b[2] * fn; 
		t1 = b[3];  b[3] = (b[2] + t2) * p - b[3] * fn; 
		b[4] = (b[3] + t1) * p - b[4] * fn; 
		b[4] = b[4] - b[4] * b[4] * b[4] * 0.166667f;    /* clipping */
		b[0] = in; 

		s[i] = /* lowpass */
		b[4] = renormal (b[4]);
	}
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
VCF_get_f (VCF * vcf)
{
	return PyFloat (vcf->f);
}

static int
VCF_set_f (VCF * vcf, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vcf->f))
		return -1;
	return 0;
}

static PyObject *
VCF_get_resonance (VCF * vcf)
{
	return PyFloat (vcf->resonance);
}

static int
VCF_set_resonance (VCF * vcf, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vcf->resonance))
		return -1;
	return 0;
}

static PyGetSetDef 
VCF_getsets [] =
{
	{"f", (getter) VCF_get_f, (setter) VCF_set_f, 
	 "cutoff frequency, Hertz."},
	{"resonance", (getter) VCF_get_resonance, (setter) VCF_set_resonance, 
	 "value from 0 .. 1."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<VCF>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = VCF_getsets;
}

char * Py<VCF>::__doc__ = 
"VCF ([f, resonance]) -- Moog VCF, Variation 1, code by Paul "
"Kellett. Lowpass only.";

void
Py<VCF>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"f", "resonance", 0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|dd", keys,
				&f, &resonance))
		return;
}

/* /////////////////////////////////////////////////////////////////////// */

static Type * RandomSVF_type;

class RandomSVF
: public Plugin
{
	public:
		SVF svf;

		double f, Q;
		double t;
	
		int range[2];

		int remain;

		RandomSVF();

		virtual void process_audio (int frames);
		virtual void set_state (Graph * g, int s);
};

/* /////////////////////////////////////////////////////////////////////// */

RandomSVF::RandomSVF()
: Plugin (RandomSVF_type, Audio)
{
	/* initial f is the f of NOTE (= 'Eb4') */
	f = n2f (NOTE);
	Q = 0.94;
	t = 0.2;

	range[0] = 50;
	range[1] = 100;

	add_in (new In (this, CyclicAudio_type, 4, "input"));
	add_out (new Out (this, CyclicAudio_type, "output"));
}

void
RandomSVF::set_state (Graph * g, int s)
{
	if (state == Plugin::Prepared)
	{
		svf.set_out (SVF::Band);
		remain = 0;
	}

	Plugin::set_state (g, s);
}

void 
RandomSVF::process_audio (int frames)
{
	CyclicAudio * event = ins[0]->pop_mix();
	
	if (!event)
		return;

	d_sample * s = event->data;

	while (frames > 0)
	{
		if (remain == 0)
		{
			double fs = graph->sample_rate;
			int r = random() % (range[1] - range[0]);
			double f = n2f (range[0] + r) / fs;
			svf.set_f_Q (f, Q);
			remain = (int) (t * graph->sample_rate);
		}

		int n = min (remain, frames);
		
		svf.one_cycle (s, n);
		
		frames -= n;
		remain -= n;
		s += n;
	}

	outs[0]->push (event);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
RandomSVF_get_f (RandomSVF * vcf)
{
	return PyFloat (vcf->f);
}

static int
RandomSVF_set_f (RandomSVF * vcf, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vcf->f))
		return -1;
	return 0;
}

static PyObject *
RandomSVF_get_Q (RandomSVF * vcf)
{
	return PyFloat (vcf->Q);
}

static int
RandomSVF_set_Q (RandomSVF * vcf, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vcf->Q))
		return -1;
	return 0;
}

static PyObject *
RandomSVF_get_t (RandomSVF * vcf)
{
	return PyFloat (vcf->t);
}

static int
RandomSVF_set_t (RandomSVF * vcf, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vcf->t))
		return -1;
	return 0;
}

/* range */
static int
RandomSVF_set_range (RandomSVF * v, PyObject * s)
{
	return !PyArg_ParseTuple (s, "ii", v->range, v->range + 1);
}

static PyObject *
RandomSVF_get_range (RandomSVF * v)
{
	return Py_BuildValue ("ii", v->range[0], v->range[1]);
}

static PyGetSetDef 
RandomSVF_getsets [] =
{
	{"f", (getter) RandomSVF_get_f, (setter) RandomSVF_set_f, 
	 "cutoff frequency, Hz."},
	{"t", (getter) RandomSVF_get_t, (setter) RandomSVF_set_t, 
	 "time between filter changes."},
	{"Q", (getter) RandomSVF_get_Q, (setter) RandomSVF_set_Q, 
	 "value from 0 .. 1."},
	{"range", (getter) RandomSVF_get_range, (setter) RandomSVF_set_range, 
	 "2 MIDI note numbers denoting the cutoff frequency range."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<RandomSVF>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = RandomSVF_getsets;
}

char * Py<RandomSVF>::__doc__ = 
"RandomSVF ([f, Q]) -- SVF stepping through random cutoff frequencies.";

void
Py<RandomSVF>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"f", "Q", "t", "range", 0};
	PyObject * r = 0;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|dddO", keys,
				&f, &Q, &t, &r))
		return;

	if (r) RandomSVF_set_range (this, r);
}

/* /////////////////////////////////////////////////////////////////////// */

static Type * SweepingSVF_type;

class SweepingSVF
: public Plugin
{
	public:
		SVF svf;
		DSP::Sine lfo;
		DSP::Lorenz lorenz;

		d_float f, Q;
	
		SweepingSVF();

		virtual void process_audio (int frames);
		virtual void set_state (Graph * g, int s);
};

/* /////////////////////////////////////////////////////////////////////// */

SweepingSVF::SweepingSVF()
: Plugin (SweepingSVF_type, Audio)
{
	f = n2f (NOTE);
	Q = 0.2;

	add_in (new In (this, CyclicAudio_type, 4, "input"));
	add_out (new Out (this, CyclicAudio_type, "output"));
}

void
SweepingSVF::set_state (Graph * g, int s)
{
	lorenz.init (.001, (double) random() / (double) RAND_MAX);
	svf.set_out (SVF::Low);

	if (g)
		lfo.set_f (.3 * 32, g->sample_rate, 0);

	Plugin::set_state (g, s);
}

void 
SweepingSVF::process_audio (int frames)
{
	CyclicAudio * event = ins[0]->pop_mix();
	
	if (!event)
		return;

	d_sample * s = event->data;

	while (frames > 0)
	{
		/* subcycle in 32 frame units. */
		int n = min (32, frames);

		double fc = lorenz.get() * 8000 + 2000 + f;
		svf.set_f_Q (fc / (double) graph->sample_rate, Q);

		svf.one_cycle (s, n);
		
		frames -= n;
		s += n;
	}

	outs[0]->push (event);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
SweepingSVF_get_f (SweepingSVF * vcf)
{
	return PyFloat (vcf->f);
}

static int
SweepingSVF_set_f (SweepingSVF * vcf, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vcf->f))
		return -1;
	return 0;
}

static PyObject *
SweepingSVF_get_Q (SweepingSVF * vcf)
{
	return PyFloat (vcf->Q);
}

static int
SweepingSVF_set_Q (SweepingSVF * vcf, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vcf->Q))
		return -1;
	return 0;
}

static PyGetSetDef 
SweepingSVF_getsets [] =
{
	{"f", (getter) SweepingSVF_get_f, (setter) SweepingSVF_set_f, 
	 "cutoff frequency, Hz."},
	{"Q", (getter) SweepingSVF_get_Q, (setter) SweepingSVF_set_Q, 
	 "value from 0 .. 1."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<SweepingSVF>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = SweepingSVF_getsets;
}

char * Py<SweepingSVF>::__doc__ = 
"SweepingSVF ([f, Q]) -- LFO-modulated SVF";

void
Py<SweepingSVF>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"f", "Q", 0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|dd", keys,
				&f, &Q))
		return;
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_VCF_types (PyObject * dict)
{
	VCF_type = new PyType<VCF> ("VCF", dict);
	SweepingSVF_type = new PyType<SweepingSVF> ("SweepingSVF", dict);
	RandomSVF_type = new PyType<RandomSVF> ("RandomSVF", dict);
}
