/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include "basics.h"
#include "Graph.h"
#include "PyType.h"

/* more sincs enable higher frequencies. needs to be a power of two unless
 * you change all index clamp code to use '%' instead of '/'. */
#define N_SINCS	16
#define SINC(n) sinc[(n) & (N_SINCS - 1)]

/* number of samples per sinc. if you change this, you will need to
 * recalculate SUM, which is the average sum over the sinc window. */
#define WINDOW 32 
#define SUM -0.994756

#define LEAK 0.999
#define SCALE 0.8

#define WINDOW2 ((WINDOW + 1) / 2)

#define SUM_OVER_SINCS	1

#if SUM_OVER_SINCS
static double sum = 0.0;
static int sincs = 0;
#endif

static double window[WINDOW];

static inline double
blackman (int i)
{
	return 0.42 
		- 0.5 * cos (2 * (double) i * M_PI / WINDOW) 
		+ 0.08 * cos (4 * (double) i * M_PI / WINDOW);
}

class Sinc
{
	public:
		float data[WINDOW];
		double scale;
		int n;
		
		Sinc()
			{
				memset (data, 0, sizeof (data));
				
				n = WINDOW;

				/* alternating signs in the window array means we save work 
				 * in sinc(). */
				for (int i = 0; i < WINDOW; i += 2)
					window[i] = blackman (i);

				for (int i = 1; i < WINDOW; i += 2)
					window[i] = -blackman (i);
			}
	
		/* call instead of sinc ;) */
		void alias (int period, double phase)
			{
				n = -period;

				data[WINDOW2] = SUM;
			}
		
		void sinc (int period, double phase)
			{
				n = -period;

				double 
					x = (-WINDOW2 - phase) * M_PI,
					sine = sin (phase * M_PI);

				int i;
				for (i = 0; i < WINDOW2; ++i)
					data[i] = window[i] * sine / x,
					x += M_PI;

				data[i] = window[i] * (phase ? sine / x : 1.0);
				x += M_PI;

				while (++i < WINDOW)
					data[i] = window[i] * sine / x,
					x += M_PI;

#if SUM_OVER_SINCS
				for (int j = 0; j < WINDOW; ++j)
					sum += data[j];
				sincs++;
#endif
			}

		inline void get (d_sample * s, int frames)
			{
				int minus_n = -n;
				
				if (minus_n > 0)
				{
					if (minus_n >= frames)
					{
						n += frames;
						return;
					}

					frames += n;
					s -= n;
					n = 0;
				}

				if (n < WINDOW)
				{
					int i = min (WINDOW - n, frames);
					frames -= i;

					for (int j = 0; j < i; ++j)
						s[j] += scale * data[n++];
				}

				n += min (frames, WINDOW - n);
			}
	
		inline int left()
			{
				return WINDOW - n;
			}
};

/* /////////////////////////////////////////////////////////////////////// */

static Type * VCO_type;

class VCO
: public Plugin
{
	public:
		int lo, hi, active;
		Sinc sinc[N_SINCS];

		int period;
		double fract, delta, out;

	public:
		double f, shape, pw;

		VCO();

#if SUM_OVER_SINCS
		~VCO()
			{
				fprintf (stderr, "%.12f\n",  sum / sincs);
			}
#endif

		virtual void set_state (Graph * g, int s);
		virtual void process_audio (int frames);

	private:
		inline void one_cycle (d_sample * s, int frames);
		inline void start_period (double sample_rate, double f);
		inline int start_sinc (double samples, double scale);
};

VCO::VCO()
: Plugin (VCO_type, Audio)
{
	f = n2f (60);
	shape = 0.;
	pw = 1.;

	period = 0;
	fract = 0.;

	out = 0.;
	delta = 0.;

	lo = hi = active = 0;

	add_in (new In (this, FloatEvent_type, 1, "f"));
	add_out (new Out (this, CyclicAudio_type, "out"));
}

void
VCO::set_state (Graph * g, int s)
{
	if (s < Plugin::Activated)
	{
		hi = lo = active = 0;
		out = 0.;
	}

	Plugin::set_state (g, s);
}

int
VCO::start_sinc (double samples, double scale)
{
	if (active == N_SINCS)
		throw AssertionError() << "out of sincs.";

	int p = (int) samples;
	double f = samples - p;

	sinc[hi].sinc (p, f);
	sinc[hi].scale = scale * SCALE;

	++active;
	++hi;
	hi &= N_SINCS - 1;

	return p;
}

void 
VCO::start_period (double sample_rate, double _f)
{
	f = _f;
	
	/* maintain active sinc indices */
	while (active && SINC(lo).left() <= 0)
		++lo,
		--active;
	lo &= N_SINCS - 1;

	/* depending on shape, we schedule one or two sinc window(s) for this
	 * period. there's always one at the end of the period, and somewhere
	 * inbetween we have one that turns the triangle into a square (rect)
	 * wave. */
	double samples;

	if (shape)
	{
		samples = fract + pw * sample_rate / (2. * f);

		start_sinc (samples, -shape);
	}

	/* end-period sinc */
	samples = fract + sample_rate / f;
	
	/* aliasing at higher frequencies can be remedied a little by giving
	 * this sinc 1/2 weight and starting another sinc with 1/2 weight at
	 * samples + 1. */
	period = start_sinc (samples, 1.);

	fract = samples - period;

	delta = (shape - 1.) * SCALE * SUM / period;
}

inline void 
VCO::one_cycle (d_sample * s, int frames)
{
	for (int i = 0; i < frames; ++i)
		s[i] = delta;

	for (int j = lo; j < lo + active; ++j)
		SINC(j).get (s, frames);

	for (int i = 0; i < frames; ++i)
		s[i] = out = LEAK * out + s[i];

	period -= frames;
	if (period == 0)
#if 1
		/* unswept */
		start_period (graph->sample_rate, f);
#else
		/* sweep */
		// set_f (graph->sample_rate, f > 20 ? f - 1 : 3400);
		start_period (graph->sample_rate, f < 3000 ? f + .5 : 30);
#endif
}

void 
VCO::process_audio (int frames)
{
	CyclicAudio * event = (CyclicAudio *) outs[0]->cache->pop();
	
	d_sample * s = event->data;

	d_frame ff = event->frame = graph->frame;
	
	while (frames > 0)
	{
		int n = min (period, frames);

		d_tick tick = graph->frame2tick (ff + n);
		ins[0]->pop_value (tick, &f);

		one_cycle (s, n);

		/* update the counters. */
		frames -= n;
		s += n;

		/* loop jump. */
		if (graph->is_loop_frame (ff, ff + n))
			++loop_cycle,
			ff -= graph->loop.frame[1] - graph->loop.frame[0];

		ff += n;
	}

	outs[0]->push (event);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
VCO_get_f (VCO * vco)
{
	return PyFloat (vco->f);
}

static int
VCO_set_f (VCO * vco, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vco->f))
		return -1;
	return 0;
}

static PyObject *
VCO_get_shape (VCO * vco)
{
	return PyFloat (vco->shape);
}

static int
VCO_set_shape (VCO * vco, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vco->shape))
		return -1;
	return 0;
}

static PyObject *
VCO_get_pw (VCO * vco)
{
	return PyFloat (vco->pw);
}

static int
VCO_set_pw (VCO * vco, PyObject * value)
{
	if (!PyArg_Parse (value, "d", &vco->pw))
		return -1;
	return 0;
}

static PyGetSetDef 
VCO_getsets [] =
{
	{"f", (getter) VCO_get_f, (setter) VCO_set_f, 
	 "f, in Hertz."},
	{"shape", (getter) VCO_get_shape, (setter) VCO_set_shape, 
	 "shape (0 = sawtooth .. 1 = square)."},
	{"pw", (getter) VCO_get_pw, (setter) VCO_set_pw, 
	 "pulse width (0 .. 1), if 'shape' != 0."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

void
Py<VCO>::type_init (Type * t, PyObject * dict)
{
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_base = Plugin_type;
	t->tp_getset = VCO_getsets;
}

char * Py<VCO>::__doc__ = 
"VCO ([f = n2f ('c'), shape = 0, pw = 1]) -- value-controlled oscillator.\n"
"the current implementation does not sweep and allows the oscillation "
"shape to be a thing between sawtooth (0) and square (1), with varying "
"pulse width. Basic anti-aliasing is provided; however frequencies beyond "
"about 2 kHz will cause audibly aliased output.";

void
Py<VCO>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"f", "shape", "pw", 0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|ddd", keys,
				&f, &shape, &pw))
		return;
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_VCO_types (PyObject * dict)
{
	VCO_type = new PyType<VCO> ("VCO", dict);
}
