/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include <sys/mman.h>

#include "basics.h"
#include "WaveFile.h"

WaveFile::WaveFile()
: SoundFile (WaveFile_type)
{
	at = 0;
	map = 0;
}

WaveFile::~WaveFile()
{
	if (mode != O_RDONLY && file.fd != -1)
		write_header();
}

void
WaveFile::open (char * path, int _mode, int _channels, int sr, int bits)
{
	mode = _mode;

	file.open (path, mode == O_RDONLY ? mode : (O_RDWR | O_CREAT | O_TRUNC));
	
	if (mode == O_RDONLY)
		read_header();
	else
	{
		channels = _channels;
		sample_rate = sr;

		if (bits == 0)
		{
			format = 3;
			bytes_per_sample = sizeof (d_sample);
		}
		else if (bits != 8 && bits != 16 && bits != 32)
			throw ValueError() << "bit depth must be in (32, 16, 8, 0).";
		else
		{
			format = 1;
			bytes_per_sample = bits / 8;
		}
		
		frame_size = channels * bytes_per_sample;
		write_header();

		data = file.tell();
	}
}

void
WaveFile::read_header()
{
	if (!file.read_four ("RIFF"))
		throw IOError() << get_path() << " is not a RIFF file.";
		
	/* we expect data to be the last chunk and fetch size via tell(). */
	file.read_le<int32>(); 
	
	if (!file.read_four ("WAVE"))
		throw IOError() << get_path() << " is not a WAVE file.";
	
	format = -1;

	/* chunk loop */
	int32 size;

	while (1)
	{
		char sign[4];
		file.read (sign, 4);
		
		size = file.read_le<int32>();
		
		if (!strncmp (sign, "data", 4))
			break;

		if (strncmp (sign, "fmt ", 4))
		{
			file.seek (size, SEEK_CUR);
			continue;
		}

		/* read the format header */
		format = file.read_le<int16>();
		if (format != 1 && format != 3)
			throw IOError() << get_path() << " comes in an unsupported sample format.";
			
		channels = file.read_le<int16>();
		sample_rate = file.read_le<int32>(); 

		/* bytes / second isn't important to us. */
		file.read_le<int32>(); 

		frame_size = file.read_le<int16>();
		
		/* determine bytes per sample */
		int bits_per_sample = file.read_le<int16>(); 
		
		bytes_per_sample = 1;
		while (bits_per_sample > 8)
			++bytes_per_sample,
			bits_per_sample -= 8;

		if (bytes_per_sample > 4 || bytes_per_sample == 3)
			throw IOError() << get_path() << " has an unsupported number of bytes/sample.";

		file.seek (size - 16, SEEK_CUR);
	}

	if (format == -1)
		throw IOError() << get_path() << " has no 'fmt ' before 'data'.";

	data = file.tell();
	frames = size / frame_size;

#if 0
	msg(), get_path(), " has ", frames, "frames @ ", data;
#endif
}

void
WaveFile::write_header()
{
	off_t size = file.seek (0, SEEK_END);
	
	file.seek (0, SEEK_SET);
	
	file.write ("RIFF", 4);
	file.write_le ((int32) size);
	
	file.write ("WAVEfmt ", 8);
	file.write_le ((int32) 16); /* sizeof fmt chunk */

	file.write_le ((int16) format);
	file.write_le ((int16) channels);
	file.write_le ((int32) sample_rate);
	file.write_le ((int32) (sample_rate * bytes_per_sample * channels));
	file.write_le ((int16) frame_size); 
	file.write_le ((int16) (bytes_per_sample * 8));

	file.write ("data", 4);
#if 0
	/* this assumes all the data chunk we've written are frames, which isn't 
	 * true because we mmap more space for pagesize alignment. */
	size -= file.tell() + 4;
	file.write_le ((int32) size);
#else
	file.write_le ((int32) (frames * frame_size));
#endif
}

void
WaveFile::seek (d_frame f)
{
	at = f;
}

d_frame
WaveFile::tell()
{
	return at;
}

void
WaveFile::read (d_sample ** dest, int nframes)
{
	nframes = min ((int) (frames - at), nframes);

	if (format == 3)
		read_format_3 (dest, nframes);
	else
	{
		if (bytes_per_sample == 1)
			read_format_1<byte> (dest, nframes, 0x7F, -1.);
		else if (bytes_per_sample == 2)
			read_format_1<int16> (dest, nframes, 0x7FFF);
		else if (bytes_per_sample == 4)
			read_format_1<int32> (dest, nframes, 0x7FFFFFFF);
	}

	at += nframes;
}

void
WaveFile::write (d_sample ** src, int nframes)
{
	if (format == 3)
		write_format_3 (src, nframes);
	else
	{
		if (bytes_per_sample == 1)
			write_format_1<byte> (src, nframes, 0x7F, -1.);
		else if (bytes_per_sample == 2)
			write_format_1<int16> (src, nframes, 0x7FFF);
		else if (bytes_per_sample == 4)
			write_format_1<int32> (src, nframes, 0x7FFFFFFF);
	}

	at += nframes;

	if (at > frames)
		frames = at;
}

/* file read routines
 * 
 * read format 3 (little-endian IEEE float) samples 
 */
void
WaveFile::read_format_3 (d_sample ** dest, int frames)
{
	d_sample * map = (d_sample *) map_frames (frames);

	for (int channel = 0; channel < channels; ++channel)
	{
		d_sample * s = map + channel;
		d_sample * d = dest[channel];

		for (int i = 0; i < frames; ++i)
			*d++ = le2host (*s),
			s += channels;
	}

	unmap_frames();
}

/* read format 1 (little-endian signed integer) samples */
template <class T>
void
WaveFile::read_format_1 (d_sample ** dest, int frames, d_sample factor, d_sample add)
{
	T * map = (T *) map_frames (frames);

	for (int channel = 0; channel < channels; ++channel)
	{
		T * s = map + channel;
		d_sample * d = dest[channel];

		for (int i = 0; i < frames; ++i)
			*d++ = add + ((d_sample) le2host (*s)) / factor,
			s += channels;
	}

	unmap_frames();
}

/* file write routines
 * 
 * write format 3 (little-endian IEEE float) samples 
 */
void
WaveFile::write_format_3 (d_sample ** src, int frames)
{
	d_sample * map = (d_sample *) map_frames (frames, true);

	for (int channel = 0; channel < channels; ++channel)
	{
		d_sample * d = map + channel;
		d_sample * s = src [channel];

		for (int i = 0; i < frames; ++i)
			*d = host2le (*s++),
			d += channels;
	}

	unmap_frames();
}

/* write format 1 (little-endian signed integer) samples */
template <class T>
void
WaveFile::write_format_1 (d_sample ** src, int frames, d_sample factor, d_sample sub)
{
	T * map = (T *) map_frames (frames, true);

	for (int channel = 0; channel < channels; ++channel)
	{
		T * d = map + channel;
		d_sample * s = src [channel];

		for (int i = 0; i < frames; ++i)
			*d = host2le ((T) ((*s++ - sub) * factor)),
			d += channels;
	}

	unmap_frames();
}

/* memory-map frames at the 'file pointer' */
void *
WaveFile::map_frames (int frames, bool touch)
{
	int prot = PROT_READ, flags = MAP_PRIVATE;
	
	if (mode != O_RDONLY)
			prot |= PROT_WRITE,
			flags = MAP_SHARED;
			
	/* mmap wants file offsets to be multiples of page_size, so we map
	 * additional space around the real window that is never accessed.
	 */
	size_t pagesize = getpagesize();

	off_t map_at = (frame_size * at) + data;
	off_t offset = map_at % pagesize;
	map_at -= offset;

	map_length = frame_size * frames + offset;
	map_length += pagesize - (map_length % pagesize);

	if (mode != O_RDONLY && touch)
	{
		file.seek (map_at + map_length - 1, SEEK_SET);
		file.write (this, 1);
	}

	map = mmap (
			0, 
			map_length,
			prot, flags, 
			file.fd, 
			map_at);
	
	if (map == MAP_FAILED)
		throw IOError (errno) << "mmapping " << this;
	
	return ((char *) map) + offset;
}

/* unmap it. */
void
WaveFile::unmap_frames()
{
	if (map == 0)
		return;
	
	if (munmap (map, map_length))
		throw IOError (errno) << "munmapping " << this;
	
	map = 0;
}

