/*
	dsp/ComplexIIR.h
	
	Copyright 2003-4 tim goetze <tim@quitte.de>

	http://quitte.de/dsp/

	code for calculating Chebyshev and Butterworth filter prototypes.
	rewrite of Dr Anthony J. Fisher's mkfilter routines:
 	
	www-users.cs.york.ac.uk/~fisher/software/mkfilter 

	more code inspiration from numerical recipes and maarten de boer's polarbear 
*/
/*
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA or point your web browser to http://www.gnu.org.
*/

#ifndef _DSP_COMPLEX_IIR_H_
#define _DSP_COMPLEX_IIR_H_

#include "complex.h"

/* max order */
#define N 42

namespace DSP {

static inline complex blt (complex z)
{
	return (complex (2) + z) / (complex (2) - z);
}

static void
find_coefs (double * ab, complex * roots, int n)
{
	complex c[n + 1];

	/* nr part */
	c[n] = -roots [0];
	c[n - 1] = complex (1, 0);
	
	for (int j = 1; j < n; ++j)
	{
		c[n - j - 1] = complex (1, 0);

		for (int i = n - j; i < n; ++i)
			c[i] = c[i + 1] - (roots[j] * c[i]);

		c[n] = -(roots [j] * c [n]);
	}

	/* extract real to coefficient */
	for (int i = 0; i <= n; ++i)
		ab[i] = c[i].re;
}

class PZs
{
	public:
		enum {
			MaxPoles = N,
			MaxZeros = N
		};

		complex poles [N], zeros [N];
		int npoles, nzeros;

		PZs()
			{ 
				clear();
			}

		void clear()
			{
				npoles = nzeros = 0; 
			}

		void add_pole (complex z)
			{
				poles [npoles++] = z;
			}

		void add_zero (complex z)
			{
				zeros [nzeros++] = z;
			}

		void dump_poles (char * s)
			{
				for (int i = 0; i < npoles; ++i)
					msg(), s, i, poles[i].re, "j ", poles[i].im;
			}

		void dump_zeros (char * s)
			{
				for (int i = 0; i < nzeros; ++i)
					msg(), s, i, zeros[i].re, "j ", zeros[i].im;
			}
	
		int compute_ab (double * a, double * b, int n)
			{
				int nn = max (npoles, nzeros) + 1;
				if (n < nn)
					throw AssertionError() << "too many poles/zeros.";

				if (nzeros)
					find_coefs (a, zeros, nzeros);
				else
					a[0] = 1;
				for (int i = nzeros + 1; i < nn; ++i)
					a[i] = 0;

				b[0] = 0;
				find_coefs (b, poles, npoles);
				for (int i = npoles + 1; i < nn; ++i)
					b[i] = 0;

				return nn;
			}

		/* template so we can choose from float or double. */
		template <class T>
		void compute_amplitude_response (int n, T * data)
			{
				double d = M_PI / (double) n;

				for (int i = 0; i < n; ++i)
				{
					double num = 1, den = 1;
					polar_complex omega (1, i * d);

					for (int j = 0; j < nzeros; ++j)
					{
						complex c = zeros[j] - omega;
						num *= c.abs();
					}

					for (int j = 0; j < npoles; ++j)
					{
						complex c = poles[j] - omega;
						den *= c.abs();
					}

					data[i] = (T) (num / den);
				}
			}
};

class ComplexIIR
{
	public:
		PZs splane, zplane;

		complex top [N + 1], bot [N + 1];
			
		double a[N], b[N];

	public:
		ComplexIIR()
			{ 
				for (int i = 0; i < N; ++i)
					a[i] = b[i] = 0;
			}

		void compute_ab()
			{
				// splane.dump_poles ("s-plane pole ");
				// splane.dump_zeros ("s-plane zero ");
				compute_z();
				// zplane.dump_poles ("z-plane pole ");
				// zplane.dump_zeros ("z-plane zero ");
				expandpoly();
			}
	
		void compute_z()
			{
				for (int i = 0; i < splane.npoles; ++i)
					zplane.add_pole (blt (splane.poles[i]));
				
				for (int i = 0; i < splane.nzeros; ++i)
					zplane.add_zero (blt (splane.zeros[i]));

				while (zplane.nzeros < zplane.npoles)
					zplane.zeros[zplane.nzeros++] = -1;
			}
		
		void expandpoly()
			{
				expand (zplane.zeros, zplane.nzeros, top);
				expand (zplane.poles, zplane.npoles, bot);

				for (int i = 0; i <= zplane.nzeros; ++i)
					a[i] = +(top[i].re / bot[zplane.npoles].re);

				for (int i = 0; i <= zplane.npoles; ++i)
					b[i] = -(bot[i].re / bot[zplane.npoles].re);
			}
		
		void expand (complex z[], int n, complex c[])
			{
				c[0] = 1.;
				for (int i = 0; i < n; ++i)
					c[i + 1] = 0;
				for (int i = 0; i < n; ++i)
					multin (z[i], n, c);

				for (int i = 0; i < n + 1; ++i)
					if (fabs (c[i].im) > 1e-10)
						throw AssertionError() << "coefficient z^" << -i << " not real.";
			}

		void multin (complex w, int n, complex c[])
			{
				complex mw = -w;
				for (int i = n; i >= 1; --i)
					c[i] = mw * c[i] + c[i - 1];
				c[0] = mw * c[0];
			}

		complex evaluate (complex top[], int nz, complex bot[], int np, complex z)
			{
				return eval (top, nz, z) / eval (bot, np, z);
			}

		complex eval (complex c[], int n, complex z)
			{
				complex sum (0);

				for (int i = n; i >= 0; --i)
					sum = sum * z + c[i];

				return sum;
			}
};

class ChebyshevIIR
: public ComplexIIR
{
	public:
		double ripple, alpha;
		int order, hp;

		struct {
			complex dc, fc, hf;
		} gain;

	public:
		ChebyshevIIR (double _alpha, int _hp, double _ripple, int _order)
			{ 
				order = _order;
				hp = _hp;
				ripple = _ripple;

				/* prewarp */
				alpha = tan (M_PI * _alpha) / M_PI;

				compute_s();
				normalize();
				compute_ab();
				compute_gain();
				msg(), "gain.dc=", hypot (gain.dc);
				msg(), "gain.fc=", hypot (gain.fc);
				msg(), "gain.hf=", hypot (gain.hf);
			}

		void compute_s()
			{
				for (int i = 0; i < 2 * order; ++i)
				{
					double theta = ((order & 1) ? (i * M_PI) : ((i + .5) * M_PI)) / order;
					complex z = expj (theta);
					if (z.re < 0)
						splane.add_pole (z);
				}
				
				double r = pow (10, ripple * .1);
				double eps = sqrt (r - 1);
				double y = asinh (1 / eps) / (double) order;

				if (y <= 0)
					throw AssertionError() << "y < 0";

				double sh = sinh (y);
				double ch = cosh (y);
				
				for (int i = 0; i < splane.npoles; ++i)
				{
					splane.poles[i].re *= sh;
					splane.poles[i].im *= ch;
				}
			}

		void normalize()
			{
				double w1 = 2 * M_PI * alpha;

				if (hp)
				{
					for (int i = 0; i < splane.npoles; ++i)
						splane.poles[i] = w1 / splane.poles[i],
						splane.zeros[i] = 0;
					splane.nzeros = splane.npoles;
				}
				else
				{
					for (int i = 0; i < splane.npoles; ++i)
						splane.poles[i] = splane.poles[i] * w1;
				}
			}
			
			void compute_gain()
			{
				gain.dc = evaluate (top, zplane.nzeros, bot, zplane.npoles, 1);
				gain.hf = evaluate (top, zplane.nzeros, bot, zplane.npoles, -1);

				double theta = M_PI * (2 * alpha);
				gain.fc = evaluate (top, zplane.nzeros, bot, zplane.npoles, expj (theta));
			}
};	

} /* namespace DSP */

#undef N

#endif /* _DSP_COMPLEX_IIR_H_ */
