/*
	dsp/Granular.h
	
	Copyright 2003-4 Tim Goetze <tim@quitte.de>
	
	http://quitte.de/dsp/

	simple synchronous grain generator, no pitch scaling or other tricks.
*/
/*
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA or point your web browser to http://www.gnu.org.
*/

#ifndef _DSP_GRANULAR_H_
#define _DSP_GRANULAR_H_

#include "Sine.h"

namespace DSP {
	
class Grain
{
	public:
		d_sample * src;
		d_sample gain;

		/* squared cos over [pi, 2 pi] for the fade envelope */
		DSP::Sine fade; 
		
		int frame;
		int state;
		enum {
			Off = 0,
			Attack = 1,
			Body = 2,
			Release = 3,
		};
		
		int waypoint [5];

		Grain()
			{
				state = Off;
				frame = 0;
			}

		inline void start 
				(d_sample * s, int fade_length, int body_length, d_sample _gain);
		inline void get (d_sample * s, int frames);
		
	private:
		inline int get_body (d_sample * d, int frames);
		inline int get_faded (d_sample * d, int frames);
};

class Granular
{
	public:
		d_sample * wave;
		int size;

		double jitter;
		d_sample gain, gain_jitter; 
		double fs;

		int fade, body;
		
		enum {
			GRAINS = 6,
		};
		
		Grain grains [GRAINS];

		int current;
		int step, remain;
		
		Granular()
			{
				gain = 1;
				gain_jitter = 0;
				current = 0;
				remain = 0;
				fade = 4000;
				body = 0;
				step = 8000;
			}

		inline void init (double _fs, d_sample * _wave, int _size);

		inline void start_grain();
		inline void one_cycle (d_sample * s, int frames);
};

/* /////////////////////////////////////////////////////////////////////// */

inline void 
Grain::start (d_sample * s, int fade_length, int body_length, d_sample _gain)
{
	src = s;
	state = Attack;
	frame = 0;
	
	/* quarter cycle / fade_length, start half-cycle */
	fade.set_f (.5, fade_length, M_PI);

	waypoint [Attack] = fade_length;
	waypoint [Body] = waypoint [Attack] + body_length;
	waypoint [Release] = waypoint [Body] + fade_length;

	gain = _gain;
}
		
inline void 
Grain::get (d_sample * s, int frames)
{
	while (state && frames)
	{
		int n = 0;

		if (state == Attack)
			n = get_faded (s, frames);
		else if (state == Body)
			n = get_body (s, frames);
		else if (state == Release)
			n = get_faded (s, frames);

		frames -= n;
		s += n;
	}
}

inline int 
Grain::get_body (d_sample * d, int frames)
{
	int to = frame + frames;

	if (to >= waypoint[Body])
	{
		state = Release;
		to = waypoint[Body];
	}

	int n = to - frame;

	register d_sample * s = src + frame;

	for (int i = 0; i < n; ++i)
		d[i] += gain * s[i];

	frame += n;
	return n;
}

inline int 
Grain::get_faded (d_sample * d, int frames)
{
	int to = frame + frames;

	if (to >= waypoint[state])
	{
		to = waypoint[state];
		state = (state + 1) % (Release + 1);
	}

	int n = to - frame;

	register d_sample * s = src + frame;

	for (int i = 0; i < n; ++i)
	{
		register double f = fade.get();
		d[i] += gain * f * s[i];
	}
	
	frame += n;
	return n;
}

/* /////////////////////////////////////////////////////////////////////// */

inline void
Granular::init (double _fs, d_sample * _wave, int _size)
{
	fs = _fs;
	wave = _wave;
	size = _size;
}

inline void
Granular::start_grain()
{
	/* calculate possible room for jitter */
	double n = size - 2 * fade - body - 1;
	
	/* not enough data for even 1 grain, bail out */
	if (n < 0) return;

	/* offset by positional jitter */
	n *= jitter;
	int offset = lrint ((n * (double) random()) / (double) RAND_MAX);

	/* modulate gain according to jitter */
	d_sample g = gain - gain * gain_jitter * frandom();

	/* go. */
	grains[current].start (wave + offset, fade, body, g);

	current = (current + 1) % GRAINS;
}

inline void
Granular::one_cycle (d_sample * s, int frames)
{
	while (frames)
	{
		if (remain == 0)
		{
			start_grain();
			remain = step;
		}

		int n = min (frames, remain);

		for (int i = 0; i < GRAINS; ++i)
			grains[i].get (s, n);

		s += n;
		frames -= n;
		remain -= n;
	}
}

} /* ~namespace DSP */

#endif /* _DSP_GRANULAR_H_ */

