/* Copyright 2001-3 tim goetze <tim@quitte.de> -- see 'COPYING'. */

#ifndef _DSP_NOISE_H_
#define _DSP_NOISE_H_

namespace DSP {
	
/* pink noise generation credits to phil burk. 

	patest_pink.c

	Generate Pink Noise using Gardner method.
	Optimization suggested by James McCartney uses a tree
	to select which random value to replace.

    x x x x x x x x x x x x x x x x 
     x   x   x   x   x   x   x   x   
       x       x       x       x       
           x               x               
                   x                               

    Tree is generated by counting trailing zeros in an increasing index.
	When the index is zero, no random number is selected.

*/

#define PINK_MAX_RANDOM_ROWS   (30)
#define PINK_RANDOM_BITS       (24)
#define PINK_RANDOM_SHIFT      ((sizeof (long) * 8) - PINK_RANDOM_BITS)

class PinkNoise
{
	public:
		long rows [PINK_MAX_RANDOM_ROWS];
		long sum;   /* Used to optimize summing of generators. */
		int index;        /* Incremented each sample. */
		int mask;    /* Index wrapped by ANDing with this mask. */
		float scalar;       /* Used to scale within range of -1.0 to +1.0 */

		int n_rows;

		uint32 seed;
		

		PinkNoise (int r = 12)
			{
				seed = 2222;
				init (r);
			}

		inline uint32 random()
			{
				return seed = (seed * 196314165) + 907633515;
			}

		void init (int r)
			{
				n_rows = r;

				index = 0;
				mask = (1 << n_rows) - 1;
				
				/* Calculate maximum possible signed random value. 
				 * Extra 1 for white noise always added. */
				long pmax = (n_rows + 1) * (1 << (PINK_RANDOM_BITS - 1));
				
				scalar = 1. / pmax;
				
				for (int i = 0; i < n_rows; ++i) 
					rows[i] = 0;
				
				sum = 0;
			}

		inline d_sample
		get_white()
		{
			return 1. - (double) random() / (UINT_MAX / 2.);
		}
		
		inline d_sample
		get_pink()
		{
			/* Increment and mask index. */
			index = (index + 1) & mask;

			/* If index is zero, don't update any random values. */
			if (index)
			{
				/* Determine how many trailing zeros in PinkIndex. */
				/* The algorithm may hang if n==0 so test first. */
				int num0s = 0;

				/* Added gcc inline assembly is by me (tim). Does *not* improve
				 * performance on my k6-3 but is hoped to do so on other chips. 
				 */
				#ifdef __i386__
				asm ("bsf %1, %0" : "=ir" (num0s) : "ir" (index));
				#else
				int i = index;
				while ((i & 1) != 1)
					i >>= 1,
					num0s++;
				#endif

				/* Replace the indexed ROWS random value.
				 * Subtract and add back to RunningSum instead of adding all the random
				 * values together. Only one changes each time.
				 */
				sum -= rows [num0s];
				long r = ((long) random()) >> PINK_RANDOM_SHIFT;
				sum += r;
				rows [num0s] = r;
			}
			
			/* Add extra white noise value. */
			long r = ((long) random()) >> PINK_RANDOM_SHIFT;
			return sum + r;
		}
};

class OneOverFNoisz
{
	public:
		double y;

		OneOverFNoisz()
		{
			y = fmod (M_PI, 1.); /* arbitrary */
		}

		d_sample get()
		{
			return y = 1. - 2. * fmod (y * (1 + y), 1.0);
		}
};

} /* namespace DSP */

#endif /* _DSP_NOISE_H_ */
