/* posted to music-dsp by J. Stein
 *
 * minor adaptions for coding conventions.
 *
 * not overly useful, but interesting.
 */

/* phd.c  Pisarenko Harmonic Decomposition */

/* This routine estimates the frequency
   of a single sinusoid in white noise using the P.H.D.

   Input: array of floats containing the signal, and its length
   Returned: frequency estimate in digital units

   Digital units of frequency are in the range 0 < f < 1/2
   to convert to physical frequency in Hz multiply by the sampling rate.

   The accuracy of the PHD improves with buffer size and SNR.

   To use in real-time collect the incoming samples in a buffer
   of sufficient length (dependent on the amount of noise averaging needed).
   To increase time resolution you can overlap these buffers.

   For further details see DSP-CSP Section 13.5   */

/* Jonathan (Y) Stein */

#ifndef _PISARENKO_H_
#define _PISARENKO_H_

namespace DSP {
	
double 
pisarenko1 (d_sample * x, int N) 
{
	int i ;
	double r1, r2, a, g;

	/* first, the needed autocorrelations */
	for (r1 = 0.0, i=0; i < N-1; i++)  
		r1 += x[i] * x[i+1];
	for (r2 = 0.0, i=0; i < N-2; i++)  
		r2 += x[i] * x[i+2];

	/* now for the magic! */
	a = r2 / (2 * r1) ;
	if (r1 >= 0.0)
		g = a + sqrt (a * a + 2);
	else
		g = a - sqrt (a * a + 2);

	/* convert from angular and return */
	return acos (g / 2.0) / (2 * M_PI);
}

}; /* namespace DSP */

#ifdef TEST
/* g++ -DTEST pisarenko.h */
#include <stdio.h>
#include <stdlib.h>


int main (int argc, char ** argv) {
/* test PHD on artificial signal of various frequencies */
	float *x ;
	float f, fp ;
	int t, N=128 ;

	/* allocate array of length N */
	x = (float *) malloc(N * sizeof(float)) ;

	/* titles for printout */
	printf ( "\n\n  f        f_phd    err\n") ;

	/* loop on frequencies */
	for (f=0.05;f<=0.5;f+=0.05) {

		/* create noisy signal */
		for (t=0;t<N;t++) x[t] = sin(TwoPi*f*t) + (float)rand()/RAND_MAX - 0.5 ;

		/* perform PHD */
		fp = DSP::pisarenko1 (x, N) ;

		/* report */
		printf ( "%f %f %f\n", f, fp, fabs(f-fp)  ) ;

	}

}
#endif

#endif /* _PISARENKO_H_ */
