'MidiFile -- read a standard midi file.'

import sys
from nam import *

def _int (s):
	'make an int from binary little-endian data in string s.'
	i = 0
	for c in s:
		i <<= 8
		i |= ord (c)
	return i

class Binary:
	'read methods for midi peculiarities and not-so-peculiarities.'
	def read_four (f, signature):
		r = f.read (4)
		if r == signature: return
		raise TypeError ('%s is not %s.' % (r, signature))

	def read_int (f, n):
		v = 0
		while n:
			v <<= 8
			v |= ord (f.read (1))
			n -= 1
		return v;	

class KeyEvent (Event):
	def __init__ (self, signs, minor, tick = 0):
		Event.__init__ (self, tick)
		self.signs = signs
		self.minor = minor

class SMPTEEvent (Event):
	def __init__ (self, h, m, s, frame, fract, tick = 0):
		Event.__init__ (self, tick)
		self.h = h
		self.m = m
		self.s = s
		self.frame = frame
		self.fract = fract

class EL (EventList):
	pass

class MidiFile (file, Binary):
	'reads a midi file.'
	def __init__ (f, name, mode = 'r'):
		f.path = name
		file.__init__ (f, name, mode)
		f.time = TimeMap()
		f.tempo = EventList (TempoEvent)
		f.tracks = []
		f.read_all()
		f.cue_points = EventList (TextEvent)

		if not len (f.time):
			f.time.add (TimeEvent (4, 4))

	def read_all (f):
		f.read_four ('MThd')
	
		length = f.read_int (4)
		
		f.format = f.read_int (2)
		tracks = f.read_int (2)

		f.quarter = f.read_int (2)
		if f.quarter & 0x8000:
			raise "cannot handle second-based ticks"

		f.seek (length - 6, 1 or 'SEEK_CUR') # skip SMPTE info.

		# try to read all, but return gracefully if at least 1 track has
		# been read.
		n = 0
		for i in range (tracks):
			f.read_track()
			n += 1

	def read_track (f):
		f.read_four ('MTrk')
		EOT = f.read_int (4) + f.tell()

		p = EL (MidiEvent)
		f.tracks += [p]

		f.status = None
		f.tick = 0

		while f.tell() < EOT:
			f.read_event (p)

	def read_event (f, p):
		f.tick += read_var (f)
		c = read_int (f)
		
		if c < 0x80: 
			if not f.status:
				raise "running status without prior channel event at offset 0x%x." % f.tell()

		elif c < 0xF0:
			f.status = c
			c = read_int (f)

		elif c == 0xF0:
			length = read_var (f)
			data = '\xF0' + f.read (length)
			return p.insert (SysexEvent (data), f.tick)

		elif c == 0xFF:
			return f.read_meta (p)

		elif c > 0xF0:
			raise ValueError ('%02x is not a valid event status.' % c)

		e = f.read_channel_event (c)
		return p.insert (e, f.tick)

	def read_channel_event (f, c):
		'read a channel event'
		e = MidiEvent (f.status, c)
		if (len (e) == 3):
			e[2] = read_int (f)
		return e
		
	def read_meta (f, p):
		meta = read_int (f)
		data = f.read (read_var (f))
		
		if meta == 0:
			f.sequence_number = _int (data)

		elif meta == 1: 
			# 'text event'
			p.name = data

		elif meta == 2:
			print >> sys.stderr, "copying notice: ", repr (data)

		elif meta == 3: p.name = data
		elif meta == 4: p.name = data # actually, instrument.
		elif meta == 7: f.cue_points.add (TextEvent (data, f.tick))
		elif meta == 9: p.device = data
		elif meta == 8: p.program_name = data

		elif meta == 0x20: p.channel = _int (data) + 1
		elif meta == 0x21: p.port = _int (data)

		elif meta == 0x2f: 
			# end-of-track indicator
			pass
			
		elif meta == 0x51: 
			# tempo in 10 ** -6 s / quarter -> bpm
			t = TempoEvent (60. * 10 ** 6 / _int (data))
			return f.tempo.add (t, f.tick)

		elif meta == 0x54: 
			e = SMPTEEvent (*data)
			e.tick = f.tick
			# nowhere to store it yet.

		elif meta == 0x59:
			e = KeyEvent (data[0], data[1], f.tick)
			# nowhere to store it yet.

		elif meta == 0x58:
			# time measure -- the 3rd and 4th members (ticks & 32nds/quarter)
			# don't seem reliable.
			n = ord (data[0])
			d = 1 << ord (data[1])
			ticks = (f.quarter * 4) / d
			e = TimeEvent (n, d, ticks)
			return f.time.add (e, f.tick)
			
		else:
			print >> sys.stderr, 'unknown meta event %02x' % meta, repr (data)

