#! /usr/bin/env python
import sys, os, time

sys.path += ['.']
from nam import *
import config

pcm = config.pcm
midi = config.midi_in
quarter = 192

class AlsaGraph (Graph):
	def __init__ (self):
		Graph.__init__ (self, quarter)
		self.add (RTClock())
		self.clock = AlsaClock (44100, os.geteuid() and 2048 or 128)
		self.add (self.clock)
		self.add (ScriptScheduler())
		self.pcm = AlsaOut (pcm)
		self.add (self.pcm, 10.0)
		self.state = Plugin.Activated

class Midi2Frequency (ScriptPlugin):
	"""
		A simple MIDI to Frequency converter. The input stream is routed 
		through, and for every note-on event frequency values are emitted;
		two to control twin oscillators and another to control a filter
		cutoff. 
	"""
	def __init__ (self):
		ScriptPlugin.__init__ (self)
		
		self.new_in ('notes in', MidiEvent)
		self.new_out ('notes out', MidiEvent)
		self.new_out ('f1', FloatEvent)
		self.new_out ('f2', FloatEvent)
		self.new_out ('filter f', FloatEvent)

		self.ffadd = 38000
		self.transpose = -12

		# second oscillator detune, cent units
		self.detune = 1

	def note_on (self, e):
		"handle a note-on event"
		# 'f1' output
		f = self.outs[1].cache.pop()
		f.__init__ (n2f (e.note), e.tick)
		f.loop = e.loop
		self.outs[1].push (f) 
		 
		# 'f2' output
		f = self.outs[2].cache.pop()
		f.__init__ (n2f (e.note + self.detune / 100.), e.tick)
		f.loop = e.loop
		self.outs[2].push (f) 
		 
		 # 'filter f' output
		f = self.outs[3].cache.pop()
		f.__init__ (n2f (e.note + 12) + self.ffadd, e.tick)
		f.loop = e.loop
		self.outs[3].push (f)

	def process (self, port, events):
		for e in events:
			try: e.note += self.transpose
			except: pass
			if e.is_note_on:
				self.note_on (e)
			self.outs[0].push (e)

class Plate (LadspaPlugin):
	def __init__ (self, size = 6., damping = .01):
		LadspaPlugin.__init__ (self, 'plate_1423.so')
		self.ins[0].value = size
		self.ins[1].value = damping

	def dump (self):
		for i in self.ins:
			print "%40s %5.4f" % (i.name, i.value), i.bounds
		for o in self.outs:
			print o.name
	
graph = AlsaGraph()

tune = MidiIn (midi)
graph.add (tune, 0.)

m2f = Midi2Frequency()
graph.add (m2f, 1.)

osc1 = LadspaPlugin ('caps.so', 'VCOs')
graph.add (osc1, 2.)
osc2 = LadspaPlugin ('caps.so', 'VCOs')
graph.add (osc2, 2.)
	
adsr = ADSR()
graph.add (adsr, 3.)

mod = Modulator (mul = 0.3, add = 0.6)
graph.add (mod, 3.1)

drive = Gain()
graph.add (drive, 3.4)

rat = Gain()
graph.add (rat, 3.5)

vcf = VCF()
graph.add (vcf, 4.)

lfo = LFO (rate = 192 * 8, wave = (0, 1, 0.7, 1, 0.2, 0.5))
graph.add (lfo, 4.5)

lfomod = Modulator (mul = 0.2, add = 0.8)
graph.add (lfomod, 4.6)

gain = Gain (0.) # silent at first
graph.add (gain, 5.)

plate = Plate()
graph.add (plate, 9.)

# connection list: (plug-out, port #, plug-in, port #)
connect = (
	# event cabling
	(tune, 0, m2f, 'notes in'), 
	(m2f, 'notes out', adsr, 'notes in'),
	(m2f, 'f1', osc1, 'f'),
	#(m2f, 'f2', osc2, 'f'),
	(m2f, 'filter f', vcf, 'f'),
	(adsr, 'cv', mod, 'cv dry in'), 
	(mod, 'cv wet out', vcf, 'resonance'), 
	(lfo, 'cv out', lfomod, 'cv dry in'), 
	(lfomod, 'cv wet out', gain, 'factor'), 
	# audio cabling
	(osc1, 0, adsr, 0), 
	#(osc2, 0, adsr, 0), 
	(adsr, 0, drive, 0),
	(drive, 0, rat, 0),
	(rat, 0, vcf, 0),
	(vcf, 0, gain, 0),
	(gain, 0, plate, 3),
	(plate, 0, graph.pcm, 0),
	(plate, 1, graph.pcm, 1),
)

for o, n, i, m in connect:
	o = type (n) == int and o.outs[n] or o[n]
	i = type (m) == int and i.ins[m] or i[m]
	o.connect (i)

gain.factor = 1.0
graph.state = Plugin.Rolling

