#include <basics.h>
#include <EventList.h>
#include <Graph.h>
#include <PyType.h>

#include "Gtk.h"
#include "Pixels.h"

#include "DataView.h"

Type * DataView_type;
Type * PoleZeroView_type;

/* /////////////////////////////////////////////////////////////////////// */

DataView::DataView()
: RGBView (DataView_type),
	signals ("signals", 12),
	marks ("marks", 12)
{
	gc = 0;
	y0 = 0;
	amp_per_pixel = 0.01;
	pixels_per_frame = 2;
	y_align = .5;
	
	x_scale = 64;
	y_scale[0] = 1.;
	y_scale[1] = .5;
	y_scale[2] = .25;

	x_offset = 0;
	y_offset = 0;

	clamped = 0;
}

DataView::~DataView()
{
	if (gc) gdk_gc_destroy (gc);
}

void
DataView::configure (const Rect<int> r)
{
	RGBView::configure (r);

	if (gc == 0)
		gc = gdk_gc_new (this->PixelView::window());

	if (clamped)
		amp_per_pixel = clamped / (double) r.height;

	y0 = (int) (y_align * r.height);

	/* called from Gtk, which doesn't know about the lock */
	ScopedLock lock (mutex);

	complete_redraw();
}

void
DataView::redraw (const Rect<int> & r)
{
	rgb->clip = r;
	rgb->fill (r, Color (gdk_base()));

	draw_scales();

	if (pixels_per_frame == 1)
	{
		for (int i = 0; i < signals; ++i)
			draw_signal_1 (i, r);
	}
	else
	{
		for (int i = 0; i < signals; ++i)
			draw_signal (i, r);
	}

	for (int i = 0; i < marks; ++i)
		draw_mark (marks[i]);

	RGBView::rgb_to_pixels (r);
}

void
DataView::complete_redraw()
{
	// hide_marks (0);
	redraw();
	// recalc_marks();
	// show_marks (0);
	
	blit();
	gdk_flush();
}

void
DataView::draw_mark (DataViewMark & mark)
{
	float x = (mark.index - (float) x_offset) * (float) pixels_per_frame;
	double y = sample2y (mark.amplitude);

	rgb->disc (x + .5, y, 3, 2.5, mark.color, 192);
}

void
DataView::draw_scales()
{
	AlphaColor c ((uint32) 0xcccccc, 128);
	AlphaColor text ((uint32) 0x000000, 128);
	AlphaColor dimtext ((uint32) 0x000000, 64);

	/* x */
	int y0, y1;
	
	if (clamped)
		y0 = 0, y1 = pixels->height;
	else
		y0 = (int) sample2y (y_scale[0]),
		y1 = (int) sample2y (-y_scale[0]);

	Rect<int> rx (0, y0, 1, y1 - y0);

	int f = x2frame (0);
	f -= (f % x_scale);
	
	rx.x = frame2x (f);
	int dx = x_scale * pixels_per_frame;
	int ty = min (pixels->height, rx.y + rx.height) - 2;

	while (rx.x < pixels->width)
	{
		rgb->fill (rx, c);
		
		if (rx.x >= 0)
		{
			char s [24];
			sprintf (s, "%d", x2frame (rx.x));
			if (rx.x + rgb->text_width (s) < pixels->width - rgb->text_width ("0123456"))
				rgb->text (rx.x + 2, ty, s, dimtext);
		}
		
		rx.x += dx;
	}
	
	/* y */
	Rect<int> ry (0, 0, pixels->width, 1);

	ry.y = (int) sample2y (0);
	rgb->fill (ry, c);

	for (int i = 0; i < 3; ++i)
	{
		char s[24];
		
		/* > 0 */
		ry.y = (int) sample2y (y_scale[i]);
		rgb->fill (ry, c);

		sprintf (s, "%+.2f", y_scale[i]);
		rgb->text (pixels->width - rgb->text_width (s) - 2, ry.y + rgb->font_height() - 1, s, text);

		/* < 0 */
		ry.y = (int) sample2y (-y_scale[i]);
		rgb->fill (ry, c);

		sprintf (s, "%+.2f", -y_scale[i]);
		rgb->text (pixels->width - rgb->text_width (s) - 2, ry.y - 2, s, text);
	}
}

static unsigned int colors[] = 
{
	0x297994, // output blue
	0xefaa5a, // nude44 flesh
	0x7b8242, // le mans dirt green
	0x940c18, // toyota gpone red
	0x005184, // chamberlain viper blue
	0xdecb7b, // headlight yellow
	0x94967b, // single gray
	0x84a68c, // curbstone green
	0x426939, // helmet green
	0x106db5, // riley blue
	0x6b92ad, // light curb blue
	0x1861de, // electric blue
	0xdecf94, // sand yellow
	0x844539, // gravel sunset brown
	0xf78a52, // neon orange
	0x84a631, // tree green
	0x5aa6ff, // light curb blue II
	0x4275a5, // rubber blue
	0
};

void
DataView::draw_signal (int sig, const Rect<int> & rect)
{
	AudioSignal * signal = signals[sig];
	Color c (colors [sig % 12]);

	Rect<int> r = rect;

	int i = max (x2frame (r.x), 0); /* sample index */
	int x = frame2x (i); /* pixel column */
	int w = r.width - x;
	
	int n = i + min (signal->size - i, w / pixels_per_frame);
	double yo = sample2y (y_offset);

	while (i < n)
	{
		d_sample s = signal->data[i];
		
		if (!finite (s))
		{
			char * s = "NAN";
			rgb->text (x + 2, y0, s, AlphaColor (c, 96));
			break;
		}

		if (s > y_offset)
			rgb->columnf (x, sample2y (s), yo, c, 96);
		else
			rgb->columnf (x, yo, sample2y (s), c, 96);
		
		++i;
		x += pixels_per_frame;
	}
}

void
DataView::draw_signal_1 (int sig, const Rect<int> & rect)
{
	AudioSignal * signal = signals[sig];
	Color c (colors [sig % 12]);

	Rect<int> r = rect;

	int i = max (x2frame (r.x), 0); /* sample index */
	int x = frame2x (i); /* pixel column */
	int w = r.width - x;
	
	int n = i + min (signal->size - i, w / pixels_per_frame);

	d_sample last_s = (i >= 0 && i < signal->size) ? signal->data[i] : 0;
	double last_y = sample2y (last_s);
	
	while (i < n) /* < n */
	{
		d_sample s = signal->data[i];
		
		if (!finite (s))
		{
			char * s = "NAN";
			rgb->text (x + 2, y0, s, AlphaColor (c, 96));
			break;
		}
		
		double y = sample2y (s);

		rgb->fade_column (x - 1, last_y, y, c, 96, 0);
		rgb->fade_column (x, last_y, y, c, 0, 96);

		int iy = (int) y;
		int a = (int) (96. * (y - iy));
		rgb->put (x, iy, c, 96 - a); 
		rgb->put (x, iy + 1, c, a); 

		last_y = y;
		last_s = s;

		++i;
		x += pixels_per_frame;
	}
}

/* // Py<DataView> /////////////////////////////////////////////////////// */

static PyObject * 
DataView_get_y_align (DataView * v)
{
	return PyFloat (v->y_align);
}

static int
DataView_set_y_align (DataView * v, PyObject * x)
Py_TRY
{
	double d;
	
	if (!PyArg_Parse (x, "d", &d))
		return -1;

	if (d < -5 || d > 5)
		throw ValueError() << "out of bounds.";
	
	v->y_align = d;
	v->y0 = (int) (d * v->pixels->height);

	return 0;
}
Py_CATCH (-1);

static PyObject * 
DataView_get_pixels_per_frame (DataView * v)
{
	return PyFloat (v->pixels_per_frame);
}

static PyObject * 
DataView_get_amp_per_pixel (DataView * v)
{
	return PyFloat (v->amp_per_pixel);
}

static int
DataView_set_amp_per_pixel (DataView * v, PyObject * x)
Py_TRY
{
	double d;
	
	if (!PyArg_Parse (x, "d", &d))
		return -1;

	if (d < .00001 || d > 1)
		throw ValueError() << "out of bounds.";
	
	v->amp_per_pixel = d;

	return 0;
}
Py_CATCH (-1);

static PyObject * 
DataView_get_x_offset (DataView * v)
{
	return PyInt (v->x_offset);
}

static int
DataView_set_x_offset (DataView * v, PyObject * x)
Py_TRY
{
	int i;
	
	if (!PyArg_Parse (x, "i", &i))
		return -1;

	if (i < -200)
		throw ValueError() << "offset must be >= 0.";
	
	v->x_offset = i;

	return 0;
}
Py_CATCH (-1);

static PyObject * 
DataView_get_y_offset (DataView * v)
{
	return PyFloat (v->y_offset);
}

static int
DataView_set_y_offset (DataView * v, PyObject * y)
Py_TRY
{
	if (!PyArg_Parse (y, "d", &v->y_offset))
		return -1;
	return 0;
}
Py_CATCH (-1);

static PyObject * 
DataView_get_x_scale (DataView * v)
{
	return PyInt (v->x_scale);
}

static int
DataView_set_x_scale (DataView * v, PyObject * x)
Py_TRY
{
	int i;
	
	if (!PyArg_Parse (x, "i", &i))
		return -1;

	if (i < 4)
		throw ValueError() << "x_scale must be >= 4.";
	
	v->x_scale = i;

	return 0;
}
Py_CATCH (-1);

static PyObject * 
DataView_get_y_scale (DataView * v)
{
	return Py_BuildValue ("fff", v->y_scale[0], v->y_scale[1], v->y_scale[2]);
}

static int
DataView_set_y_scale (DataView * v, PyObject * y)
Py_TRY
{
	if (!PyArg_ParseTuple (y, "fff", v->y_scale, v->y_scale + 1, v->y_scale + 2))
		return -1;
	return 0;
}
Py_CATCH (-1);

static PyObject *
DataView_get_marks (DataView * v)
{
	PyObject * l = PyTuple_New (v->marks);
	if (l == 0) return 0;
	
	for (int i = 0; i < v->marks; ++i)
	{
		DataViewMark & m = v->marks[i];
		
		PyObject * v = Py_BuildValue ("dfi", m.index, m.amplitude, m.color.packed());
		if (v == 0)
		{
			::decref (l, true);
			return 0;
		}
		
		PyTuple_SetItem (l, i, v);
	}

	return l;
}

static int
DataView_set_marks (DataView * v, PyObject * args)
Py_TRY
{
	if (!PySequence_Check (args))
		throw TypeError() << "needs a sequence";

	int n = PySequence_Length (args);
	
	DataViewMark marks[n];
	
	for (int i = 0; i < n; ++i)
	{
		PyObject * v = PySequence_GetItem (args, i);
		uint32 c;

		if (!PyArg_ParseTuple (v, "dfi", &marks[i].index, &marks[i].amplitude, &c))
		{
			::decref (v, true);
			return -1;
		}
		
		::decref (v, true);
		marks[i].color = c;
	}
	
	v->marks.clear();
	v->marks.insert_items (marks, n, List::End);

	return 0;
}
Py_CATCH (-1);

static PyGetSetDef 
DataView_getsets [] = 
{
	{"amp_per_pixel", (getter) DataView_get_amp_per_pixel, (setter) DataView_set_amp_per_pixel, 
	 "the y zoom (amplitude / pixel)."},
	{"pixels_per_frame", (getter) DataView_get_pixels_per_frame, (setter) 0, 
	 "the x zoom."},
	{"y_align", (getter) DataView_get_y_align, (setter) DataView_set_y_align, 
	 "the y alignment."},
	{"y_offset", (getter) DataView_get_y_offset, (setter) DataView_set_y_offset, 
	 "offset all amplitudes by this value."},
	{"x_offset", (getter) DataView_get_x_offset, (setter) DataView_set_x_offset, 
	 "the frame at which the display window opens on the signals."},
	{"x_scale", (getter) DataView_get_x_scale, (setter) DataView_set_x_scale, 
	 "draw vertical grid lines every 'x_scale' pixels."},
	{"y_scale", (getter) DataView_get_y_scale, (setter) DataView_set_y_scale, 
	 "a tuple of three sample values indicating where to draw horizontal "
	 "grid lines."},
	{"marks", (getter) DataView_get_marks, (setter) DataView_set_marks, 
	 "((index, amplitude, color), ...) -- dataview draws a mark in 'color' "
	 "at sample 'index' at amplitude 'amplitude'."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

/* sequence methods */

static int 
DataView_sq_length (DataView * v)
{
	ScopedLock lock (v->mutex);
	return v->signals;
}

static PyObject *
DataView_sq_item (DataView * v, int i) 
Py_TRY
{
	ScopedLock lock (v->mutex);
	return ::incref (v->signals [i]); 
} 
Py_CATCH (0);

static PySequenceMethods 
DataView_as_sequence = 
{
	(inquiry) DataView_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) DataView_sq_item, /* sq_item */
	(intintargfunc) 0, /* sq_slice */
	(intobjargproc) 0, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
DataView_add (DataView * v, AudioSignal * arg) 
Py_TRY
{
	if (!AudioSignal_type->is_derived (arg))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (arg);

	ScopedLock lock (v->mutex);
	
	if (v->signals.find (arg) != -1)
		throw AssertionError() << "already in list.";

	v->signals.add_object (arg);

	return PyNone();
} 
Py_CATCH (0);

static PyObject *
DataView_remove (DataView * v, AudioSignal * arg) 
Py_TRY
{
	if (!AudioSignal_type->is_derived (arg))
		throw TypeError() << "need an AudioSignal, not " << get_type_name (arg);

	ScopedLock lock (v->mutex);
	int i = v->signals.find (arg);

	if (i < 0)
		throw LookupError() << "not in this view.";
	
	v->signals.remove_items (i, 1);

	return PyNone();
} 
Py_CATCH (0);

static PyObject *
DataView_clear (DataView * v)
{
	ScopedLock lock (v->mutex);
	v->signals.clear_and_decref();

	return PyNone();
}

static PyObject *
DataView_get_length (DataView * v)
Py_TRY
{
	int n = 0;
	
	ScopedLock lock (v->mutex);
	for (int i = 0; i < v->signals; ++i)
		if (n < v->signals[i]->size)	
			n = v->signals[i]->size;

	return PyInt (n);
} 
Py_CATCH (0);

static PyObject * 
DataView_redraw (DataView * v)
Py_TRY
{
	ScopedLock lock (v->mutex);

	v->complete_redraw();
		
	return PyNone();
}
Py_CATCH (0);

/* translation methods */

static PyObject *
DataView_x2frame (DataView * v, PyObject * arg)
{
	int x;

	if (!PyArg_Parse (arg, "i", &x))
		return 0;

	return PyFloat (v->x2frame (x));
}

static PyObject *
DataView_frame2x (DataView * v, PyObject * arg)
{
	double f;

	if (!PyArg_Parse (arg, "d", &f))
		return 0;

	return PyFloat (v->frame2x ((int) f));
}

static PyObject *
DataView_sample2y (DataView * v, PyObject * arg)
{
	d_sample n;
	
	if (!PyArg_Parse (arg, "f", &n))
		return 0;
	
	return PyFloat (v->sample2y (n));
}

static PyObject *
DataView_y2sample (DataView * v, PyObject * arg)
{
	int y;

	if (!PyArg_Parse (arg, "i", &y))
		return 0;

	return PyFloat (v->y2sample (y));
}

static PyObject *
DataView_clamp (DataView * v, PyObject * args)
{
	d_sample upper, lower;

	if (!PyArg_ParseTuple (args, "ff", &upper, &lower))
		return 0;

	v->clamped = fabs (upper - lower);

	return PyNone();
};

static PyMethodDef 
DataView_methods [] = 
{
	{"add", (PyCFunction) DataView_add, METH_O,
	 "add (signal) -- add one signal to the list."},
	{"remove", (PyCFunction) DataView_remove, METH_O,
	 "remove a signal from the list."},
	{"clear", (PyCFunction) DataView_clear, METH_NOARGS,
	 "remove all signals from the display list."},
	{"redraw", (PyCFunction) DataView_redraw, METH_NOARGS,
	 "redraw() -- complete redraw."},
	{"x2frame", (PyCFunction) DataView_x2frame, METH_O,
	 "x2frame (x) -- translate 'x' into a frame."},
	{"frame2x", (PyCFunction) DataView_frame2x, METH_O,
	 "frame2x (x) -- translate 'frame' index into an x coordinate."},
	{"sample2y", (PyCFunction) DataView_sample2y, METH_O,
	 "sample2y (sample) -- translate 'sample' into a pixel offset."},
	{"y2sample", (PyCFunction) DataView_y2sample, METH_O,
	 "y2sample (sample) -- translate 'y' into a sample pitch."},
	{"get_length", (PyCFunction) DataView_get_length, METH_NOARGS,
	 "get_length() -- return the length of the longest displayed signal."},
	{"clamp", (PyCFunction) DataView_clamp, METH_VARARGS,
	 "clamp (upper, lower) -- adjust the display window."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

char * Py<DataView>::__doc__ = 
"DataView() -- a RGBView displaying the contents of Audio event lists.";

void
Py<DataView>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {"signal", 0};

	AudioSignal * s = 0;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "|O!", keys,
				AudioSignal_type, &s))
		throw blewit();

	PixelView::init();

	gtk_widget_set_name (widget, "DataView");

	if (s) DataView_add (this, s);
}

void
Py<DataView>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = RGBView_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = DataView_getsets;
	t->tp_methods = DataView_methods;
	t->tp_as_sequence = &DataView_as_sequence;
}

/* /////////////////////////////////////////////////////////////////////// */

PoleZeroView::PoleZeroView()
: RGBView (PoleZeroView_type)
{
	y0 = x0 = 0;
	scale = 1.;
}

PoleZeroView::~PoleZeroView()
{
}

void
PoleZeroView::configure (const Rect<int> r)
{
	RGBView::configure (r);

	double s = max (r.width, r.height);
	scale = (s / 2) * .8;

	x0 = r.width / 2;
	y0 = r.height / 2;

	complete_redraw();
}

void
PoleZeroView::redraw (const Rect<int> & r)
{
	rgb->clip = r;
	rgb->fill (r, Color (gdk_base()));

	draw_unit_circle();
	draw_poles_zeros();

	RGBView::rgb_to_pixels (r);
}

void
PoleZeroView::complete_redraw()
{
	redraw();
	
	blit();
	gdk_flush();
}

void
PoleZeroView::draw_poles_zeros()
{
	Color pole_color (0xFF0000);
	Color zero_color (0x0000FF);
	Color mark_color (0x000001);

	for (int i = 0; i < pzs.npoles; ++i)
	{
		double x = real2x (pzs.poles[i].re);
		double y = imag2y (pzs.poles[i].im);

		rgb->disc (x, y, 3, 0, pole_color, 192);
	}

	for (int i = 0; i < pzs.nzeros; ++i)
	{
		double x = real2x (pzs.zeros[i].re);
		double y = imag2y (pzs.zeros[i].im);

		rgb->disc (x, y, 3, 0, zero_color, 192);
	}

	double x = real2x (mark.re);
	double y = imag2y (mark.im);

	rgb->disc (x, y, 6, 0, mark_color, 32);
}

void
PoleZeroView::draw_unit_circle()
{
	AlphaColor c ((uint32) 0xcccccc, 128);

	rgb->fill (Rect<int> (x0, 0, 1, pixels->height), c);
	rgb->fill (Rect<int> (0, y0, pixels->width, 1), c);

	rgb->disc (x0, y0, scale, scale - 2, Color (0xcccccc), 192);
}

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
PoleZeroView_get_poles (PoleZeroView * v)
{
	PyObject * l = PyTuple_New (v->pzs.npoles);
	if (l == 0) return 0;
	
	for (int i = 0; i < v->pzs.npoles; ++i)
	{
		PyObject * c = PyComplex_FromDoubles (
				v->pzs.poles[i].re, v->pzs.poles[i].im);
		if (c == 0)
		{
			::decref (l, true);
			return 0;
		}
		
		PyTuple_SetItem (l, i, c);
	}

	return l;
}

static PyObject *
PoleZeroView_get_zeros (PoleZeroView * v)
{
	PyObject * l = PyTuple_New (v->pzs.nzeros);
	if (l == 0) return 0;
	
	for (int i = 0; i < v->pzs.nzeros; ++i)
	{
		PyObject * c = PyComplex_FromDoubles (
				v->pzs.zeros[i].re, v->pzs.zeros[i].im);
		if (v == 0)
		{
			::decref (l, true);
			return 0;
		}
		
		PyTuple_SetItem (l, i, c);
	}

	return l;
}

static int
PoleZeroView_set_poles (PoleZeroView * v, PyObject * args)
Py_TRY
{
	if (!PySequence_Check (args))
		throw TypeError() << "needs a sequence";

	int n = PySequence_Length (args);
	
	if (n >= DSP::PZs::MaxPoles)
		throw OverflowError() << DSP::PZs::MaxPoles << " poles at max.";

	DSP::complex poles[n];

	for (int i = 0; i < n; ++i)
	{
		PyObject * v = PySequence_GetItem (args, i);

		Py_complex c = PyComplex_AsCComplex (v);

		if (PyErr_Occurred())
		{
			::decref (v, true);
			return -1;
		}
		
		::decref (v, true);
		poles[i].re = c.real;
		poles[i].im = c.imag;
	}

	v->pzs.npoles = 0;
	for (int i = 0; i < n; ++i)
		v->pzs.add_pole (poles[i]);

	return 0;
}
Py_CATCH (-1);

static int
PoleZeroView_set_zeros (PoleZeroView * v, PyObject * args)
Py_TRY
{
	if (!PySequence_Check (args))
		throw TypeError() << "needs a sequence";

	int n = PySequence_Length (args);
	
	if (n >= DSP::PZs::MaxZeros)
		throw OverflowError() << DSP::PZs::MaxZeros << " zeros at max.";

	DSP::complex zeros[n];

	for (int i = 0; i < n; ++i)
	{
		PyObject * v = PySequence_GetItem (args, i);

		Py_complex c = PyComplex_AsCComplex (v);

		if (PyErr_Occurred())
		{
			::decref (v, true);
			return -1;
		}
	
		::decref (v, true);
		zeros[i].re = c.real;
		zeros[i].im = c.imag;
	}

	v->pzs.nzeros = 0;
	for (int i = 0; i < n; ++i)
		v->pzs.add_zero (zeros[i]);

	return 0;
}
Py_CATCH (-1);

static PyObject *
PoleZeroView_get_mark (PoleZeroView * v)
{
	return PyComplex_FromDoubles (v->mark.re, v->mark.im);
}

static int
PoleZeroView_set_mark (PoleZeroView * v, PyObject * arg)
{
	Py_complex c = PyComplex_AsCComplex (arg);
	
	if (PyErr_Occurred())
		return -1;

	v->mark.re = c.real;
	v->mark.im = c.imag;

	return 0;
}

static PyGetSetDef 
PoleZeroView_getsets [] = 
{
	{"poles", (getter) PoleZeroView_get_poles, (setter) PoleZeroView_set_poles, 
	 "the poles (tuple of complex numbers)."},
	{"zeros", (getter) PoleZeroView_get_zeros, (setter) PoleZeroView_set_zeros, 
	 "the zeros (tuple of complex numbers)."},
	{"mark", (getter) PoleZeroView_get_mark, (setter) PoleZeroView_set_mark, 
	 "complex number where a mark is drawn."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject * 
PoleZeroView_redraw (PoleZeroView * v)
Py_TRY
{
	v->complete_redraw();
		
	return PyNone();
}
Py_CATCH (0);

/* translation methods */

static PyObject *
PoleZeroView_x2real (PoleZeroView * v, PyObject * arg)
{
	int x;

	if (!PyArg_Parse (arg, "i", &x))
		return 0;

	return PyFloat (v->x2real (x));
}

static PyObject *
PoleZeroView_real2x (PoleZeroView * v, PyObject * arg)
{
	double f;

	if (!PyArg_Parse (arg, "d", &f))
		return 0;

	return PyFloat (v->real2x (f));
}

static PyObject *
PoleZeroView_imag2y (PoleZeroView * v, PyObject * arg)
{
	double n;
	
	if (!PyArg_Parse (arg, "d", &n))
		return 0;
	
	return PyFloat (v->imag2y (n));
}

static PyObject *
PoleZeroView_y2imag (PoleZeroView * v, PyObject * arg)
{
	int y;

	if (!PyArg_Parse (arg, "i", &y))
		return 0;

	return PyFloat (v->y2imag (y));
}

static PyObject *
PoleZeroView_get_amplitude_response (PoleZeroView * v, PyObject * N)
Py_TRY
{
	int n;

	if (!PyArg_Parse (N, "i", &n))
		return 0;
	if (n < 4)
		throw ValueError() << "no fewer than 4 f bins";

	AudioSignal * s = NEW (AudioSignal);
	if (s == 0) return 0;

	s->set_size (n);
	
	v->pzs.compute_amplitude_response (n, s->data);

	return s;
}
Py_CATCH (0);

static PyMethodDef 
PoleZeroView_methods [] = 
{
	{"x2real", (PyCFunction) PoleZeroView_x2real, METH_O,
	 "x2real (x) -- translate 'x' into a real."},
	{"real2x", (PyCFunction) PoleZeroView_real2x, METH_O,
	 "real2x (x) -- translate 'real' into an x coordinate."},
	{"imag2y", (PyCFunction) PoleZeroView_imag2y, METH_O,
	 "imag2y (imag) -- translate 'imag' into a pixel offset."},
	{"y2imag", (PyCFunction) PoleZeroView_y2imag, METH_O,
	 "y2imag (imag) -- translate 'y' into a imag."},
	{"redraw", (PyCFunction) PoleZeroView_redraw, METH_NOARGS,
	 "redraw() -- full redraw."},
	{"get_amplitude_response", (PyCFunction) PoleZeroView_get_amplitude_response, METH_O,
	 "get_amplitude_response (n) -- compute the amplitude response over 'n' "
	 "frequency bins. returns an AudioSignal."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

char * Py<PoleZeroView>::__doc__ = 
"PoleZeroView() -- a RGBView displaying poles and zeros and the unit circle.";

void
Py<PoleZeroView>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();

	PixelView::init();

	gtk_widget_set_name (widget, "PoleZeroView");
}

void
Py<PoleZeroView>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = RGBView_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = PoleZeroView_getsets;
	t->tp_methods = PoleZeroView_methods;
}

/* /////////////////////////////////////////////////////////////////////// */

void
init_DataView_types (PyObject * dict)
{
	DataView_type = new PyType<DataView::DataView> ("DataView", dict);
	PoleZeroView_type = new PyType<PoleZeroView::PoleZeroView> ("PoleZeroView", dict);
}
