#include <basics.h>
#include <Object.h>
#include <PyType.h>

#include "Gtk.h"
#include "RGB.h"
#include "FreeType.h"
#include FT_OUTLINE_H

/* /usr/local/include/freetype2/freetype/config/ftheader.h */
static FT_Library library;

void FT (char * s, int e)
{
	if (e) throw blewit() << s << ": " << e;
}

/* //////////////////////////////////////////////////////////////////////// */

FreeType::FreeType()
: Object (FreeType_type)
{
	FT ("FT_Init_FreeType", FT_Init_FreeType (&library));
	move (0, 0);
	rotate (0);
}

FreeType::~FreeType()
{
	FT_Done_Face (face);
	/* FIXME: this segfaults. do we have to refcount freetype? */
#if 0
	FT_Done_FreeType (library);
#endif
}

void
FreeType::init (char * path, int n)
{
	FT ("FT_New_Face", FT_New_Face (library, path, n, &face));
	set_size (12);
}

/* //////////////////////////////////////////////////////////////////////// */

void 
FreeType::set_size (int y)
{
	FT ("FT_Set_Pixel_Sizes", FT_Set_Pixel_Sizes (face, 0, size = y));
}

int 
FreeType::get_size()
{
	return size;
}

void 
FreeType::move (float x, float y)
{
	position.x = (int) x;
	position.y = (int) y;
	
	x = (x - position.x) * (1 << 6);
	y = (y - position.y) * (1 << 6);

	translate.x = (int) x;
	translate.y = (int) y;

	/* reset kerning */
	last = 0;
}

void
FreeType::rotate (float a)
{
	float sin = ::sin (a);
	float cos = ::cos (a);

	transform.xx = (FT_Fixed) ( cos * 0x10000);
	transform.xy = (FT_Fixed) (-sin * 0x10000);
	transform.yx = (FT_Fixed) ( sin * 0x10000);
	transform.yy = (FT_Fixed) ( cos * 0x10000);
}

template <class C>
void
FreeType::draw (RGB * rgb, C * text, int n)
{
	for (int i = 0; i < n; ++i)
	{
		FT_UInt gi = FT_Get_Char_Index (face, text[i]);
		draw_glyph (rgb, gi);
	}
}

void
FreeType::load_glyph (FT_UInt gi)
{
	FT_UInt flags = FT_LOAD_NO_HINTING | FT_LOAD_PEDANTIC;
	FT ("FT_Load_Glyph", FT_Load_Glyph (face, gi, flags));
}

void
FreeType::draw_glyph (RGB * rgb, FT_UInt gi)
{
	load_glyph (gi);

	if (last)
	{
		FT_Vector delta;
		FT_Get_Kerning (face, last, gi, ft_kerning_default, &delta);
	
		FT_Vector_Transform (&delta, &transform);
		translate.x += delta.x;
		translate.y += delta.y;
	}
		
	FT_Glyph glyph;
	
	FT ("FT_Get_Glyph", FT_Get_Glyph (face->glyph, &glyph));
	FT ("FT_Glyph_Transform", 
			FT_Glyph_Transform (glyph, &transform, &translate));
	FT ("FT_Glyph_To_Bitmap", 
			FT_Glyph_To_Bitmap (&glyph, ft_render_mode_normal, 0, 1));

	FT_BitmapGlyph b = (FT_BitmapGlyph) glyph;
	
	copy (rgb, &b->bitmap, position.x + b->left, position.y - b->top);
	
	FT_Done_Glyph (glyph);

	/* for some reason, face->glyph->advance has the correct advance vector,
	 * while glyph->advance hasn't. */
	FT_Vector_Transform (&face->glyph->advance, &transform);
	translate.x += face->glyph->advance.x;
	translate.y += face->glyph->advance.y;
	
	/* remember glyph index for kerning */
	last = gi;
}


void
FreeType::copy (RGB * rgb, FT_Bitmap * b, int x0, int y0)
{
	Rect<int> r (x0, y0, b->width, b->rows);

	if (!rgb->size.clip (r))
		return;

	int dx = r.x - x0;
	int dy = r.y - y0;

	for (int y = 0; y < r.height; ++y)
	{
		byte * a = b->buffer + b->pitch * (y + dy) + dx;

		Pixel p (rgb->pointer (r.x, r.y + y));

		for (int x = 0; x < r.width; ++x)
		{
			p.put (AlphaColor (color, (*a++ * alpha) >> 8));
			p += 3;
		}
	}
}

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
FreeType_attach (FreeType * ft, PyObject * args)
Py_TRY
{
	char * path;

	if (!PyArg_ParseTuple (args, "s", &path))
		throw blewit();

	FT ("FT_Attach_File", FT_Attach_File (ft->face, path));

	return PyNone();
}
Py_CATCH (0);

static PyObject *
FreeType_draw (FreeType * ft, PyObject * args)
Py_TRY
{
	RGB * rgb;
	Py_UNICODE * s;
	int n, size = 0;

	if (!PyArg_ParseTuple (args, "O!u#|i", RGB_type, &rgb, &s, &n, &size))
		throw blewit();

	if (size)
		ft->set_size (size);

	ft->draw (rgb, s, n);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
FreeType_glyph (FreeType * ft, PyObject * args)
Py_TRY
{
	RGB * rgb;
	int gi;

	if (!PyArg_ParseTuple (args, "O!i", RGB_type, &rgb, &gi))
		throw blewit();

	ft->draw_glyph (rgb, gi);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
FreeType_move (FreeType * ft, PyObject * args)
{
	float x, y;

	if (!PyArg_ParseTuple (args, "ff", &x, &y))
		return 0;

	ft->move (x, y);

	return PyNone();
}

static PyObject *
FreeType_get_glyph_name (FreeType * ft, PyObject * args)
Py_TRY
{
	FT_UInt gi;

	if (!PyArg_ParseTuple (args, "i", &gi))
		return 0;

	char name[256];
	FT ("FT_Get_Glyph_Name", FT_Get_Glyph_Name (ft->face, gi, name, 255));
	
	return PyString (name);
}
Py_CATCH (0);

static PyObject *
FreeType_get_char_index (FreeType * ft, PyObject * args)
{
	int c;

	if (!PyArg_ParseTuple (args, "i", &c))
		return 0;
	
	return PyInt (FT_Get_Char_Index (ft->face, c));
}

static PyObject *
FreeType_rotate (FreeType * ft, PyObject * args)
Py_TRY
{
	float angle;

	if (!PyArg_ParseTuple (args, "f", &angle))
		return 0;

	ft->rotate (angle);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
FreeType_shift (FreeType * ft, PyObject * args)
Py_TRY
{
	float dx, dy;

	if (!PyArg_ParseTuple (args, "ff", &dx, &dy))
		return 0;

	ft->translate.x += (int) (dx * (1 << 6));
	ft->translate.y += (int) (dy * (1 << 6));

	return PyNone();
}
Py_CATCH (0);

static PyObject *
FreeType_get_glyph (FreeType * ft, PyObject * args)
{
	int c;

	if (!PyArg_ParseTuple (args, "i", &c))
		return 0;
	
	FT_UInt i = FT_Get_Char_Index (ft->face, c);
	ft->load_glyph (i);

	Glyph * g = NEW(Glyph);
	FT ("FT_Get_Glyph", FT_Get_Glyph (ft->face->glyph, &g->glyph));

	return g;
}

static PyMethodDef
FreeType_methods[] =
{
	{"attach", (PyCFunction) FreeType_attach, METH_VARARGS,
	 "attach (path) -- used to load .afm font metrics."},
	{"draw", (PyCFunction) FreeType_draw, METH_VARARGS,
	 "draw (rgb, u'text', size) -- render a string to an RGB."},
	{"glyph", (PyCFunction) FreeType_glyph, METH_VARARGS,
	 "glyph (rgb, glyph, size) -- render a glyph to an RGB."},
	{"get_glyph", (PyCFunction) FreeType_get_glyph, METH_VARARGS,
	 "get_glyph (char) -- return the glyph representing 'char'."},
	{"get_char_index", (PyCFunction) FreeType_get_char_index, METH_VARARGS,
	 "get_char_index (char) -- return the glyph index of a char."},
	{"get_glyph_name", (PyCFunction) FreeType_get_glyph_name, METH_VARARGS,
	 "get_glyph_name (glyph) -- return the name of a glyph."},
	{"move", (PyCFunction) FreeType_move, METH_VARARGS,
	 "move (x, y) -- move the baseline start, resets the relative position."},
	{"shift", (PyCFunction) FreeType_shift, METH_VARARGS,
	 "shift (x, y) -- shift the next glyph by (x, y)."},
	{"rotate", (PyCFunction) FreeType_rotate, METH_VARARGS,
	 "rotate (angle) -- set the glyph transform matrix."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
FreeType_get_ps_name (FreeType * ft)
{
	const char * s = FT_Get_Postscript_Name (ft->face);
	return PyString (s ? s : "");
}

static PyObject *
FreeType_get_size (FreeType * ft)
{
	return PyInt (ft->get_size());
}

static int
FreeType_set_size (FreeType * ft, PyObject * args)
Py_TRY
{
	int i;

	if (!PyArg_Parse (args, "i", &i))
		return -1;
	
	ft->set_size (i);
	return 0;
}
Py_CATCH (0);

static PyObject *
FreeType_get_x (FreeType * ft)
{
	return PyInt (ft->position.x);
}

static PyObject *
FreeType_get_y (FreeType * ft)
{
	return PyInt (ft->position.y);
}

static int 
FreeType_set_color (FreeType * ft, PyObject * args)
{
	Color c;
	
	if (!PyArg_ParseTuple (args, "iii|i", &c.r, &c.g, &c.b, &ft->alpha))
		return -1;
	
	ft->color = c;

	return 0;
}

static PyObject *
FreeType_get_alpha (FreeType * ft)
{
	return PyInt (ft->alpha);
}

static int 
FreeType_set_alpha (FreeType * ft, PyObject * args)
{
	if (!PyArg_Parse (args, "i", &ft->alpha))
		return -1;
	return 0;
}

static PyObject *
FreeType_get_first_char (FreeType * ft)
Py_TRY
{
	FT_UInt i;
	FT_ULong l = FT_Get_First_Char (ft->face, &i);
	return Py_BuildValue ("ii", l, i);
}
Py_CATCH (0);

static PyGetSetDef
FreeType_getsets[] =
{
	{"ps_name", (getter) FreeType_get_ps_name, (setter) 0, 
		"the PostScript name of the font face."},
	{"x", (getter) FreeType_get_x, (setter) 0, 
		"current x of the pen"},
	{"y", (getter) FreeType_get_y, (setter) 0, 
		"current x of the pen baseline"},
	{"alpha", (getter) FreeType_get_alpha, (setter) FreeType_set_alpha, 
		"current transparency"},
	{"color", (getter) 0, (setter) FreeType_set_color, 
		"current (r, g, b[, alpha])"},
	{"size", (getter) FreeType_get_size, (setter) FreeType_set_size, 
		"current pixel size"},
	{"first_char", (getter) FreeType_get_first_char, (setter) 0, 
		"first char in charmap"},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

Type * FreeType_type;

char * Py<FreeType>::__doc__ = 
"FreeType() -- a PixelView displaying the contents of a RGB buffer.";

void
Py<FreeType>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"path", "n", 0};
	char * path;
	int n = 0;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|i", keys, &path, &n))
		throw blewit();

	init (path, n);
}

void
Py<FreeType>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Object_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_methods = FreeType_methods;
	t->tp_getset = FreeType_getsets;
}

/* //////////////////////////////////////////////////////////////////////// */

static inline double
ft2d (FT_Pos ft)
{
	return ((double) ft) * (1. / (1 << 6));
}

static PyObject *
Glyph_get_advance (Glyph * g)
{
	double x = ft2d (g->glyph->advance.x);
	double y = ft2d (g->glyph->advance.y);

	return Py_BuildValue ("ff", x, y);
}

static PyObject *
Glyph_get_points (Glyph * g)
Py_TRY
{
	FT_Outline & o = g->outline();
	
	PyRef l (PyTuple_New (o.n_points));

	for (int i = 0; i < o.n_points; ++i)
	{
		double x = ft2d (o.points[i].x);
		double y = ft2d (o.points[i].y);

		PyObject * p = Py_BuildValue ("(ff)i", x, y, o.tags[i]);
		if (p == 0) throw blewit();

		PyTuple_SetItem (l, i, p);
	}

	return ::incref (l);
}
Py_CATCH (0);

static PyObject *
Glyph_get_contours (Glyph * g)
Py_TRY
{
	FT_Outline & o = g->outline();
	
	PyRef c (PyTuple_New (o.n_contours));

	int n = 0;
	for (int i = 0; i < o.n_contours; ++i)
	{
		int m = o.contours[i];
		PyRef l (PyTuple_New (m - n));

		for (int j = 0; j < m - n; ++j)
		{
			double x = ft2d (o.points[j].x);
			double y = ft2d (o.points[j].y);

			PyObject * p = Py_BuildValue ("(ff)i", x, y, o.tags[j]);
			if (p == 0) throw blewit();

			PyTuple_SetItem (l, j, p);
		}

		PyTuple_SetItem (c, i, ::incref (l));
	}

	return ::incref (c);
}
Py_CATCH (0);

static PyGetSetDef
Glyph_getsets[] =
{
	{"advance", (getter) Glyph_get_advance, (setter) 0, 
		"advance vector"},
	{"points", (getter) Glyph_get_points, (setter) 0, 
		"outline points"},
	{"contours", (getter) Glyph_get_contours, (setter) 0, 
		"outline contours, a list of individual outlines in the glyph, each of "
		"them a list of points. A point is ((x, y), tags) where 'tags' has bit "
		"0 set if the point is on the line. If it is off the line (clear bit) "
		"then bit 1 turns it into a 3rd-order control point, otherwise it is "
		"2nd-order."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

class GlyphRender
{
	public:
		RGB * rgb;

		GlyphRender (RGB * _rgb)
			{	
				rgb = _rgb;
			}

		void
		run (int x, int y, int w, int alpha)
			{
				msg(), "( ", x, y, ") x ", w, " : ", alpha;
				Pixel p (rgb->pointer (x, y));
				AlphaColor c ((uint) 0, alpha);
				p.run (w, c);
			}
};

static void
gray_spans (int y, int count, FT_Span * spans, void * _r)
{
	GlyphRender * r = (GlyphRender *) _r;
	for (int i = 0; i < count; ++i)
		r->run (spans[i].x, y, spans[i].len, spans[i].coverage);
}

static void
black_spans (int y, int count, FT_Span * spans, void * _r)
{
	GlyphRender * r = (GlyphRender *) _r;
	for (int i = 0; i < count; ++i)
		r->run (spans[i].x, y, spans[i].len, spans[i].coverage);
}

static PyObject *
Glyph_render (Glyph * g, PyObject * args)
Py_TRY
{
	RGB * rgb;

	if (!PyArg_ParseTuple (args, "O!", RGB_type, &rgb))
		return 0;
	
	GlyphRender r (rgb);

	FT_Outline & o = g->outline();

	FT_Raster_Params raster = {
		0, /* target */
		0, /* source */
		ft_raster_flag_aa | ft_raster_flag_clip | ft_raster_flag_direct, /* flags */
		gray_spans, 
		black_spans,
		0, /* bit_test */
		0, /* bit_set */
		&r /* user */
	};

	raster.clip_box.xMin = rgb->clip.x;
	raster.clip_box.xMax = rgb->clip.x + rgb->clip.width;
	raster.clip_box.yMin = rgb->clip.y;
	raster.clip_box.yMax = rgb->clip.y + rgb->clip.height;

	FT ("FT_Outline_Render", FT_Outline_Render (library, &o, &raster));

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
Glyph_methods[] =
{
	{"render", (PyCFunction) Glyph_render, METH_VARARGS,
	 "render (rgb) -- render the outline into an rgb buffer"},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

Type * Glyph_type;

char * Py<Glyph>::__doc__ = 
"Glyph() -- one glyph.";

void
Py<Glyph>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();

	throw NotImplementedError();
}

void
Py<Glyph>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Object_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_methods = Glyph_methods;
	t->tp_getset = Glyph_getsets;
}

/* //////////////////////////////////////////////////////////////////////// */

void
init_FreeType_types (PyObject * dict)
{
	FreeType_type = new PyType<FreeType> ("FreeType", dict);
	Glyph_type = new PyType<Glyph> ("Glyph", dict);
}
