#include <basics.h>
#include <EventList.h>
#include <Graph.h>
#include <PyType.h>

#include "Gtk.h"
#include "Pixels.h"
#include "MidiView.h"

Type * MidiView_type;

namespace MidiView {

/* minimum pixels between adjacent tick marks for the time grid: */
#define TIME_STEP_PIXELS 26

MidiView::MidiView (Type * t)
: RGBView (t),
	lists ("midi view lists"),
	marks ("midi view marks")
{
	graph = 0;
	gc = 0;

	init_note_tables();
	
	x0 = y0 = -50;

	pixels_per_note = 3;
	ticks_per_pixel = pow (2, 2.5);

	snap = 48;

	staff_limit[0] = 24;
	staff_limit[1] = 96;

	time_map = NEW (TimeMap);
}

MidiView::~MidiView()
{
	if (gc) gdk_gc_destroy (gc);

	::decref (time_map, true);
	::decref (graph, true);
}

void
MidiView::configure (const Rect<int> r)
{
	/* called from Gtk, who doesn't know about the lock */
	ScopedLock lock (mutex);

	RGBView::configure (r);

	clip = r;

	if (gc == 0)
		gc = gdk_gc_new (this->window());

	complete_redraw();
}

/* since we are drawing to the RGB buffer, we do not need the Gtk lock
 * around drawing routines. it so happens that rgb_to_pixels does no X
 * calls either.
 *
 * the view mutex should be held however, and overlays (marks) hidden. 
 */
void
MidiView::redraw (const Rect<int> & r)
{
	clip = rgb->clip = r;
	rgb->fill (clip, Color (gdk_base()));

	TimeGrid (this).draw();
	draw_stafflines();
	draw_notes();

	RGBView::rgb_to_pixels (clip);
}

void
MidiView::scroll (int dx, int dy)
{
	if (!(dx | dy))
		return;

	dy = -dy;

	x0 += dx;
	y0 += dy;
	
	dx = -dx;

	int adx = abs (dx);
	int ady = abs (dy);

	/* size of the scrolled area */
	int width = pixels->width - adx;
	int height = pixels->height - ady;

	/* scroll more than one screenful: redraw everything. */
	if (width < 0 || height < 0)
	{
		complete_redraw();
		return;
	}

	hide_marks (1);

	/* source and dest coordinates for copy_area: */
	int xx0 = (dx < 0) ? -dx : 0;
	int xx1 = (dx < 0) ? 0 : dx;
	
	int yy0 = (dy < 0) ? -dy : 0;
	int yy1 = (dy < 0) ? 0 : dy;
	
	/* we know how to scroll the pixmap now: */
	{
		gdk_window_copy_area
			(
					pixmap(), fg_gc(),
					xx1, yy1,
					pixmap(),
					xx0, yy0,
					width, height
			);
	}

	/* newly axposed areas: */
	Rect<int> rx (0, 0, 0, 0), ry (0, 0, 0, 0);

	if (dx)
	{
		int x = (dx < 0) ? pixels->width + dx : 0;
		rx.set_from (x, 0, adx, pixels->height);
	}
	if (dy)
	{
		int y = (dy < 0) ? pixels->height + dy : 0;
		ry.set_from (0, y, pixels->width, ady);
	}

	/* exposed area sizes: */
	int arx = (int) rx;
	int ary = (int) ry;

	/* seeing that we loop over a lot of events for every redraw, i think it's
	 * more efficient to do a complete redraw if both x and y are scrolled. 
	 */
	if (arx && ary)
		redraw (Rect<int> (0, 0, pixels->width, pixels->height));
	else 
	{
		if (arx)
			redraw (rx);
		else
			redraw (ry);
	}

	/* scroll the window */
	gdk_window_copy_area
		(
				window(), fg_gc(),
				xx1, yy1,
				window(),
				xx0, yy0,
				width, height
		);
	
	recalc_marks();
	show_marks (1);

	/* blit the newly exposed areas */
	if (arx)
		blit (rx);
	if (ary)
		blit (ry);

	/* unfortunately necessary */
	gdk_flush();
}

void
MidiView::complete_redraw()
{
	clip = rgb->size;

	hide_marks (0);
	redraw();
	recalc_marks();
	show_marks (0);
	
	blit();
	gdk_flush();
}

/* /////////////////////////////////////////////////////////////////////// */

void
MidiView::show_marks (bool blit)
{
	for (int i = 0; i < marks; ++i)
		marks[i]->show (blit);
}

void
MidiView::hide_marks (bool blit)
{
	for (int i = 0; i < marks; ++i)
		marks[i]->hide (blit);
}

void
MidiView::recalc_marks()
{
	for (int i = 0; i < marks; ++i)
		marks[i]->make_rect();
}

/* /////////////////////////////////////////////////////////////////////// */

bool
MidiView::hit_test (int x, int y, Object ** part, Object ** event)
{
	Rect<int> r;

	for (int i = 0; i < lists; ++i)
	{
		EventList & l = *(lists[i]);
		
		ScopedLock lock (l.mutex);
		
		int stop = l.index_of_tick (x2tick (x + 1));
		
		for (int j = 0; j < stop; ++j)
		{
			Note * note = (Note *) l[j];
			
			if (get_type (note) != Note_type)
				continue;
			
			note_rect (note, r);

			if (!r.has (x, y))
				continue;
			
			*part = &l;
			*event = note;
			return true;
		}
	}

	return false;
}

/* drawing routines */

/* iterate all lists and draw the contents */
void
MidiView::draw_notes()
{
	d_tick stop = x2tick (pixels->width);

	for (int i = 0; i < lists; ++i)
	{
		EventList & l = *(lists[i]);
		
		ScopedLock lock_list (l.mutex);

		for (int j = 0; j < l; ++j)
		{
			Note * note = (Note *) l[j];
			
			if (note->tick > stop)
				break;

			if (get_type (note) == Note_type)
				draw_note (note, l.color);
		}
	}
}

/* draw one note */
void
MidiView::draw_note (Note * note, unsigned int color)
{
	Rect<int> r;

	note_rect (note, r);

	/* fill, checking if r is empty */
	if (!rgb->fill (r, AlphaColor (color, 128)))
		return;

	/* draw helper lines from staves to note: */
	int n = lrintf (note->note()) & 127;
	int g = gravity[n];

	for (int i = n - g; i >= 0 && i < 128; i += g)
	{
		if (!gravity[i])
			break;

		if (!is_staff[i])
			continue;

		int y = note2y (i);
		int dx = (i == n) ? -2 : +2;
			
		rgb->fill 
			(
				Rect<int> (r.x + dx, y, r.width - 2 * dx, 1), 
				AlphaColor (gdk_dark(), 50)
			);
	}
}

void
MidiView::draw_stafflines()
{
	int system_alpha = 24;

	if (system_alpha)
	{
		int up, lo;

		/* f (bass) staff background fill */
		up = note2y (57);
		lo = note2y (43) + 1;

		rgb->fill 
			(
				Rect<int> (clip.x, up, clip.width, lo - up),
				AlphaColor (gdk_dark(), system_alpha)
			);

		/* g (soprano) staff background fill */
		up = note2y (77);
		lo = note2y (64) + 1;

		rgb->fill 
			(
				Rect<int> (clip.x, up, clip.width, lo - up),
				AlphaColor (gdk_dark(), system_alpha)
			);
	}

	/* individual staff lines */
	Rect<int> line (-1, -1, pixels->width, 1);
	if (line.x < 0) line.x = 0;

	for (int n = staff_limit[0]; n < staff_limit[1]; ++n)
	{
		int a = alpha[n];
		if (a == 0)
			continue;
	
		line.x = (a >= 128) ? 0 : max (tick2x (0), 0);
		line.y = note2y (n);

		rgb->fill (line, AlphaColor (gdk_dark(), a));
	}
}

/* /////////////////////////////////////////////////////////////////////// */

/* set up tables describing the staff line system */

void
MidiView::init_note_tables()
{
	memset (is_staff, 0, sizeof (is_staff));
	memset (gravity, 0, sizeof (gravity));
	memset (alpha, 0, sizeof (gravity));

	int zone_index = 0;
	int	current_y = 0;

	for (int note = 0; note < 128; note++)
	{
		/* staff lines are only drawn for particular pitches. */
		switch ((note + 12) % 24)
		{
			case 0:
			case 7:
			case 14:		/* 'D' actually is a 3 fold zone */
			case 17:
				is_staff[note] = true;
				zone_index = 0;
				break;
			case 4:
			case 11:
			case 21:
				is_staff[note] = true;
				zone_index = 1;
				break;
		}

		if (is_staff[note])
			current_y += 4,
			n2y[note] = current_y;
			
		else
			n2y[note] = current_y + zone_index;
			
		zone_index++;

		/* gravity indicates whether to go up or down when drawing helper lines,
		 * depending on whether the note is below, between or above the two
		 * staves. 
		 */
		if (note < 43)
			gravity [note] = +1;
			
		else if (note <= 57)
			;
			
		else if (note <= 60)
			gravity [note] = +1;
			
		else if (note < 63)
			gravity [note] = -1;

		else if (note <= 77)
			;

		else
			gravity [note] = -1;

		/* transparency values for the staff lines */
		if (is_staff[note])
		{
			if (!gravity [note])
				alpha [note] = 144;
			else switch (note % 12)
			{
				case 0:
					alpha[note] = 0x60; break;
				case 5:
				case 7:
					#if 0
					alpha[note] = 0x30; break;
					#endif
				default:
					alpha[note] = 0x00; break;
			}
		}
	}

	/* normalize to 0 up left. */
	for (int n = 0; n < 128; ++n)
		n2y[n] = - (n2y[n] - current_y);
}

/* // TimeGrid ///////////////////////////////////////////////////////////// */

TimeGrid::TimeGrid (MidiView * v)
: color (v->gdk_dark())
{
	view = v;
	
	int x0 = view->tick2x (0);

	last_mark = x0 - 4 * TIME_STEP_PIXELS;

	stop = (int) view->x2tick (
			 	view->clip.x + view->clip.width + 2 * TIME_STEP_PIXELS);

	font_height = view->rgb->font_height();

	/* set up the line rect: */
	int y0 = view->note2y (view->staff_limit[1]);
	int y1 = view->note2y (0); //view->staff_limit[0]);

	line.set_from (0, y0, 1, y1 - y0);
}

void
TimeGrid::draw()
{
	bar = beat = 0;
	parse_time (&default_time);

	Iterator<TimeEvent *> it (&view->time_map->events, 0);
	d_tick next;

	if (it.have()) /* have time events */
		next = it.get()->tick;
	else
	{
		next = stop;
		draw_time (&default_time);
	}
	
	tick = 0;

	while (1)
	{
		while (tick < next)
			proceed();

		if (tick > stop)
			break;
		
		if (tick == stop)
		{
			proceed();
			break;
		}

		while (it.have() && tick >= next)
		{
			parse_time (it.get());
			draw_time (it.get());
			it.next();
			next = it.have() ? (int) it.get()->tick : stop;
		}

		tick = (int) current->tick;
	}	
}

void 
TimeGrid::proceed()
{
	register int ticks = (int) current->ticks;

	draw_mark (view->tick2x (tick));

	fract += step;
	weight = (weight + 1) % weight_max;

	while (fract >= ticks)
	{
		if (++beat >= current->n)
		{
			++bar,
			beat = 0;
		}

		fract -= ticks;
	}

	tick += step;
}

void
TimeGrid::parse_time (TimeEvent * e)
{
	current = e;
	
	weight = 0;
	fract = 0;

	int beat = (int) e->ticks;

	/* choose 'step' for a reasonable step size for the time grid. 
	 * first, try denominator units: */
	step = beat;
	weight_max = 1;
	int pixels = lrint ((d_tick) step / view->ticks_per_pixel);

	/* step is too small to draw beat units: */
	if (pixels < TIME_STEP_PIXELS)
	{
		sub_bar_labels = 0;

		/* try bar units. */
		step *= e->n;
	}
	else /* step is large enough to check for sub-beat units: */
	{
		int snap = (int) view->snap;

		while (step > snap && pixels >= 2 * TIME_STEP_PIXELS)
		{
			int n = step / snap, m = step % snap;
			if (m) step -= m;
			else if (n & 1) step -= snap;
			else step /= 2;
			pixels = lrint ((d_tick) step / view->ticks_per_pixel);
		}

		weight_max = e->ticks / step;

		sub_bar_labels = 1;
	}
}

void
TimeGrid::draw_time (TimeEvent * e)
{
	AlphaColor ac (color, 160);
	Rect<int> r (view->tick2x (e->tick), view->note2y (127), 1, font_height * 2);

	view->rgb->fill (r, ac);

	char s[12];
	sprintf (s, "%d", e->n);
	view->rgb->text (r.x + 2, r.y + font_height, s, ac);
	sprintf (s, "%d", e->d);
	view->rgb->text (r.x + 2, r.y + 2 * font_height, s, ac);
}

void
TimeGrid::draw_mark (int x)
{
	if (!sub_bar_labels && (x - last_mark) < TIME_STEP_PIXELS)
		return;

	last_mark = x;
	
	/* simple visibility preselect (note that labels matter [-48]): */
	if (x < -48 || x > view->clip.x + view->clip.width)
		return;

	line.x = x;

	int a;
	Rect<int> r (line);

	if (beat || fract)
		a = (weight ? 36 : 72);

	else /* beat one of a bar */
	{
		a = 160;

		r.y -= view->pixels_per_note;
		r.height += view->pixels_per_note;
	}	

	view->rgb->fill (r, AlphaColor (color, a));

	if (fract == 0)
	{
		int y = line.y + font_height;

		char s[24];

		if (beat == 0) 
		{
			sprintf (s, "%d", bar);
			view->rgb->text (x + 2, y, s, AlphaColor (color, a));
		}
		else if (sub_bar_labels)
		{
			sprintf (s, "%d", beat + 1);
			view->rgb->text (x + 2, y, s, AlphaColor (color, a));
		}
	}
}

/* Mark ////////////////////////////////////////////////////////////////// */

Mark::Mark()
{
	visible = 0;
	tick = 0;
	rect.y = 0;
	rect.width = 1;
}
		
Mark::~Mark()
{
	gdk_colormap_free_colors (gdk_colormap_get_system(), &color, 1);
}

void
Mark::init (MidiView * v, char * name, d_tick t, const Color & c)
{
	set_object_name (name);

	tick = t;

	color.red = c.r << 8;
	color.green = c.g << 8;
	color.blue = c.b << 8;
	gdk_colormap_alloc_color (gdk_colormap_get_system(), &color, true, true);

	view = v;
	
	make_rect();
}

void Mark::make_rect()
{
	rect.x = view->tick2x (tick);
	rect.height = view->rgb->size.height;
}

void
Mark::hide (bool blit)
{
	if (!visible)
		return;

	visible = false;

	Rect<int> r;
	if (draw (r) && blit)
		view->blit (r);
}

void
Mark::show (bool blit)
{
	if (visible)
		return;

	visible = true;

	Rect<int> r;
	if (draw (r) && blit)
		view->blit (r);
}

bool
Mark::draw (Rect<int> & r)
{
	r = rect;

	if (view->gc == 0)
		return false;
	
	gdk_gc_set_foreground (view->gc, &color);
	gdk_gc_set_function (view->gc, GDK_XOR);
	gdk_draw_line (view->pixmap(), view->gc, r.x, r.y, r.x, r.y + r.height);

	return true;
}

/* Py<MidiView::Mark> ////////////////////////////////////////////////////// */

static PyObject *
Mark_get_tick (Mark * m)
{
	return PyFloat (m->tick);
};

static int
Mark_set_tick (Mark * m, PyObject * arg)
{
	d_tick t;

	if (!PyArg_Parse (arg, "f", &t))
		return -1;

	/* easy way out in case there's nothing to do */
	if (m->tick == t)
		return 0;
	
	m->tick = t;
	
	ScopedLock lock (m->view->mutex);

	int x = m->view->tick2x (t);
	if (x == m->rect.x)
		return 0;

	m->hide (1);
	m->rect.x = x;
	m->show (1);

	gdk_flush();

	return 0;
}

static PyObject *
Mark_get_color (Mark * m)
{
	return Py_BuildValue ("iii", 
			m->color.red >> 8,
			m->color.green >> 8,
			m->color.blue >> 8);
}

static PyObject *
Mark_get_visible (Mark * m)
{
	return PyInt (m->visible);
};

static PyGetSetDef 
Mark_getsets[] = {
	{"tick", (getter) Mark_get_tick, (setter) Mark_set_tick, 
	 "when."},
	{"color", (getter) Mark_get_color, (setter) 0, 
	 "used to draw the mark."},
	{"visible", (getter) Mark_get_visible, (setter) 0, 
	 "whether the mark is visible."},
	{ }
};

/* Mark::methods */

static PyObject *
Mark_show (Mark * m)
{
	ScopedLock lock (m->view->mutex);

	m->show (1);
	
	gdk_flush();

	return PyNone();
}

static PyObject *
Mark_hide (Mark * m)
{
	ScopedLock lock (m->view->mutex);

	m->hide (1);

	gdk_flush();

	return PyNone();
}

static PyObject *
Mark_move (Mark * m, PyObject * arg)
{
	return Mark_set_tick (m, arg) ? 0 : PyNone();
}

static PyMethodDef 
Mark_methods[] = {
	{"show", (PyCFunction) Mark_show, METH_NOARGS, "show() -- show the mark."},
	{"hide", (PyCFunction) Mark_hide, METH_NOARGS, "hide() -- hide the mark."},
	{"move", (PyCFunction) Mark_move, METH_O, "move (tick) -- move the mark."},
	{ }
};

/* type (Mark) */

Type * Mark_type;

char * Py<Mark>::__doc__ = 
"MidiView.Mark() -- marks a point in time within a MidiView.";

void
Py<Mark>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "use MidiView.new_mark()";
}

void
Py<Mark>::type_init (Type * t, PyObject * dict)
{
	/* not sure if i find this nice; gcc 2.95 wants it en tous cas */
	using namespace MidiView;

	t->tp_base = NamedObject_type;
	t->tp_getset = Mark_getsets;
	t->tp_methods = Mark_methods;
}

/* //////////////////////////////////////////////////////////////////////// */

/* Py<MidiView> */

static PyObject * 
MidiView_get_x0 (MidiView * v)
{
	return PyInt (v->x0);
}

static PyObject * 
MidiView_get_y0 (MidiView * v)
{
	return PyInt (v->y0);
}

static int
MidiView_set_x0 (MidiView * v, PyObject * x)
{
	if (!PyArg_Parse (x, "i", &v->x0))
		return -1;
	return 0;
}

static int
MidiView_set_y0 (MidiView * v, PyObject * y)
{
	if (!PyArg_Parse (y, "i", &v->y0))
		return -1;
	return 0;
}

static PyObject * 
MidiView_get_snap (MidiView * v)
{
	return PyInt ((int) v->snap);
}

static int
MidiView_set_snap (MidiView * v, PyObject * x)
Py_TRY
{
	int i;

	if (!PyArg_Parse (x, "i", &i))
		return -1;

	if (i < 1)
		throw ValueError() << "snap must be >= 1";
	if (i > QUARTER * QUARTER)
		throw ValueError() << "snap must be < " << QUARTER * QUARTER;

	v->snap = i;

	return 0;
}
Py_CATCH (-1);

static PyObject * 
MidiView_get_time_map (MidiView * v)
{
	return ::incref (v->time_map);
}

static int
MidiView_set_time_map (MidiView * v, PyObject * x)
{
	TimeMap * map;

	if (!PyArg_Parse (x, "O!", TimeMap_type, &map))
		return -1;

	ScopedLock (v->mutex);
	::decref (v->time_map);
	::incref (v->time_map = map);

	return 0;
}

static PyObject *
MidiView_get_marks (MidiView * v)
{
	return PyTuple (&v->marks);
}

static PyGetSetDef 
MidiView_getsets [] = 
{
	{"x0", (getter) MidiView_get_x0, (setter) MidiView_set_x0, 
	 "the left pixel offset."},
	{"y0", (getter) MidiView_get_y0, (setter) MidiView_set_y0, 
	 "the top pixel offset."},
	{"snap", (getter) MidiView_get_snap, (setter) MidiView_set_snap, 
	 "the basic tick grid step."},
	{"time_map", (getter) MidiView_get_time_map, (setter) MidiView_set_time_map, 
	 "the time map."},
	{"marks", (getter) MidiView_get_marks, (setter) 0, 
	 "the marks."},
	{ }
};

/* sequence methods */

static int 
MidiView_sq_length (MidiView * v)
{
	ScopedLock lock (v->mutex);
	return v->lists;
}

static PyObject *
MidiView_sq_item (MidiView * v, int i) 
Py_TRY
{
	ScopedLock lock (v->mutex);
	return ::incref (v->lists [i]); 
} 
Py_CATCH (0);

static PySequenceMethods 
MidiView_as_sequence = 
{
	(inquiry) MidiView_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) MidiView_sq_item, /* sq_item */
	(intintargfunc) 0, /* sq_slice */
	(intobjargproc) 0, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject * 
MidiView_redraw (MidiView * v)
Py_TRY
{
	ScopedLock lock (v->mutex);

	v->complete_redraw();
		
	return PyNone();
}
Py_CATCH (0);

static PyObject *
MidiView_new_mark (MidiView * v, PyObject * args)
{
	char * name;
	d_tick t;
	Color c (0, 0, 0);
	
	if (!PyArg_ParseTuple (args, "sf(iii)", &name, &t, &c.r, &c.g, &c.b))
		return 0;

	/* we xor to the screen, assumed white, so xor now and get a reasonable
	 * result. */
	c.r ^= 255;
	c.g ^= 255;
	c.b ^= 255;

	Mark * m = NEW (Mark);

	ScopedLock lock (v->mutex);
	m->init (v, name, t, c);
	v->marks.add_object (m);
	
	return m;
}

static PyObject *
MidiView_remove_mark (MidiView * v, Mark * m)
Py_TRY
{
	if (!Mark_type->is_derived (m))
		throw TypeError() << "need a Mark, not " << get_type_name (m);

	ScopedLock lock (v->mutex);
	int i = v->marks.find (m);

	if (i < 0)
		throw LookupError() << "not in this view.";
	
	v->marks.remove_items (i, 1);

	return PyNone();
}
Py_CATCH (0);

static PyObject *
MidiView_add (MidiView * v, EventList * arg) 
Py_TRY
{
	if (!EventList_type->is_derived (arg))
		throw TypeError() << "need an EventList, not " << get_type_name (arg);

	ScopedLock lock (v->mutex);
	v->lists.add_object (arg);

	return PyNone();
} 
Py_CATCH (0);

static PyObject *
MidiView_remove (MidiView * v, EventList * arg) 
Py_TRY
{
	if (!EventList_type->is_derived (arg))
		throw TypeError() << "need an EventList, not " << get_type_name (arg);

	ScopedLock lock (v->mutex);
	int i = v->lists.find (arg);

	if (i < 0)
		throw LookupError() << "not in this view.";
	
	v->lists.remove_items (i, 1);

	return PyNone();
} 
Py_CATCH (0);

/* scroll */

static PyObject * 
MidiView_scroll (MidiView * v, PyObject * args)
Py_TRY
{
	int dx, dy = 0;

	if (!PyArg_ParseTuple (args, "i|i", &dx, &dy))
		return 0;

	ScopedLock lock (v->mutex);

	v->scroll (dx, dy);

	return PyNone();
}
Py_CATCH (0);

/* zoom */

static PyObject * 
MidiView_zoom_x (MidiView * v, PyObject * args)
Py_TRY
{
	ScopedLock lock (v->mutex);

	double delta;
	int x_align;

	if (!PyArg_ParseTuple (args, "di", &delta, &x_align))
		return 0;

	d_tick t_align = v->x2tick (x_align);
	double scale = sqrt (2.);

	int z = lrint (log (v->ticks_per_pixel) / log (scale));
	z -= delta < 0 ? -1 : +1;
	if (z <= -8) z = -8;

	d_tick tpp = pow (scale, z);

	if (tpp == v->ticks_per_pixel)
		return PyNone();

	v->ticks_per_pixel = tpp;

	/* find out how much the alignment point moved */
	int dx = x_align - v->tick2x (t_align);
	v->x0 -= dx;

	v->complete_redraw();

	return PyNone();
}
Py_CATCH (0);

static PyObject * 
MidiView_zoom_y (MidiView * v, PyObject * args)
Py_TRY
{
	ScopedLock lock (v->mutex);

	double delta;
	int align_y;

	if (!PyArg_ParseTuple (args, "di", &delta, &align_y))
		return 0;

	int z = v->pixels_per_note;
	z += delta < 0 ? -1 : +1;
	if (z <= 0) z = 1;

	if (z == v->pixels_per_note)
		return PyNone();

	v->pixels_per_note = z;

	v->complete_redraw();

	return PyNone();
}
Py_CATCH (0);

/* translation methods */

static PyObject *
MidiView_tick2x (MidiView * v, PyObject * arg)
{
	d_tick t;
	
	if (!PyArg_Parse (arg, "f", &t))
		return 0;
	
	return PyInt (v->tick2x (t));
}

static PyObject *
MidiView_x2tick (MidiView * v, PyObject * arg)
{
	int x;

	if (!PyArg_Parse (arg, "i", &x))
		return 0;

	return PyFloat (v->x2tick (x));
}

static PyObject *
MidiView_note2y (MidiView * v, PyObject * arg)
{
	d_note n;
	
	if (!PyArg_Parse (arg, "f", &n))
		return 0;
	
	return PyInt (v->note2y (n));
}

static PyObject *
MidiView_y2note (MidiView * v, PyObject * arg)
{
	int y;

	if (!PyArg_Parse (arg, "i", &y))
		return 0;

	return PyFloat (v->y2note (y));
}

static PyObject *
MidiView_hit_test (MidiView * v, PyObject * args)
{
	int x, y;
	
	if (!PyArg_ParseTuple (args, "ii", &x, &y))
		return 0;

	Object * p, * e;
	
	if (!v->hit_test (x, y, &p, &e))
		return PyNone();

	return Py_BuildValue ("OO", p, e);
}

static PyObject *
MidiView_get_marks_around (MidiView * v, PyObject * arg)
{
	if (v->marks < 3)
		return PyNone();
	
	d_tick t, t0 = 0, t1 = D_TICK_MAX;
	
	if (!PyArg_Parse (arg, "f", &t))
		return 0;

	/* i = 1: skip 'tick' mark */
	for (int i = 1; i < v->marks; ++i)
	{
		d_tick tm = v->marks[i]->tick;

		if (tm < t)
		{
			if (tm > t0)
				t0 = tm;
		}
		else if (tm > t)
		{
			if (tm < t1)
				t1 = tm;
		}
	}

	/* allow t0 = 0 */
	if (t1 == D_TICK_MAX)
		return PyNone();

	return Py_BuildValue ("ff", t0, t1);
}

static PyMethodDef 
MidiView_methods [] = 
{
  {"add", (PyCFunction) MidiView_add, METH_O,
	 "add (eventlist) -- add an eventlist to the view."},
  {"new_mark", (PyCFunction) MidiView_new_mark, METH_VARARGS,
	 "new_mark (name, tick, color = (0, 0, 0)) -- create a mark at 'tick'. returns "
	 "the mark object."},
	{"remove", (PyCFunction) MidiView_remove, METH_O,
	 "remove (eventlist) -- remove 'eventlist'."},
	{"remove_mark", (PyCFunction) MidiView_remove_mark, METH_O,
	 "remove_mark (mark) -- remove 'mark'."},
	{"get_marks_around", (PyCFunction) MidiView_get_marks_around, METH_O,
	 "get_marks_around (tick) -- returns (t0, t1) ticks of the closest previous "
	 "and next mark around 'tick', or None."},
	{"hit_test", (PyCFunction) MidiView_hit_test, METH_VARARGS,
	 "hit_test (x, y) -- return the object at (x, y) or None."},
	{"redraw", (PyCFunction) MidiView_redraw, METH_NOARGS,
	 "redraw() -- complete redraw."},
	{"scroll", (PyCFunction) MidiView_scroll, METH_VARARGS,
	 "scroll (dx, dy) -- scroll the view by (dx, dy) pixels."},
	{"tick2x", (PyCFunction) MidiView_tick2x, METH_O,
	 "tick2x (tick) -- translate 'tick' into a pixel offset."},
	{"x2tick", (PyCFunction) MidiView_x2tick, METH_O,
	 "x2tick (x) -- translate 'x' into a tick."},
	{"note2y", (PyCFunction) MidiView_note2y, METH_O,
	 "note2y (note) -- translate 'note' into a pixel offset."},
	{"y2note", (PyCFunction) MidiView_y2note, METH_O,
	 "y2note (note) -- translate 'y' into a note pitch."},
	{"zoom_x", (PyCFunction) MidiView_zoom_x, METH_VARARGS,
	 "zoom_x (delta, x) -- zoom in or out horizontally, depending on the sign "
	 "of 'delta'. other than that, the value of 'delta' is ignored. 'x' is "
	 "consulted for realigning the view (should be the mouse position)."},
	{"zoom_y", (PyCFunction) MidiView_zoom_y, METH_VARARGS,
	 "zoom_y (delta) -- zoom in or out vertically, depending on the sign "
	 "of 'delta'. other than that, the value of 'delta' is ignored. 'y' is "
	 "consulted for realigning the view (should be the mouse position)."},
	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

char * Py<MidiView>::__doc__ = 
"MidiView() -- a RGBView displaying the contents of Midi event lists.";

void
Py<MidiView>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys[] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();

	PixelView::init();

	GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_FOCUS);
	gtk_widget_set_name (widget, "midi");
}

void
Py<MidiView>::type_init (Type * t, PyObject * dict)
{
	/* not sure if i find this nice; gcc 2.95 wants it en tous cas */
	using namespace MidiView;

	t->tp_base = RGBView_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = MidiView_getsets;
	t->tp_methods = MidiView_methods;
	t->tp_as_sequence = &MidiView_as_sequence;

	/* needed to have t->tp_dict */
	PyType_Ready (t);

	Mark_type = new PyType<Mark> ("Mark", t->tp_dict);
}

}; /* namespace MidiView */

/* /////////////////////////////////////////////////////////////////////// */

void
init_MidiView_types (PyObject * dict)
{
	MidiView_type = new PyType<MidiView::MidiView> ("MidiView", dict);
}
