#ifndef _MIDI_VIEW_H_
#define _MIDI_VIEW_H_

#include "TimeMap.h"

/* for time/pitch - spatial translation, it would be nice to stay in 
 * the float domain, but for pixel-oriented displays we lose some 
 * potential to use color shades to discern different event weights.
 * 
 * if, for example, a vertical line expressing a point in time 
 * is antialiased correctly, ie spreading visual energy across two
 * columns of pixels on-screen, we cannot discern from two points
 * in time that are assigned lesser weight, ie visual energy. 
 *
 * i think that giving up some resolution in the time domain for
 * the possibility of gathering multiple layers of expression is
 * the better way, given that note height and tick positions 
 * usually are discrete numbers, although both are typed floats. 
 */

/* should be visible outside the namespace */
extern Type * MidiView_type;

namespace MidiView {
	
extern Type * Mark_type;

class MidiView;

class Mark
: public NamedObject
{
	public:
		d_tick tick;

		GdkColor color;
		int visible;
		Rect<int> rect;

		/* not incref'd, lives and dies with the view */
		MidiView * view;

		Mark();
		~Mark();

		void init (MidiView * mv, char * name, d_tick t, const Color & c);
		void make_rect();

		void hide (bool blit = 0);
		void show (bool blit = 0);
		void move (d_tick tick);
		bool draw (Rect<int> & r); /* returns false if widget not drawable yet */
};

class Selection
: public EventList
{
};

class MidiView
: public RGBView
{
	public:
		Mutex mutex;
		TypedObjectList<EventList> lists;
		TypedObjectList<Mark> marks;
		TimeMap * time_map;
		Graph * graph;
		GdkGC * gc;

	public:
		MidiView (Type * t = MidiView_type);
		~MidiView();	

		void init();

		int x0, y0;
		
		d_tick snap; 

		d_tick ticks_per_pixel; /* x */
		int pixels_per_note;    /* y, note: this is = (actual height + 1) / 2 */

		int staff_limit[2];

		int tick2x (d_tick t)
			{
				return (int) (t / ticks_per_pixel) - x0;
			}
		
		d_tick x2tick (int x)
			{
				return (d_tick) (x + x0) * ticks_per_pixel;
			}
		
		/* note - y translation needs to be set up. */
		int n2y[128], is_staff[128], alpha[128], gravity[128];
		
		void init_note_tables();

		int note2y (d_note n)
			{
				return note2y (lrintf (n));
			}
		
		int note2y (long int i)
			{
				return note2y ((int) i);
			}

		int note2y (int i)
			{
				return y0 + (int) (pixels_per_note * n2y [i & 127]);
			}
		
		d_note y2note (int y)
			{
				/* todo. */
				return 127;
			}
		
		void scroll (int dx, int dy);
		
	public:
		Rect<int> clip;

		void redraw (const Rect<int> &);
		void redraw()
			{
				redraw (Rect<int> (0, 0, pixels->width, pixels->height));
			}
		
		void complete_redraw();

		bool hit_test (int x, int y, Object ** part, Object ** event);

	public:
		void draw_stafflines();
		void draw_note (Note *, unsigned int color);
		void draw_notes();

		/* mark management */
		void show_marks (bool blit = 1);
		void hide_marks (bool blit = 1);
		void recalc_marks();

  protected:
		virtual void configure (const Rect<int> size);

		inline void note_rect (Note * note, Rect<int> & r)
			{
				r.x = tick2x (note->tick);
				r.y = note2y (note->note()) - pixels_per_note - 1;
				r.width = tick2x (note->tick + note->l) - r.x;
				r.height = 2 * pixels_per_note + 3;
			}
};

/* /////////////////////////////////////////////////////////////////////// */

/* TimeGrid steps through an EventList of TimeEvents and draws vertical tick
 * marks into a MidiView RGB buffer, counting musical time and drawing bar
 * and beat numbers next to the marks.
 *
 * ticks are counted as ints instead of d_tick (float) to maintain precision
 * when counting over long ranges.
 */

class TimeGrid
{
	public:
		MidiView * view;
		TimeMap * map;
		
		TimeEvent default_time;
		TimeEvent * current;

		Color color;

		int	tick, /* current */
				step, /* set in parse_time() */
				stop; /* tick at x == width */

		int bar, 
				beat, 
				fract;

		int	weight, /* counts steps mod weight_max. */
				weight_max;

		int last_mark; /* x of last drawn tick mark. */

		bool sub_bar_labels; /* text labels on a sub-beat scale. */

		int font_height;

		Rect<int> line;
		
		TimeGrid (MidiView * v);

		inline void draw();
		inline void proceed();
		inline void draw_mark (int x);
		inline void parse_time (TimeEvent * e);

		void draw_time (TimeEvent * e);
};

} /* namespace MidiView */

#endif /* _MIDI_VIEW_H_ */
