#include <basics.h>
#include <PythonThread.h>
#include <PyType.h>
#include <AudioSignal.h>

#define NO_IMPORT_PYGOBJECT
#include <pygobject.h>

#include "Gtk.h"
#include "Pixels.h"

Color::Color (PyObject * arg)
{
	if (PyString_Check (arg))
	{
		char * s = PyString_AsString (arg);
		
		if (s == 0)
			throw blewit();
		
		if (sscanf (s, "#%02x%02x%02x", &r, &g, &b) != 3)
			throw TypeError() << "cannot parse '" << s << "' into a color.";
	}
	else
	{
		if (!PyArg_Parse (arg, "iii", &r, &g, &b))
			throw blewit();
	}
}

/* //////////////////////////////////////////////////////////////////////// */

Rect<int>::Rect (PyObject * arg)
{
	if (!PyArg_Parse (arg, "iiii", &x, &y, &width, &height))
		throw blewit();
}

/* Pixels ///////////////////////////////////////////////////////////////// */

void
Pixels::init (int w, int h, GtkWidget * widget)
{
	gtk_object_ref ((GtkObject *) self.widget = widget);

	if (w > 0 && h > 0)
		resize (w, h);
}

void
Pixels::resize (int w, int h)
{
	if (self.widget == 0)
		throw AssertionError() << "no widget, and no stored widget.";

	width = w;
	height = h;

	if (self.pixmap) gdk_pixmap_unref (self.pixmap);

	gtk_widget_push_visual (gdk_rgb_get_visual());
	gtk_widget_push_colormap (gdk_rgb_get_cmap());

	self.pixmap = gdk_pixmap_new (self.widget->window, width, height, -1);

	gtk_widget_pop_visual();
	gtk_widget_pop_colormap();
}

void
Pixels::draw (RGB * rgb, int x, int y, int w, int h, int x_rgb, int y_rgb)
{
	if (rgb->pixels == 0)
		throw AssertionError() << "no pixels in " << rgb;

	if (self.pixmap == 0)
		throw AssertionError() << "no pixmap in " << this;

	if (w <= 0 || h <= 0)
		return;

	if (x_rgb < 0) x_rgb = x;
	if (y_rgb < 0) y_rgb = y;

	if (w + x_rgb > rgb->size.width) w = rgb->size.width - x_rgb;
	if (h + y_rgb > rgb->size.height) h = rgb->size.height - y_rgb;

	gdk_draw_rgb_image 
		(
			self.pixmap,
			self.widget->style->base_gc[0],
			x, y,
			w, h,
			GDK_RGB_DITHER_NORMAL,
			rgb->pointer (x_rgb, y_rgb),
			rgb->rowstride
		);
}

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
Pixels_get_width (Pixels * p)
{
	return PyInt (p->width);
}

static PyObject *
Pixels_get_height (Pixels * p)
{
	return PyInt (p->height);
}

static PyGetSetDef
Pixels_getsets [] = 
{
	{"width", (getter) Pixels_get_width, (setter) 0,
		"."},
	{"height", (getter) Pixels_get_height, (setter) 0,
		"."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
Pixels_draw (Pixels * p, RGB * rgb)
Py_TRY
{
	if (!RGB_type->is_derived (rgb))
		throw TypeError() << "need a " << RGB_type->tp_name 
											<< ", not " << get_type_name (rgb);
	
	p->draw (rgb, 0, 0, p->width, p->height);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
Pixels_methods [] =
{
	{"draw", (PyCFunction) Pixels_draw, METH_O,
		"draw (rgb) -- draw the contents of 'rgb'."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

Type * Pixels_type;

char * Py<Pixels>::__doc__ = 
"Pixels (widget, width, height) -- a Gdk pixmap.";

void
Py<Pixels>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"widget", "width", "height", 0};
	
	PyObject * w;
	int width, height;

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "Oii", keys, 
				&w, &width, &height))
		throw blewit();

	init (width, height, get_widget_from_pygtk (w));
}

void
Py<Pixels>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = NamedObject_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = Pixels_getsets;
	t->tp_methods = Pixels_methods;
}

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
PixelView_get_pixels (PixelView * v)
{
	return ::incref (v->pixels);
}

static PyObject *
PixelView_get_widget (PixelView * v)
{
	return pygobject_new (G_OBJECT (v->widget));
}

static int
PixelView_set_widget (PixelView * v, PyObject * arg)
Py_TRY
{
	if (arg != Py_None)
		throw ValueError() << "can only set widget to None.";

	gtk_widget_unref (v->widget);
	v->widget = 0;

	return 0;
}
Py_CATCH (-1);

static PyGetSetDef
PixelView_getsets [] = 
{
	{"pixels", (getter) PixelView_get_pixels, (setter) 0,
	 "the Pixels viewed."},
	{"widget", (getter) PixelView_get_widget, (setter) PixelView_set_widget, 
		"the pygtk widget."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
PixelView_blit (PixelView * v)
Py_TRY
{
	v->blit();
	gdk_flush();
	
	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
PixelView_methods [] =
{
	{"blit", (PyCFunction) PixelView_blit, METH_NOARGS,
		"blit() -- copy the pixels to the widget."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

void 
PixelView::init()
{
	widget = GTK_WIDGET (gtk_drawing_area_new());

	gtk_widget_ref (widget);

	gtk_widget_set_name (*this, "PixelView");

	pixels = NEW (Pixels);
	pixels->init (0, 0, widget);

	gtk_signal_connect 
		(
			*this, "configure-event", 
			(GtkSignalFunc) PixelView::configure_signal, 
			this
		);

	gtk_signal_connect 
		(
			*this, "expose-event", 
			(GtkSignalFunc) PixelView::expose_signal, 
			this
		);
}

gint
PixelView::configure_signal (GtkWidget * w, GdkEventConfigure * e, PixelView * self)
{
	self->configure (Rect<int> (0, 0, e->width, e->height));
	return true;
}

/* unused, we set the pixmap to be the static background for the widget. */
gint
PixelView::expose_signal (GtkWidget * w, GdkEventExpose * e, PixelView * self)
{
	Rect<int> r (e->area.x, e->area.y, e->area.width, e->area.height);
	
	// msg(), "expose ", self, r;
	self->expose (r);
	
	return true;
}

void
PixelView::expose (const Rect<int> r)
{
	// blit (r);
}

void
PixelView::configure (const Rect<int> r)
{
	if (r.width == pixels->width && r.height == pixels->height)
		return;
		
	pixels->resize (r.width, r.height);

	gdk_window_set_back_pixmap (window(), pixmap(), 0);
}

/* //////////////////////////////////////////////////////////////////////// */

Type * PixelView_type;

char * Py<PixelView>::__doc__ = 
"PixelView() -- a DrawingArea displaying Pixels.";

void
Py<PixelView>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {0};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();

	init();
}

void
Py<PixelView>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = Object_type;

	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = PixelView_getsets;
	t->tp_methods = PixelView_methods;
}

/* // CurveView /////////////////////////////////////////////////////////// */

Type * CurveView_type;

class CurveView
: public PixelView
{
	public:
		int size;
		float * data;
		int border;
		char label [100];
		bool fftw_done;
		bool clamp; /* try to confine to [-1, 1] */
		
		CurveView()
			: PixelView (CurveView_type)
			{ 
				size = 0;
				data = 0;
				border = 2;
				*label = 0;
				clamp = 1;
			}

		~CurveView()
			{
				if (data)
					free (data);
			}

		void init (int n)
			{
				if (n < 4 || n > 65536)
					throw ValueError() << this << ".size not in (4 .. 65536)";

				size = n;
				
				if (data == 0)
					data = (float *) malloc (n * sizeof (float));
				else
					data = (float *) realloc (data, n * sizeof (float));
			}
		
		void set_data (PyObject *, int offset);

		template <class F>
		void set_data (F * src, int n)
			{
				if (size == 0)
					throw AssertionError() << this << " has not been opened.";
				
				if (n > size)
					msg(), "limiting ", n, "points of data to ", (n = size);
				
				while (--n >= 0)
					data[n] = (float) src[n];

				fftw_done = 0;
			}

		void redraw();

	protected:
		virtual void configure (const Rect<int> size);
};

static PyObject *
CurveView_set (CurveView * v, PyObject * args)
Py_TRY
{
	int offset = 0;
	PyObject * arg;
	
	if (!PyArg_ParseTuple (args, "O|i", &arg, &offset))
		return 0;

	if (offset < 0 || offset >= PySequence_Size (arg))
		throw ValueError() << "'offset' must be within sample sequence size.";

	v->set_data (arg, offset);
	v->redraw();
	v->blit();

	return PyNone();
}
Py_CATCH (0);

static PyObject *
CurveView_redraw (CurveView * v)
Py_TRY
{
	v->redraw();
	v->blit();

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
CurveView_methods [] = 
{
	{"set", (PyCFunction) CurveView_set, METH_VARARGS,
	 "set (data object[, offset]) -- set the data viewed."},
	{"redraw", (PyCFunction) CurveView_redraw, METH_NOARGS,
	 "redraw() -- redraw the displayed data."},
	{ }
};

/* as sequence */

static int 
CurveView_sq_length (CurveView * v)
{
	return v->size;
}

static PyObject *
CurveView_sq_item (CurveView * v, int i) 
Py_TRY
{
	if (i < 0 || i >= v->size)
		throw IndexError() << "no sample " << i << " in " << v;
	
	return PyFloat (v->data [i]); 
} 
Py_CATCH (0);

static int
CurveView_sq_ass_item (CurveView * v, int i, PyObject * value) 
Py_TRY
{
	if (i < 0 || i >= v->size)
		throw IndexError() << "no sample " << i << " in " << v;
	
	if (!PyArg_Parse (value, "f", v->data + i))
		return -1;

	return 0;
} 
Py_CATCH (-1);

static PySequenceMethods 
CurveView_as_sequence = 
{
	(inquiry) CurveView_sq_length, /* sq_length */
	(binaryfunc) 0, /* sq_concat */
	(intargfunc) 0, /* sq_repeat */
	(intargfunc) CurveView_sq_item, /* sq_item */
	(intintargfunc) 0, /* sq_slice */
	(intobjargproc) CurveView_sq_ass_item, /* sq_ass_item */
	(intintobjargproc) 0, /* sq_ass_slice */
	(objobjproc) 0, /* sq_contains */
	/* Added in release 2.0 */
	(binaryfunc) 0, /* sq_inplace_concat */
	(intargfunc) 0 /* sq_inplace_repeat */
};

/* //////////////////////////////////////////////////////////////////////// */

void
CurveView::set_data (PyObject * arg, int offset)
{
	int n = 0;

	if (AudioSignal_type->is_derived (arg))
	{
		AudioSignal * e = (AudioSignal *) arg;
		
		n = min (e->size - offset, size);
		
		memcpy (data, e->data + offset, n * sizeof (d_sample));
	}
	
	/* the PySequence protocol chews more cycles than PyList and PyTuple. */
	else if (PyList_Check (arg))
	{
		n = min (PyList_Size (arg) - offset, size);

		for (int i = 0; i < n; ++i) 
		{
			PyObject * f = PyList_GetItem (arg, i + offset);
			
			if (!PyFloat_Check (f))
				throw TypeError() << "need float items from " << arg;
			
			data[i] = (float) PyFloat_AS_DOUBLE (f);
		}
	}

	else if (PyTuple_Check (arg))
	{
		n = min (PyTuple_Size (arg) - offset, size);

		for (int i = 0; i < n; ++i) 
		{
			PyObject * f = PyTuple_GetItem (arg, i + offset);
			
			if (!PyFloat_Check (f))
				throw TypeError() << "need float items from " << arg;
			
			data[i] = (float) PyFloat_AS_DOUBLE (f);
		}
	}

	else if (PySequence_Check (arg))
	{
		n = min (PySequence_Size (arg) - offset, size);

		for (int i = 0; i < n; ++i) 
		{
			PyObject * f = PySequence_GetItem (arg, i + offset);
			
			if (!PyFloat_Check (f))
				throw TypeError() << "need float items from " << arg;
			
			data[i] = (float) PyFloat_AS_DOUBLE (f);

			::decref (f, true);
		}
	}

	else throw TypeError() << "don't know how to set data from '"
												 << get_type_name (arg) << "' objects.";

	if (size - n)
		memset (data + n, 0, (size - n) * sizeof (*data)); 

	fftw_done = 0;
}

void
CurveView::configure (const Rect<int> r)
{
	PixelView::configure (r);

	redraw();
	blit();
}

void
CurveView::redraw()
{
	pixels->clear (bg_gc());

	if (size < 2)
		return;

	int height = pixels->height - 2 * border;
	int width = pixels->width - 2 * border;
	
	if (height <= 4 || width <= 2)
		return;

	pixels->draw_rect (dark_gc(), Rect<int> (border, border, width, height));

	/* determine value range */
	register float fmin = FLT_MAX, fmax = FLT_MIN;
	
	for (int i = 0; i < size; ++i)
	{
		register float f = data[i];
		
		if (f < fmin) fmin = f;
		if (f > fmax) fmax = f;
	}

	register double scale, f0;
	gint16 y0;
	
	/* normalize scale if min & max are in [-1, +1]: */
	if (clamp && fmin >= -1 && fmin <= 1 && fmax >= -1 && fmax <= 1)
	{
		f0 = 1.;
		scale = 2. / (double) height;
		y0 = border + height / 2;
	}
	else
	{
		f0 = fmax;
		scale = (double) (fmax - fmin) / (double) height;
		y0 = border - (gint16) (-fmax / scale);
	}

	/* center line */
	if (y0 > border && y0 < border + height)
		pixels->draw_line (dark_gc(), border, y0, width + border, y0);

	/* 4 x scale lines */
	for (double i = 0.25; i < 1; i += 0.25)
	{
		int x = lrint (i * (double) width);
		pixels->draw_line (dark_gc(), x, border, x, border + height);
	}
	
	/* print fmin, fmax */
#if 1
	char minmax [24] = "0";
	int dy = 6; /* pixel height of font */

	sprintf (minmax, "%.3f", fmax);
	pixels->draw_text (widget, fg_gc(), border + 1, border + dy, minmax);
	sprintf (minmax, "%.3f", fmin);
	pixels->draw_text (widget, fg_gc(), border + 1, border + height - 1, minmax);
#endif

	int sz = size;
	register double step = (double) width / (double) sz;

	/* batch and draw lines */
	GdkPoint points [sz];

	double x = 0.0;
	int p = 0;

	for (int i = 0; i < sz; ++i, x += step)
	{
		points[i].x = border + (gint16) x;
		points[i].y = border - (gint16) (((double) data[p + i] - f0) / scale);
	}

	pixels->draw_lines (fg_gc(), points, sz);
}

/* //////////////////////////////////////////////////////////////////////// */

static int
CurveView_set_size (CurveView * v, PyObject * size)
Py_TRY
{
	int i;

	if (!PyArg_Parse (size, "i", &i))
		return -1;

	v->init (i);
	return 0;
}
Py_CATCH (-1);

static PyObject * 
CurveView_get_size (CurveView * v)
{
	return PyInt (v->size);
}

static int
CurveView_set_clamp (CurveView * v, PyObject * clamp)
Py_TRY
{
	if (!PyArg_Parse (clamp, "i", &v->clamp))
		return -1;

	return 0;
}
Py_CATCH (-1);

static PyObject * 
CurveView_get_clamp (CurveView * v)
{
	return PyInt (v->clamp);
}

static int
CurveView_set_label (CurveView * v, PyObject * label)
Py_TRY
{
	char * l;

	if (!PyArg_Parse (label, "s", &l))
		return -1;

	if (strlen (l) > 100 - 1)
		throw OverflowError() << "len (label) must be < 80.";

	strcpy (v->label, l);

	return 0;
}
Py_CATCH (-1);

static PyObject * 
CurveView_get_label (CurveView * v)
{
	return PyString (v->label);
}

static PyGetSetDef
CurveView_getsets [] =
{
	{"label", (getter) CurveView_get_label, (setter) CurveView_set_label,
	 "is displayed along with the data set."},
	{"size", (getter) CurveView_get_size, (setter) CurveView_set_size,
	 "the number of sample points."},
	{"clamp", (getter) CurveView_get_clamp, (setter) CurveView_set_clamp,
	 "whether to try and clamp the y scale to [-1, 1]."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

char * Py<CurveView>::__doc__ = 
"CurveView (size[, label = '']) -- a PixelView displaying <= 'size' data points.";

void
Py<CurveView>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"size", "label", 0};
	int size;
	char * l = 0;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "i|s", keys, &size, &l))
		throw blewit();

	if (l)
	{
		if (strlen (l) > 80 - 1)
			throw OverflowError() << "len ('label') must be < 80.";
		strcpy (label, l);
	}

	init (size);
	
	PixelView::init();

	GtkLock lock;
	gtk_widget_set_name (*this, "curve");
}

void
Py<CurveView>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = PixelView_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = CurveView_getsets;
	t->tp_methods = CurveView_methods;
	t->tp_as_sequence = &CurveView_as_sequence;
}

/* //////////////////////////////////////////////////////////////////////// */

void
init_Pixel_types (PyObject * dict)
{
	Pixels_type = new PyType<Pixels> ("Pixels", dict);
	PixelView_type = new PyType<PixelView> ("PixelView", dict);
	CurveView_type = new PyType<CurveView> ("CurveView", dict);
}

