#ifndef _PIXELS_H_
#define _PIXELS_H_

#include <gdk/gdk.h>

#include "RGB.h"

extern Type * Pixels_type;

class Pixels
: public NamedObject
{
	public:
		int width, height;

		struct {
			GdkPixmap * pixmap;
			GtkWidget * widget;
		} self;

	public:
		Pixels (GdkPixmap * _self = 0)
			: NamedObject (Pixels_type)
			{ 
				width = height = 0;

				self.pixmap = 0;
				self.widget = 0;
			}

		~Pixels()
			{
				if (self.pixmap) gdk_pixmap_unref (self.pixmap);
				if (self.widget) gtk_object_unref ((GtkObject *) self.widget);
			}

		void init (int w, int h, GtkWidget * w);
		void resize (int w, int h);

		void clear (GdkGC * gc)
			{
				gdk_draw_rectangle (self.pixmap, gc, 1, 0, 0, width, height);
			}
			
		void draw_rect (GdkGC * gc, const Rect<int> & r)
			{
				gdk_draw_rectangle (self.pixmap, gc, 0, r.x, r.y, r.width, r.height);
			}

		void draw_line (GdkGC * gc, int x1, int y1, int x2, int y2)
			{
				gdk_draw_line (self.pixmap, gc, x1, y1, x2, y2);
			}

		void draw_lines (GdkGC * gc, GdkPoint * points, int n)
			{
				gdk_draw_lines (self.pixmap, gc, points, n);
			}

		void draw_text (GtkWidget * w, GdkGC * gc, int x, int y, char * text)
			{
				PangoLayout * layout = gtk_widget_create_pango_layout (w, text);
				gdk_draw_layout (self.pixmap, gc, x, y, layout);
				g_object_unref (G_OBJECT (layout));
			}

		operator GdkPixmap * () 
			{ return self.pixmap; }

		void draw (RGB * rgb, 
				int x = 0, int y = 0, int w = 0, int h = 0, 
				int x_rgb = -1, int y_rgb = -1);
};

/* //////////////////////////////////////////////////////////////////////// */

extern Type * PixelView_type;

#define WIDGET_STATE GTK_STATE_NORMAL

class PixelView
: public Object
{
	public:
		Pixels * pixels;
		GtkWidget * widget;
		
		PixelView (Type * t = PixelView_type)
			: Object (t)
			{ 
				pixels = 0;
				widget = 0;
			}

		~PixelView()
			{
				::decref (pixels, true);
				if (widget) gtk_widget_unref (widget);
			}

		void init();

		void blit()
			{
				blit (Rect<int> (0, 0, pixels->width, pixels->height));
			}

		void blit (Rect<int> r)
			{
				/* because 'pixels' is the back pixmap, clear area does this very
				 * well. (gdk_draw_pixmap isn't executed right away btw.)
				 */
				gdk_window_clear_area (window(), r.x, r.y, r.width, r.height);
			}

		GdkPixmap * pixmap()
			{
				return pixels->self.pixmap;
			}

	protected:
		static gint configure_signal (GtkWidget *, GdkEventConfigure *, PixelView *);
		static gint expose_signal (GtkWidget *, GdkEventExpose *, PixelView *);

		/* called in a signal handler, with the Gtk lock held. */
		virtual void configure (const Rect<int> size);
		virtual void expose (const Rect<int> r);

	public:
		operator GtkWidget * ()
			{ return widget; }

		operator GtkObject * ()
			{ return GTK_OBJECT (widget); }

		/* access to sub-objects */
		
		GdkWindow * window() 
			{ return widget->window; }

		GtkStyle * style()
			{ return widget->style; }

		/* access to drawing resources */
		
		GdkGC * fg_gc (int i = WIDGET_STATE)
			{
				return style()->fg_gc[i];
			}

		GdkGC * bg_gc (int i = WIDGET_STATE)
			{
				return style()->bg_gc[i];
			}

		GdkGC * base_gc (int i = WIDGET_STATE)
			{
				return style()->base_gc[i];
			}

		GdkGC * light_gc (int i = WIDGET_STATE)
			{
				return style()->light_gc[i];
			}

		GdkGC * mid_gc (int i = WIDGET_STATE)
			{
				return style()->mid_gc[i];
			}

		GdkGC * dark_gc (int i = WIDGET_STATE)
			{
				return style()->dark_gc[i];
			}

		GdkColor * gdk_fg()
			{
				return style()->fg;
			}

		GdkColor * gdk_bg()
			{
				return style()->bg;
			}

		GdkColor * gdk_base()
			{
				return style()->base;
			}

		GdkColor * gdk_light()
			{
				return style()->light;
			}

		GdkColor * gdk_mid()
			{
				return style()->mid;
			}

		GdkColor * gdk_dark()
			{
				return style()->dark;
			}

		GdkColor * gdk_text()
			{
				return style()->text;
			}

};

/* //////////////////////////////////////////////////////////////////////// */

extern Type * RGBView_type;

class RGBView
: public PixelView
{
	public:
		RGB * rgb;

		RGBView (Type * t = RGBView_type)
			: PixelView (t)
			{ 
				rgb = NEW (RGB);
			}

		~RGBView()
			{
				::decref (rgb);
			}

		void rgb_to_pixels (const Rect<int> r)
			{
				pixels->draw (rgb, r.x, r.y, r.width, r.height, r.x, r.y);
			}

	protected:
		virtual void configure (const Rect<int> size);
};

#endif /* _PIXELS_H_ */
