#include <basics.h>
#include <PyType.h>

#include "Gtk.h"
#include "Pixels.h"

#include "font.h"

/* log every fill call: */
#define DBG_FILL	0

#if defined (WITH_JPEG)

#undef HAVE_PROTOTYPES
#undef HAVE_STDDEF_H
#undef HAVE_STDLIB_H

#include <setjmp.h>
extern "C" 
{
#include <jpeglib.h>
}

static jmp_buf jpeg_err_jump;

static void
jpg_error (j_common_ptr cinfo)
{
	jpeg_error_mgr *err = (jpeg_error_mgr *) cinfo->err;
	
	char message[200];
	err->format_message	(cinfo, message);
	fprintf (stderr, "%s\n", message);

	longjmp (jpeg_err_jump, 1);
}

void 
RGB::read_jpeg (char * file)
{
	FILE * f = fopen (file, "r");
	if (f == 0)
		throw OSError (errno) << file;

	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;

	cinfo.err = jpeg_std_error (&jerr);
	jerr.error_exit = jpg_error;

	JSAMPLE * scanline = 0;

	if (setjmp (jpeg_err_jump)) 
	{
		/* this is where we get on error */
		if (scanline)
			delete [] scanline;

		jpeg_destroy_decompress (&cinfo);
		fclose (f);

		throw RuntimeError();
	}

	jpeg_create_decompress (&cinfo);
	jpeg_stdio_src (&cinfo, f);
	jpeg_read_header (&cinfo, true);

	jpeg_start_decompress	(&cinfo);

	resize (cinfo.output_width, cinfo.output_height);
	msg(), cinfo.output_width, cinfo.output_height, size, sizeof (JSAMPLE);

	scanline = new JSAMPLE [cinfo.output_width * 3];

	for (int y = 0; y < size.height; ++y) 
	{
		jpeg_read_scanlines (&cinfo, &scanline, 1);

		byte * scan = (byte *) scanline;
		byte * pixel = pointer (0, y);

		if (cinfo.num_components == 1) 
		{
			for (int x = 0; x < size.width; x++) 
			{
				*pixel++ = *scan;
				*pixel++ = *scan;
				*pixel++ = *scan++;
			}
		} 
		else 
		{			
			for (int x = 0; x < size.width; x++) 
			{
				*pixel++ = *scan++;
				*pixel++ = *scan++;
				*pixel++ = *scan++;
			}
		}
	}

	delete [] scanline;

	jpeg_finish_decompress (&cinfo);
	jpeg_destroy_decompress (&cinfo);

	fclose (f);
}

void 
RGB::read_half_jpeg (char * file)
{
	FILE * f = fopen (file, "r");
	if (f == 0)
		throw OSError (errno) << file;

	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;

	cinfo.err = jpeg_std_error (&jerr);
	jerr.error_exit = jpg_error;

	JSAMPLE * scanline = 0;

	if (setjmp (jpeg_err_jump)) 
	{
		/* this is where we get on error */
		if (scanline)
			delete [] scanline;

		jpeg_destroy_decompress (&cinfo);
		fclose (f);
		return;
	}

	jpeg_create_decompress (&cinfo);
	jpeg_stdio_src (&cinfo, f);
	jpeg_read_header (&cinfo, true);

	jpeg_start_decompress	(&cinfo);

	resize (cinfo.output_width / 2, cinfo.output_height / 2);

	scanline = new JSAMPLE [cinfo.output_width * 3];

	for (int y = 0; y < size.height; ++y) 
	{
		jpeg_read_scanlines (&cinfo, &scanline, 1);

		byte * scan = (byte *) scanline;
		byte * pixel = pointer (0, y);

		for (int x = 0; x < size.width; ++x, scan += 6) 
		{
			*pixel++ = (scan[0] + scan[3]) >> 1;
			*pixel++ = (scan[1] + scan[4]) >> 1;
			*pixel++ = (scan[2] + scan[5]) >> 1;
		}

		jpeg_read_scanlines (&cinfo, &scanline, 1);
	}

	delete [] scanline;

	jpeg_finish_decompress (&cinfo);
	jpeg_destroy_decompress (&cinfo);

	fclose (f);
}

static PyObject *
RGB_read_jpeg (RGB * rgb, PyObject * file)
Py_TRY
{
	char * f;
	
	if (!PyArg_Parse (file, "s", &f))
		return 0;

	rgb->read_jpeg (f);
	return PyNone();
}
Py_CATCH (0);

static PyObject *
RGB_read_half_jpeg (RGB * rgb, PyObject * file)
Py_TRY
{
	char * f;
	
	if (!PyArg_Parse (file, "s", &f))
		return 0;

	rgb->read_half_jpeg (f);
	return PyNone();
}
Py_CATCH (0);
#endif /* WITH_JPEG */

void
RGB::blend (RGB * a, RGB * b, double ab)
{
	Rect<int> r (size);
	a->size.clip (r);
	b->size.clip (r);

	int ma = min (255, (int) (256 * ab)) & 255;
	int mb = 255 ^ ma;
	
	for (int y = 0; y < r.height; ++y)
	{
		byte * p = pointer (0, y);
		byte * pa = a->pointer (0, y);
		byte * pb = b->pointer (0, y);

		for (int x = 0; x < r.width; ++x)
		{
			*p++ = (((int) *pa++) * ma + ((int) *pb++) * mb) >> 8;
			*p++ = (((int) *pa++) * ma + ((int) *pb++) * mb) >> 8;
			*p++ = (((int) *pa++) * ma + ((int) *pb++) * mb) >> 8;
		}
	}
}

void
RGB::copy (RGB * rgb)
{
	Rect<int> r (size);
	rgb->size.clip (r);
	
	for (int y = 0; y < size.height; ++y)
		memcpy (pointer (0, y), rgb->pointer (0, y), size.width * 3);
}

void
RGB::resize (int width, int height)
{
	size.width = width;
	size.height = height;

	rowstride = 3 * size.width;

	while (rowstride & 3)
		++rowstride;

	if (pixels)
		pixels = (byte *) realloc (pixels, rowstride * size.height);
	else
		pixels = (byte *) malloc (rowstride * size.height);
		
	clip = size;
	
	if (pixels == 0)
		throw MemoryError() << "allocating RGB of " << size;
}

void
RGB::rect (const Rect<int> & rect, const Color & c)
{
	fill (Rect<int> (rect.x, rect.y, rect.width, 1), c);
	fill (Rect<int> (rect.x, rect.y + rect.height, rect.width, 1), c);
	
	Rect<int> left (rect.x, rect.y + 1, 1, rect.height - 1);
	Rect<int> right (rect.x + rect.width - 1, rect.y + 1, 1, rect.height - 1);

	if (clip.clip (left))
	{
		Pixel p (pointer (left.x, left.y));
		p.run (left.height, c, rowstride);
	}

	if (clip.clip (right))
	{
		Pixel p (pointer (right.x, right.y));
		p.run (right.height, c, rowstride);
	}
}

bool
RGB::fill (const Rect<int> & rect, const Color & c)
{
	#if DBG_FILL
	msg(), this, "fill ", rect, c.r, c.g, c.b;
	#endif
	
	Rect<int> r (rect);
	clip.clip (r);

	if (r.width == 0 || r.height == 0) return 0;

	while (r.height-- > 0)
	{
		Pixel p (pointer (r.x, r.y++));
		p.run (r.width, c);
	}

	return true;
}

bool
RGB::fill (const Rect<int> & rect, const AlphaColor & c)
{
	#if DBG_FILL
	msg(), this, "fill alpha ", rect, c.r, c.g, c.b;
	#endif
	
	Rect<int> r (rect);
	clip.clip (r);

	if (r.width == 0 || r.height == 0) return 0;

	if (r.width == 1)
	{
		Pixel p (pointer (r.x, r.y++));
		p.run (r.height, c, rowstride);
	}
	else while (r.height-- > 0)
	{
		Pixel p (pointer (r.x, r.y++));
		p.run (r.width, c);
	}

	return true;
}

void
RGB::pixelf (float fx, float fy, const Color & color, int alpha)
{
	fx -= .5;
	fy -= .5;

	int ix = (int) fx;
	int iy = (int) fy;
	
	fx -= ix;
	fy -= iy;
	
	for (int y = 0; y < 1; ++y)
	{
		register int py = iy + y;
		fy = 1 - fy;

		for (int x = 0; x < 1; ++x)
		{
			register int px = ix + x;
			fx = 1 - fx;

			if (!clip.has (ix + x, iy + y))
				continue;

			int a = (int) ((float) alpha * ((fx + fy) * .5));

			Pixel p (pointer (px, py));
			p.put (color, a);
		}
	}
}

void 
RGB::disc (float cx, float cy, float ro, float ri, const Color & color, int alpha)
{
	int iro = (int) ro;
	Rect<int> r ((int) (cx - ro), (int) (cy - ro), 2 * iro + 1, 2 * iro + 1);

	if (!clip.clip (r))
		return;

	float ro2 = ro * ro;
	float ro2aa = (ro - 1) * (ro - 1); /* threshold for antialiasing */

	if (ri <= 0.001)
		ri = .001;

	float ri2 = ri * ri;
	float ri2aa = ri > 1 ? (ri - 1) * (ri - 1) : 0; /* threshold for antialiasing */

	AlphaColor ac (color, alpha);

	for (int y = 0; y < r.height; ++y)
	{
		Pixel p (pointer (r.x, r.y + y));
		
		for (int x = 0; x < r.width; ++x, p += 3)
		{
			float dx = (float) (x + r.x) - cx;
			float dy = (float) (y + r.y) - cy;

			float d2 = dx * dx + dy * dy;

			if (d2 > ro2)
				continue;
			
			float h; /* 1 = opaque, 0 = transparent */
			
			if (d2 > ro2aa) /* rim */
				h = ro - sqrt (d2); 
			else if (d2 < ri2aa)
				continue; /* inside, fully transparent */
			else if (d2 < ri2)
				h = 1 - (ri - sqrt (d2));
			else
			{
				p.put (ac);
				continue;
			}

			int a = (int) ((float) alpha * h);
			p.put (color, a);
		}
	}
}

/* draws a column in a color, fading the color at the top and bottom */
void
RGB::columnf (int x, float y0, float y1, const Color & color, int alpha)
{
	if (x < 0 || x >= size.width || y0 >= size.height) 
		return; 

	y0 = max (y0, 0.f);
	y1 = min (y1, (float) size.height - 1);

	int y0i = (int) y0;
	float y0f = 1. - (y0 - y0i);
	
	int y1i = (int) y1;
	float y1f = (y1 - y1i);

	int dy = y1i - y0i;
	
	if (dy <= 1)
	{
		if (y0i < 0) return; 

		Pixel p (pointer (x, y0i));
		p.put (color, alpha);

		/* easy case, just 1 pixel to draw. */
		return;
	}

	if (y0i >= 0 && y0i < size.height)
	{
		Pixel p (pointer (x, y0i));
		p.put (color, (int) (alpha * y0f));
	}

	Rect<int> r (x, y0i + 1, 1, y1i - y0i - 1);
	if (size.clip (r))
		fill (r, AlphaColor (color, alpha));

	if (y1i >= 0 && y1i < size.height)
	{
		Pixel p (pointer (x, y1i));
		p.put (color, (int) (alpha * y1f));
	}
}

/* draws a column in a color, fading the color at the top and bottom */
void
RGB::fade_column (int x, float y0, float y1, const Color & color, int a0, int a1)
{
	if (x < 0 || x >= size.width) 
		return; 

	/* clamp to size */
	double dy0 = 0;
	if (y0 < 0) dy0 = -y0;
	else if (y0 >= size.height) dy0 = y0 - size.height;

	double dy1 = 0;
	if (y1 < 0) dy1 = -y1;
	else if (y1 >= size.height) dy1 = y1 - size.height;

	int y0i = (int) (y0 + dy0 + .5);
	int y1i = (int) (y1 + dy1 + .5);
	
	if (y1i == y0i)
	{
		// put (x, (int) y0, color, max (a0, a1));
		return;
	}

	/* direction to trace pixels */
	int dir;
	double step;
	if (y1 > y0)
		dir = 1,
		step = y1 - y0;
	else
		dir = -1,
		step = y0 - y1;
	
	step = (double) 1 / step;
	//msg(), "-- ", y0, y1, dir, step;

	Pixel p0 (pointer (x, y0i));
	Pixel p1 (pointer (x + 1, y0i));

	int r = rowstride * dir;
	double da0 = a0 * step;
	double da1 = a1 * step;

	for (int y = y0i; y != y1i; y += dir)
	{
		double alpha0 = a0 + da0 * ((double) y + .5 - y0);
		double alpha1 = a1 + da1 * ((double) y + .5 - y1);
		// msg(), x, y, y - y0, step * (y + .5 - y0), a, (int) (a + .5);
		p0.put (color, (int) (alpha0 + .5));
		p1.put (color, (int) (alpha1 + .5));
		p0 += r;
		p1 += r;
	}
}

/* //////////////////////////////////////////////////////////////////////// */

int
RGB::font_width()
{
	return fixed_width;
}

int
RGB::font_height()
{
	return sizeof (fixed[0]);
}

int 
RGB::text_width (const char * s)
{
	return strlen (s) * font_width();
}

/* draws a text using the bitmaps from 'font.h' which can be made from 
 * a .bdf-encoded bitmap font, see the Makefile. the font needs to be
 * fixed-width for this simple renderer to work.
 *
 * you can fetch your own .bdf from your font server by using the
 * program 'fstobdf' which usually comes with an X11 distribution.
 */
void
RGB::text (int x, int y0, char * s, const AlphaColor & c)
{
	int l = strlen (s);
	
	int rows = font_height();
	
	y0 -= (rows + fixed_descent);

	/* first hit-test the whole thing to check whether we can lazily return: */
	Rect<int> r (x, y0, l * fixed_width, rows);

	if (!clip.clip (r))
		return;

	int test_bit = 1 << fixed_width;

	for (int i = 0; i < l; ++i, x += fixed_width)
	{
		r.x = x;
		r.width = fixed_width;
		
		if (!clip.clip (r))
			continue;

		byte * bits = fixed [(byte) s[i]];
		bits += (r.y - y0);

		int shift = r.x - x;
		
		for (int y = 0; y < r.height; ++y, ++bits)
		{
			byte * p = pointer (r.x, r.y + y);
			int b = *bits << shift;

			for (int j = 0; j < r.width; ++j)
			{
				if (b & test_bit)
					c.alpha_put (p);
				b <<= 1;
				p += 3;
			}
		}
	}
}

/* Targa image writer ///////////////////////////////////////////////////// */

template <class T>
void
tga_write (int fd, T t)
{
	t = host2le (t);
	if (::write (fd, &t, sizeof (T)) != sizeof (T))
		throw OSError (errno);
}

template <class P>
void
tga_write_pixels (int fd, P * p, int n)
{
	int s = (int) sizeof (P);
	byte b[n * s];
	byte * bb = b;

	/* stupid tga wants bgr pixels. */
	for (int x = 0; x < n; ++x)
	{
		byte * pp = (byte *) (p + x);
		for (int i = 0; i < s; ++i)
			bb[i] = pp[s - i - 1];
		bb += s;
	}

	write (fd, b, n * s);
}

/* write a scanline in tga rle format */
template <class T>
static void
write_rle_line (int fd, T * pixels, int width)
{
	int x = 1, x0 = 0;
	enum {RLE_MAX = 128};

	while (x < width)
	{
		if (pixels[x] == pixels[x0])
		{
			int count = 1;
			do {
				++x, ++count;
			} while (pixels[x] == pixels[x0] && count < RLE_MAX && x < width);

			tga_write (fd, (uint8) ((count - 1) | RLE_MAX));
			tga_write_pixels (fd, pixels + x0, 1);
		}
		else
		{
			int count = 0;
			do {
				++x, ++count;
			} while (pixels[x] != pixels[x - 1] && count < RLE_MAX && x < width);
			--x;

			tga_write (fd, (uint8) (count - 1));
			tga_write_pixels (fd, pixels + x0, count);
		}

		x0 = x;
		x += 1;
	}

	/* pixels left? */
	if (x0 < width)
	{
		tga_write (fd, (uint8) 0);
		tga_write_pixels (fd, pixels + x0, 1);
	}
}

class RGBPixel {
	public:
		byte r, g, b;

		bool operator == (RGBPixel & p)
			{
				return r == p.r && g == p.g && b == p.b;
			}

		bool operator != (RGBPixel & p)
			{
				return r != p.r && g != p.g && b != p.b;
			}
};


void
RGB::write_tga (char * path)
{
	int fd = open (path, O_WRONLY | O_CREAT | O_TRUNC, 0644);
	if (fd < 0)
		throw OSError (errno);

	bool RLE = 1;

	/* write header */
	tga_write (fd, (uint8) 0); /* id */
	tga_write (fd, (uint8) 0); /* no color map */
	tga_write (fd, (uint8) (RLE ? 10 : 2)); /* RLE-compressed RGB */
	tga_write (fd, (uint16) 0); /* colormap start */
	tga_write (fd, (uint16) 0); /* colormap end */
	tga_write (fd, (uint8) 0); /* no color map entries */
	tga_write (fd, (uint16) 0); /* offset x */
	tga_write (fd, (uint16) 0); /* offset y */
	tga_write (fd, (uint16) size.width); /* width */
	tga_write (fd, (uint16) size.height); /* height */
	tga_write (fd, (uint8) 24); /* 24 bits/pixel */
	tga_write (fd, (uint8) 0x20); /* (0, 0) = up-left, 0 alpha bits */

	if (RLE)
	{
		for (int y = 0; y < size.height; ++y)
			write_rle_line (fd, (RGBPixel *) pointer (0, y), size.width);
	}
	else
	{
		for (int y = 0; y < size.height; ++y)
			tga_write_pixels (fd, (RGBPixel *) pointer (0, y), size.width);
	}

	close (fd);
}

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
RGB_write_tga (RGB * rgb, PyObject * args)
Py_TRY
{
	char * s;

	if (!PyArg_ParseTuple (args, "s", &s))
		return 0;

	rgb->write_tga (s);
	return PyNone();
} 
Py_CATCH (0);

static PyObject *
RGB_fill (RGB * rgb, PyObject * args)
Py_TRY
{
	Rect<int> r;
	PyObject * rgba;

	if (!PyArg_ParseTuple (args, "iiiiO", &r.x, &r.y, &r.width, &r.height, &rgba))
		return 0;

	int c[4] = {0xff, 0xef, 0xcf, -1};

	if (!PyArg_ParseTuple (rgba, "iii|i", c, c+1, c+2, c+3))
		return 0;
	
	if (c[3] < 0 || c[3] > 255)
		rgb->fill (r, Color (c));
	else
		rgb->fill (r, AlphaColor (c));
				
	return PyNone();
}
Py_CATCH (0);

static PyObject *
RGB_blend (RGB * rgb, PyObject * args)
{
	RGB * a, * b;
	double ab;

	if (!PyArg_ParseTuple (args, "O!O!d", RGB_type, &a, RGB_type, &b, &ab))
		return 0;

	rgb->blend (a, b, ab);
	return PyNone();
}

static PyObject *
RGB_copy (RGB * rgb, RGB * source)
Py_TRY
{
	if (!RGB_type->is_derived (source))
		throw TypeError() << "need an rgb, not " << get_type_name (source);

	rgb->copy (source);
	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
RGB_methods[] =
{
	{"fill", (PyCFunction) RGB_fill, METH_VARARGS,
	 "fill (x, y, width, height, (r, g, b[, a])) -- "
	 "fill a rectangle with a color."},
	{"blend", (PyCFunction) RGB_blend, METH_VARARGS,
	 "blend (rgb_a, rgb_b, ab) -- render the 'ab'-weighted combination of "
	 "'rgb_a' and 'rgb_b' into this image. 'ab' is from [0 .. 1]"},
	{"copy", (PyCFunction) RGB_copy, METH_O,
	 "copy (rgb) -- copy pixels from 'rgb'."},
	{"write_tga", (PyCFunction) RGB_write_tga, METH_VARARGS,
	 "write_tga (path) -- write an RLE-compressed targa image file."},
#if defined (WITH_JPEG)
	{"read_jpeg", (PyCFunction) RGB_read_jpeg, METH_O,
	 "read_jpeg ('foo.jpeg') -- read an image file."},
	{"read_half_jpeg", (PyCFunction) RGB_read_half_jpeg, METH_O,
	 "read_half_jpeg ('foo.jpeg') -- read an image file, scaling to half x and y."},
#endif
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
RGB_get_width (RGB * rgb)
{
	return PyInt (rgb->size.width);
}

static PyObject *
RGB_get_height (RGB * rgb)
{
	return PyInt (rgb->size.height);
}

static PyGetSetDef
RGB_getsets [] = 
{
	{"width", (getter) RGB_get_width, (setter) 0,
		"."},
	{"height", (getter) RGB_get_height, (setter) 0,
		"."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

Type * RGB_type;

char * Py<RGB>::__doc__ = 
"RGB (width, height) -- a RGB buffer.";

void
Py<RGB>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {"width", "height", 0};
	int width, height;
	
	if (!PyArg_ParseTupleAndKeywords (args, kwds, "ii", keys, &width, &height))
		throw blewit();

	resize (width, height);
}

void
Py<RGB>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = NamedObject_type;
	t->tp_getset = RGB_getsets;
	t->tp_methods = RGB_methods;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
}

/* //////////////////////////////////////////////////////////////////////// */

/* RGBView */

void
RGBView::configure (const Rect<int> r)
{
	rgb->resize (r.width, r.height);

	PixelView::configure (r);
}

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
RGBView_get_rgb (RGBView * v)
{
	return ::incref (v->rgb);
}

static PyGetSetDef
RGBView_getsets[] =
{
	{"rgb", (getter) RGBView_get_rgb, (setter) 0,
	 "the rgb buffer associated."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

static PyObject *
RGBView_redraw (RGBView * v, PyObject * args)
Py_TRY
{
	if (v->rgb == 0 || v->pixels == 0)
		return PyNone();

	Rect<int> r (0, 0, v->pixels->width, v->pixels->height);
	
	if (PySequence_Size (args) && !PyArg_ParseTuple (args, "iiii", 
				&r.x, &r.y, &r.width, &r.height))
		return 0;

	v->pixels->draw (v->rgb, r.x, r.y, r.width, r.height, r.x, r.y);

	return PyNone();
}
Py_CATCH (0);

static PyMethodDef
RGBView_methods[] =
{
	{"redraw", (PyCFunction) RGBView_redraw, METH_VARARGS,
	 "redraw ([x, y, width, height]) -- redraw a rectangle from the "
	 "contents of the associated RGB buffer. if called without arguments, "
	 "the complete image is redrawn. otherwise, you must specify all four "
	 "measures."},
	{ }
};

/* //////////////////////////////////////////////////////////////////////// */

Type * RGBView_type;

char * Py<RGBView>::__doc__ = 
"RGBView() -- a PixelView displaying the contents of a RGB buffer.";

void
Py<RGBView>::__init__ (PyObject * args, PyObject * kwds)
{
	static char * keys [] = {};

	if (!PyArg_ParseTupleAndKeywords (args, kwds, "", keys))
		throw blewit();

	PixelView::init();
}

void
Py<RGBView>::type_init (Type * t, PyObject * dict)
{
	t->tp_base = PixelView_type;
	t->tp_flags |= Py_TPFLAGS_BASETYPE;
	t->tp_getset = RGBView_getsets;
	t->tp_methods = RGBView_methods;
}

/* //////////////////////////////////////////////////////////////////////// */

void
init_RGB_types (PyObject * dict)
{
	RGB_type = new PyType<RGB> ("RGB", dict);
	RGBView_type = new PyType<RGBView> ("RGBView", dict);
}
