#ifndef _RGB_H_
#define _RGB_H_

class Color
{
	public:
		int r, g, b;

		Color() {}
					
		Color (int _r, int _g, int _b) 
			{
				set_from (_r, _g, _b);
			}
		
		Color (uint c) 
			{
				*this = c;
			}
		
		void operator = (uint c)
			{
				set_from (c >> 16, (c >> 8) & 255, c & 255);
			}

		Color (byte * p) 
			{
				set_from (p);
			}

		Color (PyObject *);

		Color (GdkColor * gdk)
			{
				r = gdk->red >> 8;
				g = gdk->green >> 8;
				b = gdk->blue >> 8;
			}

		Color (int * rgb)
			{
				r = *rgb++;
				g = *rgb++;
				b = *rgb;
			}

		void set_from (int _r, int _g, int _b) 
			{
				r = _r;
				g = _g;
				b = _b;
			}
		
		void set_from (byte * p) 
			{
				r = p[0];
				g = p[1];
				b = p[2];
			}
		
		inline void put (byte * p) const 
			{
				p[0] = r;
				p[1] = g;
				p[2] = b;
			}
		
		inline void put (byte ** pp) const 
			{
				byte * p = *pp;
				*pp += 3;

				p[0] = r;
				p[1] = g;
				p[2] = b;
			}
		
		Color & operator = (const Color & that) 
			{
				set_from (that.r, that.g, that.b); 
				return *this;
			}

		Color & operator = (byte * p) 
			{
				set_from (p); 
				return *this;
			}

		unsigned int packed() 
			{
				return (r << 16) | (g << 8) | b;
			}

		inline Color & operator *= (int alpha) 
			{
				r *= alpha;
				g *= alpha;
				b *= alpha;
				
				return *this;
			}

		inline Color & operator >>= (int shift) 
			{
				r >>= shift;
				g >>= shift;
				b >>= shift;
				
				return *this;
			}
		
		inline Color & operator += (const Color & that) 
			{
				r += that.r;
				g += that.g;
				b += that.b;

				return *this >>= 1;
			}
};

/* //////////////////////////////////////////////////////////////////////// */

class AlphaColor
:	public Color
{
	public:
		int alpha, inverse;
		
		AlphaColor (unsigned int c, int a)
			: Color (c)
			{
				set_alpha (a);
			}
			
		AlphaColor (const Color & c, int a)
			: Color (c)
			{
				set_alpha (a);
			}
			
		AlphaColor (GdkColor * g, int a)
			: Color (g)
			{
				set_alpha (a);
			}

		AlphaColor (int * rgba)
			: Color (rgba)
			{
				set_alpha (rgba[3]);
			}

		void set_alpha (int a)
			{
				alpha = a;
				inverse = 256 - alpha;
				(*this) *= alpha;
			}

		inline void alpha_put (byte * p) const 
			{
				p[0] = (byte) ((r + inverse * p[0]) >> 8);
				p[1] = (byte) ((g + inverse * p[1]) >> 8);
				p[2] = (byte) ((b + inverse * p[2]) >> 8); 
			}
};

/* //////////////////////////////////////////////////////////////////////// */

class Pixel
{
	public:
		byte * pointer;
		
		Pixel (byte * _p)
			{
				pointer = _p;
			}

		inline void run (int n, const Color & c, int stride = 3) 
			{
				while (n-- > 0) 
				{
					c.put (pointer);
					pointer += stride;
				}
			}

		inline void run (int n, const AlphaColor & c, int stride = 3) 
			{
				while (n-- > 0) 
				{
					c.alpha_put (pointer);
					pointer += stride;
				}
			}

		inline void put (const Color & c) 
			{
				c.put (pointer);
			}

		inline void put (const AlphaColor & c) 
			{
				c.alpha_put (pointer);
			}

		inline void put (const Color & c, int alpha) 
			{
				AlphaColor (c, alpha).alpha_put (pointer);
			}

		inline void operator += (int i)
			{
				pointer += i;
			}

		inline uint32 get()
			{
				return (pointer[0] << 16) | (pointer[1] << 8) | pointer[0];
			}
};

/* //////////////////////////////////////////////////////////////////////// */

extern Type * RGB_type;

class RGB
: public NamedObject
{
	public:
		byte * pixels;
		int rowstride;
		
		Rect<int> size;
		Rect<int> clip;

		Color fore, back;

		RGB()
			: NamedObject (RGB_type)
			{
				pixels = 0;
			}

		~RGB() 
			{
				if (pixels) free (pixels);
			}

		void resize (int width, int height);

		inline byte * pointer (int x, int y) 
			{
				return pixels + 3 * x + y * rowstride;
			}
	
		void rect (const Rect<int> &, const Color &);
		void rect (const Rect<int> &, const AlphaColor &);

		bool fill (const Rect<int> &, const Color &);
		bool fill (const Rect<int> &, const AlphaColor &);

		void line (float fx0, float fy0, float fx1, float fy1, const Color & c);
		
		void disc (float cx, float cy, float ro, float ri, const Color & color, int alpha);

		void text (int x, int y, char * s, const AlphaColor & a);
		int text_width (const char * s);

		void pixelf (float x, float y, const Color & c, int alpha);
		inline void put (int x, int y, const Color & c, int alpha)
			{
				if (!clip.has (x, y))
					return;
				Pixel p (pointer (x, y));
				p.put (c, alpha);
			}

		void columnf (int x, float y0, float y1, const Color & c, int alpha);
		void fade_column (int x, float y0, float y1, const Color & c, int a0, int a1);

		int font_width();
		int font_height();

		void blend (RGB * a, RGB * b, double ab);
		void copy (RGB *);

		void write_tga (char * path);

#if defined (WITH_JPEG)
		void read_jpeg (char * file);
		void read_half_jpeg (char * file);
#endif
};

#endif /* _RGB_H_ */
