/* (c) 2001, 2002 tim goetze <tim@quitte.de> -- see the file 'COPYING.gz'. */

#include <sys/mman.h>
#include <dlfcn.h>

#include "basics.h"

#include "Thread.h"
#include "PythonThread.h"

#include "Event.h"
#include "ObjectFifo.h"

#include "PyType.h"

extern void init_Event_types (PyObject * dict);
extern void init_Audio_types (PyObject * dict);	
extern void init_ObjectFifo_types (PyObject * dict);

extern void init_Graph_types (PyObject * dict);
extern void init_Scheduler_types (PyObject * dict);
extern void init_ScriptPlugin_types (PyObject * dict);
extern void init_Ladspa_types (PyObject * dict);
extern void init_MidiStream_types (PyObject * dict);

/* also has AudioClock. */
extern void init_AlsaAudio_types (PyObject * dict);

extern void init_RTClock_types (PyObject * dict);
extern void init_EventList_types (PyObject * dict);
extern void init_Part_types (PyObject * dict);
extern void init_DiskStream_types (PyObject * dict);
extern void init_SoundFile_types (PyObject * dict);
extern void init_Gain_types (PyObject * dict);
extern void init_LFO_types (PyObject * dict);
extern void init_Modulator_types (PyObject * dict);
extern void init_Noise_types (PyObject * dict);
extern void init_Jack_types (PyObject * dict);
extern void init_IIWU_types (PyObject * dict);
extern void init_Sine_types (PyObject * dict);
extern void init_VCF_types (PyObject * dict);
extern void init_VCO_types (PyObject * dict);
extern void init_Delay_types (PyObject * dict);
extern void init_IIR_types (PyObject * dict);
extern void init_FIR_types (PyObject * dict);
extern void init_MpegFile_types (PyObject * dict);
extern void init_SamplePlayer_types (PyObject * dict);
extern void init_Resampler_types (PyObject * dict);
extern void init_Granular_types (PyObject * dict);
extern void init_Spice_types (PyObject * dict);

extern void init_Cycler_types (PyObject * dict);
extern void init_AlsaMidi_types (PyObject * dict);

extern void init_Map_types (PyObject * dict);

#ifdef WITH_GTK
extern void init_Gtk_types (PyObject * dict);
extern void fini_Gtk_types();
#endif /* WITH_GTK */

#ifdef WITH_PVOC
extern void init_PhaseVocoder_types (PyObject * dict);
#endif /* WITH_PVOC */

extern void init_OSS_types (PyObject * dict);

/* /////////////////////////////////////////////////////////////////////// */

static PyObject *
read_var (PyObject * self, PyObject * file)
Py_TRY
{
	FILE * fp = PyFile_AsFile (file);
	
	if (fp == 0)
		throw TypeError() << "need a file, not " << get_type_name (file);

	/* inspired by code from Will Ware, posted to comp.lang.python iirc. */ 
	 
	long v = 0;

	while (1)
	{
		byte c = (byte) fgetc (fp);
		byte d = c & 0x7F;
		
		v = (v << 7) | d;
		
		if (c == d) /* ie c has no 0x80 */
			return PyInt (v);
	}

	return PyNone();
}
Py_CATCH (0);

static PyObject *
read_int (PyObject * self, PyObject * file)
Py_TRY
{
	FILE * fp = PyFile_AsFile (file);
	
	if (fp == 0)
		throw TypeError() << "need a file, not " << get_type_name (file);

	return PyInt (fgetc (fp));
}
Py_CATCH (0);

/* /////////////////////////////////////////////////////////////////////// */

#define A 440.0

d_float 
n2f (d_note n)
{
	return A * pow (2.0, (n - 69.0) / 12.0);
}

d_note
f2n (d_float f)
{
	/* 0.69314718055994529 = log (2) */
	return 69.0 + 12.0 * log (f / A) / 0.69314718055994529; 
}

d_note 
PyNote2n (PyObject * note)
{
	if (PyString_Check (note))
		return MidiEvent::string2note (PyString_AsString (note));

	d_note n;

	if (!PyArg_Parse (note, "d", &n))
		throw TypeError() << "note must be a string or a number.";
	
	if (n < 0 || n > 127) 
		throw ValueError() << "note must be >= 0 and <= 127.";
	
	return n;
}

static PyObject *
module_n2n (PyObject * self, PyObject * note)
Py_TRY
{
	if (PyString_Check (note))
		return PyFloat (MidiEvent::string2note (PyString_AsString (note)));
	
	double d;
	
	if (!PyArg_Parse (note, "d", &d))
		return 0;
	
	char s[12];
	return PyString (MidiEvent::note2string (d, s));
}
Py_CATCH (0);

static PyObject *
module_n2f (PyObject * self, PyObject * note)
Py_TRY
{
	return PyFloat (n2f (PyNote2n (note)));
}
Py_CATCH (0);

static PyObject *
module_f2n (PyObject * self, PyObject * arg)
Py_TRY
{
	double f;

	if (!PyArg_Parse (arg, "d", &f))
		return 0;

	return PyFloat (f2n (f));
}
Py_CATCH (0);

static PyObject *
module_exp2 (PyObject * self, PyObject * arg)
{
	if (sizeof (float) != sizeof (int))
		return PyNone();

	float x;

	if (!PyArg_Parse (arg, "f", &x))
		return 0;

	long * px = (long *) (&x); // store address of float as long pointer
	const float tx = (x - .5f) + (3 << 22); // temporary value for truncation
	const long  lx = *((long *) &tx) - 0x4b400000; // integer power of 2
	const float dx = x - (float) (lx); // float remainder of power of 2
	x = 1.f + dx * (0.6960656421638072f + // cubic apporoximation of 2^x
						dx * (0.224494337302845f +  // for x in the range [0, 1]
						dx * (0.07944023841053369f)));
	*px += (lx << 23); // add integer power of 2 to exponent

	return PyFloat (x);
}

/* (used for chebyshev) polynomial calculator */
static PyObject *
module_poly (PyObject * self, PyObject * args)
Py_TRY
{
	double x;

	if (!PyArg_ParseTuple (args, "dO", &x, &args))
		return 0;

	if (!PyList_Check (args))
		throw TypeError() << "need a float and a list.";
	
	int n = PyList_Size (args);

	if (n == 0)
		return PyFloat (0);

	double a[n];

	for (int i = 0; i < n; ++i)
	{
		PyObject * f = PyList_GetItem (args, i);
			
		if (!PyFloat_Check (f))
			throw TypeError() << "need a list of floats.";

		a[i] = PyFloat_AS_DOUBLE (f);
	}

	/*
	for (int i = 0; i < frames; ++i)
	{
		int n = HARMONICS - 1;
		float x = s[i], y = c[n];
		
		while (n > 0)
		{
			y *= x;
			y += c[--n];
		}

		s[i] = y;
	}
	*/
	--n;
	double y = a[n];
	
	while (n > 0)
	{
		y *= x;
		y += a[--n];
	}

	return PyFloat (y);
}
Py_CATCH (0);

static PyObject *
module_db2lin (PyObject * self, PyObject * arg)
{
	double db;

	if (!PyArg_Parse (arg, "d", &db))
		return 0;
	
	return PyFloat (pow (10., db / 20.));
}

static PyObject *
module_lin2db (PyObject * self, PyObject * arg)
{
	double lin;

	if (!PyArg_Parse (arg, "d", &lin))
		return 0;
	
	return PyFloat (20 * log10 (lin));
}

static PyObject *
module_str2hex (PyObject * self, PyObject * arg)
Py_TRY
{
	static char * x = "0123456789ABCDEF";

	unsigned char * s; int n;
	if (PyString_AsStringAndSize (arg, (char **) &s, &n))
		throw TypeError() << "need a string.";

	char t[n * 3];
	int j = 0;
	for (int i = 0; i < n; ++i)
	{
		unsigned int c = s[i];
		t[j++] = x[(c & 0xF0) >> 4];
		t[j++] = x[(c & 0x0F)];
		t[j++] = (i & 7) != 7 ? ' ' : '\n';
	}

	return PyString_FromStringAndSize ((char *) t, n * 3 - 1);
}
Py_CATCH (0);

static PyObject *
module_finite (PyObject * self, PyFloatObject * arg)
Py_TRY
{
	if (!PyFloat_Check (arg))
		throw TypeError() << "needs a float.";

	return PyInt (finite (arg->ob_fval));
}
Py_CATCH (0);

/* /////////////////////////////////////////////////////////////////////// */

#ifdef DEBUG

static PyObject *
module_curexc (PyObject * self)
{
	PyObject * o = PyErr_Occurred();
	if (o == 0)
		return PyNone();
	return ::incref (o);
}

static PyObject *
module_trap (PyObject * self)
{
	TRAP;

	return PyNone();
}

static PyObject *
module_object (PyObject * self, PyObject * arg)
{
	return ::incref ((PyObject *) PyInt_AsLong (arg));
}

#endif

static PyObject *
module_ob_refcnt (PyObject * self, PyObject * arg)
{
	return PyInt (arg->ob_refcnt);
}

static PyMethodDef  
module_methods [] = 
{
	{"read_var", (PyCFunction) read_var, METH_O,
	 "read_var (file) -- read a variable-length MIDI file entity."},
	{"read_int", (PyCFunction) read_int, METH_O,
	 "read_int (file) -- read one byte from a file."},
	{"n2f", (PyCFunction) module_n2f, METH_O,
	 "n2f (note) -- return the frequency (Hz) for a given midi note."},
	{"n2n", (PyCFunction) module_n2n, METH_O,
	 "n2n (note) -- return a string describing a note number or "
		"a note number from a string."},
	{"f2n", (PyCFunction) module_f2n, METH_O,
	 "f2n (frequency) -- return the midi note for a given frequency (Hz)."},
	{"exp2", (PyCFunction) module_exp2, METH_O,
	 "exp2 (x) -- return a fast approximation for pow (2, x)."},
	{"poly", (PyCFunction) module_poly, METH_VARARGS,
	 "poly (x, (c0, c1, ...)) -- return c0 + c1 * x + c2 * x * x ..."},
	{"db2lin", (PyCFunction) module_db2lin, METH_O,
	 "db2lin (db) -- return linear amplification factor for 'db'."},
	{"lin2db", (PyCFunction) module_lin2db, METH_O,
	 "lin2db (lin) -- return dB for linear amplification factor."},
	{"str2hex", (PyCFunction) module_str2hex, METH_O,
	 "str2hex (string) -- return a hex representation of a string."},
	{"ob_refcnt", (PyCFunction) module_ob_refcnt, METH_O,
	 "ob_refcnt (object) -- return the reference count on 'object'."},
	{"finite", (PyCFunction) module_finite, METH_O,
	 "finite (float) -- tell whether float is finite."},

	#ifdef DEBUG
	{"_curexc", (PyCFunction) module_curexc, METH_NOARGS,
	 "_curexc() -- fetch mysterious exceptions."},
	{"_trap", (PyCFunction) module_trap, METH_NOARGS,
	 "_trap() -- step on a hardware trap to activate the debugger. "
	 "will cause the process to exit unconditionally if not running under "
	 "gdb (or whatever you may be using instead)."},
	{"_object", (PyCFunction) module_object, METH_O,
	 "_object (address) -- return (PyObject *) address. don't try this "
	 "at home, kids."},
	#endif

	{ }
};

/* /////////////////////////////////////////////////////////////////////// */

Type * ModuleReference_type = 0;

/* The cleanup needed at exit (disposing of all objects still in a thread-
 * local trash fifo) should be done while the Python interpreter hasn't 
 * shut down yet; to go sure it's executed on module unload.
 */
class ModuleReference
: public Object
{
	public:
		ModuleReference()
			: Object (ModuleReference_type)
			{ }

		~ModuleReference()
			{
#ifdef WITH_GTK
				fini_Gtk_types();
#endif /* WITH_GTK */

				/* deactivate the trash fifo of the initial (this) thread. */
				PythonThread::get()->active = 0;

				PythonThread::exit();

				/* release the PL to give threads waiting for it a chance to finish */
				{
					PythonUnlock unlock;
					PythonThread::housekeep (true);
				}
				
				msg(), 
						(int) Object::objects_alive - 1, "object(s), ",
						(int) Thread::threads_started, "thread(s), ",
						(int) PythonThread::all->length, "trash fifo(s).";
			}
};

char * Py<ModuleReference>::__doc__ = 
"ModuleReference() -- one object of this type is created at module "
"initialization time and inserted into the module dict, performing various "
"cleanup functions when the module is unloaded.";

void
Py<ModuleReference>::__init__ (PyObject * args, PyObject * kwds)
{
	throw NotImplementedError() << "you cannot make this.";
}

void
Py<ModuleReference>::type_init (Type * type, PyObject * dict)
{
	type->tp_base = Object_type;
}

#ifdef DEBUG
static void
at_exit()
{
	if (Object::objects_alive)
		fprintf (stderr, "%d object(s).\n", (int) Object::objects_alive);
}
#endif

/* /////////////////////////////////////////////////////////////////////// */

extern char * __version__; /* __version__.cc */

extern "C" 
{
	void
	initnam()
	{
		#ifdef WITH_STDMSG
		bool have_stdmsg = _msg::init ("stdmsg");
		#else
		bool have_stdmsg = _msg::init ("/dev/null");
		#endif

		_msg(), "\n";
		_msg(), "*** nam ", __version__, 
				" * Copyright 2001-4 Tim Goetze <tim@quitte.de> ***";

		/* try to disable paging, unconditionally. */
		if (mlockall (MCL_CURRENT | MCL_FUTURE))
			msg(), "cannot lock memory: ", strerror (errno);

		/* initialize the gettimeofday call. */
		gettimeofday();

		#ifdef DEBUG
		atexit (at_exit);
		#endif

		/* enable plugin modules to find symbols defined in this .so */
		dlopen (0, RTLD_LAZY | RTLD_GLOBAL);
		
		PyEval_InitThreads();
		Thread::init();

		PyObject * self = Py_InitModule4 
		(
			"nam", 
			module_methods,
			"http://quitte.de/nam.html",
			(PyObject *) 0,
			PYTHON_API_VERSION
		);

		PyObject * dict = PyModule_GetDict (self);

		PyDict_SetItemString (dict, "__version__", 
				PyString_FromString (__version__));
		if (have_stdmsg)
			PyDict_SetItemString (dict, "stdmsg", _msg::stdmsg);
		
		/* instantiate the basic types. */
		Object_type = new PyType<Object> ("Object", dict);
		NamedObject_type = new PyType<NamedObject> ("NamedObject", dict);

		ModuleReference_type = new PyType<ModuleReference> ("ModuleReference", dict);

		Object * r = new ModuleReference();
		PyDict_SetItemString (dict, "_module_reference", r);
		::decref (r, true);

		/* python doesn't define these, but one occasionally needs them. */
		PyDict_SetItemString (dict, "nan", PyFloat (NAN));
		PyDict_SetItemString (dict, "inf", PyFloat (HUGE_VALL));

		/* core types, order is important. */
		init_ObjectFifo_types (dict);
		init_Graph_types (dict);
		
		init_Event_types (dict);
		init_Audio_types (dict);

		init_RTClock_types (dict);

		init_EventList_types (dict);
		init_Part_types (dict);

		init_Scheduler_types (dict);

		init_AlsaAudio_types (dict);
		init_MidiStream_types (dict);

		init_DiskStream_types (dict);
		init_SoundFile_types (dict);
		init_ScriptPlugin_types (dict);

		init_Jack_types (dict);
		init_Ladspa_types (dict);
		init_IIWU_types (dict);

		/* */

		init_Gain_types (dict);
		
		init_LFO_types (dict);
		init_Modulator_types (dict);
		init_Noise_types (dict);

		init_Sine_types (dict);
		init_VCF_types (dict);
		init_VCO_types (dict);
		init_Delay_types (dict);
		init_IIR_types (dict);
		init_FIR_types (dict);

		init_MpegFile_types (dict);
		init_SamplePlayer_types (dict);
		init_Resampler_types (dict);
		init_Granular_types (dict);
		init_Spice_types (dict);
		
		init_Cycler_types (dict);
		init_AlsaMidi_types (dict);

		init_Map_types (dict);

		init_OSS_types (dict);

#ifdef WITH_PVOC
		init_PhaseVocoder_types (dict);
#endif

#ifdef WITH_GTK
		PyObject * gtk = PyModule_New ("gtkplus");
		PyDict_SetItemString (dict, "gtkplus", gtk);

		try {
			init_Gtk_types (PyModule_GetDict (gtk));
		} catch (blewit & b) {
			msg(), "cannot init gtk:\n  ", b;
		}
#endif /* WITH_GTK */
	}
} /* extern "C" */

